! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module cyl_grid
  use types
  use vector_type
  use number_utils
  implicit none
  type cyl_grid_type
     private 
     real(kind=d_t) :: r_delta, z_delta
     ! z_min is the minimum z location
     ! Assuming that cyclider is centered on axis.
     real(kind=d_t) :: z_min
     !Radius size, z size, slice size
     integer :: s_size,r_size,z_size
  end type cyl_grid_type

  interface create_grid
     module procedure create_grid_c
  end interface

  interface zone_in
     module procedure zone_in_c
  end interface

  interface inside_grid
     module procedure inside_grid_c
  end interface

  interface first_zone
     module procedure first_zone_c
  end interface

  interface number_of_zones
     module procedure number_of_zones_c
  end interface

  interface number_of_surfaces
     module procedure number_of_surfaces_c
  end interface

  interface get_adjacent_zone
     module procedure get_adjacent_zone_c
  end interface

  interface area_of_surface
     module procedure area_of_surface_c
  end interface

  interface grid_zone_center
     module procedure grid_zone_center_c
  end interface

  interface next_zone
     module procedure next_zone_c
  end interface

  interface get_point_normal
     module procedure get_point_normal_c
  end interface

  interface get_point
     module procedure get_point_c
  end interface

  interface print_surface_type
     module procedure print_surface_type_c
  end interface

  private
  public :: create_grid,number_of_zones,zone_in,maximum_surrounding_zones
  public :: surrounding_zones,number_of_surfaces,area_of_surface
  public :: get_point,get_point_normal,cyl_grid_type,first_zone,inside_grid
  public :: get_adjacent_zone,next_zone,grid_zone_center,print_surface_type
contains
  
  subroutine create_grid_c(grid,r_delta,z_delta,z_min,s_size,r_size,z_size)
    type(cyl_grid_type),intent(out) :: grid
    real(kind=d_t),intent(in) :: r_delta,z_delta,z_min
    integer,intent(in) :: s_size,r_size,z_size
    grid%r_delta = r_delta
    grid%z_delta = z_delta
    grid%z_min = z_min
    grid%s_size = s_size
    grid%r_size = r_size
    grid%z_size = z_size
  end subroutine create_grid_c

  ! Returns the total number of zones for the grid.
  integer function number_of_zones_c(grid)
    type(cyl_grid_type),intent(in) :: grid
    ! Since the middle radius has only one point, need to compensate 
    !  for that missing location.
    number_of_zones_c = grid%z_size * (grid%s_size * (grid%r_size - 1) + 1)
  end function number_of_zones_c

  ! Converts the integer s,r,z location to an integer zone_id
  integer pure function integer_location_to_zone_id(grid,s,r,z)
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: s,r,z
    ! z range 0 to z_size - 1
    ! r range 0 to r_size - 1
    ! s range 0 to s_size - 1
    integer_location_to_zone_id = s*grid%r_size*grid%z_size + r*grid%z_size + z
  end function integer_location_to_zone_id

  ! Figures out the s,r,z integers for a given zone_id
  subroutine decompose_zone_id(grid,zone_id,s,r,z)
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id
    integer,intent(out) :: s,r,z
    integer :: left
    z = mod(zone_id,grid%z_size)
    left = zone_id/grid%z_size
    r = mod(left,grid%r_size)
    s = left/grid%r_size
  end subroutine decompose_zone_id
  
  ! Figures out the integer location numbers for the given real valued 
  ! location
  pure subroutine get_integer_location(grid,x,y,z,s_i,r_i,z_i)
    type(cyl_grid_type),intent(in) :: grid
    real(kind=d_t),intent(in) :: x,y,z
    ! z_i range 0 to z_size - 1
    ! r_i range 0 to r_size - 1
    ! s_i range 0 to s_size - 1
    integer,intent(out) :: s_i,r_i,z_i
    real(kind=d_t) :: r,t
    z_i = floor((z - grid%z_min)/grid%z_delta)
    r = sqrt(x**2+y**2)
    ! returns value in range -pi to pi  
    r_i = floor(r/grid%r_delta)
    if(r_i < 1) then 
       s_i = 0
    else
       t = atan2(y,x)
       s_i = floor(grid%s_size*(t+pi)/(2*pi))
    end if
  end subroutine get_integer_location

    ! Returns the zone that the location is in.
  integer pure function zone_in_c(grid,x,y,z)
    type(cyl_grid_type),intent(in) :: grid
    real(kind=d_t),intent(in) :: x,y,z
    integer :: s_i,r_i,z_i
    call get_integer_location(grid,x,y,z,s_i,r_i,z_i)
    zone_in_c = integer_location_to_zone_id(grid,s_i,r_i,z_i)
  end function zone_in_c

  logical pure function inside_grid_c(grid,x,y,z)
    type(cyl_grid_type),intent(in) :: grid
    real(kind=d_t),intent(in) :: x,y,z
    integer :: s_i,r_i,z_i
    call get_integer_location(grid,x,y,z,s_i,r_i,z_i)
    inside_grid_c = (0 <= z_i .and. z_i < grid%z_size) .and. &
         (0 <= r_i .and. r_i < grid%r_size) .and. &
         (0 <= s_i .and. s_i < grid%s_size)
  end function inside_grid_c

  ! Gets the maximum possible number of surrounding zones for the 
  !  given grid
  integer function maximum_surrounding_zones(grid) 
    type(cyl_grid_type),intent(in) :: grid
    maximum_surrounding_zones = max(6,2+grid%s_size)
  end function maximum_surrounding_zones

  subroutine add_zone(grid,zones,num_of_zones,s,r,z)
    type(cyl_grid_type),intent(in) :: grid
    integer,dimension(:),intent(inout) :: zones
    integer,intent(inout) :: num_of_zones
    integer,intent(in) :: s,r,z
    num_of_zones = num_of_zones + 1
    zones(num_of_zones) = integer_location_to_zone_id(grid,s,r,z)
    
  end subroutine add_zone

  ! Returns a list of all the zones that surround the given zone_id in 
  !  the grid.
  subroutine surrounding_zones(grid,zone_id,zones,num_of_zones)
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id
    integer,intent(out) :: num_of_zones
    integer,dimension(:),intent(out) :: zones
    integer :: s_i,r_i,z_i,cz,cs
    num_of_zones = 0
    call decompose_zone_id(grid,zone_id,s_i,r_i,z_i)    
    if(r_i == 0) then
       ! At center, so just add all the zones that are one more than this
       do cs = 0,grid%s_size - 1
          call add_zone(grid,zones,num_of_zones,cs,r_i+1,z_i)
       end do
    else 
       ! Add slices to the side
       call add_zone(grid,zones,num_of_zones,&
            modulo(s_i+1,grid%s_size),r_i,z_i)
       call add_zone(grid,zones,num_of_zones,&
            modulo(s_i-1,grid%s_size),r_i,z_i)       
       ! add inner and outer grid locations
       if(r_i - 1 == 0) then
          ! Inner zone only has one slice.
          call add_zone(grid,zones,num_of_zones,0,r_i - 1,z_i)
       else 
          call add_zone(grid,zones,num_of_zones,s_i,r_i - 1,z_i)
       end if
       if(r_i + 1 < grid%r_size) then
          call add_zone(grid,zones,num_of_zones,s_i,r_i + 1,z_i)
       end if
    end if
    ! Top and bottom
    if( z_i > 0 )then
       cz = z_i - 1
       call add_zone(grid,zones,num_of_zones,s_i,r_i,cz)
    end if
    if( z_i + 1 < grid%z_size )then
       cz = z_i + 1
       call add_zone(grid,zones,num_of_zones,s_i,r_i,cz)
    end if
    
  end subroutine surrounding_zones
  
  ! Returns the number of surfaces that given zone_id is surrounded by.
  integer function number_of_surfaces_c(grid,zone_id) 
    integer,dimension(:),allocatable :: junk
    integer,intent(in) :: zone_id
    type(cyl_grid_type),intent(in) :: grid
    integer :: num_of_zones
    allocate(junk(maximum_surrounding_zones(grid)))
    call surrounding_zones(grid,zone_id,junk,num_of_zones)
    number_of_surfaces_c = num_of_zones
    deallocate(junk)
  end function number_of_surfaces_c
  

  ! Returns the zone id of the adjectent zone that has the given surface number
  integer function get_adjacent_zone_c(grid,zone_id,surface_number) 
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id,surface_number
    integer,dimension(:),allocatable :: other_zones
    integer :: num_of_zones
    allocate(other_zones(maximum_surrounding_zones(grid)))
    call surrounding_zones(grid,zone_id,other_zones,num_of_zones)
    get_adjacent_zone_c = other_zones(surface_number)    
    deallocate(other_zones)
  end function get_adjacent_zone_c

  ! Returns the area of the surface.
  real(kind=d_t) function area_of_surface_c(grid,zone_id,surface_number)
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id,surface_number
    integer :: other_zone
    ! Integer location for this on (si,ri,zi) and integer 
    !  locations for other zone (so,ro,zo)
    integer :: si,ri,zi,so,ro,zo,ds,dr,dz
    real(kind=d_t) :: radius_inside, radius_outside
    other_zone = get_adjacent_zone(grid,zone_id,surface_number)
    call decompose_zone_id(grid,zone_id,si,ri,zi)
    call decompose_zone_id(grid,other_zone,so,ro,zo)
    ds = so - si
    dr = ro - ri
    dz = zo - zi
    area_of_surface_c = -1.0
    if(dr /= 0 .and. ds /= 0) then
       ! This is an artifact of ri at radius zero always being zero
       ds = 0
    end if
    if(ri == 0) then 
       if(abs(ds) /= 0) then 
          print *,"ERROR should not have non zero ds",ds,dr,dz,si,ri,zi
       else if(abs(dz) /= 0) then
          ! Circle on top or bottom
          area_of_surface_c = pi*(grid%r_delta)**2
       else if(abs(dr) /= 0) then 
          ! Cylinder surface.
          area_of_surface_c = 2*pi*grid%r_delta*grid%z_delta/grid%s_size
       end if
    else
       radius_inside = ri*grid%r_delta
       radius_outside = (ri+1)*grid%r_delta
       if(abs(ds) /= 0) then 
          ! Surface is between slices
          area_of_surface_c = grid%r_delta*grid%z_delta
       else if(abs(dz) /= 0) then
          ! Circle slice part on top or bottom
          area_of_surface_c = pi*(radius_outside**2-radius_inside**2)/grid%s_size
       else if(dr < 0) then 
          ! Inner Cylinder surface.
          area_of_surface_c = 2*pi*radius_inside*grid%z_delta/grid%s_size
       else if(dr > 0) then
          ! Outer Cylinder surface
          area_of_surface_c = 2*pi*radius_outside*grid%z_delta/grid%s_size
       else 
          print *,"ERROR weird ds,dr,and,dz",ds,dr,dz,si,ri,zi
       end if
    end if
  end function area_of_surface_c

  ! Returns a point on the parameterized surface.  The u and v are used
  !  to choose the point on the surface.  u ranges from 0 to max_u and v ranges
  !  ranges from 0 to max_v.  x,y,z are the point on the surface.
  !  area is the area that the point has.  
  subroutine get_point_normal_c(grid,u,v,max_u,max_v,zone_id,surface_number,&
       x,y,z,area,normal)
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: u,v,max_u,max_v,zone_id,surface_number
    real(kind=d_t),intent(out) :: x,y,z,area
    type(vector),intent(out) :: normal
    integer :: other_zone
    ! Integer location for this on (si,ri,zi) and integer 
    !  locations for other zone (so,ro,zo)
    integer :: si,ri,zi,so,ro,zo,ds,dr,dz
    real(kind=d_t) :: radius_inside, radius_outside
    real(kind=d_t) :: theta,radius,rad_in,rad_out,theta_low,theta_high
    other_zone = get_adjacent_zone(grid,zone_id,surface_number)
    call decompose_zone_id(grid,zone_id,si,ri,zi)
    call decompose_zone_id(grid,other_zone,so,ro,zo)
    ds = so - si
    dr = ro - ri
    dz = zo - zi
    if(dr /= 0 .and. ds /= 0) then
       ! This is an artifact of ri at radius zero always being zero
       ds = 0
    end if
    if(ri == 0) then 
       if(abs(ds) /= 0) then 
          print *,"ERROR should not have non zero ds",ds,dr,dz,si,ri,zi
       else if(abs(dz) /= 0) then
          theta = (2*pi*u)/max_u
          ! Circle on top or bottom          
          radius = (grid%r_delta*v)/max_v
          if(v > 0) then 
             rad_in = (grid%r_delta*(v-0.5))/max_v
          else 
             rad_in = (grid%r_delta*0)/max_v
          end if
          if(v >= max_v) then 
             rad_out = (grid%r_delta*(v))/max_v
          else
             rad_out = (grid%r_delta*(v+0.5))/max_v
          end if
          x = radius*cos(theta)
          y = radius*sin(theta)
          if(dz > 0) then             
             z = grid%z_min+(zi+1)*grid%z_delta
             normal = vector(0.0,0.0,1.0_d_t)
          else 
             z = grid%z_min+zi*grid%z_delta
             normal = vector(0.0,0.0,-1.0_d_t)
          end if          
          area = pi*(rad_out**2-rad_in**2)/max_u
          if(u == 0 .or. u == max_u) then
             area = 0.5*area
          end if
       else if(abs(dr) /= 0) then 
          ! Cylinder surface.
          theta_high = (2*pi*(so+1))/grid%s_size
          theta_low = (2*pi*so)/grid%s_size
          theta = theta_low + ((theta_high - theta_low)*u)/max_u
          radius = grid%r_delta
          x = radius*cos(theta)
          y = radius*sin(theta)
          z = grid%z_min+zi*grid%z_delta+(grid%z_delta*v)/max_v
          normal = vector(cos(theta),sin(theta),0.0)
          area = 2*pi*grid%r_delta*grid%z_delta/(max_u*max_v*grid%s_size)
          if(u == 0 .or. u == max_u) then
             area = 0.5*area
          end if
          if(v == 0 .or. v == max_v) then
             area = 0.5*area
          end if
       end if
    else
       radius_inside = ri*grid%r_delta
       radius_outside = (ri+1)*grid%r_delta
       theta_high = (2*pi*(si+1))/grid%s_size
       theta_low = (2*pi*si)/grid%s_size
       if(abs(ds) /= 0) then 
          ! Surface is between slices
          radius = radius_inside + (grid%r_delta*u)/max_u
          if(abs(ds) /= 1) then
             ! This occurs if the ds crosses the 0 to s_size - 1 boundary.
             !  If this occurs, then the ds is the wrong sign and too big.
             if(ds > 0) then
                ds = -1
             else 
                ds = 1
             end if
          end if
          if(ds > 0) then 
             theta = theta_high
          else 
             theta = theta_low
          end if
          z = grid%z_min+zi*grid%z_delta+(grid%z_delta*v)/max_v
          x = radius*cos(theta)
          y = radius*sin(theta)
          normal = real(ds)*vector(cos(theta+pi/2.0),sin(theta+pi/2.0),0.0)
          area = grid%r_delta*grid%z_delta/(max_u*max_v)
          if(u == 0 .or. u == max_u) then
             area = 0.5*area
          end if
          if(v == 0 .or. v == max_v) then
             area = 0.5*area
          end if
       else if(abs(dz) /= 0) then
          ! Circle slice part on top or bottom
          theta = theta_low + ((theta_high - theta_low)*u)/max_u
          radius = radius_inside + (grid%r_delta*v)/max_v
          x = radius*cos(theta)
          y = radius*sin(theta)
          if(dz > 0) then
             z = grid%z_min+(zi+1)*grid%z_delta
             normal = vector(0.0,0.0,1.0_d_t)
          else 
             z = grid%z_min+zi*grid%z_delta
             normal = vector(0.0,0.0,-1.0_d_t)
          end if
          if(v > 0) then 
             rad_in = radius_inside + (grid%r_delta*(v-0.5))/max_v
          else 
             rad_in = radius_inside + (grid%r_delta*0)/max_v
          end if
          if(v >= max_v) then 
             rad_out = radius_inside + (grid%r_delta*(v))/max_v
          else
             rad_out = radius_inside + (grid%r_delta*(v+0.5))/max_v
          end if
          area = pi*(rad_out**2-rad_in**2)/(grid%s_size*max_u)
          if(u == 0 .or. u == max_u) then
             area = 0.5*area
          end if
       else 
          if(dr < 0) then 
             ! Inner Cylinder surface.
             radius = radius_inside
          else 
             ! Outer Cylinder surface
             radius = radius_outside
          end if
          theta = theta_low + ((theta_high - theta_low)*u)/max_u
          z = grid%z_min+zi*grid%z_delta+(grid%z_delta*v)/max_v
          x = radius*cos(theta)
          y = radius*sin(theta)
          normal = real(dr)*vector(cos(theta),sin(theta),0.0)
          area = 2*pi*radius*grid%z_delta/(max_u*max_v*grid%s_size)
          if(u == 0 .or. u == max_u) then
             area = 0.5*area
          end if
          if(v == 0 .or. v == max_v) then
             area = 0.5*area
          end if
       end if
    end if
    
  end subroutine get_point_normal_c

  ! Returns a point on the parameterized surface.  The u and v are used
  !  to choose the point on the surface.  u ranges from 0 to max_u and v ranges
  !  ranges from 0 to max_v.  x,y,z are the point on the surface.
  !  area is the area that the point has.  
  subroutine print_surface_type_c(grid,zone_id,surface_number)
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id,surface_number
    integer :: other_zone
    ! Integer location for this on (si,ri,zi) and integer 
    !  locations for other zone (so,ro,zo)
    integer :: si,ri,zi,so,ro,zo,ds,dr,dz
    other_zone = get_adjacent_zone(grid,zone_id,surface_number)
    call decompose_zone_id(grid,zone_id,si,ri,zi)
    call decompose_zone_id(grid,other_zone,so,ro,zo)
    ds = so - si
    dr = ro - ri
    dz = zo - zi
    if(dr /= 0 .and. ds /= 0) then
       ! This is an artifact of ri at radius zero always being zero
       ds = 0
    end if
    if(ri == 0) then 
       if(abs(ds) /= 0) then 
          print *,"ERROR should not have non zero ds",ds,dr,dz,si,ri,zi
       else if(abs(dz) /= 0) then
          ! Circle on top or bottom          
          if(dz > 0) then             
             print *,"Central zone top"
          else 
             print *,"Central zone bottom"
          end if          
       else if(abs(dr) /= 0) then 
          print *,"Central zone edge"
       end if
    else
       if(abs(ds) /= 0) then 
          ! Surface is between slices
          print *,"Between slice"
       else if(abs(dz) /= 0) then
          ! Circle slice part on top or bottom
          if(dz > 0) then
             print *,"Circle slice top"
          else 
             print *,"Circle slice bottom"
          end if
       else 
          if(dr < 0) then 
             ! Inner Cylinder surface.
             print *,"Inner cyclinder surface"
          else 
             ! Outer Cylinder surface
             print *,"Outer cyclinder surface"
          end if
       end if
    end if
    
  end subroutine print_surface_type_c


  ! Returns a point on the parameterized surface.  The u and v are used
  !  to choose the point on the surface.  u ranges from 0 to max_u and v ranges
  !  ranges from 0 to max_v.  x,y,z are the point on the surface.
  !  area is the area that the point has.  
  subroutine get_point_c(grid,u,v,max_u,max_v,zone_id,surface_number,x,y,z,area)
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: u,v,max_u,max_v,zone_id,surface_number
    real(kind=d_t),intent(out) :: x,y,z,area
    type(vector) :: normal
    call get_point_normal(grid,u,v,max_u,max_v,zone_id,surface_number,&
         x,y,z,area,normal)    
  end subroutine get_point_c

  ! Returns the first zone_id that can be used for iterating through
  !  all the zones.
  integer function first_zone_c(grid)
    type(cyl_grid_type),intent(in) :: grid
    first_zone_c = integer_location_to_zone_id(grid,0,0,0)
  end function first_zone_c

  ! Returns the next zone in the iteration
  integer function next_zone_c(grid,zone_id)
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id
    integer :: s_i,r_i,z_i
    call decompose_zone_id(grid,zone_id,s_i,r_i,z_i)
    s_i = s_i + 1
    if(s_i >= grid%s_size .or. (s_i > 0 .and. r_i == 0)) then
       s_i = 0
       r_i = r_i + 1
       if(r_i >= grid%r_size) then
          r_i = 0
          z_i = z_i + 1
       end if
    end if
    next_zone_c = integer_location_to_zone_id(grid,s_i,r_i,z_i)
  end function next_zone_c

  ! Return the center of the zone with the given id.
  subroutine grid_zone_center_c(grid,zone_id,x,y,z)
    type(cyl_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id
    real(kind=d_t),intent(out) :: x,y,z
    integer :: s_i,r_i,z_i
    real(kind=d_t) :: theta,radius
    call decompose_zone_id(grid,zone_id,s_i,r_i,z_i)
    z = grid%z_min+(z_i+0.5)*grid%z_delta
    if(r_i == 0) then
       x = 0.0
       y = 0.0
    else
       theta = (2*pi*(s_i+0.5))/grid%s_size
       radius = (r_i+0.5)*grid%r_delta
       x = radius*sin(theta)
       y = radius*cos(theta)
    end if
  end subroutine grid_zone_center_c


end module cyl_grid

