! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module earthquake
  use types
  use vector_type
  use number_utils
  use str_utils

  implicit none

  type earthquake_wave
     !What time this wave starts and ends
     real(kind=d_t) :: start_time, end_time
     ! What direction is the displacement in
     !  the amplitude is the length of this
     type(vector) :: displacement
     ! period of the wave
     real(kind=d_t) :: period
     ! initial cycle to start in
     ! should range from 0 to 2*pi
     real(kind=d_t) :: initial_cycle
     !This is the offset of the displacement
     ! 0 means none, 1 means displacement ranges from 
     ! 0 to 2*displacement, -1 means displacement ranges from
     ! -2*displacement to 0
     real(kind=d_t) :: offset
  end type earthquake_wave

  
  !Earthquake parameters
  ! Shuts off earthquakes at compile time.
  logical, parameter :: ct_earthquake = .true.
  logical :: is_earthquake
  ! The time that the first wave starts and the last wave ends
  real(kind=d_t) :: first_earthquake_start
  real(kind=d_t) :: last_earthquake_end
  ! The number of earthquake waves that are created
  integer :: number_of_earthquake_waves
  ! the number of earthquake waves that are allocated.
  integer :: max_number_of_earthquake_waves
  ! The array of earthquake waves
  type(earthquake_wave), allocatable, dimension(:) :: earthquake_waves
  ! An array of displacements
  type(vector), allocatable, dimension(:) :: earthquake_displacement
  ! When the array of displacements starts in time
  real(kind=d_t) :: earthquake_table_start
  ! How much time is between the array of displacements
  real(kind=d_t) :: earthquake_table_delta
  ! How many values in the array of displacements
  integer :: earthquake_table_length
  ! The end time of the earthquake data.
  real(kind=d_t) :: earthquake_table_end
  ! Use tabular data
  logical :: tabular_earthquake
  ! The file name to load the tabular data from.
  character(256) :: tabular_earthquake_filename

  INTEGER, PARAMETER :: LOAD_EQ_UNIT = 210

  private
  public :: calc_earthquake_displacement, set_earthquake_defaults
  public :: initialize_earthquake, enable_earthquake_waves
  public :: parse_earthquake_sine_wave_line
  public :: parse_eq_sine_wave_offset_line
  public :: parse_tabular_earthquake, free_earthquake
  public :: get_earthquake_displacement, print_earthquake_info

contains

  subroutine set_earthquake_defaults
    is_earthquake = .false.
    first_earthquake_start = -1.0
    last_earthquake_end = -1.0
    number_of_earthquake_waves = 0
    max_number_of_earthquake_waves = 0
    tabular_earthquake = .false.

  end subroutine set_earthquake_defaults

  subroutine print_earthquake_info
    integer :: i
    if(is_earthquake) then
       PRINT *,"#earthquake",is_earthquake .and. ct_earthquake,&
            "first_start",first_earthquake_start,"last_end",last_earthquake_end
       do i=1,number_of_earthquake_waves
          print *,"#wave start",earthquake_waves(i)%start_time,"end",&
               earthquake_waves(i)%end_time,"displacement",&
               earthquake_waves(i)%displacement,"period",&
               earthquake_waves(i)%period,"initial_cycle",&
               earthquake_waves(i)%initial_cycle,"offset",&
               earthquake_waves(i)%offset
       end do
    end if
    if(tabular_earthquake) then
       print *,"#Tabular earthquake ",earthquake_table_start," to ",&
            earthquake_table_end," timestep ",earthquake_table_delta
       print *,"#Tabular "," number ",earthquake_table_length," filename ",&
            tabular_earthquake_filename
       
    end if
    if(.not. is_earthquake .and. .not. tabular_earthquake) then 
       PRINT *,"#earthquake off"
    end if

  end subroutine print_earthquake_info

  subroutine initialize_earthquake
    if(tabular_earthquake) then
       call load_earthquake_table()
    end if
  end subroutine initialize_earthquake

  subroutine free_earthquake
    if(tabular_earthquake) then
       deallocate(earthquake_displacement)
    end if
  end subroutine free_earthquake

  subroutine load_earthquake_table() 
    integer :: i,openstatus
    real(kind=d_t) :: x,y,z
    open(LOAD_EQ_UNIT,file=tabular_earthquake_filename,action="READ",&
         iostat=openstatus)
    if(openstatus > 0) then 
       print *,"ERROR Cannot open file to load earthquake table"
       return
    end if
    do i=1,earthquake_table_length
       read(LOAD_EQ_UNIT,*) x,y,z
       earthquake_displacement(i) = vector(x,y,z)
       !print *,i,earthquake_displacement(i)
    end do
    close(LOAD_EQ_UNIT)
  end subroutine load_earthquake_table


  subroutine enable_earthquake_waves(in_max_number_waves)
    integer,intent(in) :: in_max_number_waves
    max_number_of_earthquake_waves = in_max_number_waves
    allocate(earthquake_waves(max_number_of_earthquake_waves))
    is_earthquake = .true.
  end subroutine enable_earthquake_waves

  subroutine parse_earthquake_sine_wave_line(line,start)
    character(*), intent(in) :: line
    integer,intent(in) :: start
    integer :: i, at
    character(52) :: word
    at = start
    
    number_of_earthquake_waves = number_of_earthquake_waves + 1        
    i = number_of_earthquake_waves
    if(.not. ct_earthquake) then 
       print *,"ERROR earthquake not enabled in this compile"
    end if
    if(.not. is_earthquake) then 
       print *,"ERROR earthquake not enabled"
    end if
    if(i > max_number_of_earthquake_waves) then
       print *,"ERROR not enough earthquake waves allocated"
    end if
    call get_param(line,word,at)
    earthquake_waves(i)%start_time = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%end_time = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%displacement%x = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%displacement%y = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%displacement%z = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%period = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%initial_cycle = str_to_double(word)        
    earthquake_waves(i)%offset = 0.0
    if(i == 1) then 
       first_earthquake_start = earthquake_waves(i)%start_time
       last_earthquake_end = earthquake_waves(i)%end_time
    else 
       if(earthquake_waves(i)%start_time < first_earthquake_start) then
          first_earthquake_start = earthquake_waves(i)%start_time
       end if
       if(last_earthquake_end < earthquake_waves(i)%end_time) then
          last_earthquake_end = earthquake_waves(i)%end_time
       end if
    end if

  end subroutine parse_earthquake_sine_wave_line
    
  subroutine parse_eq_sine_wave_offset_line(line,start)
    character(*), intent(in) :: line
    integer,intent(in) :: start
    integer :: i, at
    character(52) :: word
    at = start
    number_of_earthquake_waves = number_of_earthquake_waves + 1        
    i = number_of_earthquake_waves
    if(.not. ct_earthquake) then 
       print *,"ERROR earthquake not enabled in this compile"
    end if
    if(.not. is_earthquake) then 
       print *,"ERROR earthquake not enabled"
    end if
    if(i > max_number_of_earthquake_waves) then
       print *,"ERROR not enough earthquake waves allocated"
    end if
    call get_param(line,word,at)
    earthquake_waves(i)%start_time = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%end_time = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%displacement%x = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%displacement%y = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%displacement%z = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%period = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%initial_cycle = str_to_double(word)        
    call get_param(line,word,at)
    earthquake_waves(i)%offset = str_to_double(word)
    if(i == 1) then 
       first_earthquake_start = earthquake_waves(i)%start_time
       last_earthquake_end = earthquake_waves(i)%end_time
    else 
       if(earthquake_waves(i)%start_time < first_earthquake_start) then
          first_earthquake_start = earthquake_waves(i)%start_time
       end if
       if(last_earthquake_end < earthquake_waves(i)%end_time) then
          last_earthquake_end = earthquake_waves(i)%end_time
       end if
    end if
    
  end subroutine parse_eq_sine_wave_offset_line

  subroutine parse_tabular_earthquake(line,start)
    character(*), intent(in) :: line
    integer,intent(in) :: start
    integer :: at
    character(52) :: word
    at = start
    tabular_earthquake = .true.
    call get_param(line,word,at)
    earthquake_table_start = str_to_double(word)
    call get_param(line,word,at)
    earthquake_table_delta = str_to_double(word)
    call get_param(line,word,at)
    earthquake_table_length = str_to_int(word)
    call get_param(line,word,at)
    tabular_earthquake_filename = word
    allocate(earthquake_displacement(earthquake_table_length))
    earthquake_table_end = earthquake_table_start + &
         (earthquake_table_length-1)*earthquake_table_delta

  end subroutine parse_tabular_earthquake

  pure type(vector) function get_earthquake_displacement(time)
    real(kind=d_t), intent(in) :: time
    type(vector) :: displacement, wall_velocity
    logical :: displaced
    call calc_earthquake_displacement(displacement,wall_velocity,time,displaced)
    get_earthquake_displacement = displacement
  end function get_earthquake_displacement

  pure subroutine calc_earthquake_displacement(displacement,wall_velocity,time,displaced)
    real(kind=d_t), intent(in) :: time
    type(vector), intent(out) :: displacement, wall_velocity
    logical, intent(out) :: displaced
    displaced = .true.
    if(ct_earthquake .and. is_earthquake .and. first_earthquake_start <= time &
         .and.  time <= last_earthquake_end) then 
       displacement = calc_displacement(time)
       wall_velocity = calc_velocity(time)
       !call default_base_force(pos+displacement,velocity+wall_velocity, omega,&
       !     force,torque,slip,slip_normal,current_wear)
    else if(ct_earthquake .and. tabular_earthquake .and. &
         earthquake_table_start < time .and. time < earthquake_table_end) then
       call calc_tabular_eq_data(displacement,wall_velocity,time)
       !call default_base_force(pos+displacement,velocity+wall_velocity, omega,&
       !     force,torque,slip,slip_normal,current_wear)
    else 
       displaced = .false.
       displacement = vector(0.0,0.0,0.0)
       wall_velocity = vector(0.0,0.0,0.0)
       !call default_base_force(pos,velocity, omega, force, torque, slip,&
       !     slip_normal,current_wear)
    end if
    
  end subroutine calc_earthquake_displacement

  pure subroutine calc_tabular_eq_data(displacement,wall_velocity,time)
    real(kind=d_t), intent(in) :: time
    type(vector), intent(out) :: displacement, wall_velocity
    real(kind=d_t) :: k_real, offset
    type(vector) :: d_k,d_kp1
    integer :: k
    if(.not. tabular_earthquake .or. time <= earthquake_table_start .or. earthquake_table_end <= time) then
       displacement = vector(0.0d0,0.0d0,0.0d0)
       wall_velocity = vector(0.0d0,0.0d0,0.0d0)
    else 
       k_real = (time - earthquake_table_start)/earthquake_table_delta + 1
       k = floor(k_real)
       offset = k_real - k
       d_k = earthquake_displacement(k)
       d_kp1 = earthquake_displacement(k+1)
       displacement = (1.0-offset)*d_k+offset*d_kp1
       wall_velocity = (1.0/earthquake_table_delta)*(d_kp1 - d_k)
    end if
  end subroutine calc_tabular_eq_data


  pure type(vector) function calc_displacement(time) 
    real(kind=d_t), intent(in) :: time
    integer :: i
    type(vector) :: displacement
    type(earthquake_wave) :: wave
    displacement = vector(0.0,0.0,0.0)
    do i=1,number_of_earthquake_waves
       wave = earthquake_waves(i)
       if(wave%start_time < time .and. time < wave%end_time) then
          displacement = displacement + &
               (sin((time - wave%start_time)*&
               2.0*pi/wave%period+wave%initial_cycle)+&
               wave%offset)*&
               wave%displacement
       end if
    end do
    calc_displacement = displacement
  end function calc_displacement

  pure type(vector) function calc_velocity(time) 
    real(kind=d_t), intent(in) :: time
    integer :: i
    type(vector) :: velocity
    type(earthquake_wave) :: wave
    velocity = vector(0.0,0.0,0.0)
    do i=1,number_of_earthquake_waves
       wave = earthquake_waves(i)
       if(wave%start_time < time .and. time < wave%end_time) then
          velocity = velocity + &
             (2.0*pi/wave%period)*&
             cos((time-wave%start_time)*2.0*pi/wave%period+wave%initial_cycle)&
             *wave%displacement
       end if
    end do
    calc_velocity = velocity
  end function calc_velocity


end module earthquake
