! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module force_tally
  use types
  use vector_type
  use number_utils
  use constants
  implicit none 

  type(vector),dimension(:),allocatable :: force_sum
  real(kind=d_t),dimension(:),allocatable :: pp_sum, z_sum, surface_sum
  real(kind=d_t),dimension(:,:),allocatable :: rz_pp_sum, rz_z_sum, rz_n_sum
  integer(kind=li),dimension(:),allocatable ::force_count,pp_count,surface_count
  integer(kind=li),dimension(:,:),allocatable:: rz_count
  real(kind=d_t) :: z_low,z_high,z_delta,r_delta,r_max

  type wall_tally_type
     private
     real(kind=d_t),dimension(:),pointer :: wall_sum, vertical_sum, transverse_sum
     integer(kind=li),dimension(:),pointer :: wall_count
  end type wall_tally_type

  type(wall_tally_type) :: wall_tally
  integer :: number_of_timesteps

  private
  public :: tally_force, initialize_force_tally,print_force_tally,tally_pp_force
  public :: tally_pebble_wall_force,wall_tally_type
  public :: begin_new_force_tally_timestep

contains
  subroutine tally_force(force,pp,index)
    type(vector), intent(in) :: force
    integer, intent(in) :: index
    type(vector),dimension(:),pointer :: pp
    integer :: i
    i = iclamp(lbound(force_sum,1),int((pp(index)%z-z_low)/z_delta),&
         ubound(force_sum,1))
    call atomic_increment(force_sum(i),force)
    !$OMP atomic 
    force_count(i) = force_count(i)+1
  end subroutine tally_force

  subroutine tally_pebble_wall_force(normal_force,transverse_force,position)
    type(vector),intent(in) :: normal_force,position,transverse_force
    !type(wall_tally_type),intent(inout) :: wall_tally
    integer :: i
    i = iclamp(lbound(wall_tally%wall_sum,1),int((position%z-z_low)/z_delta),&
         ubound(wall_tally%wall_sum,1))
    !$OMP atomic
    wall_tally%wall_sum(i) = wall_tally%wall_sum(i) + abs(normal_force)
    !$OMP atomic
    wall_tally%vertical_sum(i) = wall_tally%vertical_sum(i) + transverse_force%z
    !$OMP atomic
    wall_tally%transverse_sum(i) = wall_tally%transverse_sum(i) + abs(transverse_force)
    !$OMP atomic
    wall_tally%wall_count(i) = wall_tally%wall_count(i) + 1
  end subroutine tally_pebble_wall_force
  

  subroutine tally_pp_force(pfj,pfk,ppj,ppk)
    type(vector),intent(in) :: pfj,pfk,ppj,ppk
    type(vector) :: n_jk
    integer :: j,k,r_j,r_k, min_jk
    real(kind=d_t) :: mag_j,mag_k
    j = iclamp(1,int((ppj%z-z_low)/z_delta),size(force_sum))
    k = iclamp(1,int((ppk%z-z_low)/z_delta),size(force_sum))
    r_j = iclamp(1,int(sqrt(ppj%x**2+ppj%y**2)/r_delta),ubound(rz_count,2))
    r_k = iclamp(1,int(sqrt(ppk%x**2+ppk%y**2)/r_delta),ubound(rz_count,2))  
    n_jk = normalize(ppj-ppk)
    mag_j = abs(pfj)
    mag_k = abs(pfk)
    !$OMP atomic
    pp_sum(j) = pp_sum(j) + mag_j
    !$OMP atomic 
    pp_sum(k) = pp_sum(k) + mag_k
    !$OMP atomic
    pp_count(j) = pp_count(j) + 1
    !$OMP atomic
    pp_count(k) = pp_count(k) + 1
    !$OMP atomic
    z_sum(j) = z_sum(j) + abs(pfk%z)
    !$OMP atomic
    z_sum(k) = z_sum(k) + abs(pfj%z)
    !$OMP atomic
    rz_pp_sum(j,r_j) = rz_pp_sum(j,r_j) + mag_j
    !$OMP atomic
    rz_pp_sum(k,r_k) = rz_pp_sum(k,r_k) + mag_k
    !$OMP atomic
    rz_z_sum(j,r_j) = rz_z_sum(j,r_j) + abs(pfj%z)
    !$OMP atomic
    rz_z_sum(k,r_k) = rz_z_sum(k,r_k) + abs(pfk%z)
    !$OMP atomic
    rz_n_sum(j,r_j) = rz_n_sum(j,r_j) + abs(pfj .dot. n_jk)
    !$OMP atomic
    rz_n_sum(k,r_k) = rz_n_sum(k,r_k) + abs(pfk .dot. n_jk)
    !$OMP atomic
    rz_count(j,r_j) = rz_count(j,r_j) + 1
    !$OMP atomic
    rz_count(k,r_k) = rz_count(k,r_k) + 1
    
    if(j /= k) then
       min_jk = min(j,k)
       !$OMP atomic 
       surface_count(min_jk) = surface_count(min_jk) + 1
       if(j < k) then
          !$OMP atomic
          surface_sum(min_jk) = surface_sum(min_jk) + pfk%z
       else
          !$OMP atomic
          surface_sum(min_jk) = surface_sum(min_jk) + pfj%z
       end if
    end if
  end subroutine tally_pp_force

  subroutine initialize_force_tally(x_high_d,x_low_d,y_high_d,y_low_d,z_high_d,z_low_d)
    real(kind=d_t), intent(in) :: z_high_d,z_low_d,x_high_d,x_low_d,y_high_d,y_low_d
    integer :: size, r_size
    r_max = sqrt(max(abs(x_high_d),abs(x_low_d))**2 &
         +max(abs(y_high_d),abs(y_low_d))**2)
    z_low = z_low_d
    z_high = z_high_d
    z_delta = 2.0*pebble_radius
    r_delta = 2.0*pebble_radius
    size = ceiling((z_high-z_low)/z_delta)+1
    r_size = ceiling(r_max/r_delta)+1
    allocate(force_sum(size),force_count(size),pp_sum(size),pp_count(size),&
         surface_count(size))
    allocate(z_sum(size),rz_pp_sum(size,r_size),rz_z_sum(size,r_size),&
         surface_sum(size))
    allocate(rz_count(size,r_size),rz_n_sum(size,r_size))
    allocate(wall_tally%wall_sum(size),wall_tally%wall_count(size))
    allocate(wall_tally%vertical_sum(size),wall_tally%transverse_sum(size))
    force_sum(:) = vector(0.0,0.0,0.0)
    force_count(:) = 0
    pp_sum(:) = 0.0
    pp_count(:) = 0
    z_sum(:) = 0.0
    rz_pp_sum(:,:) = 0.0
    rz_z_sum(:,:) = 0.0
    rz_n_sum(:,:) = 0.0
    rz_count(:,:) = 0
    wall_tally%wall_sum(:) = 0.0
    wall_tally%vertical_sum(:) = 0.0
    wall_tally%transverse_sum(:) = 0.0
    wall_tally%wall_count(:) = 0
    surface_count(:) = 0
    surface_sum(:) = 0.0
    number_of_timesteps = 0
  end subroutine initialize_force_tally
  
  subroutine begin_new_force_tally_timestep()
    number_of_timesteps = number_of_timesteps + 1
  end subroutine begin_new_force_tally_timestep

  subroutine print_force_tally
    integer :: i, r
    integer(kind=li) :: n,n2
    do i=lbound(force_sum,1),ubound(force_sum,1)
       if(force_count(i) > 0) then 
          n = force_count(i)
          n2 = pp_count(i)
          print *,"#force_sum",i,z_low+z_delta*i,n,(1.0/n)*force_sum(i),&
               n2,pp_sum(i)/n2,z_sum(i)/n2
       end if
    end do
    do i=lbound(rz_count,1),ubound(rz_count,1)
       do r=lbound(rz_count,2),ubound(rz_count,2)
          if(rz_count(i,r) > 0) then
             n = rz_count(i,r)
             print *,"#rz_force_sum",i,r,z_low+z_delta*i,r*r_delta,n,&
                  rz_pp_sum(i,r)/n,rz_z_sum(i,r)/n,rz_n_sum(i,r)/n
          end if
       end do
    end do
    do i=lbound(wall_tally%wall_sum,1),ubound(wall_tally%wall_sum,1)
       if(wall_tally%wall_count(i) > 0) then
          n = wall_tally%wall_count(i)
          n2 = number_of_timesteps
          print *,"#wall_force_sum",i,z_low+z_delta*i,n,wall_tally%wall_sum(i),&
               wall_tally%wall_sum(i)/n2,wall_tally%vertical_sum(i)/n2,&
               wall_tally%transverse_sum(i)/n2
       end if
    end do
    do i=lbound(surface_count,1),ubound(surface_count,1)
       if(surface_count(i) > 0) then
          n = surface_count(i)
          print *,"#z_surface_force_sum",i,z_low+z_delta*(i+1),n,surface_sum(i),&
               surface_sum(i)/number_of_timesteps
       end if
    end do
  end subroutine print_force_tally
  
end module force_tally
