! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module geometry
  use types
  use constants
  use vector_type
  use number_utils
  use force_tally
  use str_utils
  use wear_tally
  implicit none

  ! Rectangular geometry stuff.
  REAL(KIND=D_T) :: x_wall_distance, y_wall_distance
  logical :: rectangular_vat


  REAL(kind=d_t) :: cone_location
  REAL(kind=d_t) :: cone_hole_radius
  REAL(kind=d_t) :: cone_radius
  REAL(kind=d_t) :: cone_slope
  REAL(kind=d_t) :: cone_bottom
  integer :: number_of_cones
  REAL(kind=d_t) :: middle_vessel_radius

  logical :: offset_inlet_and_cone


  ! Cone packing stuff
  real(kind=d_t) :: packing_cone_radius, packing_cone_bottom,packing_cone_top
  real(kind=d_t) :: packing_chute_length
  logical :: use_cone_packing

  ! hole_size is the ratio of the hole to pebble radius
  ! hole_depth is the depth of the hole to the door location in meters
  real(kind=d_t) :: hole_size, hole_depth


  REAL(kind=d_t) :: door_location
  logical :: door_open

  REAL(KIND=D_T) :: floor_location

  REAL(KIND=D_T) :: inside_vessel_radius
  REAL(KIND=D_T) :: outside_vessel_radius

  REAL(kind=d_t) :: recirculation_height

  REAL(kind=d_t),allocatable, dimension(:) :: cone_x_offsets, cone_y_offsets
  !If this is false, than the inlets need to be calculated.
  logical :: inlet_starts_specified = .false.
  type(vector),allocatable, dimension(:) :: inlet_positions, inlet_velocities

  integer,parameter :: max_contacts = 10

  type geometry_contact_type
     integer :: number_of_contacts
     type(vector),dimension(max_contacts) :: normal
     real(kind=d_t),dimension(max_contacts) :: inside_dist
     logical :: in_intersection_mode = .false. 
     logical :: empty_intersection 
     integer :: intersection_start_index
  end type geometry_contact_type

  type block_type
     !Creates a rectangular parallel piped
     !Stores the minimum and maximum dimensions in (x,y,z)
     real(kind=d_t),dimension(3) :: min,max
  end type block_type

  type cylinder_type
     !Creates a vertical cylinder
     real(kind=d_t) :: radius,x_center,y_center,z_min,z_max
  end type cylinder_type

  type plane_type
     !The normal n is also a,b,c
     !the formula for the plane is a*x+b*y+c*z+d = 0
     !n is normalized (i.e. a**2+b**2+c**2 = 1 )
     type(vector) :: n
     real(kind=d_t) :: d
  end type plane_type

  type combined_type
     !XXX when f2003 is more widely available, should just be a super
     ! class for block_type and cylinder_type
     !use the first non-null type as the kind.
     type(block_type),pointer :: block => null()
     type(cylinder_type),pointer :: cylinder => null()
     type(plane_type),pointer :: plane => null()
  end type combined_type

  type plane_union_type
     type(plane_type),pointer,dimension(:) :: list
  end type plane_union_type

  type indentation_row_type
     !This goes from this row's z_min to the next row's z_min.
     !The indentation will be centered on each angle in the angle list.
     real(kind=d_t) :: z_min
     real(kind=d_t),dimension(:),pointer :: angle_list
  end type indentation_row_type

  type indentation_type 
     !Put the plane union type pattern on each row, at the 
     ! angles in the row.  
     ! Only do this between z_min and z_max
     type(plane_union_type) :: pattern
     type(indentation_row_type),dimension(:),pointer :: row_list
     real(kind=d_t) :: z_min,z_max
     real(kind=d_t) :: wall_radius
  end type indentation_type

  type(indentation_type) :: indentation

  type group_type
     type(combined_type),pointer,dimension(:) :: list
  end type group_type
     
  !List of all the union object that are available.
  type(group_type),allocatable,dimension(:) :: union_list

  !List of all the intersection objects that are available
  type(group_type),allocatable,dimension(:) :: intersection_list

  private
  public :: default_base_force, initialize_geometry, set_geometry_defaults
  public :: set_cone, set_floor_location, set_hole, set_vessel_radius
  public :: get_vessel_area, get_vessel_fill_edge, get_loading_height
  public :: print_geometry_info, get_potential_energy_zero, set_rectangular_vat
  public :: set_number_of_cones, set_offset_inlet_and_cone
  public :: get_number_of_recirc_locs, get_recirculation_location
  public :: is_pebble_through_door, door_open, setup_cone_packing
  public :: get_max_xy_extents, default_base_force_sd, is_contacting_wall
  public :: get_base_loading_height, load_geometry_section
  public :: set_recirculation_height, get_recirculation_velocity

contains

  pure subroutine add_contact(geometry_contact,inside_dist,normal)
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: normal
    real(kind=d_t),intent(in) :: inside_dist
    integer :: n
    n = geometry_contact%number_of_contacts
    if(inside_dist < 0.0) then
       !No force, so don't add.
       call add_no_contact(geometry_contact)
       return
    end if
    if(n >= max_contacts) then
       geometry_contact%normal(n+1) = vector(pi,pi,pi)
       !stop "ERROR too many geometry contacts.  Increase max_contacts in geometry"
    else
       n = n + 1
       geometry_contact%number_of_contacts = n
       geometry_contact%normal(n) = normal
       geometry_contact%inside_dist(n) = inside_dist
    end if
  end subroutine add_contact
    
  pure subroutine add_no_contact(geometry_contact)
    type(geometry_contact_type),intent(inout) :: geometry_contact
    if(geometry_contact%in_intersection_mode) then
       geometry_contact%empty_intersection = .true.
    end if
  end subroutine add_no_contact

  pure subroutine start_intersection_mode(geometry_contact)
    type(geometry_contact_type),intent(inout) :: geometry_contact
    geometry_contact%in_intersection_mode = .true.
    geometry_contact%intersection_start_index = geometry_contact%number_of_contacts + 1
    geometry_contact%empty_intersection = .false.
  end subroutine start_intersection_mode

  pure subroutine end_intersection_mode(geometry_contact)
    type(geometry_contact_type),intent(inout) :: geometry_contact
    integer :: i, least_inside_index, first
    real(kind=d_t) :: minimum_inside_dist
    geometry_contact%in_intersection_mode = .false.
    if(geometry_contact%empty_intersection) then
       !Need to remove the new contacts added
       geometry_contact%number_of_contacts = geometry_contact%intersection_start_index - 1
    else if(geometry_contact%number_of_contacts < geometry_contact%intersection_start_index) then
       ! pass because no intersecting bodies found
    else
       ! Need to find smallest inside distance, and only keep that one.
       first = geometry_contact%intersection_start_index
       least_inside_index = first
       minimum_inside_dist = geometry_contact%inside_dist(least_inside_index)       
       do i=first+1,geometry_contact%number_of_contacts
          if(geometry_contact%inside_dist(i) < minimum_inside_dist )then
             least_inside_index = i
             minimum_inside_dist = geometry_contact%inside_dist(i)
          end if
       end do
       geometry_contact%inside_dist(first)  = geometry_contact%inside_dist(least_inside_index)
       geometry_contact%normal(first) = geometry_contact%normal(least_inside_index)
       geometry_contact%number_of_contacts = first
    end if
  end subroutine end_intersection_mode

  function is_pebble_through_door(n,loc) 
    logical :: is_pebble_through_door
    integer, intent(in) :: n !Which door to use
    type(vector), intent(in) :: loc
    is_pebble_through_door = .false.
    if(loc%z < door_location + pebble_radius) then 
       if(number_of_cones > 1) then
          if(closest_plane_location(loc%x,loc%y,cone_x_offsets,&
               cone_y_offsets) == n) then
             ! Multiple doors and pebble through correct door.
             is_pebble_through_door = .true.
          end if
       else 
          is_pebble_through_door = .true.
       end if
    end if
  end function is_pebble_through_door

  function get_number_of_recirc_locs()
    integer :: get_number_of_recirc_locs
    get_number_of_recirc_locs =  size(inlet_positions)
  end function get_number_of_recirc_locs

  subroutine set_recirculation_height(height)
    real(kind=d_t),intent(in) :: height
    recirculation_height = height
  end subroutine set_recirculation_height   

  function get_recirculation_location(n)
    type(vector) :: get_recirculation_location
    integer, intent(in) :: n !Which recirculation location to use
    get_recirculation_location = inlet_positions(n)
    return 
  end function get_recirculation_location

  function get_recirculation_velocity(n)
    type(vector) :: get_recirculation_velocity
    integer, intent(in) :: n !Which recirculation location to use
    get_recirculation_velocity = inlet_velocities(n)
    return 
  end function get_recirculation_velocity

  subroutine setup_cone_packing(cone_radius,cone_bottom,chute_length)
    real(kind=d_t), intent(in) :: cone_radius,cone_bottom,chute_length
    use_cone_packing = .true.
    packing_cone_radius = cone_radius
    packing_cone_bottom = cone_bottom
    packing_chute_length = chute_length
  end subroutine setup_cone_packing

  subroutine get_max_xy_extents(x_high,y_high)
    real(kind=d_t), intent(out) :: x_high, y_high
    if(rectangular_vat) then 
       x_high = x_wall_distance+4.0*pebble_radius
       y_high = y_wall_distance+4.0*pebble_radius
    else 
       x_high = outside_vessel_radius+4.0*pebble_radius
       y_high = x_high
    end if

  end subroutine get_max_xy_extents

  subroutine set_offset_inlet_and_cone()
    offset_inlet_and_cone = .true.
  end subroutine set_offset_inlet_and_cone

  subroutine set_number_of_cones(number_of_cones_d)
    integer, intent(in) :: number_of_cones_d
    number_of_cones = number_of_cones_d
  end subroutine set_number_of_cones

  subroutine set_rectangular_vat(x_distance,y_distance)
    real(kind=d_t), intent(in) :: x_distance,y_distance
    rectangular_vat = .true.
    x_wall_distance = x_distance
    y_wall_distance = y_distance
  end subroutine set_rectangular_vat

  ! Returns the location that can be a zero value for calculating 
  !  gravitational energy.  All pebbles should be at or above this location
  function get_potential_energy_zero()
    real(kind=d_t) :: get_potential_energy_zero
    get_potential_energy_zero = floor_location
  end function get_potential_energy_zero

  !Gets the height were the vessel is uniform (above the chutes)
  function get_loading_height()
    real(kind=d_t) :: get_loading_height
    if(use_cone_packing) then
       get_loading_height = packing_cone_top
    else
       get_loading_height = max(cone_location,floor_location)
    end if
  end function get_loading_height

  !Gets the height where the loading can start.  May be in the chutes.
  function get_base_loading_height()
    real(kind=d_t) :: get_base_loading_height
    if(use_cone_packing) then
       get_base_loading_height = packing_cone_bottom
    else
       get_base_loading_height = max(door_location,floor_location)
    end if
  end function get_base_loading_height
    

  function get_vessel_area()
    real(kind=d_t) :: get_vessel_area
    if(rectangular_vat) then
       get_vessel_area = (2.0*x_wall_distance)*(2.0*y_wall_distance)
    else 
       get_vessel_area = pi*(outside_vessel_radius**2-inside_vessel_radius**2)    
    end if
       
  end function get_vessel_area

  function get_vessel_fill_edge()
    real(kind=d_t) :: get_vessel_fill_edge
    if(rectangular_vat) then
       get_vessel_fill_edge = max(x_wall_distance,y_wall_distance)
    else 
       get_vessel_fill_edge = outside_vessel_radius
    end if

  end function get_vessel_fill_edge

  subroutine set_cone(cone_location_d,cone_slope_d)
    real(kind=d_t), intent(in) :: cone_location_d,cone_slope_d
    cone_location = cone_location_d
    cone_slope = cone_slope_d
  end subroutine set_cone

  subroutine set_floor_location(floor_location_d) 
    real(kind=d_t), intent(in) :: floor_location_d
    floor_location = floor_location_d
  end subroutine set_floor_location

  subroutine set_hole(hole_size_d,hole_depth_d)
    real(kind=d_t), intent(in) :: hole_size_d, hole_depth_d
    hole_size = hole_size_d
    hole_depth = hole_depth_d
  end subroutine set_hole

  subroutine set_vessel_radius(inside,outside)
    real(kind=d_t), intent(in) :: inside, outside
    inside_vessel_radius = inside
    outside_vessel_radius = outside
    indentation%wall_radius = outside
    rectangular_vat = .false.
  end subroutine set_vessel_radius

  subroutine set_geometry_defaults
    number_of_cones = 1
    use_cone_packing = .false.

    cone_location = -4.0
    cone_slope = 0.25
    floor_location = 0.0
    hole_size = 0.0
    hole_depth = 0.0
    inside_vessel_radius = 0.0
    outside_vessel_radius = 1.0
    indentation%wall_radius = outside_vessel_radius
    offset_inlet_and_cone = .false.
    packing_chute_length = 0.0
    rectangular_vat = .false.
    recirculation_height = 10.0

    allocate(union_list(0),indentation%pattern%list(0),intersection_list(0))
  end subroutine set_geometry_defaults

  subroutine load_plane(plane,line)
    character(*),intent(in) :: line
    type(plane_type),intent(out) :: plane
    real(kind=d_t) :: a,b,c,d, length
    read(line(6:),*) a,b,c,d
    length = sqrt(a**2+b**2+c**2)
    plane%n = (1.0/length)*vector(a,b,c)
    plane%d = d/length
  end subroutine load_plane

  subroutine load_vector_point_plane(plane,line)
    character(*),intent(in) :: line
    type(plane_type),intent(out) :: plane
    real(kind=d_t) :: a,b,c,px,py,pz,d,length
    !format of line 'plane_vp vx vy vz px py pz'
    read(line(9:),*) a,b,c,px,py,pz
    d = -a*px-b*py-c*pz
    length = sqrt(a**2+b**2+c**2)
    plane%n = (1.0/length)*vector(a,b,c)
    plane%d = d/length    
    
  end subroutine load_vector_point_plane
    

  subroutine load_geometry_section(settings_unit,echo_input)
    integer,intent(in) :: settings_unit
    logical,intent(in) :: echo_input
    type(string_queue) :: sq
    character(256) :: line
    type(string_node),pointer :: current, group_start
    character,pointer,dimension(:) :: temp
    integer :: at, unions, object_count, current_union, inlets, current_inlet
    integer :: intersections, current_intersection
    character(52) :: word
    real(kind=d_t) :: px,py,pz,vx,vy,vz
    call initialize_sq(sq)
    unions = 0
    intersections = 0
    inlets = 0
    io_loop: do
       read(settings_unit,'(A256)') line
       if(echo_input) then
          print *,"#input ",trim(line)
       end if
       at = 1
       call get_param(line, word, at)
       if(str_equal('end_geometry ',word)) then
          exit io_loop
       else if(str_equal('rem ',word)) then
          cycle io_loop
       else if(str_equal('start_union ',word)) then
          call increment(unions,1)
       else if(str_equal('start_intersection ',word)) then
          call increment(intersections,1)    
       else if(str_equal('inlet ',word)) then
          call increment(inlets,1)
       end if
       temp => str_char_to_array(trim(line))
       call sq_append(sq,temp)
       nullify(temp)
    end do io_loop
    
    deallocate(union_list)
    allocate(union_list(unions))
    deallocate(intersection_list)
    allocate(intersection_list(intersections))
    if(inlets > 0) then
       allocate(inlet_positions(inlets),inlet_velocities(inlets))
       inlet_starts_specified = .true.
    end if

    !print *,"#inlets",inlets

    object_count = -1
    current_union = 0
    current_intersection = 0
    current_inlet = 1
    call sq_first(sq,current)
    geom_loop: do
       if(.not. associated(current)) exit geom_loop
       call sn_value(current,temp)
       !line = str_array_to_char(temp)       
       !print *,"Read value ",temp
       if(ca_starts_with('start_union',temp)) then
          group_start => current
          object_count = 0
          call increment(current_union,1)
       else if(ca_starts_with('start_intersection',temp)) then
          group_start => current
          object_count = 0
          call increment(current_intersection,1)          
       else if(ca_starts_with('end_union',temp)) then          
          call load_group_section(union_list, current_union, object_count, group_start)
          nullify(group_start)
       else if(ca_starts_with('end_intersection',temp)) then          
          call load_group_section(intersection_list, current_intersection, object_count, group_start)
          nullify(group_start)
       else if(ca_starts_with('start_indentation',temp)) then
          call load_indentation_section(current)
       else if(ca_starts_with('inlet',temp)) then
          line = str_array_to_char(temp)
          read(line(6:),*) px,py,pz,vx,vy,vz
          inlet_positions(current_inlet) = vector(px,py,pz)
          inlet_velocities(current_inlet) = vector(vx,vy,vz)
          !print *,"#inlet",current_inlet,inlet_positions(current_inlet),inlet_velocities(current_inlet),trim(line)
          call increment(current_inlet,1)
       else
          !increment object count,so second pass can allocate correct union size
          call increment(object_count,1)
       end if
       call sq_inplace_next(current)
    end do geom_loop

    
  end subroutine load_geometry_section

  subroutine load_group_section(group_list, current_group, object_count, group_start)
    integer, intent(in) :: object_count, current_group
    type(group_type),dimension(:), intent(inout) :: group_list
    type(string_node),pointer :: group_start
    character(256) :: line
    character,pointer,dimension(:) :: temp
    integer :: i
    allocate(group_list(current_group)%list(object_count))
    !Skip 'start_union', goto first object
    call sq_inplace_next(group_start)
    group_loop: do i = 1,object_count
       call sn_value(group_start,temp)
       !Need characters for read.
       line = str_array_to_char(temp)
       if(ca_starts_with('block ',temp)) then
          allocate(group_list(current_group)%list(i)%block)
          !print *,"block:",line(7:)
          read (line(7:),*) group_list(current_group)%list(i)%block%min,group_list(current_group)%list(i)%block%max
          if(any(group_list(current_group)%list(i)%block%min > &
               group_list(current_group)%list(i)%block%max)) then
             print *,"ERROR max less than min in ",trim(line)
             !union_list(current_union)%list(i)%block%min,&
             !union_list(current_union)%list(i)%block%max
             stop
          end if
       else if(ca_starts_with('cylinder ',temp)) then
          allocate(group_list(current_group)%list(i)%cylinder)
          !print *,"cylinder:",trim(line)
          read (line(9:),*) group_list(current_group)%list(i)%cylinder%radius,&
               group_list(current_group)%list(i)%cylinder%x_center,&
               group_list(current_group)%list(i)%cylinder%y_center,&
               group_list(current_group)%list(i)%cylinder%z_min,&
               group_list(current_group)%list(i)%cylinder%z_max
          if(group_list(current_group)%list(i)%cylinder%z_max < &
               group_list(current_group)%list(i)%cylinder%z_min) then
             print *,"ERROR max less than min in ",trim(line)
             !group_list(current_group)%list(i)%cylinder%z_min,&
             !group_list(current_group)%list(i)%cylinder%z_max
             stop
          end if
       else if(ca_starts_with('plane ',temp)) then
          allocate(group_list(current_group)%list(i)%plane)
          call load_plane(group_list(current_group)%list(i)%plane,line)
       else if(ca_starts_with('plane_vp',temp)) then
          allocate(group_list(current_group)%list(i)%plane)
          call load_vector_point_plane(group_list(current_group)%list(i)%plane,line)
       end if
       call sq_inplace_next(group_start)
    end do group_loop
  end subroutine load_group_section

  subroutine load_indentation_section(current)
    type(string_node),pointer :: current
    type(string_node),pointer :: pattern_start, row_start, row_count_pointer
    character,pointer,dimension(:) :: temp
    character(256) :: line
    integer :: pattern_count, angle_count, row_count,row_index,i,angles_count,j
    real(kind=d_t) :: a

    !Skip start_indentation line
    call sq_inplace_next(current)

    row_count_pointer => current
    
    row_count = 0
    !Find out how many rows there are
    row_count_loop: do
       if(.not. associated(row_count_pointer)) exit row_count_loop
       call sn_value(row_count_pointer,temp)
       
       if(ca_starts_with('end_indentation',temp)) then
          exit row_count_loop
       else if(ca_starts_with('start_row',temp)) then
          call increment(row_count,1)
       end if
       call sq_inplace_next(row_count_pointer)
    end do row_count_loop
    
    allocate(indentation%row_list(row_count))
    row_index = 0

    indent_loop: do 
       if(.not. associated(current)) exit indent_loop
       call sn_value(current,temp)
       
       !read in the data.
       if(ca_starts_with('end_indentation',temp)) then
          exit indent_loop
       else if(ca_starts_with('start_pattern',temp)) then
          pattern_start => current
          pattern_count = 0
       else if(ca_starts_with('plane',temp)) then
          call increment(pattern_count,1)
       else if(ca_starts_with('end_pattern',temp)) then
          allocate(indentation%pattern%list(pattern_count))
          !Skip 'start_pattern'
          call sq_inplace_next(pattern_start)
          
          pattern_loop: do i = 1,pattern_count
             call sn_value(pattern_start,temp)
             line = str_array_to_char(temp)
             !print *,"pattern_loop",i,temp,trim(line)
             if(ca_starts_with('plane ',temp)) then
                call load_plane(indentation%pattern%list(i),line)
             else if(ca_starts_with('plane_vp',temp)) then
                call load_vector_point_plane(indentation%pattern%list(i),line)
             else
                print *,"#Error expected plane inside pattern, not:",trim(line)
                stop
             end if
             call sq_inplace_next(pattern_start)
          end do pattern_loop
          nullify(pattern_start)
       else if(ca_starts_with('min_max',temp)) then
          line = str_array_to_char(temp)
          read(line(8:),*) indentation%z_min,indentation%z_max
          if(indentation%z_min >= indentation%z_max) then
             print *,"ERROR max less than min in ",trim(line)
             stop
          end if
       else if(ca_starts_with('start_row',temp)) then
          row_start => current
          angle_count = 0
          call increment(row_index,1)
       else if(ca_starts_with('row_min',temp)) then
          line = str_array_to_char(temp)
          read(line(8:),*) indentation%row_list(row_index)%z_min
          if(row_index > 1) then 
             if(indentation%row_list(row_index-1)%z_min >= &
               indentation%row_list(row_index)%z_min) then
                print *,"ERROR rows must be in increasing order by row_min",trim(line)
                stop
             end if
          end if
          if(.not. between(indentation%z_min-1.0e-10,&
               indentation%row_list(row_index)%z_min,indentation%z_max)) then
             print *,"ERROR row_min must be between indentation min and max",trim(line)
             stop
          end if
       else if(ca_starts_with('angle ',temp)) then
          call increment(angle_count,1)
       else if(ca_starts_with('angles',temp)) then
          line = str_array_to_char(temp)
          read(line(7:),*) angles_count
          call increment(angle_count,angles_count)
       else if(ca_starts_with('end_row',temp)) then
          allocate(indentation%row_list(row_index)%angle_list(angle_count))
          !skip start_row
          call sq_inplace_next(row_start)
          i = 1
          angles_loop: do
             if(i > angle_count) exit angles_loop
             call sn_value(row_start,temp)
             if(ca_starts_with('angle ',temp)) then
                line = str_array_to_char(temp)
                read(line(6:),*) a
                !Convert to radians
                indentation%row_list(row_index)%angle_list(i) = a*2*pi/360.0
                !Note that this only increments when we get an angle
                call increment(i,1)
             else if(ca_starts_with('angles',temp)) then
                line = str_array_to_char(temp)
                read(line(7:),*) angles_count
                read(line(7:),*) j,indentation%row_list(row_index)%angle_list(i:i+angles_count-1)
                do j=i,i+angles_count-1
                   !convert from degrees to radians
                   indentation%row_list(row_index)%angle_list(j) = &
                        indentation%row_list(row_index)%angle_list(j)*2*pi/360.0
                end do
                call increment(i,angles_count)
             end if
             call sq_inplace_next(row_start)
          end do angles_loop
          
       end if
       call sq_inplace_next(current)
    end do indent_loop

  end subroutine load_indentation_section

  subroutine print_geometry_info()
    integer :: i,j
    if(rectangular_vat) then
       PRINT *,"#Rectangle x_wall",x_wall_distance," y_wall ",y_wall_distance
    else 
       PRINT *,"#ID: ",inside_vessel_radius, &
            &     " OD: ",outside_vessel_radius, &
            &     " radius: ",pebble_radius
    end if
    PRINT *,"#floor_location",floor_location
    PRINT *,"#cone",cone_location,cone_bottom,cone_slope,cone_hole_radius  
    print *,"#number_of_cones",number_of_cones
    if(number_of_cones > 1) then
       do i=1,number_of_cones 
          print *,"#cone_point ",i,cone_x_offsets(i),cone_y_offsets(i)
       end do
    end if
    do i=1,size(inlet_positions)
       print *,"#inlet_point",i,inlet_positions(i),inlet_velocities(i)
    end do
    do i=1,size(union_list)
       print *,"#union"
       do j=1,size(union_list(i)%list)
          if(associated(union_list(i)%list(j)%block)) then
             print *,"#block ",union_list(i)%list(j)%block
          else if(associated(union_list(i)%list(j)%cylinder)) then
             print *,"#cylinder",union_list(i)%list(j)%cylinder
          else if(associated(union_list(i)%list(j)%plane)) then
             print *,"#plane",union_list(i)%list(j)%plane
          end if
       end do
    end do
    if(size(indentation%pattern%list) > 0) then
       print *,"#indentation min",indentation%z_min,"max",indentation%z_max,"radius",indentation%wall_radius
       do i=1,size(indentation%pattern%list)
          print *,"#pattern",i,indentation%pattern%list(i)
       end do
       do i=1,size(indentation%row_list)
          print *,"#row",i,indentation%row_list(i)%z_min,indentation%row_list(i)%angle_list
       end do
    end if

  end subroutine print_geometry_info

  subroutine initialize_geometry(z_low,z_high)
    integer :: i
    real(kind=d_t),intent(out) :: z_low,z_high

    middle_vessel_radius = (inside_vessel_radius + outside_vessel_radius)/2.0
    cone_hole_radius = pebble_radius * hole_size
    if(number_of_cones > 1) then 
       cone_radius = sqrt(&
            (1.0*middle_vessel_radius-cos(PI/number_of_cones)*outside_vessel_radius)**2+&
            (sin(PI/number_of_cones)*outside_vessel_radius)**2)     
    else 
       cone_radius = outside_vessel_radius
    end if
    if(use_cone_packing) then
       packing_cone_top = (cone_radius-packing_cone_radius)+packing_cone_bottom
    end if
    cone_bottom = cone_location - &
         (cone_radius - cone_hole_radius) / cone_slope
    door_location = cone_bottom - hole_depth

    if(number_of_cones > 1) then
       allocate(cone_x_offsets(number_of_cones),cone_y_offsets(number_of_cones))
       do i=1,number_of_cones
          cone_x_offsets(i) = sin(i*2.0*PI/number_of_cones)*middle_vessel_radius
          cone_y_offsets(i) = cos(i*2.0*PI/number_of_cones)*middle_vessel_radius
       end do

       if(.not. inlet_starts_specified) then
          allocate(inlet_positions(number_of_cones),inlet_velocities(number_of_cones))
          
          do i=1,number_of_cones
             if(offset_inlet_and_cone) then 
                inlet_positions(i) = vector(&
                     sin((i+0.5)*2.0*PI/number_of_cones)*middle_vessel_radius,&
                     cos((i+0.5)*2.0*PI/number_of_cones)*middle_vessel_radius,&
                     recirculation_height)
             else
                inlet_positions(i) = vector(&
                     sin(i*2.0*PI/number_of_cones)*middle_vessel_radius,&
                     cos(i*2.0*PI/number_of_cones)*middle_vessel_radius,&
                     recirculation_height)
             end if
          end do
       end if
    else if(.not. inlet_starts_specified) then
       allocate(inlet_positions(1),inlet_velocities(1))
       inlet_positions(1) = vector(0.0,0.0,recirculation_height)
    end if
    if(.not. inlet_starts_specified) then
       inlet_velocities(:) = vector(0.0,0.0,0.0)
    end if
    z_low = min(floor_location,cone_bottom) - pebble_radius*10
    if(use_cone_packing) then 
       z_high = packing_cone_top  + pebble_radius*10
    else
       z_high = max(floor_location,cone_location) + pebble_radius*10
    end if

    door_open = .false.

  end subroutine initialize_geometry

  pure SUBROUTINE pebble_surface_velocity_forcev(inside_dist,force, &
       &     velocity,normal, omega, torque, old_slip, slip_normal, &
       &     current_wear)
    REAL(KIND=D_T), intent(in) :: inside_dist
    REAL(KIND=D_T) :: raw_force  
    real(kind=d_t),intent(inout) :: current_wear
    type(vector), intent(in) :: velocity, normal, omega, old_slip, slip_normal
    type(vector), intent(inout) :: force, torque
    type(vector) :: v_normal, v_transverse, f_normal, v_surface, f_transverse
    type(vector) :: static_friction_force, adjusted_slip
    REAL(KIND=D_T) :: fn_mag,vt_mag,ft_mag, slip_mag, static_mag
    IF (inside_dist > 0.0) THEN  
       
       v_surface = velocity + (omega .cross. ((-pebble_radius) * normal))
       raw_force = pebble_vessel_hooke * inside_dist  
       v_normal = v_surface .proj. normal
       v_transverse = v_surface - v_normal
       !PRINT *,"v",velocity,"w",omega,"n",normal
       !PRINT *,"#v_s",v_surface,"v_n",v_normal,"v_t",v_transverse
       !     First figure out the normal force
       f_normal = raw_force * normal - normal_dash_pot_constant * v_normal
       !f_normal = vdamp(f_normal,max_normal_force,max_damp_factor)
       !     Then figure out the transverse force
       fn_mag = abs(f_normal)
       vt_mag = abs(v_transverse)

       ft_mag = min(fn_mag*kinetic_friction, &
            &        vt_mag*transverse_dash_pot_constant)

       if(use_static_friction .and. &
            vmag_sqr(v_transverse) < static_velocity_sqr_max .and. &
            (slip_normal .dot. normal) > 0.5) then 
          ! The dot product checks that the normal hasn't changed too much.
          slip_mag = abs(old_slip)
          if(slip_mag*surface_static_friction_hooke > 1.0e-20_d_t) then
             adjusted_slip = perp_projection(old_slip,normal)
             if(linear_static_velocity_clamp) then 
                static_mag = min(slip_mag * surface_static_friction_hooke * &
                     (1.0 - vt_mag/static_velocity_max),&
                     fn_mag * surface_static_friction)
             else 
                static_mag = min(slip_mag * surface_static_friction_hooke,&
                     fn_mag * surface_static_friction)
             end if
             static_friction_force = (-static_mag) * normalize(adjusted_slip)
          else
             static_mag = 0.0
             static_friction_force = vector(0.0,0.0,0.0)
          end if
       else 
          static_mag = 0.0
          static_friction_force = vector(0.0,0.0,0.0)
       end if

       if(calculate_pebble_wear_vol) then
          current_wear = current_wear + abs(v_transverse) * (abs(f_normal)**(get_wear_force_exponent()))
       end if
       
            
       !     PRINT *,"MAG",fn_mag,vt_mag,ft_mag
       IF (ABS(ft_mag) + ABS(static_mag) < 1.0e-10) THEN  
          force = force + f_normal
       ELSE  
          !PRINT *,"#f,t",force,torque
          if(abs(ft_mag) < 1.0e-10) then
             !ft_mag is very small, so ignore it.
             f_transverse = static_friction_force
          else
             f_transverse = (-ft_mag/vt_mag) * v_transverse + &
                  static_friction_force
          end if
          force = force + f_normal + f_transverse
          torque = torque + (((-pebble_radius) * normal) .cross. f_transverse)
          !PRINT *,"#f_t",f_transverse,"f_n",f_normal,"f",force,"t",torque
       END IF
    END IF
  END SUBROUTINE pebble_surface_velocity_forcev

  SUBROUTINE pebble_surface_force_sd2(position,inside_dist,force, &
       &     velocity,normal, omega, torque,has_slip, old_slip, &
       &     slip_delta, do_force_tally)
    REAL(KIND=D_T), intent(in) :: inside_dist
    REAL(KIND=D_T) :: raw_force
    type(vector), intent(in) :: velocity, normal, omega, old_slip, position
    type(vector), intent(inout) :: force, torque, slip_delta
    logical,intent(out) :: has_slip
    logical,intent(in) :: do_force_tally
    type(vector) :: v_normal, v_transverse, f_normal, v_surface, f_transverse
    type(vector) :: static_friction_force, adjusted_slip, static_normal
    REAL(KIND=D_T) :: fn_mag,vt_mag,ft_mag, slip_mag, static_mag
    REAL(KIND=D_T) :: ft_sum_mag2, max_transverse_force, adjusted_slip_mag
    IF (inside_dist > 0.0) THEN  
       
       v_surface = velocity + (omega .cross. ((-pebble_radius) * normal))
       raw_force = pebble_vessel_hooke * inside_dist  
       v_normal = v_surface .proj. normal
       v_transverse = v_surface - v_normal
       f_transverse = vector(0.0,0.0,0.0)

       !     First figure out the normal force
       f_normal = raw_force * normal - normal_dash_pot_constant * v_normal
       !     Then figure out the transverse force
       fn_mag = abs(f_normal)
       vt_mag = abs(v_transverse)

       ft_mag = vt_mag*transverse_dash_pot_constant
       max_transverse_force = fn_mag*kinetic_friction

       if(use_static_friction .and. &
            vmag_sqr(v_transverse) < static_velocity_sqr_max) then 
          slip_mag = abs(old_slip)
          if(slip_mag*surface_static_friction_hooke > 1.0e-20_d_t) then
             adjusted_slip = perp_projection(old_slip,normal)
             static_mag = slip_mag * surface_static_friction_hooke
             max_transverse_force = fn_mag * surface_static_friction
             !print *,"#wall_forces lsm",slip_mag * surface_static_friction_hooke * &
             !     (1.0 - vt_mag/static_velocity_max),"sm",&
             !     slip_mag * surface_static_friction_hooke,"mufn",fn_mag * surface_static_friction
             
             adjusted_slip_mag = abs(adjusted_slip)
             if(adjusted_slip_mag > 1.0e-20_d_t) then
                static_friction_force = (-static_mag) * normalize(adjusted_slip)
             else
                static_friction_force = vector(0.0,0.0,0.0)
             end if
          else
             static_mag = 0.0
             static_friction_force = vector(0.0,0.0,0.0)
          end if
       else 
          static_mag = 0.0
          static_friction_force = vector(0.0,0.0,0.0)
       end if
       
       if(vmag_sqr(v_transverse) < static_velocity_sqr_max) then
          static_normal = pebble_radius * normal 
          slip_delta = slip_delta + v_transverse
          if(vmag_sqr(old_slip) > 1.0e-20_d_t .and. &
               abs(normal .cross. old_slip) > 1.0e-20_d_t) then 
             slip_delta = slip_delta + &
                  ((old_slip .dot. normal) / (abs(normal .cross. old_slip) * abs(normal) * abs(old_slip))) * &
               ((normal .cross. old_slip) .cross. old_slip)
          end if
          !(1.0/vmag_sqr(static_normal)) * (velocity .cross. static_normal .cross. old_slip)
          has_slip = .true.
          if(zero_long_slips .and. &
               static_mag > fn_mag * surface_static_friction * long_slip_mult) then
             ! Slip direct force greater than normal force * static
             !print *,"#ZLS geom",static_mag,fn_mag,surface_static_friction,long_slip_mult
             has_slip = .false.
          end if
          if(decrease_long_slips .and. static_mag > &
               fn_mag * surface_static_friction * decrease_slip_mult) then
             !basically this add a delta that tries to decrease the 
             ! slip by the amount that it exceeds the decrease_slip_mult
             slip_delta = slip_delta - decrease_slip_slope*(static_mag - &
                  fn_mag * surface_static_friction * decrease_slip_mult)*&
                  old_slip
          end if

       else
          has_slip = .false.
       end if
       !print *,"#wall_forces hs",has_slip,"lsvc",linear_static_velocity_clamp

       if(calculate_pebble_wear_vol) then
          call tally_pebble_surface_wear(position,v_transverse,f_normal)
       end if
       
            
       IF (ABS(ft_mag) + ABS(static_mag) < 1.0e-10) THEN  
          force = force + f_normal
       ELSE  
          if(abs(ft_mag) < 1.0e-10) then
             !ft_mag is very small, so ignore it.
             f_transverse = static_friction_force
          else
             f_transverse = (-ft_mag/vt_mag) * v_transverse + &
                  static_friction_force
          end if
          ft_sum_mag2 =  vmag_sqr(f_transverse)
          !print *,"#wall_forces","ft_mag2",ft_sum_mag2,"vt",v_transverse,&
          !     "sf",static_friction_force,"kf",(-ft_mag/vt_mag) * v_transverse,&
          !     "ft",f_transverse,"mt",max_transverse_force,"as",adjusted_slip,&
          !     "sm",static_mag,"os",old_slip,"n",normal
          if(ft_sum_mag2 > max_transverse_force**2) then
             f_transverse = (max_transverse_force/sqrt(ft_sum_mag2))*f_transverse
          end if
          force = force + f_normal + f_transverse
          torque = torque + (((-pebble_radius) * normal) .cross. f_transverse)
       END IF
       if(do_force_tally) then
          call tally_pebble_wall_force(f_normal,f_transverse,position)
       end if
    END IF
  END SUBROUTINE pebble_surface_force_sd2

  pure integer function closest_plane_location(x,y,x_offsets,y_offsets) 
    real(kind=d_t), intent(in), dimension(:) :: x_offsets, y_offsets
    real(kind=d_t), intent(in) :: x,y
    real(kind=d_t) :: min_dist, temp_dist
    integer :: min_index,i
    min_index = 1
    min_dist = (x-x_offsets(1))**2 + (y-y_offsets(1))**2
    do i=2,size(x_offsets)
       temp_dist = (x-x_offsets(i))**2 + (y-y_offsets(i))**2
       if(temp_dist < min_dist) then
          min_dist = temp_dist
          min_index = i
       end if
    end do
    closest_plane_location = min_index
  end function closest_plane_location

  pure integer function closest_plane_location_v(x,y,offsets) 
    type(vector), intent(in), dimension(:) :: offsets
    real(kind=d_t), intent(in) :: x,y
    real(kind=d_t) :: min_dist, temp_dist
    integer :: min_index,i
    min_index = 1
    min_dist = (x-offsets(1)%x)**2 + (y-offsets(1)%y)**2
    do i=2,size(offsets)
       temp_dist = (x-offsets(i)%x)**2 + (y-offsets(i)%y)**2
       if(temp_dist < min_dist) then
          min_dist = temp_dist
          min_index = i
       end if
    end do
    closest_plane_location_v = min_index
  end function closest_plane_location_v

  pure subroutine ceiling_cylinder_cones_force(z_min, z_max, radius, m,b,  &
       x_offsets, y_offsets,&
       pos, geometry_contact)
    real(kind=d_t), intent(in) :: z_min,z_max,m,b, radius
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    real(kind=d_t), intent(in), dimension(:) :: x_offsets, y_offsets
    real(kind=d_t) :: x,y,z
    integer :: min_index
    x = pos%x
    y = pos%y
    z = pos%z
    if(z < z_min .or. z > z_max + pebble_radius) then 
       return
    end if
    min_index = closest_plane_location(x,y,x_offsets,y_offsets)
    call ceiling_cyl_cone_force_offset(z_min, z_max, radius, m,b, &
         x_offsets(min_index),y_offsets(min_index), &
         pos, geometry_contact)
    
  end subroutine ceiling_cylinder_cones_force


  pure subroutine ceiling_cylinder_cones_force_v(z_min, z_max, radius, m,b,  &
       offsets, pos, geometry_contact)
    real(kind=d_t), intent(in) :: z_min,z_max,m,b, radius
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    type(vector), intent(in), dimension(:) :: offsets
    real(kind=d_t) :: x,y,z
    integer :: min_index
    x = pos%x
    y = pos%y
    z = pos%z
    if(z < z_min .or. z > z_max + pebble_radius) then 
       return
    end if
    min_index = closest_plane_location_v(x,y,offsets)
    call ceiling_cyl_cone_force_offset(z_min, z_max, radius, m,b, &
         offsets(min_index)%x,offsets(min_index)%y, &
         pos, geometry_contact)
    
  end subroutine ceiling_cylinder_cones_force_v

  subroutine ceiling_cones_force(z_min, z_max, m,b, x_offsets, y_offsets,&
       pos, geometry_contact)
    real(kind=d_t), intent(in) :: z_min,z_max,m,b
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    real(kind=d_t), intent(in), dimension(:) :: x_offsets, y_offsets
    real(kind=d_t) :: x,y,z
    integer :: min_index
    x = pos%x
    y = pos%y
    z = pos%z
    if(z < z_min .or. z > z_max) then 
       return
    end if
    min_index = closest_plane_location(x,y,x_offsets,y_offsets)
    call ceiling_cone_force_offset(z_min, z_max, m,b, x_offsets(min_index),&
         y_offsets(min_index), pos, geometry_contact)
    
  end subroutine ceiling_cones_force
    
    

  subroutine cylinders_force(z_min, z_max, radius, x_offsets, y_offsets, &
       pos, geometry_contact)
    real(kind=d_t), intent(in) :: z_min,z_max,radius
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    real(kind=d_t), intent(in), dimension(:) :: x_offsets, y_offsets
    real(kind=d_t) :: x,y,z
    integer :: min_index
    x = pos%x
    y = pos%y
    z = pos%z
    if(z < z_min .or. z > z_max) then 
       return
    end if
    min_index = closest_plane_location(x,y,x_offsets,y_offsets)
    call cylinder_force_offset(z_min, z_max, radius, x_offsets(min_index),&
         y_offsets(min_index), pos, geometry_contact)
    
  end subroutine cylinders_force

  subroutine cones_force(z_min, z_max, m, b, x_offsets, y_offsets, &
       pos, geometry_contact)
    real(kind=d_t), intent(in) :: z_min,z_max,m,b
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    real(kind=d_t), intent(in), dimension(:) :: x_offsets, y_offsets
    real(kind=d_t) :: x,y,z
    integer :: min_index
    x = pos%x
    y = pos%y
    z = pos%z
    if(z < z_min .or. z > z_max) then 
       return
    end if
    min_index = closest_plane_location(x,y,x_offsets,y_offsets)
    call cone_force_offset(z_min, z_max, m,b, x_offsets(min_index),&
         y_offsets(min_index), pos, geometry_contact)
    
  end subroutine cones_force

  pure subroutine ceiling_cyl_cone_force_offset(z_min, z_max, radius, m,b,&
       x_offset, y_offset, &
       pos,geometry_contact)
    !The force is only applied if z is between z_min and z_max
    !m is the slope of the cone, and b is the offset
    ! radius = m * z + b
    real(kind=d_t), intent(in) :: z_min,z_max,m,b, radius, x_offset, y_offset
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    type(vector) :: o_pos
    o_pos = pos - vector(x_offset,y_offset,0.0)
    call ceiling_cylinder_cone_force(z_min, z_max, radius, m,b, o_pos, &
         geometry_contact)
  end subroutine ceiling_cyl_cone_force_offset

  subroutine ceiling_cone_force_offset(z_min, z_max, m,b, x_offset, y_offset, &
       pos, geometry_contact)
    real(kind=d_t), intent(in) :: z_min,z_max,m,b, x_offset, y_offset
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    type(vector) :: o_pos
    o_pos = pos - vector(x_offset,y_offset,0.0)
    call ceiling_cone_force(z_min, z_max, m,b, o_pos, geometry_contact)
    
  end subroutine ceiling_cone_force_offset


  subroutine cylinder_force_offset(z_min, z_max, radius, x_offset, y_offset, &
       pos, geometry_contact)
    real(kind=d_t), intent(in) :: z_min,z_max,radius, x_offset, y_offset
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    type(vector) :: o_pos
    o_pos = pos - vector(x_offset,y_offset,0.0)
    call cylinder_force(z_min, z_max, radius, o_pos, geometry_contact)
    
  end subroutine cylinder_force_offset
    

  subroutine cone_force_offset(z_min, z_max, m,b, x_offset, y_offset, &
       pos, geometry_contact)
    real(kind=d_t), intent(in) :: z_min,z_max,m,b, x_offset, y_offset
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    type(vector) :: o_pos
    o_pos = pos - vector(x_offset,y_offset,0.0)
    call cone_force(z_min, z_max, m,b, o_pos, geometry_contact)
    
  end subroutine cone_force_offset



  pure subroutine ceiling_cylinder_cone_force(z_min, z_max, radius, m,b, &
       pos, geometry_contact)
    !The force is only applied if z is between z_min and z_max
    !m is the slope of the cone, and b is the offset
    ! radius = m * z + b
    real(kind=d_t), intent(in) :: z_min,z_max,m,b, radius
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    real(kind=d_t) :: ceiling_dist_inside
    type(vector) :: cone_nearest, diff
    !x,y,z are the position of the pebble
    !x_c,y_c,z_c are the closest point on the cone to the pebble
    ! or most inside point if the pebble is partially inside the cone
    real(kind=d_t) :: z,x_c,y_c,z_c
    ! the radius of the pebble and the cone
    real(kind=d_t) :: r,r_c, diff_length, dist_inside
    real(kind=d_t) :: hole_sqr, cyl_dist_inside, pebble_cent, pebble_cent_sqr

    z = pos%z
    if(z < z_min .or. z > z_max + pebble_radius) then 
       return
    end if
    r = sqrt(pos%x**2+pos%y**2)
    ceiling_dist_inside = z - z_min

    z_c = (m*(r-b)+z)/(m**2+1)
    r_c = m*z_c+b
    x_c = (r_c/r) * pos%x
    y_c = (r_c/r) * pos%y
    cone_nearest = vector(x_c,y_c,z_c)
    diff = pos - cone_nearest    
    diff_length = abs(diff)

    hole_sqr = (radius - pebble_radius)**2
    pebble_cent_sqr = pos%x**2 + pos%y**2       

    if(r_c < r .or. diff_length < pebble_radius) then
       if(r_c < r) then
          !So the radius of the point on cone < radius of pebble loc
          dist_inside = diff_length + pebble_radius
          !now make sure that diff points in the correct direction
          diff = (-1.0)*diff
       else 
          dist_inside = pebble_radius - diff_length
       end if
       !Normalize diff
       diff = (1.0/diff_length)*diff
       !PRINT *,"#CONE di",dist_inside," diff ",diff," diff_length ",diff_length

       pebble_cent = sqrt(pebble_cent_sqr)
       cyl_dist_inside = pebble_cent - SQRT(hole_sqr)
       
       if(ceiling_dist_inside < 0.0 .and. dist_inside < 0.0 .and. &
            cyl_dist_inside < 0.0) then
          !The pebble is inside none of the surfaces.
          return
       end if
       
       if(ceiling_dist_inside < dist_inside&
            .and. ceiling_dist_inside < cyl_dist_inside) then
          ! Ceiling dist is closest.
          call add_contact(geometry_contact,&
               ceiling_dist_inside,vector(0.0,0.0,-1.0))
       else if(dist_inside < cyl_dist_inside) then
          ! cone distance is closest          
          call add_contact(geometry_contact,dist_inside,diff);
       else
          ! Cylinder distance is smallest.
          call add_contact(geometry_contact,cyl_dist_inside,&
               vector(-pos%x/pebble_cent,-pos%y/pebble_cent,0.0D0))
       end if
    end if

    

  end subroutine ceiling_cylinder_cone_force

  subroutine ceiling_cone_force(z_min, z_max, m,b, pos, geometry_contact)
    !The force is only applied if z is between z_min and z_max
    !m is the slope of the cone, and b is the offset
    ! radius = m * z + b
    real(kind=d_t), intent(in) :: z_min,z_max,m,b
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector), intent(in) :: pos
    real(kind=d_t) :: ceiling_dist_inside
    type(vector) :: cone_nearest, diff
    !x,y,z are the position of the pebble
    !x_c,y_c,z_c are the closest point on the cone to the pebble
    ! or most inside point if the pebble is partially inside the cone
    real(kind=d_t) :: z,x_c,y_c,z_c
    ! the radius of the pebble and the cone
    real(kind=d_t) :: r,r_c, diff_length, dist_inside
    
    z = pos%z
    if(z < z_min .or. z > z_max + pebble_radius) then 
       return
    end if
    r = sqrt(pos%x**2+pos%y**2)
    ceiling_dist_inside = z - z_min

    z_c = (m*(r-b)+z)/(m**2+1)
    r_c = m*z_c+b
    x_c = (r_c/r) * pos%x
    y_c = (r_c/r) * pos%y
    cone_nearest = vector(x_c,y_c,z_c)
    diff = pos - cone_nearest    
    diff_length = abs(diff)
    if(r_c < r .or. diff_length < pebble_radius) then
       if(r_c < r) then
          !So the radius of the point on cone < radius of pebble loc
          dist_inside = diff_length + pebble_radius
          !now make sure that diff points in the correct direction
          diff = (-1.0)*diff
       else 
          dist_inside = pebble_radius - diff_length
       end if
       !Normalize diff
       diff = (1.0/diff_length)*diff
       !PRINT *,"#CONE di",dist_inside," diff ",diff," diff_length ",diff_length
       if(ceiling_dist_inside < dist_inside) then
          ! Choose the closest distance inside.
          call add_contact(geometry_contact,&
               ceiling_dist_inside,vector(0.0,0.0,-1.0))
       else 
          call add_contact(geometry_contact,dist_inside,diff)
       end if
    end if

  end subroutine ceiling_cone_force

  subroutine ceiling_force(z_min,z_max, radius,pos, geometry_contact)
    real(kind=d_t), intent(in) :: z_min,z_max,radius
    type(vector), intent(in) :: pos
    type(geometry_contact_type),intent(inout) :: geometry_contact

    real(kind=d_t) :: z,r, dist_inside
    z = pos%z
    if(z < z_min .or. z > z_max) then 
       return
    end if
    r = sqrt(pos%x**2+pos%y**2)
    if(r < radius) then 
       return
    end if
    dist_inside = z_min - z
    if(dist_inside < 0.0) then
       call add_contact(geometry_contact,-dist_inside,vector(0.0,0.0,-1.0))
    end if
  end subroutine ceiling_force

  subroutine cylinder_force(z_min, z_max, radius, pos, geometry_contact)
    real(kind=d_t), intent(in) :: z_min,z_max,radius
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    real(kind=d_t) :: hole_sqr, dist_inside, pebble_cent, pebble_cent_sqr

    if(pos%z < z_min  .or. pos%z > z_max) then 
       return
    end if

    hole_sqr = (radius - pebble_radius)**2
    pebble_cent_sqr = pos%x**2 + pos%y**2
       
    if(pebble_cent_sqr > hole_sqr - 1.0e-10) then

       pebble_cent = sqrt(pebble_cent_sqr)
       dist_inside = pebble_cent - SQRT(hole_sqr)
       call add_contact(geometry_contact,&
            dist_inside, vector(-pos%x/pebble_cent,-pos%y/pebble_cent,0.0D0))
    end if

  end subroutine cylinder_force
  
    

  subroutine cone_force(z_min, z_max, m,b, pos, geometry_contact)
    !The force is only applied if z is between z_min and z_max
    !m is the slope of the cone, and b is the offset
    ! radius = m * z + b
    real(kind=d_t), intent(in) :: z_min,z_max,m,b
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    type(vector) :: cone_nearest, diff
    !x,y,z are the position of the pebble
    !x_c,y_c,z_c are the closest point on the cone to the pebble
    ! or most inside point if the pebble is partially inside the cone
    real(kind=d_t) :: z,x_c,y_c,z_c
    ! the radius of the pebble and the cone
    real(kind=d_t) :: r,r_c, diff_length, dist_inside
    z = pos%z
    if(z < z_min .or. z > z_max + pebble_radius) then 
       return
    end if
    r = sqrt(pos%x**2+pos%y**2)
    z_c = (m*(r-b)+z)/(m**2+1)
    r_c = m*z_c+b
    x_c = (r_c/r) * pos%x
    y_c = (r_c/r) * pos%y
    cone_nearest = vector(x_c,y_c,z_c)
    diff = pos - cone_nearest    
    diff_length = abs(diff)
    if(r_c < r .or. diff_length < pebble_radius) then
       if(r_c < r) then
          !So the radius of the point on cone < radius of pebble loc
          dist_inside = diff_length + pebble_radius
          !now make sure that diff points in the correct direction
          diff = (-1.0)*diff
       else 
          dist_inside = pebble_radius - diff_length
       end if
       !Normalize diff
       diff = (1.0/diff_length)*diff
       !PRINT *,"#CONE di",dist_inside," diff ",diff," diff_length ",diff_length
       call add_contact(geometry_contact,dist_inside,diff)
    end if
  end subroutine cone_force

  pure subroutine cylinder_type_force(cylinder, pos, geometry_contact)
    type(cylinder_type),intent(in) :: cylinder
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    !x_o and y_o are Positions relative to the center of the cylinder
    ! r_o is the radius of the position relative to the center of the cylinder
    real(kind=d_t) :: x_o,y_o,r_o,r_o2
    real(kind=d_t) :: low_overlap,high_overlap,radius_overlap, overlap
    type(vector) :: normal
    if(.not. between(cylinder%z_min-pebble_radius,pos%z,&
         cylinder%z_max+pebble_radius)) then
       call add_no_contact(geometry_contact)
       return
    end if
    x_o = pos%x - cylinder%x_center
    y_o = pos%y - cylinder%y_center
    r_o2 = x_o**2+y_o**2
    if(r_o2 > (cylinder%radius + pebble_radius)**2) then
       call add_no_contact(geometry_contact)
       return
    end if
    r_o = sqrt(r_o2)
    low_overlap = (pos%z + pebble_radius) - cylinder%z_min
    high_overlap = cylinder%z_max - (pos%z - pebble_radius)
    radius_overlap = (cylinder%radius + pebble_radius) - r_o
    !overlap is positive if there should be a force.
    ! Choose the smallest overlap.
    overlap = min(low_overlap,high_overlap,radius_overlap)
    if(radius_overlap == overlap) then
       normal = vector(x_o/r_o,y_o/r_o,0.0)
       call add_contact(geometry_contact,overlap,normal)
    else if(low_overlap == overlap) then
       normal = vector(0.0,0.0,-1.0)
       call add_contact(geometry_contact,overlap,normal)
    else if(high_overlap == overlap) then
       normal = vector(0.0,0.0,1.0)
       call add_contact(geometry_contact,overlap,normal)
    else
       !Throw an error here.
       r_o = union_list(-1)%list(-1)%cylinder%z_min
       !print *,"ERROR, should have found a cylinder side"
    end if
  end subroutine cylinder_type_force

  pure logical function between(low, v, high) 
    real(kind=d_t),intent(in) :: low,v,high
    between = low < v .and. v < high
  end function between

  pure subroutine block_type_force(block_data, pos, geometry_contact)
    type(block_type),intent(in) :: block_data
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    real(kind=d_t),dimension(3) :: pos_array, temp_normal,normal
    real(kind=d_t) :: min_overlap,low_overlap,high_overlap,overlap
    integer :: i
    logical :: touching
    pos_array = (/ pos%x,pos%y,pos%z /)
    touching = .true.
    !Figure out if the pebble is touching this block
    do i = 1,3
       if(.not. between(block_data%min(i) - pebble_radius,pos_array(i),&
            block_data%max(i)+pebble_radius)) then
          touching = .false.
       end if       
    end do
    if(.not. touching) then
       call add_no_contact(geometry_contact)
       return
    end if
    min_overlap = 0.0
    normal = (/ 0.0,0.0,0.0 /)
    !Find how much overlap is occuring, and keep the largest overlap.
    do i = 1,3
       if(between(block_data%min(i) - pebble_radius,pos_array(i),&
            block_data%max(i)+pebble_radius)) then
          low_overlap = pos_array(i)+pebble_radius - block_data%min(i)
          high_overlap = block_data%max(i) - (pos_array(i)-pebble_radius)
          if(low_overlap < high_overlap) then
             overlap = low_overlap
             temp_normal(:) = 0
             temp_normal(i) = -1.0
          else
             overlap = high_overlap
             temp_normal(:) = 0
             temp_normal(i) = 1.0
          end if
          if(min_overlap == 0.0 .or. overlap < min_overlap) then
             min_overlap = overlap
             normal(:) = temp_normal(:)
          end if
       end if
    end do
    call add_contact(geometry_contact,min_overlap,vector(normal(1),normal(2),normal(3)))
  end subroutine block_type_force

  pure subroutine plane_type_force_m(pos,plane,has_overlap,overlap,normal)
    type(vector),intent(in) :: pos
    type(plane_type),intent(in) :: plane
    logical,intent(out) :: has_overlap
    real(kind=d_t),intent(out) :: overlap
    type(vector),intent(out) :: normal
    real(kind=d_t) :: distance

    distance = plane%n%x*pos%x+plane%n%y*pos%y+plane%n%z*pos%z+plane%d
    if(distance < pebble_radius) then
       has_overlap = .true.
       overlap = pebble_radius - distance
       normal = plane%n
    else
       has_overlap = .false.
    end if
  end subroutine plane_type_force_m

  pure subroutine plane_type_force(plane_data, pos, geometry_contact)
    type(plane_type),intent(in) :: plane_data
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(vector),intent(in) :: pos
    logical :: has_overlap
    real(kind=d_t) :: overlap
    type(vector) :: normal
    
    call plane_type_force_m(pos,plane_data,has_overlap,overlap,normal)
    if(has_overlap) then
       call add_contact(geometry_contact,overlap,normal)
    else
       call add_no_contact(geometry_contact)
    end if
  end subroutine plane_type_force

  pure subroutine plane_union_type_force(pos,plane_union,&
       has_overlap,overlap,normal)
    type(vector),intent(in) :: pos
    type(plane_union_type),intent(in) :: plane_union
    logical,intent(out) :: has_overlap
    real(kind=d_t),intent(out) :: overlap
    type(vector),intent(out) :: normal
    real(kind=d_t) :: max_overlap,cur_overlap
    type(vector) :: max_normal,cur_normal
    logical :: union_has_overlap,cur_has_overlap

    integer :: i

    max_overlap = 0.0
    max_normal = vector(0.0,0.0,0.0)
    union_has_overlap = .false.
    do i=1,size(plane_union%list)
       call plane_type_force_m(pos,plane_union%list(i),cur_has_overlap,&
            cur_overlap,cur_normal)
       !print *,"#union",i,"pos",pos,"cho",cur_has_overlap,"co",cur_overlap,"cn",cur_normal,"p",plane_union%list(i)
       if(cur_has_overlap) then
          union_has_overlap = .true.
          if(cur_overlap > max_overlap) then
             max_overlap = cur_overlap
             max_normal = cur_normal
          end if
       end if
    end do
    has_overlap = union_has_overlap
    overlap = max_overlap
    normal = max_normal
  end subroutine plane_union_type_force

  pure subroutine rotate(x,y,delta_angle)
    real(kind=d_t),intent(in) :: delta_angle
    real(kind=d_t),intent(inout) :: x,y
    real(kind=d_t) :: current_angle,angle,radius
    current_angle = atan2(y,x)
    radius = sqrt(x**2+y**2)
    angle = current_angle + delta_angle
    x = radius*cos(angle)
    y = radius*sin(angle)
    
  end subroutine rotate

  pure subroutine indentation_type_force(pos,indentation,geometry_contact)
    type(vector),intent(in) :: pos
    type(indentation_type),intent(in) :: indentation
    type(geometry_contact_type),intent(inout) :: geometry_contact
    real(kind=d_t) :: outside_sqr,pebble_cent_sqr,wall_overlap,radius
    real(kind=d_t) :: indentation_overlap, angle, min_angle, angle_delta
    logical :: has_indentation_overlap, only_wall
    integer :: row,angle_index,i
    type(vector) :: ref_pos,adjusted_pos,indentation_normal

    has_indentation_overlap = .false.
    !First, see if inside radius. 
    pebble_cent_sqr = pos%x**2 + pos%y**2
    outside_sqr = (indentation%wall_radius - pebble_radius)**2
    if(pebble_cent_sqr > outside_sqr - 1.0e-10) then
       radius = sqrt(pebble_cent_sqr)
       wall_overlap = radius - indentation%wall_radius + pebble_radius
       indentation_overlap = wall_overlap
    else
       !Not contacting wall, so leave
       return
    end if
    
    !only_wall = .false.
    !Second, find indentation row
    if(between(indentation%z_min - pebble_radius,pos%z,indentation%z_max+pebble_radius)) then
       row = size(indentation%row_list)
       row_loop: do i = 1,size(indentation%row_list)-1
          if(between(indentation%row_list(i)%z_min,pos%z,indentation%row_list(i+1)%z_min)) then
             row = i
             exit row_loop
          end if
       end do row_loop

       !Third, find angle and check the plane union
       angle = atan2(pos%y,pos%x)
       
       min_angle = 8.0*pi
       angle_index = 0
       angle_loop: do i=1,size(indentation%row_list(row)%angle_list)
          angle_delta = abs(angle-indentation%row_list(row)%angle_list(i))
          !If difference is greater than pi (180 degrees), should actually
          ! be closer,  i.e. -179.0 and 179.0 should be 2 degrees apart
          if(angle_delta > pi) then 
             angle_delta = 2*pi - angle_delta
          end if
          !print *,"#angle_loop",i,angle_delta,min_angle,indentation%row_list(row)%angle_list(i)
          if(angle_delta < min_angle) then
             min_angle = angle_delta
             angle_index = i
          end if
       end do angle_loop
       !print *,"#angle",angle,angle_index,min_angle
       !determine angle delta and radius delta, and use those
       ! to query plane_union_type_force

       angle_delta = indentation%row_list(row)%angle_list(angle_index)
       adjusted_pos = pos
       call rotate(adjusted_pos%x,adjusted_pos%y,-angle_delta)
       ref_pos = vector(indentation%wall_radius,0.0,indentation%row_list(row)%z_min)
       adjusted_pos = adjusted_pos - ref_pos
       
       call plane_union_type_force(adjusted_pos,indentation%pattern,&
            has_indentation_overlap,indentation_overlap,indentation_normal)
       call rotate(indentation_normal%x,indentation_normal%y,angle_delta)
       !print *,"#indent pos",pos,"ap",adjusted_pos,"rp",ref_pos,"hio",&
       !     has_indentation_overlap,"io",indentation_overlap,"in",indentation_normal
       only_wall = .false.
    else
       only_wall = .true.
       
    end if

    !If there is no indentation overlap, then the pebble is in a indentation
    if(.not. has_indentation_overlap .and. .not. only_wall) then
       return
    end if

    if(only_wall .or. wall_overlap <= indentation_overlap) then
       call add_contact(geometry_contact,wall_overlap,vector(-pos%x/radius,-pos%y/radius,0.0D0))
    else if(indentation_overlap < wall_overlap) then
       call add_contact(geometry_contact,indentation_overlap,indentation_normal)
    end if

  end subroutine indentation_type_force

  pure subroutine union_force(pos,geometry_contact,union)
    type(vector),intent(in) :: pos
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(group_type),intent(in) :: union
    integer :: i
    !real(kind=d_t) :: max_overlap
    !type(vector) :: normal
    !orig_contacts = geometry_contact%number_of_contacts
    do i=1,size(union%list)
       if(associated(union%list(i)%block)) then
          call block_type_force(union%list(i)%block,pos,geometry_contact)
       end if
       if(associated(union%list(i)%cylinder)) then
          call cylinder_type_force(union%list(i)%cylinder,pos,geometry_contact)
       end if
       if(associated(union%list(i)%plane)) then
          call plane_type_force(union%list(i)%plane,pos,geometry_contact)
       end if
    end do
!!$    if(geometry_contact%number_of_contacts > orig_contacts + 1) then
!!$       !Need to only keep the largest overlap, otherwise could be 
!!$       ! doubling up on the force
!!$       max_overlap = 0.0
!!$       normal = vector(0.0,0.0,0.0)
!!$       do i = orig_contacts+1,geometry_contact%number_of_contacts
!!$          if(geometry_contact%inside_dist(i) > max_overlap) then
!!$             max_overlap = geometry_contact%inside_dist(i)
!!$             normal = geometry_contact%normal(i)
!!$          end if
!!$       end do
!!$       i = orig_contacts+1
!!$       geometry_contact%number_of_contacts = i
!!$       geometry_contact%inside_dist(i) = max_overlap
!!$       geometry_contact%normal(i) = normal
!!$    end if
  end subroutine union_force

  pure subroutine union_list_force(pos,geometry_contact)
    type(vector),intent(in) :: pos
    type(geometry_contact_type),intent(inout) :: geometry_contact
    integer :: i

    do i=1,size(union_list)
       call union_force(pos,geometry_contact,union_list(i))
    end do
  end subroutine union_list_force

  pure subroutine intersection_force(pos,geometry_contact,union)
    type(vector),intent(in) :: pos
    type(geometry_contact_type),intent(inout) :: geometry_contact
    type(group_type),intent(in) :: union
    integer :: i
    call start_intersection_mode(geometry_contact)
    do i=1,size(union%list)
       if(associated(union%list(i)%block)) then
          call block_type_force(union%list(i)%block,pos,geometry_contact)
       end if
       if(associated(union%list(i)%cylinder)) then
          call cylinder_type_force(union%list(i)%cylinder,pos,geometry_contact)
       end if
       if(associated(union%list(i)%plane)) then
          call plane_type_force(union%list(i)%plane,pos,geometry_contact)
       end if
    end do
    call end_intersection_mode(geometry_contact)
  end subroutine intersection_force

  pure subroutine intersection_list_force(pos,geometry_contact)
    type(vector),intent(in) :: pos
    type(geometry_contact_type),intent(inout) :: geometry_contact
    integer :: i

    do i=1,size(intersection_list)
       call intersection_force(pos,geometry_contact,intersection_list(i))
    end do
  end subroutine intersection_list_force



  pure subroutine default_geometry_calculations(pos,geometry_contact)
    type(vector),intent(in) :: pos
    type(geometry_contact_type),intent(inout) :: geometry_contact
    REAL(KIND=D_T) :: dist_from_floor, pebble_cent, pebble_cent_sqr  
    REAL(KIND=D_T) :: dist_inside, inside_sqr, outside_sqr 


    !print *,"z",force,torque
    dist_from_floor = pos%z - pebble_radius - floor_location  
    IF(dist_from_floor < 0.0) THEN  
       call add_contact(geometry_contact,-dist_from_floor,vector(0.0,0.0,1.0))
    END IF

    !print *,"q",force,torque

    dist_from_floor = pos%z - pebble_radius - door_location
    !print *,"r",dist_from_floor
    if(.not. door_open .and. dist_from_floor < 0.0) then
       call add_contact(geometry_contact,-dist_from_floor,vector(0.0,0.0,1.0))
    end if
    !print *,"a",force,torque
    !     How far is the pebble from the vessel's center
    pebble_cent_sqr = pos%x**2 + pos%y**2

    if(rectangular_vat) then 
       if(pos%x + pebble_radius > x_wall_distance) then 
          call add_contact(geometry_contact,pos%x+pebble_radius&
               -x_wall_distance,vector(-1.0,0.0,0.0))
       end if
       if(-pos%x + pebble_radius > x_wall_distance) then
          call add_contact(geometry_contact,-pos%x+pebble_radius&
               -x_wall_distance,vector(+1.0,0.0,0.0))
       end if
       if(pos%y + pebble_radius > y_wall_distance) then 
          call add_contact(geometry_contact,pos%y+pebble_radius&
               -y_wall_distance,vector(0.0,-1.0,0.0))
       end if
       if(-pos%y + pebble_radius > y_wall_distance) then
          call add_contact(geometry_contact,-pos%y+pebble_radius&
               -y_wall_distance,vector(0.0,+1.0,0.0))
       end if

    else
       inside_sqr = (inside_vessel_radius + pebble_radius)**2
       outside_sqr = (outside_vessel_radius - pebble_radius)**2
       
       IF(pebble_cent_sqr < inside_sqr + 1.0e-10) THEN
          !        Pebble is inside the vessel core area
          IF (inside_vessel_radius > 1.0e-10) THEN
             !           vessel inside radius not 0.0
             pebble_cent = SQRT(pebble_cent_sqr)
             dist_inside = SQRT(inside_sqr) - pebble_cent
             call add_contact(geometry_contact,dist_inside,vector(pos%x/pebble_cent,pos%y/pebble_cent,0.0D0))
          END IF
       ELSE IF(pebble_cent_sqr > outside_sqr - 1.0e-10) THEN
          if(size(indentation%pattern%list) > 0) then
             call indentation_type_force(pos,indentation,geometry_contact)
          else 

             !        Pebble is inside the vessel's outer area
             pebble_cent = SQRT(pebble_cent_sqr)
             dist_inside = pebble_cent - SQRT(outside_sqr)
             call add_contact(geometry_contact,dist_inside,vector(-pos%x/pebble_cent,-pos%y/pebble_cent,0.0D0))
          end if
       END IF

       !print *,"b",force,torque

    end if
    if(number_of_cones <= 1) then
       if(use_cone_packing) then
          
          call ceiling_cylinder_cone_force(packing_cone_bottom-&
               packing_chute_length,packing_cone_top,&
               packing_cone_radius,&
               1.0d0,packing_cone_radius-packing_cone_bottom,pos,&
               geometry_contact)
       end if
 
       call ceiling_cylinder_cone_force(z_min = door_location,&
            z_max = cone_location,radius = cone_hole_radius,&
            m = cone_slope, b = cone_hole_radius-cone_bottom*cone_slope,&
            pos = pos, geometry_contact = geometry_contact)
    else 
       if(use_cone_packing) then
          
          call ceiling_cylinder_cones_force_v(packing_cone_bottom-&
               packing_chute_length,packing_cone_top,&
               packing_cone_radius,&
               1.0d0,packing_cone_radius-packing_cone_bottom,&
               inlet_positions,pos,geometry_contact)
       end if

       call ceiling_cylinder_cones_force(z_min = door_location,&
            z_max = cone_location, radius = cone_hole_radius,&
            m = cone_slope, b = cone_hole_radius-cone_bottom*cone_slope,&
            x_offsets = cone_x_offsets, y_offsets = cone_y_offsets,&
            pos = pos, geometry_contact = geometry_contact)
    end if

    call union_list_force(pos,geometry_contact)
    call intersection_list_force(pos,geometry_contact)

  end subroutine default_geometry_calculations

  pure SUBROUTINE default_base_force(pos,velocity, omega,force, torque, &
       slip, slip_normal, current_wear)  
    type(vector), intent(in) :: pos, velocity, omega, slip, slip_normal
    type(vector), intent(out) :: force, torque
    real(kind=d_t), intent(inout) :: current_wear
    type(geometry_contact_type) :: geometry_contact
    integer :: i
    !    PRINT *,"#In base force"
    force = vector(0.0, 0.0, gravitational_force)
    torque = vector(0.0, 0.0, 0.0)
    geometry_contact%number_of_contacts = 0

    call default_geometry_calculations(pos,geometry_contact)

    do i = 1,geometry_contact%number_of_contacts
       call pebble_surface_velocity_forcev(geometry_contact%inside_dist(i),&
            force,velocity,geometry_contact%normal(i),omega,torque,&
            slip,slip_normal,current_wear)
    end do
    !PRINT *,"#p,v,o,f,t",pos,velocity,omega,force,torque
  END SUBROUTINE default_base_force

  subroutine default_base_force_sd(pos,velocity, omega,force, torque, &
       has_slip_delta, slip, slip_delta, do_force_tally)
    type(vector), intent(in) :: pos, velocity, omega, slip
    type(vector), intent(out) :: force, torque, slip_delta
    logical, intent(out) :: has_slip_delta
    logical, intent(in) :: do_force_tally
    type(geometry_contact_type) :: geometry_contact
    logical :: single_has_slip
    integer :: i
    !    PRINT *,"#In base force"
    force = vector(0.0, 0.0, gravitational_force)
    torque = vector(0.0, 0.0, 0.0)
    slip_delta = vector(0.0, 0.0, 0.0)
    geometry_contact%number_of_contacts = 0
    has_slip_delta = .false.

    call default_geometry_calculations(pos,geometry_contact)

    do i = 1,geometry_contact%number_of_contacts
       call pebble_surface_force_sd2(pos,geometry_contact%inside_dist(i),&
            force,velocity,geometry_contact%normal(i),omega,torque,&
            single_has_slip, slip, slip_delta,do_force_tally)
       has_slip_delta = has_slip_delta .or. single_has_slip
    end do
    !PRINT *,"#p,v,o,f,t",pos,velocity,omega,force,torque
  END SUBROUTINE default_base_force_sd

  function is_contacting_wall(pos)  
    type(vector), intent(in) :: pos
    logical :: is_contacting_wall
    type(geometry_contact_type) :: geometry_contact
    geometry_contact%number_of_contacts = 0

    call default_geometry_calculations(pos,geometry_contact)
    !print *,"#geometry_contact",geometry_contact%number_of_contacts,&
    !     geometry_contact%normal(1:geometry_contact%number_of_contacts),&
    !     geometry_contact%inside_dist(1:geometry_contact%number_of_contacts)
    
    is_contacting_wall = geometry_contact%number_of_contacts > 0
    
  END function is_contacting_wall

end module geometry

