! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.


PROGRAM marbles7
  ! $Id: pebbles7.f90 1378 2011-02-24 15:10:13Z cogljj $
  USE vector_type
  USE str_utils
  USE prob_table
  use number_utils
  use types
  use nearby_pebble_cache_module
  use geometry
  use constants
  use force_tally
  use earthquake
  use new_fortran_utils
  use wear_tally
  !$ use omp_lib
  IMPLICIT NONE

  ! the state of all the pebbles in the system.
  type pebble_state
     !pp = pebble position
     !pv = pebble velocity
     !pw = pebble angular velocity (w ~= omega)
     ! pp(number_of_pebbles)
     type(vector),pointer, dimension(:) :: pp, pv, pw
     !ppsd = pebble pebble slip
     ! ppsd(slip_zone_size,number_of_pebbles)
     type(vector),pointer, dimension(:,:) :: ppsd
     !pps_index = pebble pebble slip index
     ! which slip vector is the pebble pointing to
     ! pps_index(slip_zone_size,number_of_pebbles)
     integer,pointer,dimension(:,:) :: pps_index
     !pss = pebble surface slip
     ! pss(number_of_pebbles)
     type(vector),pointer, dimension(:) :: pss
     !psn = pebble surface normal 
     ! used to make sure that the angle change in the normal is not
     ! excessive (surface can change 90 degrees rapidly.)
     type(vector),pointer, dimension(:) :: psn
  end type pebble_state

  type pebble_delta
     !pdp = pebble d(position)  aka pebble velocity
     !pdv = pebble d(velocity) aka pebble acceleration
     !pdw = pebble d(angular velocity) aka pebble angular acceleration
     ! pdp(number_of_pebbles,multistep_order)
     type(vector),pointer, dimension(:) :: pdp, pdv, pdw
  end type pebble_delta
  
  !     The constants used throughout the program

  REAL(KIND=D_T) :: pebble_center_radius
  REAL(KIND=D_T) :: density_center
  !density in kg/m**3
  REAL(KIND=D_T) :: density_outside
  ! The fill edge is the max distance from center pebbles are filled to
  REAL(KIND=D_T) :: fill_edge 
  integer :: current_cone_drop
  !logical :: offset_inlet_and_cone
  
  !gravitational_acceleration in m/s**2
  REAL(KIND=D_T), PARAMETER :: gravitational_acceleration = -9.8
  !mass in kg
  REAL(KIND=D_T) :: mass
  !hooke forces in kg*m/(m*s**2)
  REAL(KIND=D_T) :: pebble_pebble_hooke
  !REAL(KIND=D_T), PARAMETER :: max_normal_force = pebble_pebble_hooke*0.1*pebble_radius
  !REAL(KIND=D_T), PARAMETER :: max_damp_factor = 0.1
  !REAL(KIND=D_T) :: static_speed_sqr
  REAL(KIND=D_T) :: initial_packing
  REAL(KIND=D_T), PARAMETER :: air_friction = 0.0
  !REAL(KIND=D_T) :: static_rotation_sqr
  REAL(KIND=D_T) :: moment_of_inertia

  INTEGER :: i, allocate_status, settings_unit
  INTEGER, PARAMETER :: DUMP_UNIT = 206
  INTEGER, PARAMETER :: LOAD_UNIT = 207
  INTEGER, PARAMETER :: LOAD_INFO_UNIT = 208

  ! How many adjecent pebbles can be sliding along this pebble.
  ! For perfectly hard spheres this is 12.  16 is used for paranoia reasons
  INTEGER, PARAMETER :: slip_zone_size = 16

  INTEGER(KIND=LI) :: seed  
  REAL(KIND=D_T) :: area  
  REAL(KIND=D_T) :: pebble_volume,height_needed  
!!$  INTEGER overlap_count, complete_overlap_count
!!$  REAL(KIND=D_T) overlap_total,overlap,max_overlap, overlap_sum
  INTEGER :: number_of_pebbles
  INTEGER :: runs
  REAL(KIND=D_T) :: alpha
  !How often to display the velocities
  INTEGER :: display_frequency
  !How often to display the positions
  integer :: position_frequency
  !How often to display the energy
  integer :: energy_frequency
  
  ! What is the number of the pebble we are interested in?
  integer :: pebble_of_interest
  
  !How often to display the packings and the velocity tallies
  integer :: dump_frequency
  REAL(KIND=D_T):: x_low, x_high, y_low, y_high, z_low, z_high
  REAL(KIND=D_T) :: zone_increment
  !tally_zone_inc is the size of the velocity tally boxes
  real(kind=d_t) :: tally_z_inc
  real(kind=d_t) :: tally_radial_inc
  ! The tally_sections is the number of pizza slices that are 
  !  made into different zones.
  integer, parameter :: tally_sections = 8
  INTEGER :: x_size
  INTEGER :: y_size
  INTEGER :: z_size
  INTEGER :: rt_size
  INTEGER :: zt_size
  INTEGER :: near_pebbles(near_pebbles_size),near_pebbles_count  
  !multistep_order is how many steps to keep track of
  INTEGER :: multistep_order

  !This will shut off torque if set to .false.
  ! otherwise, should be set to .true.
  logical :: torque_enable

  ! Recirculation parameters
  ! Turn recirculation on or off.
  logical :: recirculate
  !The door at the bottom will be open until a pebble goes through.
  ! then, it will be closed for door_closed_time.
  real(kind=d_t) :: door_closed_time
  ! when does the door first open
  real(kind=d_t) :: first_door_open_time 
  real(kind=d_t) :: time_till_door_open
  

  !The time before counting the velocity (so as to ignore the settling time)
  real(kind=d_t) :: first_velocity_tally_time
  integer :: velocity_tally_frequency
  !These store the data required to calculate the average velocity and 
  ! the variance.
  INTEGER,allocatable, dimension(:,:,:) :: zone_n
  type(vector),allocatable, dimension(:,:,:) :: zone_vel_sum
  REAL(kind=d_t),allocatable, dimension(:,:,:) :: zone_dot_sum
  !These store the radialized data required to calculate the average velocity
  ! and the variance.  Highly likely to be correct, but harder to interpret.
  INTEGER,allocatable, dimension(:,:,:) :: zone_rect_n
  type(vector),allocatable, dimension(:,:,:) :: zone_rect_vel_sum
  REAL(kind=d_t),allocatable, dimension(:,:,:) :: zone_rect_dot_sum
  
  !Final position dump parameters
  logical :: do_dump_positions
  character(256) :: dump_positions_filename
  ! This is used to convert to different units.
  !  if you want the data in cm, then it should be 100.0
  !  and if inches, then 39.37 and so forth
  real(kind=d_t) :: dump_positions_mult

  !Initial position loading parameters
  logical :: do_load_positions
  character(256) :: load_positions_filename
  ! This is used to convert to different units.
  !  if the incomming data is in cm, then it should be 100.0
  !  and if inches, then 39.37 and so forth
  real(kind=d_t) :: load_positions_divide

  !Initial position, velocity and such loading parameters
  logical :: do_load_pebble_info
  character(256) :: load_pebble_info_filename
  logical :: do_load_pebble_slips
  logical :: do_load_pebble_surface_slips


  !pp = pebble position
  !pv = pebble velocity
  !pw = pebble angular velocity (w ~= omega)
  ! pp(number_of_pebbles, multistep_order)
  !type(vector),allocatable, dimension(:,:) :: pp, pv, pw
  type(pebble_state), allocatable, dimension(:) :: states

  !pdp = pebble d(position)  aka pebble velocity
  !pdv = pebble d(velocity) aka pebble acceleration
  !pdw = pebble d(angular velocity) aka pebble angular acceleration
  ! pdp(number_of_pebbles,multistep_order)
  !type(vector),allocatable, dimension(:,:) :: pdp, pdv, pdw
  type(pebble_delta), allocatable, dimension(:) :: deltas

  !The time of the simulation
  real(kind=d_t) :: time = 0.0

  !Use ab algorithm
  logical :: use_ab_algorithm

  !Probability tally stuff
  logical :: run_probability_tally
  integer :: probability_tally_frequency
  real(kind=d_t) :: probability_tally_start
  real(kind=d_t) :: probability_tally_delta
  type(count_data_type) :: probability_data


  !Sort pebbles by height after loading or initializing
  logical :: sort_pebbles

  !real(kind=d_t) :: pebble_volume_wear_sum = 0.0_d_t
  !real(kind=d_t),allocatable,dimension(:) :: pebble_wear_sum_tally
  !real(kind=d_t),allocatable,dimension(:) :: surface_wear_sum_tally
  integer :: pebble_wear_vol_display_freq

  logical :: calculate_force_tally

  type(nearby_pebble_cache_type) :: nearby_pebble_cache

  character(256) :: line
  character(52) :: word, word1, word2, word3, word4, word5
  integer :: at

  !$ integer :: chunk_size

  !Defaults
  call set_geometry_defaults
  call set_earthquake_defaults
  runs = 10000
  number_of_pebbles = 1000
  kinetic_friction = 0.1
  pp_static_friction = 0.0
  surface_static_friction = 0.0
  !static_speed_sqr = 1.0e-4
  alpha = 0.0001
  normal_dash_pot_constant = 2.0
  transverse_dash_pot_constant = 2.0
  initial_packing = 0.10
  dump_frequency = 100000
  display_frequency = 1000
  position_frequency = 100000000
  energy_frequency = 100
  torque_enable = .true.
  seed = 256  
  do_dump_positions = .false.
  dump_positions_mult = 1.0
  do_load_positions = .false.
  load_positions_divide = 1.0
  do_load_pebble_info = .false.
  do_load_pebble_slips = .false.
  do_load_pebble_surface_slips = .false.
  pebble_pebble_hooke = 10000.0
  pebble_vessel_hooke = 10000.0
  pebble_radius = 0.1
  pebble_center_radius = 0.0
  density_center = 0.0
  density_outside = 2.0
  use_static_friction = .false.
  static_velocity_sqr_max = 0.1**2
  static_friction_hooke = 10000.0
  surface_static_friction_hooke = 10000.0
  zero_static_slips = .false.
  linear_static_velocity_clamp = .false.
  zero_long_slips = .false.
  shorten_long_slips = .false.
  decrease_long_slips = .false.
  long_slip_sqr = 100.0
  long_slip_mult = 1.0
  pebble_of_interest = -1
  use_ab_algorithm = .false.
  multistep_order = 4
  recirculate = .false.
  door_closed_time = 10.0
  first_door_open_time = 10.0
  first_velocity_tally_time = 1.0e6
  velocity_tally_frequency = 10000000
  run_probability_tally = .false.
  probability_tally_frequency = 1000000
  probability_tally_start = 1.0e6
  current_cone_drop = 1
  sort_pebbles = .false.
  calculate_pebble_wear_vol = .false.
  !call set_wear_force_exponent(2.0/3.0)
  pebble_wear_vol_display_freq = 0
  calculate_force_tally = .false.
  !$ chunk_size = max(number_of_pebbles/(8*omp_get_num_procs()),1)

  !call set_geometry_defaults

  settings_unit = get_settings_unit(.true.)
  input_loop: do
     read (settings_unit,'(A256)') line
     print *,"#input ",trim(line)
     at = 1
     if(len_trim(line) == 0) then
        cycle input_loop
     end if
     call get_param(line,word, at)

     if(str_equal('done ',word)) then
        exit
     else if(str_equal('rem ',word)) then
        continue
     else if(str_equal('start_geometry ',word)) then
        call load_geometry_section(settings_unit,.true.)
     else if(str_equal('vessel_radius ',word)) then
        call get_param(line,word1,at)
        !inside_vessel_radius = str_to_real(word1)
        call get_param(line,word2,at)
        !outside_vessel_radius = str_to_real(word2)        
        !rectangular_vat = .false.
        call set_vessel_radius(inside=str_to_double(word1),&
             outside=str_to_double(word2))
     else if(str_equal('rectangular_vat ',word)) then
        !rectangular_vat = .true.
        call get_param(line,word1,at)
        !x_wall_distance = str_to_real(word)
        call get_param(line,word2,at)
        !y_wall_distance = str_to_real(word)
        call set_rectangular_vat(x_distance=str_to_double(word1),&
             y_distance=str_to_double(word2))
     else if(str_equal('runs ',word)) then
        call get_param(line,word,at)
        runs = str_to_int(word)
     else if(str_equal('initial_time ',word)) then
        call get_param(line,word,at)
        time = str_to_double(word)
     else if(str_equal('number_of_pebbles ',word) .or. str_equal('number_of_spheres ',word)) then
        call get_param(line,word,at)
        number_of_pebbles = str_to_int(word)
        !$ chunk_size = max(number_of_pebbles/(8*omp_get_num_procs()),1)
     else if(str_equal('kinetic_friction ',word)) then 
        call get_param(line,word,at)
        kinetic_friction = str_to_double(word)
     else if(str_equal('static_friction ',word)) then
        print *,"ERROR: static_friction is deprecated.  use static_friction_new"
        print *,"or use pebbles6"
        !call get_param(line,word,at)
        !static_friction = str_to_double(word)
        !call get_param(line,word,at)
        !static_speed_sqr = str_to_double(word)
        stop
     else if(str_equal('alpha ',word)) then
        call get_param(line,word,at)
        alpha = str_to_double(word)
     else if(str_equal('dash_pot ',word)) then
        call get_param(line,word,at)
        normal_dash_pot_constant = str_to_double(word)
        transverse_dash_pot_constant = normal_dash_pot_constant
     else if(str_equal('dash_pot2 ',word)) then
        call get_param(line,word,at)
        normal_dash_pot_constant = str_to_double(word)
        call get_param(line,word,at)
        transverse_dash_pot_constant = str_to_double(word)
     else if(str_equal('initial_packing ',word)) then        
        call get_param(line,word,at)
        initial_packing = str_to_double(word)
     else if(str_equal('dump_frequency ',word)) then
        call get_param(line,word,at)
        dump_frequency = str_to_int(word)
     else if(str_equal('display_frequency ',word)) then
        call get_param(line,word,at)
        display_frequency = str_to_int(word)
     else if(str_equal('position_display_frequency ',word)) then
        call get_param(line,word,at)
        position_frequency = str_to_int(word)
     else if(str_equal('energy_display_frequency ',word)) then
        call get_param(line,word,at)
        energy_frequency = str_to_int(word)
     else if(str_equal('torque_enable ',word)) then 
        torque_enable = .true.
     else if(str_equal('torque_disable ',word)) then
        torque_enable = .false.
     else if(str_equal('seed ',word)) then
        call get_param(line,word,at)
        seed = str_to_int(word)
     else if(str_equal('earthquake_enable ',word)) then
        call get_param(line,word,at)
        !max_number_of_earthquake_waves = str_to_int(word)
        call enable_earthquake_waves(str_to_int(word))
     else if(str_equal('earthquake_sine_wave ',word)) then
        call parse_earthquake_sine_wave_line(line,at)
     else if(str_equal('earthquake_sine_wave_offset ',word)) then
        call parse_eq_sine_wave_offset_line(line,at)
     else if(str_equal('dump_positions ',word)) then
        call get_param(line,word,at)
        dump_positions_filename = word
        do_dump_positions = .true.
     else if(str_equal('dump_positions_mult ',word)) then
        call get_param(line,word,at)
        dump_positions_filename = word
        call get_param(line,word,at)
        dump_positions_mult = str_to_double(word)
        do_dump_positions = .true.
     else if(str_equal('load_positions ',word)) then
        call get_param(line,word,at)
        load_positions_filename = word
        do_load_positions = .true.
     else if(str_equal('load_positions_divide ',word)) then
        call get_param(line,word,at)
        load_positions_filename = word
        call get_param(line,word,at)
        load_positions_divide = str_to_double(word)
        do_load_positions = .true.
     else if(str_equal('load_pebble_info ',word) .or. str_equal('load_sphere_info ',word)) then
        call get_param(line,word,at)
        load_pebble_info_filename = word
        do_load_pebble_info = .true.
     else if(str_equal('load_pebble_slips ',word) .or. str_equal('load_sphere_slips ',word)) then
        do_load_pebble_slips = .true.
     else if(str_equal('load_pebble_surface_slips ',word) .or. str_equal('load_sphere_surface_slips ',word)) then
        do_load_pebble_surface_slips = .true.
     else if(str_equal('load_pebble_save ',word) .or. str_equal('load_sphere_save ',word)) then
        call get_param(line,word,at)
        load_pebble_info_filename = word
        do_load_pebble_info = .true.
        do_load_pebble_slips = .true.
        do_load_pebble_surface_slips = .true.
     else if(str_equal('floor_location ',word)) then
        call get_param(line,word,at)
        call set_floor_location(str_to_double(word))
     else if(str_equal('pebble_pebble_hooke ',word) .or. str_equal('sphere_sphere_hooke ',word)) then
        call get_param(line,word,at)
        pebble_pebble_hooke = str_to_double(word)
     else if(str_equal('pebble_vessel_hooke ',word) .or. str_equal('sphere_vessel_hooke ',word)) then
        call get_param(line,word,at)
        pebble_vessel_hooke = str_to_double(word)
     else if(str_equal('pebble_radius ',word) .or. str_equal('sphere_radius ',word)) then
        call get_param(line,word,at)
        pebble_center_radius = str_to_double(word)
        call get_param(line,word,at)
        pebble_radius = str_to_double(word)
     else if(str_equal('pebble_density ',word) .or. str_equal('sphere_density ',word)) then
        call get_param(line,word,at)
        density_center = str_to_double(word)
        call get_param(line,word,at)
        density_outside = str_to_double(word)
     else if(str_equal('static_friction_new ',word)) then
        use_static_friction = .true.
        call get_param(line,word,at)
        pp_static_friction = str_to_double(word)
        surface_static_friction = pp_static_friction
        call get_param(line,word,at)
        static_friction_hooke = str_to_double(word)
        surface_static_friction_hooke = static_friction_hooke
        call get_param(line,word,at)
        static_velocity_sqr_max = str_to_double(word)
     else if(str_equal('static_friction_new2 ',word)) then
        use_static_friction = .true.
        call get_param(line,word,at)
        pp_static_friction = str_to_double(word)
        surface_static_friction = pp_static_friction
        call get_param(line,word,at)
        static_friction_hooke = str_to_double(word)
        call get_param(line,word,at)
        surface_static_friction_hooke = str_to_double(word)
        call get_param(line,word,at)
        static_velocity_sqr_max = str_to_double(word)
     else if(str_equal('static_friction_new3 ',word)) then
        use_static_friction = .true.
        call get_param(line,word,at)
        pp_static_friction = str_to_double(word)
        call get_param(line,word,at)
        static_friction_hooke = str_to_double(word)
        call get_param(line,word,at)
        surface_static_friction = str_to_double(word)
        call get_param(line,word,at)
        surface_static_friction_hooke = str_to_double(word)
        call get_param(line,word,at)
        static_velocity_sqr_max = str_to_double(word)
     else if(str_equal('zero_long_slips ',word)) then
        zero_long_slips = .true.
        call get_param(line,word,at)  
        long_slip_sqr = str_to_double(word)**2
     else if(str_equal('shorten_long_slips ',word)) then
        shorten_long_slips = .true.
        call get_param(line,word,at)  
        long_slip_sqr = str_to_double(word)**2
     else if(str_equal('zero_static_slips ',word)) then
        zero_static_slips = .true.
     else if(str_equal('calculate_pebble_wear ',word) .or. str_equal('calculate_sphere_wear ',word)) then
        calculate_pebble_wear_vol = .true.
        call get_param(line,word,at)
        pebble_wear_vol_display_freq = str_to_int(word)
     else if(str_equal('normal_force_wear_exponent ',word)) then
        call get_param(line,word,at)
        call set_wear_force_exponent(str_to_double(word))
     else if(str_equal('linear_static_friction_cutoff ',word)) then
        linear_static_velocity_clamp = .true.
     else if(str_equal('pebble_of_interest ',word) .or. str_equal('sphere_of_interest ',word)) then
        call get_param(line,word,at)
        pebble_of_interest = str_to_int(word)
     else if(str_equal('use_simple_integrator ',word)) then
        use_ab_algorithm = .false.
     else if(str_equal('cone ',word)) then
        call get_param(line,word1,at)        
        !cone_location = str_to_double(word1)
        call get_param(line,word2,at)
        call set_cone(cone_location_d=str_to_double(word1), &
             cone_slope_d=str_to_double(word2))
        !cone_slope = str_to_double(word2)
     else if(str_equal('recirculate ',word)) then
        recirculate = .true.
        call get_param(line,word,at)
        call set_recirculation_height(str_to_double(word))
        call get_param(line,word,at)
        door_closed_time = str_to_double(word)
        call get_param(line,word,at)
        first_door_open_time = str_to_double(word)
        call get_param(line,word4,at)
        !hole_size = str_to_double(word)
        call get_param(line,word5,at)
        !hole_depth = str_to_double(word)
        call set_hole(hole_size_d=str_to_double(word4),&
             hole_depth_d=str_to_double(word5))
     else if(str_equal('exit_chute ',word)) then
        call get_param(line,word1,at)
        !hole_size = str_to_double(word)
        call get_param(line,word2,at)
        !hole_depth = str_to_double(word)
        call set_hole(hole_size_d=str_to_double(word1),&
             hole_depth_d=str_to_double(word2))
     else if(str_equal('recirculate_params ',word)) then
        recirculate = .true.
        call get_param(line,word,at)
        call set_recirculation_height(str_to_double(word))
        call get_param(line,word,at)
        door_closed_time = str_to_double(word)
        call get_param(line,word,at)
        first_door_open_time = str_to_double(word)
     else if(str_equal('velocity_tally ',word)) then
        call get_param(line,word,at)        
        first_velocity_tally_time = str_to_double(word)
        call get_param(line,word,at)
        velocity_tally_frequency = str_to_int(word)
     else if(str_equal('probability_table ',word)) then
        run_probability_tally = .true.
        call get_param(line,word,at)
        probability_tally_start = str_to_double(word)
        call get_param(line,word,at)
        probability_tally_frequency = str_to_int(word)
        call get_param(line,word,at)
        probability_tally_delta = str_to_double(word)
     else if(str_equal('number_of_cones ',word)) then
        call get_param(line,word,at)
        !number_of_cones = str_to_int(word)
        call set_number_of_cones(str_to_int(word))
     else if(str_equal('tabular_earthquake ',word)) then
        call parse_tabular_earthquake(line,at)
     else if(str_equal('cone_packing ',word)) then
        !use_cone_packing = .true.
        call get_param(line,word1,at)
        !packing_cone_radius = str_to_double(word)     
        call get_param(line,word2,at)
        !packing_cone_bottom = str_to_double(word)     
        call setup_cone_packing(cone_radius = str_to_double(word1),&
             cone_bottom = str_to_double(word2),chute_length = 0.0_d_t)
     else if(str_equal('cone_packing_chute ',word)) then
        !use_cone_packing = .true.
        call get_param(line,word1,at)
        !packing_cone_radius = str_to_double(word)     
        call get_param(line,word2,at)
        !packing_cone_bottom = str_to_double(word)     
        call get_param(line,word3,at)
        !packing_chute_length = str_to_double(word)
        call setup_cone_packing(cone_radius = str_to_double(word1),&
             cone_bottom = str_to_double(word2), &
             chute_length = str_to_double(word3))
     else if(str_equal('offset_inlet_holes ',word)) then
        !offset_inlet_and_cone = .true.
        call set_offset_inlet_and_cone()
     else if(str_equal('sort_pebbles ',word) .or. str_equal('sort_spheres ',word)) then
        sort_pebbles = .true.
     else if(str_equal('calculate_force_tally ',word)) then
        calculate_force_tally = .true.
     else if(str_equal('ppf_filename ',word)) then
        print *,"#ppf_filename ignored"
     else
        print *,"ERROR unknown keyword ",word
     end if
  end do input_loop

  ! Do calcuations of secondary parameters and allocations
  if(.not. use_ab_algorithm) then 
     multistep_order = 2
  end if
  static_velocity_max = sqrt(static_velocity_sqr_max)
  long_slip = sqrt(long_slip_sqr)

  area = get_vessel_area()
  fill_edge = get_vessel_fill_edge()

  pebble_volume = (4.0/3.0)*pi*pebble_radius**3  
  height_needed = REAL(number_of_pebbles)*pebble_volume/area  
  !static_rotation_sqr = static_speed_sqr/pebble_radius**2
  mass = (4.0/3.0)*pi*(density_center*pebble_center_radius**3+&
       density_outside*(pebble_radius**3 - pebble_center_radius**3))
  gravitational_force = gravitational_acceleration*mass
  moment_of_inertia = 8.0/15.0 * pi * &
       (density_center * pebble_center_radius**5 + &
       density_outside * (pebble_radius ** 5 - pebble_center_radius ** 5))
  zone_increment = pebble_radius*2.0  
  call get_max_xy_extents(x_high,y_high)
  call initialize_geometry(z_low,z_high)
  z_high = z_high + height_needed/initial_packing
  x_low = -x_high
  y_low = -y_high
  if(calculate_force_tally) then
     call initialize_force_tally(x_high,x_low,y_high,y_low,z_high,z_low)
  end if
  x_size = CEILING((x_high-x_low)/zone_increment)
  y_size = CEILING((y_high-y_low)/zone_increment)
  z_size = CEILING((z_high-z_low)/zone_increment)
  tally_z_inc = pebble_radius
  tally_radial_inc = pebble_radius
  rt_size = CEILING((y_high)/tally_radial_inc)
  zt_size = CEILING((z_high-z_low)/tally_z_inc)

  allocate(states(multistep_order),deltas(multistep_order))
  do i=1,multistep_order
     allocate(states(i)%pp(number_of_pebbles))
     allocate(states(i)%pv(number_of_pebbles))
     allocate(states(i)%pw(number_of_pebbles))
     ! If we don't allocate various fortran compilers act stupid.
     if(use_static_friction .or. .true.) then
        allocate(states(i)%ppsd(slip_zone_size,number_of_pebbles))
        allocate(states(i)%pps_index(slip_zone_size,number_of_pebbles))
        allocate(states(i)%pss(number_of_pebbles))
        allocate(states(i)%psn(number_of_pebbles))
     else 
        !nullify(states(i)%pps,states(i)%pps_index,states(i)%pss,states(i)%psn)
     end if
     allocate(deltas(i)%pdp(number_of_pebbles))
     allocate(deltas(i)%pdv(number_of_pebbles))
     allocate(deltas(i)%pdw(number_of_pebbles))
  end do
  call initialize_pebble_cache(nearby_pebble_cache,x_size,x_low,y_size,y_low,&
       z_size,z_low,zone_increment)
  !allocate(zone(x_size,y_size,z_size,zone_count),stat=allocate_status)
  !if (allocate_status /= 0) stop "ERROR unable to allocate ram"
  !Allocate the radial tally zones
  allocate(zone_n(tally_sections,rt_size,zt_size))
  allocate(zone_vel_sum(tally_sections,rt_size,zt_size))
  allocate(zone_dot_sum(tally_sections,rt_size,zt_size))
  !Allocate the rectangular tally zones
  allocate(zone_rect_n(x_size,y_size,z_size),stat=allocate_status)
  if (allocate_status /= 0) stop "ERROR unable to allocate ram"
  allocate(zone_rect_vel_sum(x_size,y_size,z_size),stat=allocate_status)
  if (allocate_status /= 0) stop "ERROR unable to allocate ram"
  allocate(zone_rect_dot_sum(x_size,y_size,z_size),stat=allocate_status)
  if (allocate_status /= 0) stop "ERROR unable to allocate ram"
  
  call initialize_wear_tallies(z_low,zone_increment,z_size)
  if(run_probability_tally) then
     call initialize_count_data(probability_data,probability_tally_delta,& 
          x_low,x_high,y_low,y_high,z_low,z_high)
  end if

  time_till_door_open = first_door_open_time

  call initialize_earthquake

  !zero the tallies
  zone_n = 0
  zone_vel_sum = vector(0.0,0.0,0.0)
  zone_dot_sum = 0.0
  zone_rect_n = 0
  zone_rect_vel_sum = vector(0.0,0.0,0.0)
  zone_rect_dot_sum = 0.0

  if(use_static_friction) then 
     call clear_pps_index(states(1)%pps_index)
  end if

  call do_runs()

  ! deallocate memory
  do i=1,multistep_order
     deallocate(states(i)%pp,states(i)%pv,states(i)%pw)
     if(use_static_friction) then 
        deallocate(states(i)%ppsd,states(i)%pps_index)
        deallocate(states(i)%pss,states(i)%psn)
     end if
     deallocate(deltas(i)%pdp,deltas(i)%pdv,deltas(i)%pdw)
  end do
  
  call free_earthquake
CONTAINS  

  subroutine initialize_pebble_data()
    integer :: initials, hits, i,j,k,l
    real(kind=d_t) :: loading_height
    logical :: done  

    loading_height = get_loading_height()
    
    if (do_load_pebble_info) then
       call load_pebble_info(states(1)%pp,states(1)%pv,states(1)%pw &
            ,load_pebble_info_filename,do_load_pebble_slips,&
            do_load_pebble_surface_slips,states(1))
    else if(do_load_positions) then 
       call load_positions(states(1)%pp,load_positions_filename,&
            load_positions_divide)
       states(1)%pv = vector(0.0,0.0,0.0)
       states(1)%pw = vector(0.0,0.0,0.0)
    else
       DO i=1,number_of_pebbles  
          CALL initialize_pebble(states(1)%pp(i),states(1)%pv(i),&
               states(1)%pw(i), &
               seed,loading_height,height_needed,&
               pebble_radius, fill_edge)
       END DO
       initials = 0  
36     CONTINUE  
       done = .TRUE.  
       hits = 0  
       call update_nearby_pebble_cache(nearby_pebble_cache,states(1)%pp)
       DO j=1,number_of_pebbles  
          call nearby_pebbles(nearby_pebble_cache,states(1)%pp(j), &
               near_pebbles, near_pebbles_size, near_pebbles_count, j)
          DO l=1,near_pebbles_count  
             k = near_pebbles(l)  
             IF (pebble_pebble_overlap(states(1)%pp(j),states(1)%pp(k),&
                  pebble_radius))  THEN
                CALL initialize_pebble(states(1)%pp(j),states(1)%pv(j),&
                     states(1)%pw(j), &
                     seed,loading_height,&
                     height_needed, pebble_radius,fill_edge)
                done = .FALSE.  
                hits = hits + 1  
             END IF
          END DO
       END DO
       PRINT *,"#Initializing",initials, hits
       initials = initials + 1
       IF (.not. done) GOTO 36
    end if

    if(sort_pebbles) then
       call sort_pebbles_by_height(states(1))
    end if

    if(.true.) then
       PRINT *,"#Initial Data",i
       call display_pebble_info(time,states(1))
    end if
  end subroutine initialize_pebble_data

  subroutine print_parameters() 
    PRINT *,"#Pebbles"  
    PRINT *,"#Version $Id: pebbles7.f90 1378 2011-02-24 15:10:13Z cogljj $"
    PRINT *,"# Prepared by Battelle Energy Alliance, LLC"
    PRINT *,"# Under Contract No. DE-AC07-05ID14517"
    PRINT *,"# With the U. S. Department of Energy"
    PRINT *,"#Pebble_Count: ",number_of_pebbles  
    PRINT *,"#alpha",alpha," grav",gravitational_force  
    PRINT *,"#runs",runs,runs*alpha,"s long",time,"s start"
    PRINT *,"#sizes",x_size,y_size,z_size  
    PRINT *,"#sp ss k friction",pp_static_friction,surface_static_friction&
         ,kinetic_friction  
    !PRINT *,"#static speed sqr",static_speed_sqr  
    if(use_static_friction) then
       PRINT *,"#static velocity max",static_velocity_max
       PRINT *,"#static friction hooke",static_friction_hooke, surface_static_friction_hooke
       PRINT *,"#linear static velocity",linear_static_velocity_clamp,zero_static_slips
    else 
       PRINT *,"#Not using static friction"
    end if
    PRINT *,"#n_dashpot",normal_dash_pot_constant," t_dashpot ",&
         transverse_dash_pot_constant," air_friction ",air_friction  
    PRINT *,"#hooke", pebble_pebble_hooke, pebble_vessel_hooke  
    PRINT *,"#d_t",d_t
    PRINT *,"#display_frequency",display_frequency,display_frequency*alpha
    PRINT *,"#mass",mass," moment_of_inertia" ,moment_of_inertia
    call print_geometry_info()
    !      call pebble_test()
    !      GO TO 25
    !     Initialization
    !pi = 4.0*ATAN(1.0)  
    if(recirculate) then 
       PRINT *,"#door_closed_time",door_closed_time,"ttdo",time_till_door_open
    else 
       PRINT *,"#recirculation off"
    end if
    PRINT *,"#first_velocity_tally_time",first_velocity_tally_time,"vtf",&
         velocity_tally_frequency
    call print_earthquake_info
    PRINT *,"#torque_enable",torque_enable
    PRINT *,"#static_friction",use_static_friction,"spf",pp_static_friction,&
         "ssf",surface_static_friction,&
         "hooke",static_friction_hooke,"v**2",static_velocity_sqr_max,&
         "surface_hooke",surface_static_friction_hooke
    PRINT *,"#display_frequency",display_frequency,display_frequency*alpha
    PRINT *,"#run_length",runs,runs*alpha
    PRINT *,"#seed",seed
    PRINT *,"#use_simple_integrator",.not. use_ab_algorithm
    print *,"#sizes",x_size,y_size,z_size
    print *,"#near_pebble_size",near_pebbles_size
    !$ print *,"#omp_procs",omp_get_num_procs()
  end subroutine print_parameters

  subroutine do_runs()
    !t_c = pebble multistep time index current
    !t_m1 = pebble multistep index previous
    !t_m2 = pebble multistep index for current time - alpha * 2
    integer :: t_c,t_m1,t_m2,t_m3

    call print_parameters()

    call initialize_pebble_data()

    !     Run the data
    t_c = 1
    run_loop: DO i=0,runs     
       PRINT *,"#Run",i
       call print_time("Start loop")
       !t_m3 is index of time three steps back
       t_m3 = 1 + mod(i-2,multistep_order)
       t_m2 = 1 + mod(i-1,multistep_order)
       t_m1 = 1 + mod(i,multistep_order)
       !t_c is the current time.  states(t_c) is where the new data is written
       t_c = 1 + mod(i+1,multistep_order)
       call begin_new_force_tally_timestep()

       if(use_ab_algorithm) then 
          if (i > 4) then 
             call get_ab_timestep(states(t_c),&
                  states(t_m1),&
                  deltas,t_c,t_m1,t_m2,t_m3,alpha,time)
          else 
             call get_rk_timestep(states(t_c),&
                  states(t_m1),&
                  deltas(t_c),&
                  alpha,time)
          end if
       else 
          call get_simple_timestep(states(t_c),states(t_m1),deltas(t_c),&
               alpha,time)
       end if
       call handle_velocity_tally(states(t_c),time,i)
       if (recirculate) then 
          call handle_recirculation(states(t_c))
       end if

       if(run_probability_tally .and. time >= probability_tally_start &
            .and. MOD(i,probability_tally_frequency) .eq. 0) then
          call handle_probability_tally(states(t_c),states(t_m1))
       end if

       IF(MOD(i,display_frequency) .eq. 0 .or. i == runs) THEN
          PRINT *,"#Display Run",i
          call display_pebble_info(time,states(t_c))
       else if(mod(i,energy_frequency) .eq. 0) then
          call display_pebble_energy(time,states(t_c))
       END IF
       if(mod(i,position_frequency) .eq. 0) then
          call display_pebble_positions(time,states(t_c))
       end if
       if(mod(i,dump_frequency) .eq. 0 .or. i == runs) then
          call display_overlaps(states(t_c)%pp)
          call display_zone_velocities()
       end if
       if(any_pebbles_of_interest()) then
          call dump_pebbles_of_interest_info(states(t_c),deltas(t_c))
       end if
       if(calculate_pebble_wear_vol .and. mod(i,pebble_wear_vol_display_freq) == 0) then
          call print_pebble_wear_data(time,alpha)
       end if
       time = time + alpha
    END DO run_loop

    if(do_dump_positions) then
       call dump_positions(states(t_c)%pp,&
            dump_positions_filename,dump_positions_mult)
    end if

    if(run_probability_tally .and. time >= probability_tally_start) then
       call print_data(probability_data)
    end if

    if(calculate_force_tally) then
       call print_force_tally
    end if
  end subroutine do_runs

  subroutine load_positions(pp,filename,divide) 
    type(vector), pointer, dimension(:) :: pp
    character(*), intent(in) :: filename
    real(kind=d_t), intent(in) :: divide
    integer :: openstatus, i
    open(LOAD_UNIT,file=filename,action="READ",iostat=openstatus)

    if(openstatus > 0) then 
       print *,"ERROR Cannot open file to load positions"
       return
    end if
    do i=1,number_of_pebbles
       read(LOAD_UNIT, *) pp(i)
       pp(i) = (1.0/divide) * pp(i)
    end do
    close(LOAD_UNIT)

  end subroutine load_positions


  subroutine dump_positions(pp,filename,mult) 
    type(vector), pointer, dimension(:) :: pp
    character(*), intent(in) :: filename
    real(kind=d_t), intent(in) :: mult
    integer :: openstatus, i
    open(DUMP_UNIT,file=filename,action="WRITE",iostat=openstatus)

    if(openstatus > 0) then 
       print *,"ERROR Cannot open file to dump positions"
       return
    end if
    do i=1,number_of_pebbles
       write(DUMP_UNIT, *) mult * pp(i)
    end do
    close(DUMP_UNIT)
    
  end subroutine dump_positions
  
  subroutine display_zone_velocities() 
    ! s_z is slice zone, r_z in radial zone, z_z is z zone
    integer :: s_z,r_z,z_z,n,x_z,y_z
    real(kind=d_t) :: cx, cy, cz
    type(vector) :: avg, sum
    real(kind=d_t) :: var, dot_sum
    print *,"#zsi ",tally_sections,rt_size,zt_size
    print *,"#zin ",tally_radial_inc,tally_z_inc,z_low 
    print *,"#zone x range y range z range n sum average dot variance"
    do s_z=1,tally_sections
       do r_z=1,rt_size
          do z_z=1,zt_size
             n = zone_n(s_z,r_z,z_z)
             if(n > 0) then 
                sum = zone_vel_sum(s_z,r_z,z_z)
                avg = (1.0/n)*sum
                dot_sum = zone_dot_sum(s_z,r_z,z_z)
                var = (1.0/(n-1))*(n*vmag_sqr(avg) &
                     -2.0*(avg .dot. sum) + dot_sum)
                call radial_zone_center(s_z,r_z,z_z,cx,cy,cz)
                print *,"#z   ",&
                     cx,cy,cz,&
                     n,sum,avg,dot_sum,var
             end if
          end do
       end do
    end do
    print *,"#rzs ",x_size,y_size,z_size
    print *,"#rzl ",x_low,y_low,z_low, zone_increment
    do z_z=1,z_size
       do y_z=1,y_size
          do x_z=1,x_size
             n = zone_rect_n(x_z,y_z,z_z)
             if(n > 0) then
                sum = zone_rect_vel_sum(x_z,y_z,z_z)
                avg = (1.0/n)*sum
                dot_sum = zone_rect_dot_sum(x_z,y_z,z_z)
                var = (1.0/(n-1))*(n*vmag_sqr(avg) &
                     -2.0*(avg .dot. sum) + dot_sum)
                cx = zone_center(x_low, zone_increment, x_z)
                cy = zone_center(y_low, zone_increment, y_z)
                cz = zone_center(z_low, zone_increment, z_z)
                print *,"#rz  ",cx,cy,cz,n,sum,avg,dot_sum,var
             end if
          end do
       end do
    end do
    
  end subroutine display_zone_velocities
  
  subroutine display_overlaps(pp)
    type(vector), pointer,dimension(:) :: pp
    real(kind=d_t) :: max_overlap, overlap_sum, overlap_total, overlap, average
    integer :: j,k,l,complete_overlap_count,overlap_count
    
    call update_nearby_pebble_cache(nearby_pebble_cache,pp)
    max_overlap = -100.0
    overlap_sum = 0.0
    complete_overlap_count = 0
    
    DO j=1,number_of_pebbles
       call nearby_pebbles(nearby_pebble_cache, pp(j),&
            &        near_pebbles, near_pebbles_size, near_pebbles_count, j)
       overlap_count = 0
       overlap_total = 0.0
       DO l=1,near_pebbles_count
          k = near_pebbles(l)
          overlap = pebble_pebble_overlap_dist(pp(j),pp(k),pebble_radius)
          IF (overlap > 0.0) THEN
             overlap_total = overlap_total + overlap
             overlap_count = overlap_count + 1
          END IF
          IF (overlap > max_overlap) THEN
             max_overlap = overlap
          END IF
       END DO
       PRINT *,"#P_overlap",j,overlap_count,overlap_total
       complete_overlap_count = complete_overlap_count + overlap_count
       overlap_sum = overlap_sum + overlap_total
    END DO
    average = overlap_sum/complete_overlap_count
    PRINT *,"#Max overlap",max_overlap
    PRINT *,"#Max overlap %",max_overlap*100/pebble_radius
    PRINT *,"#Totals",overlap_sum, complete_overlap_count
    PRINT *,"#Average",overlap_sum/complete_overlap_count
    PRINT *,"#Average %",average*100/pebble_radius
  end subroutine display_overlaps

  subroutine dump_pebbles_of_interest_info(current_state, last_delta)
    type(pebble_state), intent(in) :: current_state
    type(pebble_delta), intent(in) :: last_delta
    integer:: i
    i = pebble_of_interest
    ! i=1, p=3:4:5, v=7:8:9, w=11:12:13, dp=15:16:17,dv=19:20:21,dw=23:24:25
    print *,"#POI",i,"p",current_state%pp(i),"v",current_state%pv(i),&
         "w",current_state%pw(i),&
         "dp",last_delta%pdp(i),"dv",last_delta%pdv(i),"dw",last_delta%pdw(i)
  end subroutine dump_pebbles_of_interest_info

  subroutine sort_pebbles_by_height(state)
    type(pebble_state), intent(inout) :: state
    integer,dimension(:),allocatable :: new_ordering
    type(real_int_pair),dimension(:),allocatable :: height_list
    integer :: i
    allocate(new_ordering(number_of_pebbles))
    allocate(height_list(number_of_pebbles))
    do i=1,number_of_pebbles
       height_list(i) = real_int_pair(state%pp(i)%z,i)
    end do
    call sort_ri_pair(height_list)
    do i=1,number_of_pebbles
       !new_ordering(old_index) = new_index
       new_ordering(height_list(i)%i) = i       
    end do

    call reorder_pebbles(state, new_ordering)

    deallocate(new_ordering,height_list)
  end subroutine sort_pebbles_by_height


  ! For each pebble, moves it from old_index to new_ordering(old_index) 
  !  in the state passed in. new_index = new_ordering(old_index)
  ! So, if we have new_ordering = (/ 1, 4, 2, 3 /), then pebble 1 will 
  ! stay the same, and pebble 2_old_index will become 4_new_index
  subroutine reorder_pebbles(state, new_ordering)    
    type(pebble_state), intent(inout) :: state
    type(pebble_state) :: temp_state
    integer,dimension(:),intent(in) :: new_ordering
    integer :: i,j,old_from,old_to,new_from,new_to,empty
    type(vector) :: old_slip
    allocate(temp_state%pp(number_of_pebbles))
    allocate(temp_state%pv(number_of_pebbles))
    allocate(temp_state%pw(number_of_pebbles))
    allocate(temp_state%ppsd(slip_zone_size,number_of_pebbles))
    allocate(temp_state%pps_index(slip_zone_size,number_of_pebbles))
    allocate(temp_state%pss(number_of_pebbles))
    allocate(temp_state%psn(number_of_pebbles))
    temp_state%pps_index = 0
    do i=1,number_of_pebbles
       print *,"#reordering ",new_ordering(i)," <- ",i
       temp_state%pp(new_ordering(i)) = state%pp(i)
       temp_state%pv(new_ordering(i)) = state%pv(i)
       temp_state%pw(new_ordering(i)) = state%pw(i)
       if(use_static_friction) then
          temp_state%pss(new_ordering(i)) = state%pss(i)
          temp_state%psn(new_ordering(i)) = state%psn(i)
          do j=1,slip_zone_size
             if(state%pps_index(j,i) /= 0) then
                old_from =  i
                old_to = state%pps_index(j,i)
                new_from = new_ordering(old_from)
                new_to = new_ordering(old_to)
                old_slip = state%ppsd(j,i)
                if(new_to < new_from) then
                   call swap_int(new_to,new_from)
                   old_slip = (-1.0)*old_slip
                end if
                empty = get_empty_slip_index(temp_state%pps_index,new_from)
                temp_state%pps_index(empty,new_from) = new_to
                temp_state%ppsd(empty,new_from) = old_slip
!!$                temp_state%pps_index(j,new_ordering(i)) = &
!!$                     new_ordering(state%pps_index(j,i))
!!$                temp_state%ppsd(j,new_ordering(i)) = state%ppsd(j,i)
             end if 
          end do
       end if
    end do
    state%pp = temp_state%pp
    state%pv = temp_state%pv
    state%pw = temp_state%pw
    if(use_static_friction) then
       state%pss = temp_state%pss
       state%psn = temp_state%psn
       state%pps_index = temp_state%pps_index
       state%ppsd = temp_state%ppsd
    end if
    deallocate(temp_state%pp,temp_state%pv,temp_state%pw,temp_state%pss)
    deallocate(temp_state%psn,temp_state%pps_index,temp_state%ppsd)
  end subroutine reorder_pebbles

  subroutine get_ab_timestep(out_state, in_state, deltas,&
       t_c,t_m1,t_m2,t_m3,alpha,time)
    type(pebble_state), intent(in) :: in_state
    type(pebble_state), intent(out) :: out_state
    type(pebble_delta), dimension(:) :: deltas
    !the indexs into the dp_'s
    ! t = t index, t_m1 = t-h index, t_m2 = t-2*h index, t_m3 = t-3*h index
    integer, intent(in) :: t_c,t_m1,t_m2,t_m3
    real(kind=d_t), intent(in) :: alpha, time
    integer :: j
    
    call get_derivs(nearby_pebble_cache, in_state,&
         deltas(t_c),&
         &  number_of_pebbles, near_pebbles, near_pebbles_size, time)
    DO j=1,number_of_pebbles
       out_state%pp(j) = in_state%pp(j) + &
            (alpha/24.0)*(55.0*deltas(t_c)%pdp(j)-59.0*deltas(t_m1)%pdp(j)&
            +37.0*deltas(t_m2)%pdp(j)-9.0*deltas(t_m3)%pdp(j))
       out_state%pv(j) = in_state%pv(j) + &
            (alpha/24.0)*(55.0*deltas(t_c)%pdv(j)-59.0*deltas(t_m1)%pdv(j)&
            +37.0*deltas(t_m2)%pdv(j)-9.0*deltas(t_m3)%pdv(j))
       out_state%pw(j) = in_state%pw(j) + &
            (alpha/24.0)*(55.0*deltas(t_c)%pdw(j)-59.0*deltas(t_m1)%pdw(j)&
            +37.0*deltas(t_m2)%pdw(j)-9.0*deltas(t_m3)%pdw(j))
       
    END do
    if(use_static_friction) then 
       call update_slips(in_state, out_state, alpha, time)
    end if

  end subroutine get_ab_timestep

  subroutine get_simple_timestep(out_state,in_state,delta,alpha,time)
    type(pebble_state),intent(in) :: in_state
    type(pebble_state),intent(out) :: out_state
    type(pebble_delta), intent(out) :: delta
    real(kind=d_t), intent(in) :: alpha, time
    
    call get_derivs(nearby_pebble_cache, in_state, delta, &
         &        number_of_pebbles, near_pebbles, near_pebbles_size, time)
    call combine_step(number_of_pebbles,in_state,out_state,&
         delta,alpha, time)

  end subroutine get_simple_timestep

  subroutine get_rk_timestep(out_state,in_state,delta,alpha,time)
    type(pebble_state), intent(in) :: in_state
    type(pebble_state), intent(out) :: out_state
    type(pebble_delta), intent(out) :: delta

    real(kind=d_t), intent(in) :: alpha, time
    integer :: j
    !tpp = temporary pebble posision, tpv = temporary pebble velocity  
    type(vector),target :: tpp(number_of_pebbles), tpv(number_of_pebbles)
    !tpw = temporary pebble angular velocity
    type(vector),target :: tpw(number_of_pebbles)   
    type(vector),target :: tppsd(slip_zone_size,number_of_pebbles)
    integer,target :: tpps_index(slip_zone_size,number_of_pebbles)
    type(vector),target :: tpss(number_of_pebbles)
    type(vector),target :: tpsn(number_of_pebbles)

    type(pebble_state) :: temp_state
    type(pebble_delta) :: f2,f3,f4
    !     The Runge-Kutta values
    !f1pa = pebble acceleration 1, f1pv = pebble velocity 1
    !f1pn = pebble angular acceleration 1
    !type(vector),intent(out) :: f1pa(number_of_pebbles),f1pv(number_of_pebbles)
    !type(vector),intent(out) :: f1pn(number_of_pebbles)
    type(vector),target :: f2pa(number_of_pebbles), f2pv(number_of_pebbles)
    type(vector),target :: f2pn(number_of_pebbles)
    type(vector),target :: f3pa(number_of_pebbles), f3pv(number_of_pebbles)
    type(vector),target :: f3pn(number_of_pebbles)
    type(vector),target :: f4pa(number_of_pebbles), f4pv(number_of_pebbles)
    type(vector),target :: f4pn(number_of_pebbles)
    temp_state%pp => tpp
    temp_state%pv => tpv
    temp_state%pw => tpw
    if(use_static_friction) then
       temp_state%ppsd => tppsd
       temp_state%pps_index => tpps_index
       temp_state%pss => tpss
       temp_state%psn => tpsn
    end if
    f2%pdp => f2pv
    f2%pdv => f2pa
    f2%pdw => f2pn

    f3%pdp => f3pv
    f3%pdv => f3pa
    f3%pdw => f3pn

    f4%pdp => f4pv
    f4%pdv => f4pa
    f4%pdw => f4pn
    

    !        f1 = f(T,X)
    call get_derivs(nearby_pebble_cache, in_state, delta, &
         &        number_of_pebbles, near_pebbles, near_pebbles_size, time)
    !        f2 = f(t+h/2,x+0.5*f1*h)
    !        temps = x + 0.5*f1*h
    call combine_step(number_of_pebbles,in_state,temp_state,&
         delta,0.5*alpha, time)
    call get_derivs(nearby_pebble_cache, temp_state,f2, &
         &        number_of_pebbles, near_pebbles, near_pebbles_size, time)
    !        f3 = f(t+h/2,x+0.5*f2*h)
    !        temps = x + 0.5*f2*h
    call combine_step(number_of_pebbles,in_state,temp_state,&
         f2, 0.5*alpha, time)
    call get_derivs(nearby_pebble_cache, temp_state,f3, &
         &        number_of_pebbles, near_pebbles, near_pebbles_size, time)
    !        f4 = f(t+h,x+f3*h)
    !        temps = x + f3*h
    call combine_step(number_of_pebbles,in_state,temp_state,&
         f3,alpha, time)
    call get_derivs(nearby_pebble_cache, temp_state,f4,&
         &        number_of_pebbles, near_pebbles, near_pebbles_size, time)
    DO j=1,number_of_pebbles
!!$        if(mod(i,display_frequency) .eq. 0) then
!!$           PRINT *,"#pebble",j
!!$           PRINT *,"#Old pp,pv,pw",pp(j),pv(j),pw(j)
!!$           PRINT *,"#f1s",f1pv(j),f1pa(j),f1pn(j)
!!$           PRINT *,"#f2s",f2pv(j),f2pa(j),f2pn(j)
!!$           PRINT *,"#f3s",f3pv(j),f3pa(j),f3pn(j)
!!$           PRINT *,"#f4s",f4pv(j),f4pa(j),f4pn(j)
!!$        end if
       out_state%pp(j) = in_state%pp(j)+ &
            (alpha/6.0)*(delta%pdp(j)+2.0*f2%pdp(j)+2.0*f3%pdp(j)+f4%pdp(j)) 
       out_state%pv(j) = in_state%pv(j)+ &
            (alpha/6.0)*(delta%pdv(j)+2.0*f2%pdv(j)+2.0*f3%pdv(j)+f4%pdv(j)) 
       out_state%pw(j) = in_state%pw(j)+ &
            (alpha/6.0)*(delta%pdw(j)+2.0*f2%pdw(j)+2.0*f3%pdw(j)+f4%pdw(j)) 
       !PRINT *,"#New pp,pv,pw",pp(j)%x,pp(j)%y,pp(j)%z,pv(j)%x,pv(j)%y,pv(j)%z,pw(j)%x,pw(j)%y,pw(j)%z

    END DO
    if(use_static_friction) then
       call update_slips(in_state,out_state,alpha, time)
    end if
  end subroutine get_rk_timestep

  subroutine handle_probability_tally(current_state,old_state)
    type(pebble_state), intent(in) :: current_state, old_state
    type(vector), pointer, dimension(:) :: pp_cur,pp_old
    integer :: i
    pp_cur => current_state%pp
    pp_old => old_state%pp
    do i=1,number_of_pebbles
       call add_count(probability_data,pp_old(i),pp_cur(i))
    end do

  end subroutine handle_probability_tally
    

  subroutine handle_velocity_tally(in_state,time,run)
    type(pebble_state), intent(in) :: in_state    
    real(kind=d_t), intent(in) :: time
    integer, intent(in) :: run
    integer :: i,s_z,r_z,z_z,x_z,y_z
    real(kind=d_t) :: x,y,z
    ! The std_velocity is modified so that spiral motion and radial 
    !  motion are normalized.
    type(vector) :: std_velocity
    type(vector), pointer, dimension(:) :: pp,pv
    if(mod(run,velocity_tally_frequency) .ne. 0) then
       return
    end if
    pp => in_state%pp
    pv => in_state%pv
    if(time < first_velocity_tally_time) then 
       return
    end if
    do i=1,number_of_pebbles
       ! Use a random point in the sphere so not biased near edges 
       !  since the center of the sphere will be oversampled and the 
       !  edges will be undersampled if I just use that.
       call random_point_in_sphere(pp(i)%x,pp(i)%y,pp(i)%z,pebble_radius,&
            x,y,z,seed)
       call find_radial_zone(x,y,z,s_z,r_z,z_z)
       call regularize_radial_velocity(pp(i),pv(i),std_velocity,s_z,r_z,z_z)
       zone_n(s_z,r_z,z_z) = zone_n(s_z,r_z,z_z) + 1
       zone_vel_sum(s_z,r_z,z_z) = zone_vel_sum(s_z,r_z,z_z) + std_velocity
       zone_dot_sum(s_z,r_z,z_z) = zone_dot_sum(s_z,r_z,z_z) + &
            vmag_sqr(std_velocity)
       
       !XXX Should this be using find_zone instead of find_base_zone?
       x_z = find_base_zone(x_low, zone_increment, x_size, pp(i)%x)  
       y_z = find_base_zone(y_low, zone_increment, y_size, pp(i)%y)  
       z_z = find_base_zone(z_low, zone_increment, z_size, pp(i)%z)  
       zone_rect_n(x_z,y_z,z_z) = zone_rect_n(x_z,y_z,z_z) + 1
       zone_rect_vel_sum(x_z,y_z,z_z) = zone_rect_vel_sum(x_z,y_z,z_z) + pv(i)
       zone_rect_dot_sum(x_z,y_z,z_z) = zone_rect_dot_sum(x_z,y_z,z_z) + &
            vmag_sqr(pv(i))
    end do

  end subroutine handle_velocity_tally

  subroutine handle_recirculation(state)
    type(pebble_state), intent(inout) :: state
    type(vector), pointer,dimension(:) :: pp,pv
    type(vector), pointer,dimension(:) :: pw
    integer :: i, lowest_pebble
    real(kind=d_t) :: lowest_height
    pp => state%pp
    pv => state%pv
    pw => state%pw
    time_till_door_open = time_till_door_open - alpha
    if(time_till_door_open < 0.0) then
       if(.not. door_open) then
          print *,"#door_open",time
       end if
       door_open = .true.
    end if
    if(.not. door_open) then
       return
    end if
    lowest_pebble = 0
    lowest_height = z_high
    !Find the lowest pebble
    do i=1,number_of_pebbles
       if(is_pebble_through_door(current_cone_drop,pp(i)) .and. &
            pp(i)%z < lowest_height) then
          lowest_pebble = i
          lowest_height = pp(i)%z
       end if
    end do
    ! Drop the lowest pebble, and slam door.
    if(lowest_pebble > 0) then
       i = lowest_pebble
       pp(i) = get_recirculation_location(current_cone_drop)
       pv(i) = get_recirculation_velocity(current_cone_drop)       
       current_cone_drop = mod(current_cone_drop, get_number_of_recirc_locs())+1
       print *,"#Recirculate at ",pp(i)
       pw(i) = vector(0.0,0.0,0.0)
       door_open = .false.
       time_till_door_open = door_closed_time
       print *,"#Dropped Pebble",i
    end if
  end subroutine handle_recirculation
    
  subroutine load_pebble_info(pp,pv,pw,filename,read_slips, read_surface,&
       out_state)
    type(vector), intent(out),dimension(:) :: pp,pv,pw
    !type(vector), intent(out) :: pw(number_of_pebbles)
    type(pebble_state), intent(out), optional :: out_state
    logical, intent(in) :: read_slips, read_surface
    character(*), intent(in) :: filename
    integer :: openstatus, i, j, k, length
    character(3) :: junk_str
    character(15) :: junk_slip_str
    character(16) :: junk_surface_str
    character(9) :: junk_pp_slip_str
    type(vector) :: slip, slip_normal
    
    open(LOAD_INFO_UNIT,file=filename,action="READ",iostat=openstatus)
    
    if(openstatus > 0) then 
       print *,"ERROR Cannot open file to load pebble info"
       return
    end if

    do i=1,number_of_pebbles
       read(LOAD_INFO_UNIT,*) junk_str,j,&
            pp(i)%x,pp(i)%y,pp(i)%z,&
            pv(i)%x,pv(i)%y,pv(i)%z,&
            pw(i)%x,pw(i)%y,pw(i)%z
       if(i /= j) then
          print *,"ERROR Mismatch of pebble number",i,j
       end if
    end do
    if(read_slips) then
       call clear_pps_index(out_state%pps_index)
       slip = (/ 0.0, 0.0, 0.0 /)
       do 
          read(LOAD_INFO_UNIT,*) junk_slip_str,j,length
          if(j == 0) then
             exit
          end if
          do i=1,length
             read(LOAD_INFO_UNIT,*) junk_pp_slip_str,k,slip%x,slip%y,slip%z
             out_state%pps_index(i,j) = k
             out_state%ppsd(i,j) = slip
          end do
       end do
    end if
    if(read_surface) then
       do i=1,number_of_pebbles
          read(LOAD_INFO_UNIT,*) junk_surface_str,j,slip%x,slip%y,slip%z,&
               slip_normal%x,slip_normal%y,slip_normal%z
          if(j == i) then 
             out_state%pss(i) = slip
             out_state%psn(i) = slip_normal
          else 
             print *,"ERROR Mismatch in read surface slips ",j," /= expected",i
          end if
       end do
    end if
    close(LOAD_INFO_UNIT)

  end subroutine load_pebble_info

  subroutine display_pebble_positions(time,current_state)
    integer :: j
    type(pebble_state), intent(in) :: current_state
    real(kind=d_t), intent(in) :: time
    real(kind=d_t) :: pot_energy
    pot_energy = 0.0
    do j=1,number_of_pebbles
       print '(A4,I6,3(1X,F20.7))',"#P: ",j,current_state%pp(j)%x,&
            current_state%pp(j)%y,current_state%pp(j)%z
       pot_energy = pot_energy + &
            mass*(-gravitational_acceleration)*(current_state%pp(j)%z - get_potential_energy_zero())
    end do
    print *,"#Grav_pot_energy ",time,pot_energy
    PRINT *,"#Displacement",get_earthquake_displacement(time)
    call print_displacement(time)

  end subroutine display_pebble_positions

  SUBROUTINE display_pebble_info(time,current_state) 
    !type(vector), intent(in) :: pp,pv(number_of_pebbles)
    !type(vector), intent(in) :: pw(number_of_pebbles)    
    type(pebble_state), intent(in) :: current_state
    real(kind=d_t), intent(in) :: time
    type(vector), pointer,dimension(:) :: pp,pv
    type(vector), pointer,dimension(:) :: pw
    integer :: j
    real(kind=d_t) :: rot_energy, lin_energy, pot_energy, static_energy
    real(kind=d_t) :: wall_static
    pp => current_state%pp
    pv => current_state%pv
    pw => current_state%pw
    rot_energy = 0.0
    lin_energy = 0.0
    pot_energy = 0.0
    static_energy = 0.0
    wall_static = 0.0
    DO j=1,number_of_pebbles
       PRINT '(A3,I6,9(1X,F20.7))',"P: ",j,pp(j)%x,pp(j)%y,pp(j)%z,pv(j)%x,pv(j)%y,pv(j)%z,pw(j)%x,pw(j)%y,pw(j)%z
       rot_energy = rot_energy + 0.5*moment_of_inertia*vmag_sqr(pw(j))
       lin_energy = lin_energy + 0.5*mass*vmag_sqr(pv(j))
       pot_energy = pot_energy + mass*(-gravitational_acceleration)*(pp(j)%z - get_potential_energy_zero())
       if(use_static_friction) then
          static_energy = static_energy + get_static_potential_energy(current_state,j)*2.0
          !print *,"#p_s",j,current_state%pss(j),current_state%psn(j)
          wall_static = wall_static + 0.5 * vmag_sqr(current_state%pss(j)) * &
               surface_static_friction_hooke
       end if
          
    END DO
    !32  FORMAT (A3,I5,9(1X,F20.7))
    if(use_static_friction) then 
       call display_pebble_slips(current_state)
    end if
    PRINT '(A9,F8.3,1X,6(A5,1X,F20.7,1X))',"#Energys ",time,&
         "lin",lin_energy,"rot",rot_energy,"pot",pot_energy,&
         "pstat",static_energy,"wstat",wall_static,"tot",&
         rot_energy+lin_energy+pot_energy+static_energy+wall_static
    PRINT *,"#Displacement",get_earthquake_displacement(time)
    call print_displacement(time)

  END SUBROUTINE display_pebble_info

  SUBROUTINE display_pebble_energy(time,current_state) 
    !type(vector), intent(in) :: pp,pv(number_of_pebbles)
    !type(vector), intent(in) :: pw(number_of_pebbles)    
    type(pebble_state), intent(in) :: current_state
    real(kind=d_t), intent(in) :: time
    type(vector), pointer,dimension(:) :: pp,pv
    type(vector), pointer,dimension(:) :: pw
    integer :: j
    real(kind=d_t) :: rot_energy, lin_energy, pot_energy, static_energy
    real(kind=d_t) :: wall_static
    pp => current_state%pp
    pv => current_state%pv
    pw => current_state%pw
    rot_energy = 0.0
    lin_energy = 0.0
    pot_energy = 0.0
    static_energy = 0.0
    wall_static = 0.0
    DO j=1,number_of_pebbles
       rot_energy = rot_energy + 0.5*moment_of_inertia*vmag_sqr(pw(j))
       lin_energy = lin_energy + 0.5*mass*vmag_sqr(pv(j))
       pot_energy = pot_energy + mass*(-gravitational_acceleration)*(pp(j)%z - get_potential_energy_zero())
       if(use_static_friction) then
          static_energy = static_energy + get_static_potential_energy(current_state,j)*2.0          
          wall_static = wall_static + 0.5 * vmag_sqr(current_state%pss(j)) * &
               surface_static_friction_hooke
       end if
          
    END DO
    !32  FORMAT (A3,I5,9(1X,F20.7))
    PRINT '(A9,F8.3,1X,6(A5,1X,F20.7,1X))',"#Energys ",time,&
         "lin",lin_energy,"rot",rot_energy,"pot",pot_energy,&
         "pstat",static_energy,"wstat",wall_static,"tot",&
         rot_energy+lin_energy+pot_energy+static_energy+wall_static
    PRINT *,"#Displacement",get_earthquake_displacement(time)
    call print_displacement(time)

  END SUBROUTINE display_pebble_energy


  integer function pebble_slip_size(current_state, pebble) 
    integer, intent(in) :: pebble
    type(pebble_state),intent(in) :: current_state
    integer :: j, length
    length = 0
    do j=1,slip_zone_size
       if(current_state%pps_index(j,pebble) == 0) then
          exit
       end if
       length = length + 1
    end do
    pebble_slip_size = length
  end function pebble_slip_size

  subroutine display_pebble_slips(current_state)
    type(pebble_state), intent(in) :: current_state
    integer :: i, j, length
    do i=1,number_of_pebbles
       length = 0
       do j=1,slip_zone_size
          if(current_state%pps_index(j,i) == 0) then             
             exit
          end if
          length = length + 1
       end do
       if ( length > 0) then
          print *," #P_slip_start",i,length
          do j=1,length
             print '(A15,I7,3E25.16E3)',"#PP_slip       ",&
                  current_state%pps_index(j,i),current_state%ppsd(j,i)%x,&
                  current_state%ppsd(j,i)%y,current_state%ppsd(j,i)%z
          end do
       end if
    end do
    ! Termination data.
    print *," #P_slip_start",0,0
    do i=1,number_of_pebbles
       print '(A16,I7,6E25.16E3)',"#P_surface_slip ",i,&
            current_state%pss(i)%x,current_state%pss(i)%y,&
            current_state%pss(i)%z,current_state%psn(i)%x,&
            current_state%psn(i)%y,current_state%psn(i)%z
    end do
  end subroutine display_pebble_slips


  SUBROUTINE initialize_pebble(pos,velocity,omega,seed, &
       &     bottom,height_needed,pebble_radius, &
       &     fill_edge)
    type(vector), intent(out) :: pos, velocity,omega
    type(vector) :: force, torque, slip, slip_normal
    REAL(KIND=D_T), intent(in) :: bottom,height_needed,pebble_radius  
    REAL(KIND=D_T), intent(in) ::  fill_edge
    INTEGER(KIND=LI), intent(inout) :: seed  
    real(kind=d_t) :: dummy_current_wear
    dummy_current_wear = 0.0
    !Initialize linear and angular velocities to 0
    velocity = (/ 0.0,0.0,0.0 /)
    omega = (/ 0.0,0.0,0.0 /)
    slip = (/ 0.0,0.0,0.0 /)
    slip_normal = (/ 0.0,0.0,0.0 /)
    pos%z = random(seed)*(1.0/initial_packing)*height_needed + bottom &
         &     + pebble_radius
9   CONTINUE  
    pos%x = (random(seed)-0.5)*2.0*fill_edge
    pos%y = (random(seed)-0.5)*2.0*fill_edge

    !     Check to make sure not inside something.
    call base_force(pos, velocity, omega,force, torque, slip, slip_normal, &
         0.0_d_t, dummy_current_wear)
    IF (force%x .ne. 0.0 .or. force%y .ne. 0.0) GOTO 9  

  END SUBROUTINE initialize_pebble

  subroutine combine_step(number_of_pebbles, in_state, out_state, &
       delta, alpha, time)
    type(pebble_state), intent(in) :: in_state
    type(pebble_state), intent(out) :: out_state
    type(pebble_delta), intent(in) :: delta
    real(kind=d_t), intent(in) :: alpha, time
    integer, intent(in) :: number_of_pebbles
    integer :: j
    
    !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size)
    do j=1,number_of_pebbles
       out_state%pp(j) = in_state%pp(j) + alpha * delta%pdp(j)
       out_state%pv(j) = in_state%pv(j) + alpha * delta%pdv(j)
       out_state%pw(j) = in_state%pw(j) + alpha * delta%pdw(j)
    end do
    if(use_static_friction) then
       call update_slips(in_state,out_state,alpha,time)
    end if

  end subroutine combine_step

!!$  SUBROUTINE wall_force()  
!!$    REAL(KIND=D_T) xf,yf,zf, y  
!!$
!!$    INTEGER i  
!!$    DO i=1,1000  
!!$       y = REAL(i)*0.01 - 5.0  
!!$       CALL base_force(0.0D0,y,0.0D0,0.0D0,0.0D0,0.0D0,xf,yf,zf)  
!!$       PRINT *,y,xf,yf,zf  
!!$
!!$    END DO
!!$  END SUBROUTINE wall_force

  subroutine clear_pps_index(pps_index) 
    integer,pointer,dimension(:,:) :: pps_index
    !integer i,j
    ! This is faster.
    pps_index = 0
    return
    !do i=1,number_of_pebbles
    !   do j=1,slip_zone_size
    !      if(pps_index(i,j) /= 0) then
    !         pps_index(i,j) = 0
    !      else 
    !         exit
    !      end if
    !   end do
    !end do
  end subroutine clear_pps_index


  subroutine get_derivs(nearby_pebble_cache, state, delta, &
       number_of_pebbles, nearby_pebbles_array, nearby_pebbles_size, time)
    type(pebble_state), intent(in) :: state
    type(pebble_delta), intent(out) :: delta
    type(nearby_pebble_cache_type),intent(inout) :: nearby_pebble_cache
    INTEGER, intent(in) :: number_of_pebbles  
    INTEGER, intent(in) ::  nearby_pebbles_size  
    INTEGER, intent(inout) :: nearby_pebbles_array(nearby_pebbles_size)  
    type(vector),dimension(:),pointer :: pp, pv,pw
    type(vector),dimension(:),pointer :: opf, opv, opn
    type(vector),dimension(:,:),pointer :: ppsd
    type(vector),dimension(:),pointer :: pss
    type(vector),dimension(:),pointer :: psn
    type(vector) :: old_slip
    integer,dimension(:,:),pointer :: pps_index
    real(kind=d_t), intent(in) :: time
    LOGICAL :: changed
    REAL(kind=d_t) :: current_wear
    !LOGICAL :: in_old_static_friction
    type(vector) :: temp_opf_j, temp_opn_j,temp_opf_k,temp_opn_k

    INTEGER :: l,j,k  
    INTEGER :: nearby_pebbles_count  

    pp => state%pp
    pv => state%pv
    pw => state%pw
    !if(use_static_friction) then
    !Just copy the null.  gfortran gets confused if you don't do this everytime
    ppsd => state%ppsd
    pps_index => state%pps_index
    pss => state%pss
    psn => state%psn
    !else

    opf => delta%pdv
    opv => delta%pdp
    opn => delta%pdw

    !call print_time("#filling_zones")
    call update_nearby_pebble_cache(nearby_pebble_cache,pp)
    !call print_time("#calculating_timesteps")
    !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size) PRIVATE(current_wear)
    DO j=1,number_of_pebbles  
       current_wear = 0.0
       !         PRINT *,"#bases",j,px(j),py(j),pz(j),pvx(j),pvy(j),pvz(j)
       if(use_static_friction) then
          call base_force(pp(j),pv(j),pw(j),opf(j),opn(j),pss(j),psn(j),time,current_wear)
       else 
          call base_force(pp(j),pv(j),pw(j),opf(j),opn(j),vector(0.0,0.0,0.0),&
               vector(0.0,0.0,0.0),time,current_wear)
       end if
       if(calculate_pebble_wear_vol) then
          !XXX Should pass in a correct length
          call tally_surface_wear(pp(j),current_wear,0.0_d_t)
       end if
       if(is_pebble_of_interest(j)) then
          print *,"#PBF",j,"f",opf(j),"n",opn(j)
       end if
       call pebble_air_resistance(pv(j),opf(j))
       !         PRINT *,"#base",j,px(j),py(j),pz(j),opfx(j),opfy(j),opfz(j)
       opv(j) = pv(j)
    end do
    !$OMP PARALLEL DO SHARED(opn,opf) &
    !$OMP PRIVATE(nearby_pebbles_array,nearby_pebbles_count,temp_opf_j,temp_opf_k,temp_opn_j,temp_opn_k,changed,l,k,old_slip) &
    !$OMP SCHEDULE(dynamic,chunk_size)
    do j=1,number_of_pebbles  
       !     Now find the pebbles that are nearby (aka also in same zone)
       call nearby_pebbles(nearby_pebble_cache,pp(j), &
            &        nearby_pebbles_array, nearby_pebbles_size, &
            &        nearby_pebbles_count, j)
       DO l=1,nearby_pebbles_count  
          k = nearby_pebbles_array(l)  
          if (k < j) then 
             old_slip = get_old_slip(ppsd,pps_index,j,k)
             call pebble_pebble_velocity_force3(pp(j),pp(k),temp_opf_j,pv(j),&
                  pv(k),pw(j),pw(k),temp_opn_j,temp_opf_k,temp_opn_k,old_slip,&
                  changed)
             if(changed) then
                !If changed, then update forces.
                call atomic_increment(opf(j),temp_opf_j)
                call atomic_increment(opf(k),temp_opf_k)
                if(calculate_force_tally) then
                   call tally_pp_force(temp_opf_j,temp_opf_k,pp(j),pp(k))
                end if
                !!$OMP CRITICAL (opf_crit)
                !!opf(j) = opf(j) + temp_opf_j
                !!opf(k) = opf(k) + temp_opf_k
                !!$OMP END CRITICAL (opf_crit)
                call atomic_increment(opn(j),temp_opn_j)
                call atomic_increment(opn(k),temp_opn_k)
                !!$OMP CRITICAL (opn_crit)
                !!opn(j) = opn(j) + temp_opn_j
                !!opn(k) = opn(k) + temp_opn_k
                !!$OMP END CRITICAL (opn_crit)
                
!!$                if(is_pebble_of_interest(j)) then
!!$                   print *,"#PPF",j,k,"f",temp_opf_j,"n",temp_opn_j
!!$                end if
!!$                if(is_pebble_of_interest(k)) then
!!$                   print *,"#PPF",k,j,"f",temp_opf_k,"n",temp_opn_k
!!$                end if
                
             end if
          end if
       END DO
    END DO
    !call print_time("#converting to acceleration")
    !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size)
    DO j=1,number_of_pebbles
       !Convert force to acceleration
       opf(j) = (1.0/mass)*opf(j)
       !Convert torque to angular acceleration
       if(torque_enable) then 
          opn(j) = (1.0/moment_of_inertia) * opn(j)
       else 
          opn(j) = vector(0.0,0.0,0.0)
       end if
       if(calculate_force_tally) then
          call tally_force(opf(j),pp,j)
       end if
    END DO
    
  END SUBROUTINE get_derivs
  
!!$  SUBROUTINE pebble_test()  
!!$
!!$    REAL(KIND=D_T) xf,yf,zf  
!!$    CALL base_force(1.0D0,1.0D0,1.0D0,0.0D0,0.0D0,0.0D0,xf,yf,zf)  
!!$
!!$    PRINT *,"Empty Place",xf,yf,zf  
!!$    CALL base_force(1.0D0,0.0D0,1.0D0,0.0D0,0.0D0,0.0D0,xf,yf,zf)  
!!$
!!$    PRINT *,"Inner core",xf,yf,zf  
!!$    CALL base_force(1.05D0,0.0D0,1.0D0,0.0D0,0.0D0,0.0D0,xf,yf,zf)  
!!$
!!$    PRINT *,"Surface Inner core",xf,yf,zf  
!!$    CALL base_force(1.0D0,1.0D0,-4.0D0,0.0D0,0.0D0,0.0D0,xf,yf,zf)  
!!$
!!$    PRINT *,"Floor core",xf,yf,zf  
!!$    CALL base_force(1.0D0,1.0D0,-3.95D0,0.0D0,0.0D0,0.0D0,xf,yf,zf)  
!!$
!!$    PRINT *,"Surface floor core",xf,yf,zf  
!!$    CALL base_force(5.0D0,0.0D0,1.0D0,0.0D0,0.0D0,0.0D0,xf,yf,zf)  
!!$
!!$    PRINT *,"Outer core",xf,yf,zf  
!!$    CALL base_force(4.95D0,0.0D0,1.0D0,0.0D0,0.0D0,0.0D0,xf,yf,zf)  
!!$
!!$    PRINT *,"Surface Outer core",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 0.0  
!!$    zf = 0.0  
!!$    CALL pebble_pebble_force(1.0D0,1.0D0,1.0D0,2.0D0,2.0D0,2.0D0, &
!!$         &     xf,yf,zf)
!!$
!!$    PRINT *,"Pebble pebble",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 0.0  
!!$    zf = 0.0  
!!$    CALL pebble_pebble_force(1.0D0,1.0D0,1.0D0,1.0D0,1.0D0,1.05D0, &
!!$         &     xf,yf,zf)
!!$
!!$    PRINT *,"Pebble pebble close",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 0.0  
!!$    zf = 0.0  
!!$    CALL pebble_pebble_force(1.0D0,1.0D0,1.0D0,1.0D0,1.0D0,1.0D0, &
!!$         &     xf,yf,zf)
!!$
!!$    PRINT *,"Pebble pebble same",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 0.0  
!!$    zf = 0.0  
!!$    CALL pebble_pebble_velocity_force(1.0D0,1.0D0,1.0D0, &
!!$         &     1.0D0,1.0D0,1.05D0,xf,yf,zf,0.0D0,0.0D0,10.0D0, &
!!$         &     0.0D0,0.0D0,0.0D0)
!!$
!!$    PRINT *,"Pebble pebble close + velocity",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 0.0  
!!$    zf = 0.0  
!!$    CALL pebble_pebble_velocity_force(1.0D0,1.0D0,1.0D0, &
!!$         &     1.0D0,1.0D0,1.05D0,xf,yf,zf,0.0D0,0.0D0,-10.0D0, &
!!$         &     0.0D0,0.0D0,0.0D0)
!!$
!!$    PRINT *,"Pebble pebble close - velocity",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 0.0  
!!$    zf = 0.0  
!!$    CALL pebble_pebble_velocity_force(1.0D0,1.0D0,1.05D0, &
!!$         &     1.0D0,1.0D0,1.0D0,xf,yf,zf,0.0D0,0.0D0,10.0D0, &
!!$         &     0.0D0,0.0D0,0.0D0)
!!$
!!$    PRINT *,"Other Pebble pebble close + velocity",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 0.0  
!!$    zf = 0.0  
!!$    CALL pebble_pebble_velocity_force(1.0D0,1.0D0,1.05D0, &
!!$         &     1.0D0,1.0D0,1.0D0,xf,yf,zf,0.0D0,0.0D0,-10.0D0, &
!!$         &     0.0D0,0.0D0,0.0D0)
!!$
!!$    PRINT *,"Other Pebble pebble close - velocity",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 0.0  
!!$    zf = 0.0  
!!$    CALL pebble_pebble_velocity_force(1.0D0,1.0D0,1.0D0, &
!!$         &     1.0D0,1.0D0,1.05D0,xf,yf,zf,0.0D0,10.0D0,0.0D0, &
!!$         &     0.0D0,0.0D0,0.0D0)
!!$
!!$    PRINT *,"Pebble pebble close traverse velocity",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 0.0  
!!$    zf = 0.0  
!!$    CALL pebble_pebble_velocity_force(1.0D0,1.0D0,1.0D0, &
!!$         &     1.0D0,1.0D0,1.05D0,xf,yf,zf,0.0D0,-10.0D0,0.0D0, &
!!$         &     0.0D0,0.0D0,0.0D0)
!!$
!!$
!!$    PRINT *,"Pebble pebble close opp traverse velocity",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 0.0  
!!$    zf = 0.0  
!!$    CALL pebble_pebble_velocity_force(1.0D0,1.0D0,1.0D0, &
!!$         &     1.0D0,1.0D0,1.05D0,xf,yf,zf,0.0D0,10.0D0,10.0D0, &
!!$         &     0.0D0,0.0D0,0.0D0)
!!$
!!$    PRINT *,"Pebble pebble close both velocity",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 0.0  
!!$    zf = 0.0  
!!$    CALL pebble_pebble_velocity_force(1.0D0,1.0D0,1.0D0, &
!!$         &     1.0D0,1.0D0,1.05D0,xf,yf,zf,0.0D0,-10.0D0,-10.0D0, &
!!$         &     0.0D0,0.0D0,0.0D0)
!!$
!!$    PRINT *,"Pebble pebble close opp both velocity",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 0.0  
!!$    zf = 1.0  
!!$    CALL pebble_pebble_friction_force(1.0D0,1.0D0,1.0D0, &
!!$         &     1.0D0,1.0D0,1.05D0,xf,yf,zf)
!!$
!!$    PRINT *,"Pebble pebble friction",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 1.0  
!!$    zf = 0.0  
!!$    CALL pebble_pebble_friction_force(1.0D0,1.0D0,1.0D0, &
!!$         &     1.0D0,1.0D0,1.05D0,xf,yf,zf)
!!$
!!$    PRINT *,"Pebble pebble friction traverse",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 1.0  
!!$    zf = 1.0  
!!$    CALL pebble_pebble_friction_force(1.0D0,1.0D0,1.0D0, &
!!$         &     1.0D0,1.0D0,1.05D0,xf,yf,zf)
!!$
!!$    PRINT *,"Pebble pebble friction both",xf,yf,zf  
!!$    xf = 0.0  
!!$    yf = 1.0  
!!$    zf = 10.0  
!!$    CALL pebble_pebble_friction_force(1.0D0,1.0D0,1.0D0, &
!!$         &     1.0D0,1.0D0,1.05D0,xf,yf,zf)
!!$
!!$    PRINT *,"Pebble pebble friction both more normal",xf,yf,zf  
!!$
!!$  END SUBROUTINE pebble_test


  INTEGER FUNCTION find_base_zone(low, increment, size, value)  
    REAL(KIND=D_T), intent(in) :: low,increment,value  
    INTEGER, intent(in) :: size
    INTEGER :: zone  
    zone = find_zone(low, increment, size, value)  
    find_base_zone = iclamp(2,zone,size-1)  
  END FUNCTION find_base_zone
  
  REAL(kind=d_t) function min_zone(low, increment, bin)
    real(kind=d_t), intent(in) :: low, increment
    integer, intent(in) :: bin
    min_zone = low + increment * bin
  end function min_zone

  real(kind=d_t) function max_zone(low, increment, bin)
    real(kind=d_t), intent(in) :: low, increment
    integer, intent(in) :: bin
    max_zone = low + increment * (bin + 1)
  end function max_zone

  INTEGER FUNCTION find_zone(low, increment, size, value)  
    REAL(KIND=D_T), intent(in) :: low,increment,value  
    INTEGER, intent(in) :: size 
    INTEGER :: bin  
    bin = FLOOR((value - low) / increment)  
    find_zone = iclamp(1,bin,size)  
  END FUNCTION find_zone
  
  real(kind=d_t) function zone_center(low, increment, bin)
    REAL(KIND=D_T), intent(in) :: low,increment
    integer, intent(in) :: bin
    zone_center = low+increment * (bin + 0.5)
  end function zone_center


  subroutine find_radial_zone(x,y,z,s_z,r_z,z_z)
    REAL(kind=d_t), intent(in) :: x,y,z
    integer, intent(out) :: s_z,r_z,z_z
    REAL(kind=d_t) :: r, theta
    z_z = find_zone(z_low,tally_z_inc, zt_size, z)
    r = sqrt(x**2+y**2)
    theta = atan2(y,x)
    s_z = FLOOR(((theta + pi) / (2* pi)) * tally_sections) + 1
    r_z = find_zone(0.0_d_t,tally_radial_inc,rt_size, r)
  end subroutine find_radial_zone

  subroutine radial_zone_center(s_z,r_z,z_z,x,y,z) 
    INTEGER, intent(in) :: s_z, r_z, z_z
    REAL(kind=d_t), intent(out) :: x, y, z
    REAL(kind=d_t) :: theta, r
    theta = (real(s_z-1) / real(tally_sections) + 0.5) * 2 * pi
    r = (real(r_z - 1) + 0.5) * tally_radial_inc
    z = z_low+ tally_z_inc * (z_z + 0.5)
    x = cos(theta) * r
    y = sin(theta) * r
  end subroutine radial_zone_center

  subroutine regularize_radial_velocity(pos,velocity,radial_velocity,&
       s_z,r_z,z_z) 
    integer, intent(in) :: s_z, r_z, z_z
    type(vector), intent(in) :: pos, velocity
    type(vector), intent(out) :: radial_velocity
    real(kind=d_t) :: theta, dr,  dtheta, ctheta
    ! x plus velocity, y plus velocity
    type(vector) :: r_hat, theta_hat, c_r_hat, c_theta_hat, z_hat
    real(kind=d_t) :: cx,cy,cz
    theta = atan2(pos%y,pos%x)
    z_hat = vector(0.0,0.0,1.0)
    r_hat = vector(cos(theta),sin(theta),0.0)
    theta_hat = vector(-sin(theta),cos(theta),0.0)
    dr = velocity .dot. r_hat
    dtheta = velocity .dot. theta_hat
    !call find_radial_zone(pos%x,pos%y,pos%z,s_z,r_z,z_z)
    call radial_zone_center(s_z,r_z,z_z,cx,cy,cz)
    ctheta = atan2(cy, cx)
    c_r_hat = vector(cos(ctheta),sin(ctheta),0.0)
    c_theta_hat = vector(-sin(ctheta),cos(ctheta),0.0)
    radial_velocity = dr *  c_r_hat + dtheta * c_theta_hat + velocity%z * z_hat
  end subroutine regularize_radial_velocity

  subroutine random_point_in_sphere(x_cent,y_cent,z_cent,radius,x,y,z,seed)
    !Returns a random point somewhere in the sphere.
    integer(kind=li), intent(inout) :: seed
    real(kind=d_t), intent(in) :: x_cent,y_cent,z_cent,radius
    real(kind=d_t), intent(out) :: x,y,z
    do 
       x = (random(seed)-0.5)*2.0*radius
       y = (random(seed)-0.5)*2.0*radius
       z = (random(seed)-0.5)*2.0*radius
       if (x**2 + y**2 + z**2 <= radius**2) then
          ! found something inside sphere, so now
          ! put in terms of real coordinates and return.
          x = x + x_cent
          y = y + y_cent
          z = z + z_cent
          return
       end if
    end do
  end subroutine random_point_in_sphere

  REAL(KIND=D_T) FUNCTION RANDOM(L)  
    INTEGER(KIND=LI), intent(inout) ::  L      
    L=MOD((7**5)*L,2147483647_LI)  
    RANDOM=REAL(L)*(1.0/REAL(2147483647))  
    RETURN  
  END FUNCTION RANDOM


!!$  type(vector) FUNCTION vclamp(vec,max) 
!!$    type(vector), intent(in) :: vec
!!$    real(kind=d_t), intent(in) :: max
!!$    if(vmag_sqr(vec) <= max**2) then
!!$       vclamp = vec 
!!$       return
!!$    else 
!!$       vclamp = (max/abs(vec)) * vec
!!$    end if
!!$  end FUNCTION vclamp
!!$
!!$  type(vector) FUNCTION vdamp(vec,max,factor) 
!!$    type(vector), intent(in) :: vec
!!$    real(kind=d_t), intent(in) :: max, factor
!!$    real(kind=d_t) :: length
!!$    if(vmag_sqr(vec) <= max**2) then
!!$       vdamp = vec 
!!$       return
!!$    else 
!!$       length = abs(vec)
!!$       vdamp = (((length-max)*factor+max)/abs(vec)) * vec
!!$    end if
!!$  end FUNCTION vdamp
  
    

!!$  pure REAL(KIND=D_T) FUNCTION dot(x,y,z,other_x,other_y,other_z)  
!!$    REAL(KIND=D_T), intent(in) :: x,y,z,other_x,other_y,other_z  
!!$    dot = x*other_x+y*other_y+z*other_z  
!!$  END FUNCTION dot
!!$
!!$  pure REAL(KIND=D_T) FUNCTION mag_sqr(x,y,z)  
!!$    REAL(KIND=D_T), intent(in) :: x,y,z  
!!$    mag_sqr = x**2+y**2+z**2  
!!$  END FUNCTION mag_sqr
!!$  
!!$  REAL(KIND=D_T) FUNCTION mag(x,y,z)  
!!$    REAL(KIND=D_T), intent(in) :: x,y,z
!!$    mag = SQRT(mag_sqr(x,y,z))  
!!$  END FUNCTION mag

  function any_pebbles_of_interest()
    logical :: any_pebbles_of_interest
    any_pebbles_of_interest = pebble_of_interest > 0
  end function any_pebbles_of_interest

  function is_pebble_of_interest(pebble_number) 
    integer, intent(in) :: pebble_number
    logical :: is_pebble_of_interest
    if(pebble_number == pebble_of_interest) then
       is_pebble_of_interest = .true.
    else 
       is_pebble_of_interest = .false.
    end if
  end function is_pebble_of_interest

  pure FUNCTION pebble_pebble_overlap_dist(p,op,pebble_radius)
    REAL(KIND=D_T) :: pebble_pebble_overlap_dist
    type(vector), intent(in) :: p,op
    REAL(KIND=D_T), intent(in) :: pebble_radius
    REAL(KIND=D_T) :: dist_sqr, boundry_sqr  
    boundry_sqr = (2.0*pebble_radius)**2  
    dist_sqr = vmag_sqr(p - op)  
    pebble_pebble_overlap_dist = SQRT(boundry_sqr) - SQRT(dist_sqr)  
  END FUNCTION pebble_pebble_overlap_dist
  
  pure LOGICAL FUNCTION pebble_pebble_overlap(a,b, pebble_radius)
    REAL(KIND=D_T) :: dist_sqr 
    REAL(KIND=D_T), intent(in) :: pebble_radius  
    type(Vector), intent(in) :: a,b
    dist_sqr = vmag_sqr(a - b)
    if ( dist_sqr .ge. (2.0*pebble_radius)**2) THEN  
       pebble_pebble_overlap = .FALSE.  
    ELSE  
       pebble_pebble_overlap = .TRUE.
    END IF
  END FUNCTION pebble_pebble_overlap
    

!!$  pure SUBROUTINE projection(x,y,z,onto_x,onto_y,onto_z, &
!!$       &     out_x,out_y,out_z)
!!$    REAL(KIND=D_T), intent(in) ::  x,y,z,onto_x,onto_y,onto_z
!!$    REAL(KIND=D_T), intent(out) :: out_x,out_y,out_z  
!!$    REAL(KIND=D_T) :: mult
!!$    mult = dot(x,y,z,onto_x,onto_y,onto_z)/ &
!!$         &     mag_sqr(onto_x,onto_y,onto_z)
!!$    out_x = onto_x*mult  
!!$    out_y = onto_y*mult  
!!$    out_z = onto_z*mult  
!!$  END SUBROUTINE projection

  pure SUBROUTINE pebble_air_resistance(v,f)  
    type(vector), intent(in) :: v
    type(vector), intent(inout) :: f
    REAL(KIND=D_T) :: v_mag_sqr  
    v_mag_sqr = vmag_sqr(v)  
    f = f - air_friction*v_mag_sqr*v
  END SUBROUTINE pebble_air_resistance

  function get_static_potential_energy(current_state,i)
    !Note, ignores half the potential since it is only indexed by one 
    ! direction
    type(pebble_state), intent(in) :: current_state
    integer, intent(in) :: i
    real(kind=d_t) :: total
    real(kind=d_t) :: get_static_potential_energy
    integer :: j
    total = 0.0
    if(.not. use_static_friction) then
       get_static_potential_energy = 0.0       
       return
    end if
    do j=1,slip_zone_size
       if (current_state%pps_index(j,i) /= 0) then 
          total = total + vmag_sqr(current_state%ppsd(j,i)) * &
               static_friction_hooke / 2.0
       else
          get_static_potential_energy = total
          return
       end if
    end do
    get_static_potential_energy = total
  end function get_static_potential_energy

  pure function make_perpendicular(orig, normal) 
    type(vector), intent(in) :: orig, normal
    type(vector) :: make_perpendicular
    make_perpendicular = abs(orig) * normalize(perp_projection(orig,normal))
  end function make_perpendicular

  pure subroutine pebble_surface_slip_update(pp,time,old_psn,v,o,real_old_slip,&
       delta,new_slip,surface_normal)
    real(kind=d_t),intent(in) :: time, delta
    type(vector),intent(in) :: pp,old_psn,v,o,real_old_slip
    type(vector),intent(out) :: new_slip, surface_normal
    type(vector) :: surface_velocity, old_slip
    type(vector) :: cv
    logical :: in_contact
    real(kind=d_t) :: dummy_current_wear
    dummy_current_wear = 0.0
    call pebble_surface_normal(pp,surface_normal,&
         in_contact,time, dummy_current_wear)
    if(in_contact .and. &
         (surface_normal .dot. old_psn) > 0.5) then
       cv = v + (o .cross. ((-pebble_radius) * surface_normal))
       old_slip = real_old_slip
       if(zero_static_slips) then
          old_slip = (/ 0.0, 0.0, 0.0 /)
       else if(vmag_sqr(old_slip) > 1.0e-20_d_t) then
          old_slip = make_perpendicular(old_slip,surface_normal)
       else 
          old_slip = (/ 0.0, 0.0, 0.0 /)
       end if
       if(zero_long_slips .or. shorten_long_slips) then
          if(vmag_sqr(old_slip) > long_slip_sqr) then
             if(zero_long_slips) then
                old_slip = (/ 0.0, 0.0, 0.0 /)
             else 
                old_slip = long_slip*normalize(old_slip)
             end if
          end if
       end if
       surface_velocity = cv - (cv .proj. surface_normal)
       if(vmag_sqr(surface_velocity) < static_velocity_sqr_max) then
          new_slip = old_slip + delta * surface_velocity
       else
          new_slip = (/ 0.0, 0.0, 0.0 /)
       end if
       !print *,"#j",j,"ns",new_slip,"os",old_slip,"sv",surface_velocity,&
       !     "sn",surface_normal, "v", v, "cv", cv
    else
       new_slip = (/ 0.0, 0.0, 0.0 /)
    end if
  end subroutine pebble_surface_slip_update

  subroutine update_slips(prev_state, current_state, delta, time) 
    type(pebble_state), intent(in) :: prev_state
    type(pebble_state), intent(inout) :: current_state
    real(kind=d_t), intent(in) :: delta, time
    integer :: j, l, k, slip_index
    type(vector) :: new_slip, old_slip
    logical :: non_zero_slip
    !real(kind=d_t) :: total_slip
    !total_slip = 0.0
    call clear_pps_index(current_state%pps_index)
    call update_nearby_pebble_cache(nearby_pebble_cache,current_state%pp)

    !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size)
    do j=1,number_of_pebbles
       call pebble_surface_slip_update(current_state%pp(j),time,&
            prev_state%psn(j),current_state%pv(j),current_state%pw(j),&
            prev_state%pss(j),delta,current_state%pss(j),current_state%psn(j))
    end do
    !$OMP PARALLEL DO PRIVATE(near_pebbles,near_pebbles_count,l,k,old_slip,&
    !$OMP non_zero_slip,slip_index,new_slip) SCHEDULE(dynamic,chunk_size)
    do j=1,number_of_pebbles

       call nearby_pebbles(nearby_pebble_cache, current_state%pp(j),&
            near_pebbles, near_pebbles_size, near_pebbles_count, j)
       do l=1,near_pebbles_count
          k = near_pebbles(l)
          if(j < k) then
             if(zero_static_slips) then
                old_slip = (/ 0.0, 0.0, 0.0 /)
             else
                old_slip = get_old_slip(prev_state%ppsd,prev_state%pps_index,&
                     j,k)
             end if
             call get_new_slip(current_state%pp(j),current_state%pp(k),&
                  current_state%pv(j),current_state%pv(k),&
                  current_state%pw(j),current_state%pw(k),&
                  old_slip,delta, new_slip, non_zero_slip)
             if(non_zero_slip) then 
                !print *,"nslip",new_slip,"oslip",old_slip
                slip_index = get_empty_slip_index(current_state%pps_index, j)
                !$OMP CRITICAL (get_slip_index)
                if(current_state%pps_index(slip_index,j) /= 0)then
                   slip_index = get_empty_slip_index(current_state%pps_index, j)
                end if
                current_state%pps_index(slip_index,j) = k
                !total_slip = total_slip + abs(new_slip)
                !$OMP END CRITICAL (get_slip_index)
                current_state%ppsd(slip_index,j) = new_slip

             end if
          end if
       end do
    end do
    !print *,"#total_slip",total_slip
    
  end subroutine update_slips

  function get_empty_slip_index(pps_index, from) 
    integer,dimension(:,:),intent(in) :: pps_index
    integer,intent(in) :: from
    integer :: get_empty_slip_index
    integer :: i
    
    do i=1,slip_zone_size
       if(pps_index(i,from) == 0) then
          get_empty_slip_index = i
          return 
       end if
    end do
    print *,"ERROR Overfull slip index",from
    get_empty_slip_index = 1
  end function get_empty_slip_index

  pure function get_slip_index(pps_index,from_in,to_in) 
    integer,dimension(:,:),intent(in) :: pps_index
    integer,intent(in) :: from_in,to_in
    integer :: get_slip_index
    integer :: i, from, to
    if(.not. use_static_friction) then 
       get_slip_index = 0
       return
    end if
    if(to_in < from_in) then
       ! Swap so that from is always lower than to.
       from = to_in
       to = from_in
    else
       from = from_in
       to = to_in
    end if
    do i=1,slip_zone_size
       if(pps_index(i,from) == to) then
          get_slip_index = i
          return
       end if
    end do

    get_slip_index = 0
  end function get_slip_index


  pure function get_old_slip(ppsd,pps_index,from_in,to_in) 
    type(vector),dimension(:,:),intent(in) :: ppsd
    integer,dimension(:,:),intent(in) :: pps_index
    integer,intent(in) :: from_in,to_in    
    type(vector) :: get_old_slip
    integer :: i, from, to
    !XXX should the vector be the opposite direction if these are flipped?
    if(to_in < from_in) then
       ! Swap so that from is always lower than to.
       from = to_in
       to = from_in
    else
       from = from_in
       to = to_in
    end if

    i = get_slip_index(pps_index,from,to)
    if(i /= 0) then
       get_old_slip = ppsd(i,from)
       return
    end if

    get_old_slip = (/ 0.0,0.0,0.0 /)
    
  end function get_old_slip
  
  subroutine get_new_slip(pos, other_pos, v, other_v, omega, other_omega, &
       old_slip, delta, new_slip, non_zero_slip) 
    type(vector), intent(in) :: v, other_v
    type(vector) :: r_v
    type(vector), intent(in) :: pos, other_pos
    type(vector), intent(in) :: omega, other_omega
    type(vector), intent(in) :: old_slip
    real(kind=d_t), intent(in) :: delta
    type(vector), intent(out) :: new_slip
    logical, intent(out) :: non_zero_slip
    type(vector) :: delta_pos
    type(vector) :: v_normal, v_transverse
    type(vector) :: adjusted_slip
    real(kind=d_t) :: old_slip_mag
    type(vector) :: normal   
    delta_pos = pos - other_pos
    ! r_v is direction point on main pebble is moving.
    r_v = v - other_v - &
         ((pebble_radius * (omega + other_omega)) .cross. &
         ((1.0/(2.0*pebble_radius)) * delta_pos))
    normal = normalize((-1.0) * delta_pos)
    v_normal = r_v .proj. ((-1.0)*delta_pos)
    v_transverse = r_v - v_normal     
    if(vmag_sqr(v_transverse) < static_velocity_sqr_max .and. &
         vmag_sqr(delta_pos) < ((2.0*pebble_radius)**2)) then
       old_slip_mag = abs(old_slip)
       if(zero_long_slips .and. old_slip_mag > long_slip) then
          adjusted_slip = (/ 0.0, 0.0, 0.0 /)
       else if(shorten_long_slips .and. old_slip_mag > long_slip) then
          adjusted_slip = long_slip * &
               normalize(perp_projection(old_slip,normal))
       else if(old_slip_mag > 1.0e-20_d_t) then
          adjusted_slip = old_slip_mag * &
               normalize(perp_projection(old_slip,normal))
       else
          adjusted_slip = (/ 0.0, 0.0, 0.0 /)
       end if
       ! the -1 is since this is the slip it is going the opposite direction.
       new_slip = adjusted_slip + delta * (-1.0) * v_transverse
       non_zero_slip = .true.
    else 
       new_slip = (/ 0.0, 0.0, 0.0 /)
       non_zero_slip = .false.
    end if
    
  end subroutine get_new_slip

  SUBROUTINE pebble_pebble_velocity_force3(pos, other_pos,f,v,other_v, &
       omega, other_omega, torque, other_f, other_torque, old_slip, changed)
    REAL(KIND=D_T) :: two_pebble_radius_sqr,cent_dist_sqr  
    REAL(KIND=D_T) :: inside_dist, cent_dist, raw_force  
    REAL(KIND=D_T) :: fn_mag,vt_mag,ft_mag
    ! v is velocity, other_v is the other vectors velocity, and 
    ! r_v is the relative velocity of the two vectors
    type(vector), intent(in) :: v, other_v
    type(vector) :: r_v
    type(vector), intent(in) :: pos, other_pos
    type(vector), intent(in) :: omega, other_omega
    type(vector), intent(in) :: old_slip
    type(vector) :: delta_pos
    type(vector) :: v_normal, v_transverse
    type(vector) :: f_normal, f_transverse
    type(vector), intent(out) :: f, other_f  
    type(vector), intent(out) :: torque, other_torque
    logical,intent(out) :: changed 

    real(kind=d_t) :: static_mag, slip_mag
    type(vector) :: adjusted_slip
    type(vector) :: normal
    type(vector) :: static_friction_force
    !PRINT *,"#In pebble pebble velocity force"
    changed = .false.

    two_pebble_radius_sqr = 1.5*((2.0*pebble_radius)**2)  
    delta_pos = pos - other_pos
    normal = normalize((-1.0) * delta_pos)
    r_v = v - other_v - &
         ((pebble_radius * (omega + other_omega)) .cross. &
         ((1.0/(2.0*pebble_radius)) * delta_pos))
    cent_dist_sqr = vmag_sqr(delta_pos)
    IF(cent_dist_sqr < two_pebble_radius_sqr) THEN  
       cent_dist = SQRT(cent_dist_sqr)  
       inside_dist = cent_dist - 2.0*pebble_radius  
       IF (inside_dist < 0.0) THEN  
          raw_force = pebble_pebble_hooke * inside_dist  
          IF (ABS(cent_dist) < 1.0e-10) THEN  
             cent_dist = 1.0e-10  

          END IF
          v_normal = r_v .proj. ((-1.0)*delta_pos)
          v_transverse = r_v - v_normal
          !           First figure out the normal force
          f_normal = (raw_force / cent_dist) * ((-1.0) * delta_pos) - &
               normal_dash_pot_constant * v_normal          
          !f_normal = vdamp(f_normal,max_normal_force,max_damp_factor)
          !           Then figure out the transverse force
          

          fn_mag = abs(f_normal)
          vt_mag = abs(v_transverse)
          ft_mag = min(fn_mag*kinetic_friction,&
               &           vt_mag*transverse_dash_pot_constant)

          !print *,"#vt",v_transverse
          if(vmag_sqr(v_transverse) < static_velocity_sqr_max .and. &
               use_static_friction) then
             !calculate static friction
             slip_mag = abs(old_slip)
             if(slip_mag*static_friction_hooke > 1.0e-20_d_t) then
                adjusted_slip = perp_projection(old_slip,normal)
                if(linear_static_velocity_clamp) then
                   static_mag = min(slip_mag * static_friction_hooke * &
                        (1.0 - vt_mag/static_velocity_max),&
                        fn_mag * pp_static_friction)
                else 
                   static_mag = min(slip_mag * static_friction_hooke,&
                        fn_mag * pp_static_friction)
                end if
                static_friction_force = (-static_mag)* normalize(adjusted_slip)
                !print *,"#info sm",slip_mag,"as",adjusted_slip,"stm",static_mag,"sff",static_friction_force
             else
                !print *,"#finfo sm",slip_mag,"os",old_slip
                static_mag = 0.0
                static_friction_force = (/ 0.0, 0.0, 0.0 /)
             end if
          else 
             static_mag = 0.0
             static_friction_force = (/ 0.0 , 0.0, 0.0 /)
          end if

          if(calculate_pebble_wear_vol) then
             call tally_pebble_pebble_wear(pos,other_pos,v_transverse,f_normal)
          end if


          IF (ABS(ft_mag) + ABS(static_mag) < 1.0e-10) THEN
             f = f_normal
             other_f = (-1.0)*f_normal
             torque = (/ 0.0, 0.0, 0.0 /)
             other_torque = (/ 0.0, 0.0, 0.0 /)
             changed = .true.

          ELSE
             if(abs(ft_mag) < 1.0e-10) then 
                f_transverse = static_friction_force 
             else
                f_transverse = (-ft_mag/vt_mag) * v_transverse +&
                     static_friction_force
             end if
             f = f_normal + f_transverse
             other_f = (-1.0)*(f_normal + f_transverse)
             torque = (((-1.0/2.0) * delta_pos) .cross. f_transverse)
             other_torque = (((-1.0/2.0) * delta_pos) .cross. f_transverse)
             changed = .true.
          END IF
       END IF
    END IF
  END SUBROUTINE pebble_pebble_velocity_force3



  subroutine print_displacement(time)
    real(kind=d_t), intent(in) :: time
    type(vector) :: displacement, wall_velocity
    logical :: displaced
    call calc_earthquake_displacement(displacement,wall_velocity,time, displaced)
    print *,"#EQ_displacement ",time," disp ",displacement," wall ",wall_velocity
  end subroutine print_displacement

  pure subroutine pebble_surface_normal(pos, normal, in_contact, time, current_wear) 
    type(vector), intent(in) :: pos
    type(vector), intent(out) :: normal
    logical, intent(out) :: in_contact
    real(kind=d_t), intent(in) :: time
    real(kind=d_t), intent(inout) :: current_wear
    type(vector) :: torque, force
    ! This is ugly and cheating, but works.
    ! Just ask for the base force, subtract out gravitational force, and 
    ! the direction of the remaining force is the normal.
    call base_force(pos,vector(0.0,0.0,0.0),vector(0.0,0.0,0.0),force,torque,&
         vector(0.0,0.0,0.0),vector(0.0,0.0,0.0),time, current_wear)
    force = force - vector(0.0,0.0,gravitational_force)
    if(vmag_sqr(force) > 1.0e-10) then 
       in_contact = .true.
       normal = normalize(force)
    else 
       in_contact = .false.
       normal = (/ 0.0, 0.0, 0.0 /)
    end if
  end subroutine pebble_surface_normal

  pure subroutine base_force(pos,velocity,omega,force,torque,slip,slip_normal, time, current_wear)
    type(vector), intent(in) :: pos, velocity, omega, slip, slip_normal
    type(vector), intent(out) :: force, torque
    real(kind=d_t), intent(in) :: time
    real(kind=d_t), intent(inout) :: current_wear
    type(vector) :: displacement, wall_velocity
    logical :: displaced
    call calc_earthquake_displacement(displacement,wall_velocity,time,displaced)
    if(displaced) then
       call default_base_force(pos+displacement,velocity+wall_velocity, omega,&
            force,torque,slip,slip_normal,current_wear)
    else 
       call default_base_force(pos,velocity, omega, force, torque, slip,&
            slip_normal,current_wear)
    end if
  end subroutine base_force



END PROGRAM marbles7
