! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module marbles9_common_module
  USE vector_type
  USE str_utils
  USE prob_table
  use number_utils
  use types
  use nearby_pebble_cache_module
  use pp_slip_module
  use geometry
  use constants
  use force_tally
  use earthquake
  use new_fortran_utils
  use wear_tally
  !$ use omp_lib
  implicit none

  ! the state of all the pebbles in the system.
  type pebble_state
     !pp = pebble position
     !pv = pebble velocity
     !pw = pebble angular velocity (w ~= omega)
     ! pp(number_of_pebbles)
     type(vector),pointer, dimension(:) :: pp, pv, pw
     !Storage place for the pebble to pebble slips
     type(pp_slip_type) :: pp_slip_data
     !pss = pebble surface slip
     ! pss(number_of_pebbles)
     type(vector),pointer, dimension(:) :: pss
     !psn = pebble surface normal 
     ! used to make sure that the angle change in the normal is not
     ! excessive (surface can change 90 degrees rapidly.)
     !type(vector),pointer, dimension(:) :: psn
  end type pebble_state

  type pebble_delta
     !pdp = pebble d(position)  aka pebble velocity
     !pdv = pebble d(velocity) aka pebble acceleration
     !pdw = pebble d(angular velocity) aka pebble angular acceleration
     !pdss = pebble d(surface slip) 
     ! pdp(number_of_pebbles,multistep_order)
     type(vector),pointer, dimension(:) :: pdp, pdv, pdw, pdss
     !keep pss  is true if the slip has not turned to slide
     logical,pointer, dimension(:) :: keep_pss
     !Storage space for the pebble pebble slip delta
     type(pp_slip_type) :: pp_slip_delta
  end type pebble_delta

  !     The constants used throughout the program

  REAL(KIND=D_T) :: pebble_center_radius
  REAL(KIND=D_T) :: density_center
  !density in kg/m**3
  REAL(KIND=D_T) :: density_outside
  ! The fill edge is the max distance from center pebbles are filled to
  REAL(KIND=D_T) :: fill_edge 
  integer :: current_cone_drop
  !logical :: offset_inlet_and_cone

  !gravitational_acceleration in m/s**2
  REAL(KIND=D_T), PARAMETER :: gravitational_acceleration = -9.8
  !mass in kg
  REAL(KIND=D_T) :: mass
  !hooke forces in kg*m/(m*s**2)
  REAL(KIND=D_T) :: pebble_pebble_hooke
  REAL(KIND=D_T) :: initial_packing
  integer :: initial_rand_packing_mult
  REAL(KIND=D_T), PARAMETER :: air_friction = 0.0
  !REAL(KIND=D_T) :: static_rotation_sqr
  REAL(KIND=D_T) :: moment_of_inertia

  INTEGER(KIND=LI) :: seed  
  REAL(KIND=D_T) :: area  
  REAL(KIND=D_T) :: pebble_volume,height_needed  
  INTEGER :: number_of_pebbles
  INTEGER :: runs
  REAL(KIND=D_T) :: alpha
  !How often to display the velocities
  INTEGER :: display_frequency
  !How often to display the positions
  integer :: position_frequency
  !How often to display the energy
  integer :: energy_frequency

  ! What is the number of the pebble we are interested in?
  integer :: pebble_of_interest

  !How often to display the packings and the velocity tallies
  integer :: dump_frequency
  REAL(KIND=D_T):: x_low, x_high, y_low, y_high, z_low, z_high
  REAL(KIND=D_T) :: zone_increment
  !tally_zone_inc is the size of the velocity tally boxes
  real(kind=d_t) :: tally_z_inc
  real(kind=d_t) :: tally_radial_inc
  ! The tally_sections is the number of pizza slices that are 
  !  made into different zones.
  integer, parameter :: tally_sections = 8
  INTEGER :: x_size
  INTEGER :: y_size
  INTEGER :: z_size
  INTEGER :: rt_size
  INTEGER :: zt_size
  INTEGER :: near_pebbles(near_pebbles_size),near_pebbles_count  
  !multistep_order is how many steps to keep track of
  INTEGER :: multistep_order

  !This will shut off torque if set to .false.
  ! otherwise, should be set to .true.
  logical,parameter :: torque_enable = .true.

  logical :: zero_slips_gt_mu_fn = .false.

  ! Recirculation parameters
  ! Turn recirculation on or off.
  logical :: recirculate
  !The door at the bottom will be open until a pebble goes through.
  ! then, it will be closed for door_closed_time.
  real(kind=d_t) :: door_closed_time
  ! when does the door first open
  real(kind=d_t) :: first_door_open_time 
  real(kind=d_t) :: time_till_door_open


  !The time before counting the velocity (so as to ignore the settling time)
  real(kind=d_t) :: first_velocity_tally_time
  integer :: velocity_tally_frequency
  !These store the data required to calculate the average velocity and 
  ! the variance.
  INTEGER,allocatable, dimension(:,:,:) :: zone_n
  type(vector),allocatable, dimension(:,:,:) :: zone_vel_sum
  REAL(kind=d_t),allocatable, dimension(:,:,:) :: zone_dot_sum
  !These store the radialized data required to calculate the average velocity
  ! and the variance.  Highly likely to be correct, but harder to interpret.
  INTEGER,allocatable, dimension(:,:,:) :: zone_rect_n
  type(vector),allocatable, dimension(:,:,:) :: zone_rect_vel_sum
  REAL(kind=d_t),allocatable, dimension(:,:,:) :: zone_rect_dot_sum

  !Final position dump parameters
  logical :: do_dump_positions
  character(256) :: dump_positions_filename
  ! This is used to convert to different units.
  !  if you want the data in cm, then it should be 100.0
  !  and if inches, then 39.37 and so forth
  real(kind=d_t) :: dump_positions_mult

  !Initial position loading parameters
  logical :: do_load_positions
  character(256) :: load_positions_filename
  ! This is used to convert to different units.
  !  if the incomming data is in cm, then it should be 100.0
  !  and if inches, then 39.37 and so forth
  real(kind=d_t) :: load_positions_divide

  !Initial position, velocity and such loading parameters
  logical :: do_load_pebble_info
  character(256) :: load_pebble_info_filename
  logical :: do_load_pebble_slips
  logical :: do_load_pebble_surface_slips


  !pp = pebble position
  !pv = pebble velocity
  !pw = pebble angular velocity (w ~= omega)
  ! pp(number_of_pebbles, multistep_order)
  type(pebble_state), allocatable, dimension(:) :: states

  !pdp = pebble d(position)  aka pebble velocity
  !pdv = pebble d(velocity) aka pebble acceleration
  !pdw = pebble d(angular velocity) aka pebble angular acceleration
  ! pdp(number_of_pebbles,multistep_order)
  type(pebble_delta), allocatable, dimension(:) :: deltas

  !The time of the simulation
  real(kind=d_t) :: time = 0.0

  !Use ab algorithm
  !logical :: use_ab_algorithm

  integer, parameter :: euler_method = 1, runge_kutta_method = 2, &
       adams_bashforth_method = 3, adams_moulton_method = 4
  integer :: integration_method = euler_method

  !Probability tally stuff
  logical :: run_probability_tally
  integer :: probability_tally_frequency
  real(kind=d_t) :: probability_tally_start
  real(kind=d_t) :: probability_tally_delta
  type(count_data_type) :: probability_data


  !Sort pebbles by height after loading or initializing
  logical :: sort_pebbles

  integer :: pebble_wear_vol_display_freq

  logical :: calculate_force_tally

  type(nearby_pebble_cache_type) :: nearby_pebble_cache

  !$ integer :: chunk_size
  INTEGER, PARAMETER :: DUMP_UNIT = 206
  INTEGER, PARAMETER :: LOAD_UNIT = 207
  INTEGER, PARAMETER :: LOAD_INFO_UNIT = 208 
  integer :: extra_info_unit
  integer :: stdout_unit


contains

  subroutine set_marbles9_defaults
    runs = 10000
    number_of_pebbles = 1000
    kinetic_friction = 0.1
    pp_static_friction = 0.0
    surface_static_friction = 0.0
    !static_speed_sqr = 1.0e-4
    alpha = 0.0001
    normal_dash_pot_constant = 2.0
    transverse_dash_pot_constant = 2.0
    initial_packing = 0.10
    initial_rand_packing_mult = -1
    dump_frequency = 100000
    display_frequency = 1000
    position_frequency = 100000000
    energy_frequency = 100
    seed = 256  
    do_dump_positions = .false.
    dump_positions_mult = 1.0
    do_load_positions = .false.
    load_positions_divide = 1.0
    do_load_pebble_info = .false.
    do_load_pebble_slips = .false.
    do_load_pebble_surface_slips = .false.
    pebble_pebble_hooke = 10000.0
    pebble_vessel_hooke = 10000.0
    pebble_radius = 0.1
    pebble_center_radius = 0.0
    density_center = 0.0
    density_outside = 2.0
    use_static_friction = .false.
    static_velocity_sqr_max = 0.1**2
    static_friction_hooke = 10000.0
    surface_static_friction_hooke = 10000.0
    zero_static_slips = .false.
    linear_static_velocity_clamp = .false.
    zero_long_slips = .false.
    shorten_long_slips = .false.
    decrease_long_slips = .false.
    decrease_slip_mult = 2.0
    decrease_slip_slope = 1.0
    long_slip_sqr = 100.0
    long_slip_mult = 1.0
    pebble_of_interest = -1
    !use_ab_algorithm = .false.
    integration_method = euler_method
    multistep_order = 4
    recirculate = .false.
    door_closed_time = 10.0
    first_door_open_time = 10.0
    first_velocity_tally_time = 1.0e6
    velocity_tally_frequency = 10000000
    run_probability_tally = .false.
    probability_tally_frequency = 1000000
    probability_tally_start = 1.0e6
    current_cone_drop = 1
    sort_pebbles = .false.
    calculate_pebble_wear_vol = .false.
    pebble_wear_vol_display_freq = 0
    calculate_force_tally = .false.
    !$ chunk_size = max(number_of_pebbles/(8*omp_get_num_procs()),1)
    stdout_unit = get_stdout_unit()
  end subroutine set_marbles9_defaults


  subroutine load_positions(pp,filename,divide) 
    type(vector), pointer, dimension(:) :: pp
    character(*), intent(in) :: filename
    real(kind=d_t), intent(in) :: divide
    integer :: openstatus, i
    open(LOAD_UNIT,file=filename,action="READ",iostat=openstatus)

    if(openstatus > 0) then 
       print *,"ERROR Cannot open file to load positions"
       return
    end if
    do i=1,number_of_pebbles
       read(LOAD_UNIT, *) pp(i)
       pp(i) = (1.0/divide) * pp(i)
    end do
    close(LOAD_UNIT)

  end subroutine load_positions


  subroutine dump_positions(pp,filename,mult) 
    type(vector), pointer, dimension(:) :: pp
    character(*), intent(in) :: filename
    real(kind=d_t), intent(in) :: mult
    integer :: openstatus, i
    open(DUMP_UNIT,file=filename,action="WRITE",iostat=openstatus)

    if(openstatus > 0) then 
       print *,"ERROR Cannot open file to dump positions"
       return
    end if
    do i=1,number_of_pebbles
       write(DUMP_UNIT, *) mult * pp(i)
    end do
    close(DUMP_UNIT)

  end subroutine dump_positions

  subroutine display_zone_velocities() 
    ! s_z is slice zone, r_z in radial zone, z_z is z zone
    integer :: s_z,r_z,z_z,n,x_z,y_z
    real(kind=d_t) :: cx, cy, cz
    type(vector) :: avg, sum
    real(kind=d_t) :: var, dot_sum
    print *,"#zsi ",tally_sections,rt_size,zt_size
    print *,"#zin ",tally_radial_inc,tally_z_inc,z_low 
    print *,"#zone x range y range z range n sum average dot variance"
    do s_z=1,tally_sections
       do r_z=1,rt_size
          do z_z=1,zt_size
             n = zone_n(s_z,r_z,z_z)
             if(n > 0) then 
                sum = zone_vel_sum(s_z,r_z,z_z)
                avg = (1.0/n)*sum
                dot_sum = zone_dot_sum(s_z,r_z,z_z)
                var = (1.0/(n-1))*(n*vmag_sqr(avg) &
                     -2.0*(avg .dot. sum) + dot_sum)
                call radial_zone_center(s_z,r_z,z_z,cx,cy,cz)
                print *,"#z   ",&
                     cx,cy,cz,&
                     n,sum,avg,dot_sum,var
             end if
          end do
       end do
    end do
    print *,"#rzs ",x_size,y_size,z_size
    print *,"#rzl ",x_low,y_low,z_low, zone_increment
    do z_z=1,z_size
       do y_z=1,y_size
          do x_z=1,x_size
             n = zone_rect_n(x_z,y_z,z_z)
             if(n > 0) then
                sum = zone_rect_vel_sum(x_z,y_z,z_z)
                avg = (1.0/n)*sum
                dot_sum = zone_rect_dot_sum(x_z,y_z,z_z)
                var = (1.0/(n-1))*(n*vmag_sqr(avg) &
                     -2.0*(avg .dot. sum) + dot_sum)
                cx = zone_center(x_low, zone_increment, x_z)
                cy = zone_center(y_low, zone_increment, y_z)
                cz = zone_center(z_low, zone_increment, z_z)
                print *,"#rz  ",cx,cy,cz,n,sum,avg,dot_sum,var
             end if
          end do
       end do
    end do

  end subroutine display_zone_velocities

  subroutine display_overlaps(pp)
    type(vector), pointer,dimension(:) :: pp
    real(kind=d_t) :: max_overlap, overlap_sum, overlap_total, overlap, average
    integer :: j,k,l,complete_overlap_count,overlap_count

    call update_nearby_pebble_cache(nearby_pebble_cache,pp)
    max_overlap = -100.0
    overlap_sum = 0.0
    complete_overlap_count = 0

    DO j=1,number_of_pebbles
       call nearby_pebbles(nearby_pebble_cache, pp(j),&
            &        near_pebbles, near_pebbles_size, near_pebbles_count, j)
       overlap_count = 0
       overlap_total = 0.0
       DO l=1,near_pebbles_count
          k = near_pebbles(l)
          overlap = pebble_pebble_overlap_dist(pp(j),pp(k),pebble_radius)
          IF (overlap > 0.0) THEN
             overlap_total = overlap_total + overlap
             overlap_count = overlap_count + 1
          END IF
          IF (overlap > max_overlap) THEN
             max_overlap = overlap
          END IF
       END DO
       PRINT *,"#P_overlap",j,overlap_count,overlap_total
       complete_overlap_count = complete_overlap_count + overlap_count
       overlap_sum = overlap_sum + overlap_total
    END DO
    average = overlap_sum/complete_overlap_count
    PRINT *,"#Max overlap",max_overlap
    PRINT *,"#Max overlap %",max_overlap*100/pebble_radius
    PRINT *,"#Totals",overlap_sum, complete_overlap_count
    PRINT *,"#Average",overlap_sum/complete_overlap_count
    PRINT *,"#Average %",average*100/pebble_radius
  end subroutine display_overlaps

  subroutine dump_pebbles_of_interest_info(current_state, last_delta, time)
    type(pebble_state), intent(in) :: current_state
    type(pebble_delta), intent(in) :: last_delta
    real(kind=d_t),intent(in) :: time
    integer:: i
    i = pebble_of_interest
    ! i=1, p=3:4:5, v=7:8:9, w=11:12:13, dp=15:16:17,dv=19:20:21,dw=23:24:25
    write(extra_info_unit,*) "#POI",time,i,"p",current_state%pp(i),"v",current_state%pv(i),&
         "w",current_state%pw(i),&
         "dp",last_delta%pdp(i),"dv",last_delta%pdv(i),"dw",last_delta%pdw(i)
  end subroutine dump_pebbles_of_interest_info

  subroutine sort_pebbles_by_height(state)
    type(pebble_state), intent(inout) :: state
    integer,dimension(:),allocatable :: new_ordering
    type(real_int_pair),dimension(:),allocatable :: height_list
    integer :: i
    allocate(new_ordering(number_of_pebbles))
    allocate(height_list(number_of_pebbles))
    do i=1,number_of_pebbles
       height_list(i) = real_int_pair(state%pp(i)%z,i)
    end do
    call sort_ri_pair(height_list)
    do i=1,number_of_pebbles
       !new_ordering(old_index) = new_index
       new_ordering(height_list(i)%i) = i       
    end do

    call reorder_pebbles(state, new_ordering)

    deallocate(new_ordering,height_list)
  end subroutine sort_pebbles_by_height


  ! For each pebble, moves it from old_index to new_ordering(old_index) 
  !  in the state passed in. new_index = new_ordering(old_index)
  ! So, if we have new_ordering = (/ 1, 4, 2, 3 /), then pebble 1 will 
  ! stay the same, and pebble 2_old_index will become 4_new_index
  subroutine reorder_pebbles(state, new_ordering)    
    type(pebble_state), intent(inout) :: state
    type(pebble_state) :: temp_state
    integer,dimension(:),intent(in) :: new_ordering
    integer :: i
    allocate(temp_state%pp(number_of_pebbles))
    allocate(temp_state%pv(number_of_pebbles))
    allocate(temp_state%pw(number_of_pebbles))
    !call allocate_pp_slip(number_of_pebbles,temp_state%pp_slip_data)
    allocate(temp_state%pss(number_of_pebbles))
    !allocate(temp_state%psn(number_of_pebbles))
    do i=1,number_of_pebbles
       print *,"#reordering ",new_ordering(i)," <- ",i
       temp_state%pp(new_ordering(i)) = state%pp(i)
       temp_state%pv(new_ordering(i)) = state%pv(i)
       temp_state%pw(new_ordering(i)) = state%pw(i)
       if(use_static_friction) then
          temp_state%pss(new_ordering(i)) = state%pss(i)
          !temp_state%psn(new_ordering(i)) = state%psn(i)
       end if
    end do
    if(use_static_friction) then
       call reorder_pp_slip(new_ordering,state%pp_slip_data)
    end if
    state%pp = temp_state%pp
    state%pv = temp_state%pv
    state%pw = temp_state%pw
    if(use_static_friction) then
       state%pss = temp_state%pss
       !state%psn = temp_state%psn
    end if
    deallocate(temp_state%pp,temp_state%pv,temp_state%pw,temp_state%pss)
    !deallocate(temp_state%psn)
  end subroutine reorder_pebbles

  subroutine get_ab_timestep(out_state, in_state, deltas,&
       t_c,t_m1,t_m2,t_m3,alpha,time)
    type(pebble_state), intent(in) :: in_state
    type(pebble_state), intent(out) :: out_state
    type(pebble_delta), dimension(:) :: deltas
    !the indexs into the dp_'s
    ! t = t index, t_m1 = t-h index, t_m2 = t-2*h index, t_m3 = t-3*h index
    integer, intent(in) :: t_c,t_m1,t_m2,t_m3
    real(kind=d_t), intent(in) :: alpha, time
    integer,dimension(1:4) :: indexs 
    real(kind=d_t),dimension(1:4) :: mults

    indexs = (/ t_c, t_m1, t_m2, t_m3 /)  
    mults = (/ alpha*(55.0/24.0),alpha*(-59.0/24.0),alpha*(37.0/24.0),&
         alpha*(-9.0/24.0) /)    

    call get_derivs(nearby_pebble_cache, in_state,&
         deltas(t_c),&
         &  number_of_pebbles, near_pebbles, near_pebbles_size, time)
    call combine_steps(out_state,in_state,deltas,indexs,mults)
!!$    DO j=1,number_of_pebbles
!!$       out_state%pp(j) = in_state%pp(j) + &
!!$            (alpha/24.0)*(55.0*deltas(t_c)%pdp(j)-59.0*deltas(t_m1)%pdp(j)&
!!$            +37.0*deltas(t_m2)%pdp(j)-9.0*deltas(t_m3)%pdp(j))
!!$       out_state%pv(j) = in_state%pv(j) + &
!!$            (alpha/24.0)*(55.0*deltas(t_c)%pdv(j)-59.0*deltas(t_m1)%pdv(j)&
!!$            +37.0*deltas(t_m2)%pdv(j)-9.0*deltas(t_m3)%pdv(j))
!!$       out_state%pw(j) = in_state%pw(j) + &
!!$            (alpha/24.0)*(55.0*deltas(t_c)%pdw(j)-59.0*deltas(t_m1)%pdw(j)&
!!$            +37.0*deltas(t_m2)%pdw(j)-9.0*deltas(t_m3)%pdw(j))
!!$
!!$    END do
  end subroutine get_ab_timestep

  subroutine get_am_timestep(out_state, in_state, deltas,&
       t_c,t_m1,t_m2,t_m3,alpha,time)
    type(pebble_state), intent(in) :: in_state
    type(pebble_state), intent(out) :: out_state
    type(pebble_delta), dimension(:) :: deltas
    !the indexs into the dp_'s
    ! t = t index, t_m1 = t-h index, t_m2 = t-2*h index, t_m3 = t-3*h index
    integer, intent(in) :: t_c,t_m1,t_m2,t_m3
    real(kind=d_t), intent(in) :: alpha, time
    integer,dimension(1:4) :: indexs, indexs2 
    real(kind=d_t),dimension(1:4) :: mults, mults2
    integer :: i

    type(pebble_state) :: temp_state
    type(pebble_delta),dimension(1:4) :: f

    call allocate_state(number_of_pebbles,temp_state)
    call allocate_delta(number_of_pebbles,f(1))
    
    !y^*_n+1 = y_n + (h/24)*(55f_n-59f_n-1+37f_n-2-9f_n-3)
    indexs = (/ t_c, t_m1, t_m2, t_m3 /)  
    mults = (/ alpha*(55.0/24.0),alpha*(-59.0/24.0),alpha*(37.0/24.0),&
         alpha*(-9.0/24.0) /)    

    call get_derivs(nearby_pebble_cache, in_state, deltas(t_c),&
         &  number_of_pebbles, near_pebbles, near_pebbles_size, time)
    call combine_steps(temp_state,in_state,deltas,indexs,mults)
    !f^*_n+1 = f(x_n+1,y^*_n+1)
    call get_derivs(nearby_pebble_cache, temp_state, f(1), &
         &  number_of_pebbles, near_pebbles, near_pebbles_size, time)
    !y_n+1 = y_n + (h/24)*(9*f^*_n+1+19f_n-5f_n-1+fn-12)
    do i =2,4
       call delta_shallow_copy(f(i),deltas(indexs(i-1)))
    end do
    indexs2 = (/ 1,2,3,4 /) 
    mults2 = (/ alpha*(9.0/24.0),alpha*(19.0/24.0),alpha*(-5.0/24.0),alpha/24.0 /)
    call combine_steps(out_state,in_state,f,indexs2,mults2)
    
    call deallocate_state(temp_state)
    call deallocate_delta(f(1))
    
  end subroutine get_am_timestep

  subroutine get_simple_timestep(out_state,in_state,delta,alpha,time)
    type(pebble_state),intent(in) :: in_state
    type(pebble_state),intent(out) :: out_state
    type(pebble_delta), intent(out) :: delta
    real(kind=d_t), intent(in) :: alpha, time

    call get_derivs(nearby_pebble_cache, in_state, delta, &
         &        number_of_pebbles, near_pebbles, near_pebbles_size, time)
    call combine_step(number_of_pebbles,in_state,out_state,&
         delta,alpha)

  end subroutine get_simple_timestep

  subroutine get_simple_timestep_selected(out_state,in_state,delta,alpha,time,&
       selected, selected_count, provided, provided_count)
    type(pebble_state),intent(in) :: in_state
    type(pebble_state),intent(out) :: out_state
    type(pebble_delta), intent(out) :: delta
    integer,intent(in),dimension(:) :: selected,provided
    integer,intent(in) :: selected_count,provided_count
    real(kind=d_t), intent(in) :: alpha, time

    call write_time(extra_info_unit,"Getting Derivatives")
    call get_derivs_selected(nearby_pebble_cache, in_state, delta, &
         &        near_pebbles, near_pebbles_size, time, &
         &        selected, selected_count,&
         provided,provided_count)
    call write_time(extra_info_unit,"Combining Derivatives")
    call combine_step_selected(in_state,out_state,&
         delta,alpha, selected, selected_count)
  end subroutine get_simple_timestep_selected

  subroutine get_simple_timestep_selected2(out_state,in_state,delta,alpha,&
       time, selected, selected_count)
    type(pebble_state),intent(in) :: in_state
    type(pebble_state),intent(out) :: out_state
    type(pebble_delta), intent(out) :: delta
    integer,intent(in),dimension(:) :: selected
    integer,intent(in) :: selected_count
    real(kind=d_t), intent(in) :: alpha, time
    
    call get_derivs_selected(nearby_pebble_cache, in_state, delta, &
         &        near_pebbles, near_pebbles_size, time, &
         &        selected, selected_count,&
         selected,selected_count)
    call combine_step_selected(in_state,out_state,&
         delta,alpha, selected, selected_count)
  end subroutine get_simple_timestep_selected2


  subroutine allocate_state(number_of_pebbles,state)
    type(pebble_state), intent(out) :: state
    integer, intent(in) :: number_of_pebbles
    allocate(state%pp(number_of_pebbles))
    allocate(state%pv(number_of_pebbles))
    allocate(state%pw(number_of_pebbles))
    allocate(state%pss(number_of_pebbles))
    call allocate_pp_slip(number_of_pebbles,state%pp_slip_data)
  end subroutine allocate_state

  subroutine deallocate_state(state)
    type(pebble_state), intent(inout) :: state
    deallocate(state%pp,state%pv,state%pw,state%pss)
    call deallocate_pp_slip(state%pp_slip_data)
  end subroutine deallocate_state

  subroutine allocate_delta(number_of_pebbles,delta)
    type(pebble_delta),intent(out) :: delta
    integer,intent(in) :: number_of_pebbles
    allocate(delta%pdp(number_of_pebbles))
    allocate(delta%pdv(number_of_pebbles))
    allocate(delta%pdw(number_of_pebbles))
    allocate(delta%pdss(number_of_pebbles))
    allocate(delta%keep_pss(number_of_pebbles))
    call allocate_pp_slip(number_of_pebbles,delta%pp_slip_delta)

  end subroutine allocate_delta

  subroutine deallocate_delta(delta)
    type(pebble_delta), intent(inout) :: delta
    deallocate(delta%pdp,delta%pdv,delta%pdw,delta%pdss,delta%keep_pss)
    call deallocate_pp_slip(delta%pp_slip_delta)
  end subroutine deallocate_delta

  subroutine delta_shallow_copy(outdelta,indelta)
    type(pebble_delta),intent(in) :: indelta
    type(pebble_delta),intent(out) :: outdelta
    outdelta%pdp => indelta%pdp
    outdelta%pdv => indelta%pdv
    outdelta%pdw => indelta%pdw
    outdelta%pdss => indelta%pdss
    outdelta%keep_pss => indelta%keep_pss
    call pp_slip_shallow_copy(outdelta%pp_slip_delta,indelta%pp_slip_delta)
  end subroutine delta_shallow_copy

  subroutine get_rk_timestep(out_state,in_state,delta,alpha,time)
    type(pebble_state), intent(in) :: in_state
    type(pebble_state), intent(out) :: out_state
    type(pebble_delta), intent(inout) :: delta

    real(kind=d_t), intent(in) :: alpha, time
    integer :: j

    type(pebble_state) :: temp_state
    type(pebble_delta),dimension(1:4) :: f

    integer,dimension(1:4) :: indexs = (/ 1, 2, 3, 4 /)
    real(kind=d_t),dimension(1:4) :: mults
    
    mults = (/ alpha/6.0,alpha/3.0,alpha/3.0,alpha/6.0 /)
    !     The Runge-Kutta values
    call allocate_state(number_of_pebbles,temp_state)


    do j=2,4
       call allocate_delta(number_of_pebbles,f(j))
    end do
    !Make f(1) just the delta, so information can be gotten later.
    call delta_shallow_copy(f(1),delta)
    

    !print *,"f1"
    !        f1 = f(T,X)
    call get_derivs(nearby_pebble_cache, in_state, f(1), &
         &        number_of_pebbles, near_pebbles, near_pebbles_size, time)
    !        f2 = f(t+h/2,x+0.5*f1*h)
    !        temps = x + 0.5*f1*h
    !print *,"f2"
    call combine_step(number_of_pebbles,in_state,temp_state,f(1),0.5*alpha)
    call get_derivs(nearby_pebble_cache, temp_state,f(2), &
         &        number_of_pebbles, near_pebbles, near_pebbles_size, time)
    !        f3 = f(t+h/2,x+0.5*f2*h)
    !        temps = x + 0.5*f2*h
    !print *,"f3"
    call combine_step(number_of_pebbles,in_state,temp_state,f(2), 0.5*alpha)
    call get_derivs(nearby_pebble_cache, temp_state,f(3), &
         &        number_of_pebbles, near_pebbles, near_pebbles_size, time)
    !        f4 = f(t+h,x+f3*h)
    !        temps = x + f3*h
    !print *,"f4"
    call combine_step(number_of_pebbles,in_state,temp_state,f(3),alpha)
    call get_derivs(nearby_pebble_cache, temp_state,f(4),&
         &        number_of_pebbles, near_pebbles, near_pebbles_size, time)
    
    !print *,"all together"
    call combine_steps(out_state,in_state,f,indexs,mults)
!!$    DO j=1,number_of_pebbles
!!$       out_state%pp(j) = in_state%pp(j)+ &
!!$            (alpha/6.0)*(delta%pdp(j)+2.0*f2%pdp(j)+2.0*f3%pdp(j)+f4%pdp(j)) 
!!$       out_state%pv(j) = in_state%pv(j)+ &
!!$            (alpha/6.0)*(delta%pdv(j)+2.0*f2%pdv(j)+2.0*f3%pdv(j)+f4%pdv(j)) 
!!$       out_state%pw(j) = in_state%pw(j)+ &
!!$            (alpha/6.0)*(delta%pdw(j)+2.0*f2%pdw(j)+2.0*f3%pdw(j)+f4%pdw(j)) 
!!$
!!$    END DO

    call deallocate_state(temp_state)
    do j=2,4
       call deallocate_delta(f(j))
    end do
    call delta_shallow_copy(delta,f(1))
    
  end subroutine get_rk_timestep

  subroutine get_rk_timestep_selected(out_state,in_state,delta,alpha,time,&
       selected, selected_count, provided, provided_count)
    type(pebble_state), intent(in) :: in_state
    type(pebble_state), intent(out) :: out_state
    type(pebble_delta), intent(inout) :: delta
    integer,intent(in),dimension(:) :: selected,provided
    integer,intent(in) :: selected_count,provided_count

    real(kind=d_t), intent(in) :: alpha, time
    integer :: j

    type(pebble_state) :: temp_state
    type(pebble_delta),dimension(1:4) :: f

    integer,dimension(1:4) :: indexs = (/ 1, 2, 3, 4 /)
    real(kind=d_t),dimension(1:4) :: mults
    
    mults = (/ alpha/6.0,alpha/3.0,alpha/3.0,alpha/6.0 /)
    !     The Runge-Kutta values
    call allocate_state(number_of_pebbles,temp_state)


    do j=2,4
       call allocate_delta(number_of_pebbles,f(j))
    end do
    !Make f(1) just the delta, so information can be gotten later.
    call delta_shallow_copy(f(1),delta)
    

    !print *,"f1"
    !        f1 = f(T,X)
    call get_derivs_selected(nearby_pebble_cache, in_state, f(1), &
         near_pebbles, near_pebbles_size, time,&
         provided, provided_count,provided, provided_count)
    !        f2 = f(t+h/2,x+0.5*f1*h)
    !        temps = x + 0.5*f1*h
    !print *,"f2"
    call combine_step_selected(in_state,temp_state,f(1),&
         0.5*alpha, provided, provided_count)
    call get_derivs_selected(nearby_pebble_cache, temp_state,f(2), &
         &        near_pebbles, near_pebbles_size, time,&
         provided, provided_count,provided, provided_count)
    !        f3 = f(t+h/2,x+0.5*f2*h)
    !        temps = x + 0.5*f2*h
    !print *,"f3"
    call combine_step_selected(in_state,temp_state,f(2), &
         0.5*alpha, provided, provided_count)
    call get_derivs_selected(nearby_pebble_cache, temp_state,f(3), &
         &        near_pebbles, near_pebbles_size, time,&
         provided, provided_count,provided, provided_count)
    !        f4 = f(t+h,x+f3*h)
    !        temps = x + f3*h
    !print *,"f4"
    call combine_step_selected(in_state,temp_state,f(3),&
         alpha, provided, provided_count)
    call get_derivs_selected(nearby_pebble_cache, temp_state,f(4),&
         &        near_pebbles, near_pebbles_size, time,&
         provided, provided_count,provided, provided_count)
    
    !print *,"all together"
    call combine_steps_selected(out_state,in_state,f,indexs,mults,&
         selected, selected_count)

    call deallocate_state(temp_state)
    do j=2,4
       call deallocate_delta(f(j))
    end do
    call delta_shallow_copy(delta,f(1))
    
  end subroutine get_rk_timestep_selected


  subroutine handle_probability_tally(current_state,old_state)
    type(pebble_state), intent(in) :: current_state, old_state
    type(vector), pointer, dimension(:) :: pp_cur,pp_old
    integer :: i
    pp_cur => current_state%pp
    pp_old => old_state%pp
    do i=1,number_of_pebbles
       call add_count(probability_data,pp_old(i),pp_cur(i))
    end do

  end subroutine handle_probability_tally


  subroutine handle_velocity_tally(in_state,time,run)
    type(pebble_state), intent(in) :: in_state    
    real(kind=d_t), intent(in) :: time
    integer, intent(in) :: run
    integer :: i,s_z,r_z,z_z,x_z,y_z
    real(kind=d_t) :: x,y,z
    ! The std_velocity is modified so that spiral motion and radial 
    !  motion are normalized.
    type(vector) :: std_velocity
    type(vector), pointer, dimension(:) :: pp,pv
    if(mod(run,velocity_tally_frequency) .ne. 0) then
       return
    end if
    pp => in_state%pp
    pv => in_state%pv
    if(time < first_velocity_tally_time) then 
       return
    end if
    do i=1,number_of_pebbles
       ! Use a random point in the sphere so not biased near edges 
       !  since the center of the sphere will be oversampled and the 
       !  edges will be undersampled if I just use that.
       call random_point_in_sphere(pp(i)%x,pp(i)%y,pp(i)%z,pebble_radius,&
            x,y,z,seed)
       !x = pp(i)%x
       !y = pp(i)%y
       !z = pp(i)%z
       call find_radial_zone(x,y,z,s_z,r_z,z_z)
       call regularize_radial_velocity(pp(i),pv(i),std_velocity,s_z,r_z,z_z)
       !print *,"#rad_zones",i,in_state%pp(i),x,y,z,s_z,r_z,z_z
       zone_n(s_z,r_z,z_z) = zone_n(s_z,r_z,z_z) + 1
       zone_vel_sum(s_z,r_z,z_z) = zone_vel_sum(s_z,r_z,z_z) + std_velocity
       zone_dot_sum(s_z,r_z,z_z) = zone_dot_sum(s_z,r_z,z_z) + &
            vmag_sqr(std_velocity)

       !XXX Should this be using find_zone instead of find_base_zone?
       x_z = find_base_zone(x_low, zone_increment, x_size, pp(i)%x)  
       y_z = find_base_zone(y_low, zone_increment, y_size, pp(i)%y)  
       z_z = find_base_zone(z_low, zone_increment, z_size, pp(i)%z)  
       zone_rect_n(x_z,y_z,z_z) = zone_rect_n(x_z,y_z,z_z) + 1
       zone_rect_vel_sum(x_z,y_z,z_z) = zone_rect_vel_sum(x_z,y_z,z_z) + pv(i)
       zone_rect_dot_sum(x_z,y_z,z_z) = zone_rect_dot_sum(x_z,y_z,z_z) + &
            vmag_sqr(pv(i))
    end do

  end subroutine handle_velocity_tally

  subroutine handle_recirculation(state)
    type(pebble_state), intent(inout) :: state
    type(vector), pointer,dimension(:) :: pp,pv
    type(vector), pointer,dimension(:) :: pw
    integer :: i, lowest_pebble
    real(kind=d_t) :: lowest_height
    pp => state%pp
    pv => state%pv
    pw => state%pw
    time_till_door_open = time_till_door_open - alpha
    if(time_till_door_open < 0.0) then
       if(.not. door_open) then
          print *,"#door_open",time
       end if
       door_open = .true.
    end if
    if(.not. door_open) then
       return
    end if
    lowest_pebble = 0
    lowest_height = z_high
    !Find the lowest pebble
    do i=1,number_of_pebbles
       if(is_pebble_through_door(current_cone_drop,pp(i)) .and. &
            pp(i)%z < lowest_height) then
          lowest_pebble = i
          lowest_height = pp(i)%z
       end if
    end do
    ! Drop the lowest pebble, and slam door.
    if(lowest_pebble > 0) then
       i = lowest_pebble
       pp(i) = get_recirculation_location(current_cone_drop)
       pv(i) = get_recirculation_velocity(current_cone_drop)
       current_cone_drop = mod(current_cone_drop, get_number_of_recirc_locs())+1
       print *,"#Recirculate at ",pp(i)
       pw(i) = vector(0.0,0.0,0.0)
       door_open = .false.
       time_till_door_open = door_closed_time
       print *,"#Dropped Pebble",i
    end if
  end subroutine handle_recirculation

  subroutine handle_recirculation_selected(state, selected, selected_count)
    type(pebble_state), intent(inout) :: state
    integer,dimension(:),intent(in) :: selected
    integer,intent(in) :: selected_count
    type(vector), pointer,dimension(:) :: pp,pv
    type(vector), pointer,dimension(:) :: pw
    integer :: i,j, lowest_pebble
    real(kind=d_t) :: lowest_height
    !XXX Note that this algorithm only works if only one node has access 
    ! to the lowest pebbles.
    pp => state%pp
    pv => state%pv
    pw => state%pw
    time_till_door_open = time_till_door_open - alpha
    if(time_till_door_open < 0.0) then
       if(.not. door_open) then
          print *,"#door_open",time
       end if
       door_open = .true.
    end if
    if(.not. door_open) then
       return
    end if
    lowest_pebble = 0
    lowest_height = z_high
    !Find the lowest pebble
    do j=1,selected_count
       i = selected(j)
       if(is_pebble_through_door(current_cone_drop,pp(i)) .and. &
            pp(i)%z < lowest_height) then
          lowest_pebble = i
          lowest_height = pp(i)%z
       end if
    end do
    ! Drop the lowest pebble, and slam door.
    if(lowest_pebble > 0) then
       i = lowest_pebble
       pp(i) = get_recirculation_location(current_cone_drop)
       pv(i) = get_recirculation_velocity(current_cone_drop)
       current_cone_drop = mod(current_cone_drop, get_number_of_recirc_locs())+1
       print *,"#Recirculate at ",pp(i)       
       pw(i) = vector(0.0,0.0,0.0)
       door_open = .false.
       time_till_door_open = door_closed_time
       print *,"#Dropped Pebble",i
    end if
  end subroutine handle_recirculation_selected

  subroutine load_pebble_info(pp,pv,pw,filename,read_slips, read_surface,&
       out_state)
    type(vector), intent(out),dimension(:) :: pp,pv,pw
    !type(vector), intent(out) :: pw(number_of_pebbles)
    type(pebble_state), intent(out), optional :: out_state
    logical, intent(in) :: read_slips, read_surface
    character(*), intent(in) :: filename
    integer :: openstatus, i, j
    character(3) :: junk_str
    character(16) :: junk_surface_str
    type(vector) :: slip, slip_normal

    open(LOAD_INFO_UNIT,file=filename,action="READ",iostat=openstatus)

    if(openstatus > 0) then 
       print *,"ERROR Cannot open file to load pebble info"
       return
    end if

    do i=1,number_of_pebbles
       read(LOAD_INFO_UNIT,*) junk_str,j,&
            pp(i)%x,pp(i)%y,pp(i)%z,&
            pv(i)%x,pv(i)%y,pv(i)%z,&
            pw(i)%x,pw(i)%y,pw(i)%z
       if(i /= j) then
          print *,"ERROR Mismatch of pebble number",i,j
       end if
    end do
    if(read_slips) then
       call load_pp_slip(LOAD_INFO_UNIT,out_state%pp_slip_data)
    end if
    if(read_surface) then
       do i=1,number_of_pebbles
          read(LOAD_INFO_UNIT,*) junk_surface_str,j,slip%x,slip%y,slip%z,&
               slip_normal%x,slip_normal%y,slip_normal%z
          if(j == i) then 
             out_state%pss(i) = slip
             !out_state%psn(i) = slip_normal
          else 
             print *,"ERROR Mismatch in read surface slips ",j," /= expected",i
          end if
       end do
    end if
    close(LOAD_INFO_UNIT)

  end subroutine load_pebble_info

  subroutine display_pebble_positions(time,current_state)
    integer :: j
    type(pebble_state), intent(in) :: current_state
    real(kind=d_t), intent(in) :: time
    real(kind=d_t) :: pot_energy
    pot_energy = 0.0
    do j=1,number_of_pebbles
       print '(A4,I6,3(1X,F20.7))',"#P: ",j,current_state%pp(j)%x,&
            current_state%pp(j)%y,current_state%pp(j)%z
       pot_energy = pot_energy + &
            mass*(-gravitational_acceleration)*(current_state%pp(j)%z - get_potential_energy_zero())
    end do
    print *,"#Grav_pot_energy ",time,pot_energy
    PRINT *,"#Displacement",get_earthquake_displacement(time)
    call print_displacement(time)

  end subroutine display_pebble_positions

  SUBROUTINE display_pebble_info(time,current_state) 
    type(pebble_state), intent(in) :: current_state
    real(kind=d_t), intent(in) :: time
    type(vector), pointer,dimension(:) :: pp,pv
    type(vector), pointer,dimension(:) :: pw
    integer :: j
    real(kind=d_t) :: rot_energy, lin_energy, pot_energy, static_energy
    real(kind=d_t) :: wall_static
    pp => current_state%pp
    pv => current_state%pv
    pw => current_state%pw
    rot_energy = 0.0
    lin_energy = 0.0
    pot_energy = 0.0
    static_energy = 0.0
    wall_static = 0.0
    DO j=1,number_of_pebbles
       PRINT '(A3,I6,9(1X,F0.11))',"P: ",j,pp(j)%x,pp(j)%y,pp(j)%z,pv(j)%x,pv(j)%y,pv(j)%z,pw(j)%x,pw(j)%y,pw(j)%z
       rot_energy = rot_energy + 0.5*moment_of_inertia*vmag_sqr(pw(j))
       lin_energy = lin_energy + 0.5*mass*vmag_sqr(pv(j))
       pot_energy = pot_energy + mass*(-gravitational_acceleration)*(pp(j)%z - get_potential_energy_zero())
       if(use_static_friction) then
          static_energy = static_energy + get_static_potential_energy(current_state,j)*2.0
          !print *,"#p_s",j,current_state%pss(j),current_state%psn(j)
          wall_static = wall_static + 0.5 * vmag_sqr(current_state%pss(j)) * &
               surface_static_friction_hooke
       end if

    END DO
    if(use_static_friction) then 
       call display_pebble_slips(current_state)
    end if
    PRINT '(A9,F8.3,1X,6(A5,1X,F0.9,1X))',"#Energys ",time,&
         "lin",lin_energy,"rot",rot_energy,"pot",pot_energy,&
         "pstat",static_energy,"wstat",wall_static,"tot",&
         rot_energy+lin_energy+pot_energy+static_energy+wall_static
    PRINT *,"#Displacement",get_earthquake_displacement(time)
    call print_displacement(time)

  END SUBROUTINE display_pebble_info

  SUBROUTINE display_pebble_energy(time,current_state) 
    type(pebble_state), intent(in) :: current_state
    real(kind=d_t), intent(in) :: time
    type(vector), pointer,dimension(:) :: pp,pv
    type(vector), pointer,dimension(:) :: pw
    integer :: j
    real(kind=d_t) :: rot_energy, lin_energy, pot_energy, static_energy
    real(kind=d_t) :: wall_static
    pp => current_state%pp
    pv => current_state%pv
    pw => current_state%pw
    rot_energy = 0.0
    lin_energy = 0.0
    pot_energy = 0.0
    static_energy = 0.0
    wall_static = 0.0
    DO j=1,number_of_pebbles
       rot_energy = rot_energy + 0.5*moment_of_inertia*vmag_sqr(pw(j))
       lin_energy = lin_energy + 0.5*mass*vmag_sqr(pv(j))
       pot_energy = pot_energy + mass*(-gravitational_acceleration)*(pp(j)%z - get_potential_energy_zero())
       if(use_static_friction) then
          static_energy = static_energy + get_static_potential_energy(current_state,j)*2.0          
          wall_static = wall_static + 0.5 * vmag_sqr(current_state%pss(j)) * &
               surface_static_friction_hooke
       end if

    END DO
    PRINT '(A9,F8.3,1X,6(A5,1X,F0.9,1X))',"#Energys ",time,&
         "lin",lin_energy,"rot",rot_energy,"pot",pot_energy,&
         "pstat",static_energy,"wstat",wall_static,"tot",&
         rot_energy+lin_energy+pot_energy+static_energy+wall_static
    PRINT *,"#Displacement",get_earthquake_displacement(time)
    call print_displacement(time)

  END SUBROUTINE display_pebble_energy


  subroutine display_pebble_slips(current_state)
    type(pebble_state), intent(in) :: current_state
    integer :: i
    call display_pp_slips(current_state%pp_slip_data)
    do i=1,number_of_pebbles
       print '(A16,I7,6E25.16E3)',"#P_surface_slip ",i,&
            current_state%pss(i)%x,current_state%pss(i)%y,&
            current_state%pss(i)%z,0.0,0.0,0.0 !current_state%psn(i)%x,&
       !current_state%psn(i)%y,current_state%psn(i)%z
    end do
  end subroutine display_pebble_slips

  subroutine initialize_pebble_data()

    if (do_load_pebble_info) then
       call load_pebble_info(states(1)%pp,states(1)%pv,states(1)%pw &
            ,load_pebble_info_filename,do_load_pebble_slips,&
            do_load_pebble_surface_slips,states(1))
    else if(do_load_positions) then 
       call load_positions(states(1)%pp,load_positions_filename,&
            load_positions_divide)
       states(1)%pv = vector(0.0,0.0,0.0)
       states(1)%pw = vector(0.0,0.0,0.0)
    else if(initial_rand_packing_mult < 1) then
       call simple_random_packing(get_loading_height())
    else
       call initialize_pebs_random_points(get_base_loading_height())
    end if
    

    if(sort_pebbles) then
       call sort_pebbles_by_height(states(1))
    end if

    if(.true.) then
       PRINT *,"#Initial Data",number_of_pebbles
       call display_pebble_info(time,states(1))
    end if
  end subroutine initialize_pebble_data

  subroutine simple_random_packing(loading_height)
    logical :: done  
    integer :: initials, hits, i,j,k,l
    real(kind=d_t),intent(in) :: loading_height

    DO i=1,number_of_pebbles  
       CALL initialize_pebble(states(1)%pp(i),states(1)%pv(i),&
            states(1)%pw(i), &
            seed,loading_height,height_needed,&
            pebble_radius, fill_edge)
    END DO
    initials = 0  
36  CONTINUE  
    done = .TRUE.  
    hits = 0  
    call update_nearby_pebble_cache(nearby_pebble_cache,states(1)%pp)
    DO j=1,number_of_pebbles  
       call nearby_pebbles(nearby_pebble_cache,states(1)%pp(j), &
            near_pebbles, near_pebbles_size, near_pebbles_count, j)
       DO l=1,near_pebbles_count  
          k = near_pebbles(l)  
          IF (pebble_pebble_overlap(states(1)%pp(j),states(1)%pp(k),&
               pebble_radius))  THEN
             CALL initialize_pebble(states(1)%pp(j),states(1)%pv(j),&
                  states(1)%pw(j), &
                  seed,loading_height,&
                  height_needed, pebble_radius,fill_edge)
             done = .FALSE.  
             hits = hits + 1  
          END IF
       END DO
    END DO
    PRINT *,"#Initializing",initials, hits
    initials = initials + 1
    IF (.not. done) GOTO 36    
  end subroutine simple_random_packing

  SUBROUTINE initialize_pebble(pos,velocity,omega,seed, &
       &     bottom,height_needed,pebble_radius, &
       &     fill_edge)
    type(vector), intent(out) :: pos, velocity,omega
    REAL(KIND=D_T), intent(in) :: bottom,height_needed,pebble_radius  
    REAL(KIND=D_T), intent(in) ::  fill_edge
    INTEGER(KIND=LI), intent(inout) :: seed  
    !Initialize linear and angular velocities to 0
    velocity = vector(0.0,0.0,0.0)
    omega = vector(0.0,0.0,0.0)
    pos%z = random(seed)*(1.0/initial_packing)*height_needed + bottom &
         &     + pebble_radius
9   CONTINUE  
    pos%x = (random(seed)-0.5)*2.0*fill_edge
    pos%y = (random(seed)-0.5)*2.0*fill_edge

    !     Check to make sure not inside something.
    if(is_contacting_wall(pos)) GOTO 9  

  END SUBROUTINE initialize_pebble

  !Kloosterman and Ougouag Method of packing.
  subroutine initialize_pebs_random_points(bottom)
    integer :: num_pebbles_added, i,j, k, l, temp_pp_size
    logical :: pebble_overlaps
    type(vector),dimension(:),allocatable :: temp_pp
    type(real_int_pair),dimension(:),allocatable :: height_list
    !Determine slice size
    real(kind=d_t) :: slice_low,slice_delta, slice_volume, pebble_volume
    real(kind=d_t),intent(in) :: bottom
    
    num_pebbles_added = 0
    slice_delta = pebble_radius
    slice_low = bottom + pebble_radius
    pebble_volume = (4.0/3.0)*pi*pebble_radius**2
    slice_volume = fill_edge**2*slice_delta
    temp_pp_size = int(initial_rand_packing_mult*slice_volume/pebble_volume)+1
    
    states(1)%pv(:) = vector(0.0,0.0,0.0)
    states(1)%pw(:) = vector(0.0,0.0,0.0)
    
    !print *,"temp_pp_size",temp_pp_size,"slice_delta",slice_delta
    allocate(temp_pp(temp_pp_size),height_list(temp_pp_size))
    
    print *,"#random_pack",temp_pp_size,slice_delta,slice_volume,pebble_volume
    
    fill_loop: do     
       PRINT *,"#Initializing random pack",num_pebbles_added,slice_low

       !print *,"num_pebbles_added",num_pebbles_added,"slice_low",slice_low,"slice_delta",slice_delta
       !Generate n times as many points as will fit in a slice
       do i = 1, size(temp_pp)
          temp_pp(i)%z = random(seed)*slice_delta+slice_low
          temp_pp(i)%x = (random(seed)-0.5)*2.0*fill_edge
          temp_pp(i)%y = (random(seed)-0.5)*2.0*fill_edge
          !print *,"#Trying ",temp_pp(i)
          !foo = is_contacting_wall(temp_pp(i))
          height_list(i) = real_int_pair(temp_pp(i)%z,i)
       end do

       !sort generated positions
       call sort_ri_pair(height_list)

       do i = 1,size(temp_pp)
          !starting from lowest position, check if pebble encounters another pebble,
          ! and if not, add to nearby pebble cache and list of pebbles
          pebble_overlaps = .false.

          ! Get lowest index
          j = height_list(i)%i

          if(.not. is_contacting_wall(temp_pp(j))) then
             call nearby_pebbles(nearby_pebble_cache,temp_pp(j), &
                  near_pebbles, near_pebbles_size, near_pebbles_count, -1)

             near_pebble_loop: do l=1,near_pebbles_count
                k = near_pebbles(l)
                if(pebble_pebble_overlap(temp_pp(j),states(1)%pp(k),pebble_radius)) then
                   
                   pebble_overlaps = .true.
                   exit near_pebble_loop
                end if
             end do near_pebble_loop
             if( .not. pebble_overlaps) then
                num_pebbles_added = num_pebbles_added + 1
                states(1)%pp(num_pebbles_added) = temp_pp(j)
                call add_to_nearby_pebble_cache(nearby_pebble_cache,&
                     states(1)%pp,num_pebbles_added)
                !print *,"Adding ",num_pebbles_added,states(1)%pp(num_pebbles_added)
             end if
          end if
          !When the full number of pebbless has been found, finish.
          if(num_pebbles_added >= number_of_pebbles) then
             exit fill_loop
          end if
       end do
       slice_low = slice_low + slice_delta
    end do fill_loop
  end subroutine initialize_pebs_random_points

  subroutine combine_step(number_of_pebbles, in_state, out_state, &
       delta, alpha)
    type(pebble_state), intent(in) :: in_state
    type(pebble_state), intent(out) :: out_state
    type(pebble_delta), intent(in) :: delta
    real(kind=d_t), intent(in) :: alpha
    integer, intent(in) :: number_of_pebbles
    integer :: j, k
    type(pp_slip_single_iterator) :: pp_iterator
    type(vector) :: out_slip, delta_slip, in_slip

    call clear_pp_slip(out_state%pp_slip_data)

    !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size) &
    !$OMP PRIVATE(pp_iterator,in_slip,out_slip,delta_slip,k)
    do j=1,number_of_pebbles
       out_state%pp(j) = in_state%pp(j) + alpha * delta%pdp(j)
       out_state%pv(j) = in_state%pv(j) + alpha * delta%pdv(j)
       out_state%pw(j) = in_state%pw(j) + alpha * delta%pdw(j)
       if(delta%keep_pss(j)) then
          out_state%pss(j) = in_state%pss(j) + alpha * delta%pdss(j)
       else
          out_state%pss(j) = vector(0.0,0.0,0.0)
       end if
       !Check for slips 
       call get_pp_slip_single_iterator(pp_iterator, j)
       !Note: this zeros any slips where there is no entry in the pp_slip_delta
       do while(.not. pp_slip_single_iterator_done(delta%pp_slip_delta,&
            pp_iterator))
          call get_next_pp_slip(delta%pp_slip_delta,pp_iterator,delta_slip,k)
          in_slip = get_old_slip(in_state%pp_slip_data,j,k)
          out_slip = in_slip + alpha * delta_slip
          call add_new_slip(out_state%pp_slip_data,j,k,out_slip)
          !print *,"Out Slip",j,k,in_slip,delta_slip,out_slip,get_old_slip(out_state%pp_slip_data,j,k)
       end do

    end do
  end subroutine combine_step
  
  subroutine combine_steps(out_state,in_state,deltas,delta_indexs,delta_mults)
    type(pebble_state), intent(out) :: out_state
    type(pebble_state), intent(in) :: in_state
    type(pebble_delta),dimension(:), intent(in) :: deltas
    integer,dimension(:),intent(in) :: delta_indexs
    real(kind=d_t), dimension(:), intent(in) :: delta_mults
    integer :: i, j, k, l
    type(pp_slip_single_iterator) :: pp_iterator
    type(vector) :: out_slip, delta_slip, in_slip
    real(kind=d_t) :: alpha
    logical :: keep_pss

    call clear_pp_slip(out_state%pp_slip_data)

    !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size) &
    !$OMP PRIVATE(alpha,l,keep_pss,i)
    pebble_loop: do j=1,size(out_state%pp)
       out_state%pp(j) = in_state%pp(j)
       out_state%pv(j) = in_state%pv(j)
       out_state%pw(j) = in_state%pw(j)
       out_state%pss(j) = in_state%pss(j)
       keep_pss = .true.
       index_loop: do i=1,size(delta_indexs)
          alpha = delta_mults(i)
          l = delta_indexs(i)
          out_state%pp(j) = out_state%pp(j) + alpha * deltas(l)%pdp(j)
          out_state%pv(j) = out_state%pv(j) + alpha * deltas(l)%pdv(j)
          out_state%pw(j) = out_state%pw(j) + alpha * deltas(l)%pdw(j)
          if(deltas(l)%keep_pss(j)) then
             out_state%pss(j) = out_state%pss(j) + alpha * deltas(l)%pdss(j)
          else
             keep_pss = .false.
          end if
       end do index_loop
       if(.not. keep_pss) then
          !XXX is this correct?  Basically, this keeps the pss if any 
          ! of the pss's should be kept.  It might be safer to only keep if 
          ! all of the pss's should be kept
          out_state%pss(j) = vector(0.0,0.0,0.0)
       end if
    end do pebble_loop
    !Note loop inversion.
    index_loop2: do i=1,size(delta_indexs)
       alpha = delta_mults(i)
       l = delta_indexs(i)

       !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size) &
       !$OMP PRIVATE(pp_iterator,in_slip,out_slip,delta_slip,k,i)
       pebble_loop2: do j=1,size(out_state%pp)
          !XXX is this correct? Basically, this keeps any slips if there 
          ! are any slip deltas.

          !Check for slips 
          call get_pp_slip_single_iterator(pp_iterator, j)
          !Note: this zeros any slips where there is no entry in the pp_slip_delta
          do while(.not. pp_slip_single_iterator_done(deltas(l)%pp_slip_delta,&
               pp_iterator))
             call get_next_pp_slip(deltas(l)%pp_slip_delta,pp_iterator,delta_slip,k)
             if(.not. has_slip(out_state%pp_slip_data,j,k)) then
                in_slip = get_old_slip(in_state%pp_slip_data,j,k)
                out_slip = in_slip + alpha * delta_slip
                call add_new_slip(out_state%pp_slip_data,j,k,out_slip)
             else
                in_slip = get_old_slip(out_state%pp_slip_data,j,k)
                out_slip = in_slip + alpha * delta_slip
                call update_slip(out_state%pp_slip_data,j,k,out_slip)
             end if
             !print *,"Out Slip",j,k,in_slip,delta_slip,out_slip,get_old_slip(out_state%pp_slip_data,j,k)
          end do
       end do pebble_loop2
    end do index_loop2
  end subroutine combine_steps

  subroutine combine_steps_selected(out_state,in_state,deltas,delta_indexs,&
       delta_mults, selected, selected_count)
    type(pebble_state), intent(out) :: out_state
    type(pebble_state), intent(in) :: in_state
    type(pebble_delta),dimension(:), intent(in) :: deltas
    integer,dimension(:),intent(in) :: delta_indexs
    real(kind=d_t), dimension(:), intent(in) :: delta_mults
    integer,dimension(:),intent(in) :: selected
    integer,intent(in) :: selected_count
    integer :: i, j, k, l, m
    type(pp_slip_single_iterator) :: pp_iterator
    type(vector) :: out_slip, delta_slip, in_slip
    real(kind=d_t) :: alpha
    logical :: keep_pss

    call clear_pp_slip(out_state%pp_slip_data)

    !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size) &
    !$OMP PRIVATE(alpha,l,keep_pss,i, j, m)
    pebble_loop: do m=1,selected_count
       j = selected(m)
       out_state%pp(j) = in_state%pp(j)
       out_state%pv(j) = in_state%pv(j)
       out_state%pw(j) = in_state%pw(j)
       out_state%pss(j) = in_state%pss(j)
       keep_pss = .true.
       index_loop: do i=1,size(delta_indexs)
          alpha = delta_mults(i)
          l = delta_indexs(i)
          out_state%pp(j) = out_state%pp(j) + alpha * deltas(l)%pdp(j)
          out_state%pv(j) = out_state%pv(j) + alpha * deltas(l)%pdv(j)
          out_state%pw(j) = out_state%pw(j) + alpha * deltas(l)%pdw(j)
          if(deltas(l)%keep_pss(j)) then
             out_state%pss(j) = out_state%pss(j) + alpha * deltas(l)%pdss(j)
          else
             keep_pss = .false.
          end if
       end do index_loop
       if(.not. keep_pss) then
          !XXX is this correct?  Basically, this keeps the pss if any 
          ! of the pss's should be kept.  It might be safer to only keep if 
          ! all of the pss's should be kept
          out_state%pss(j) = vector(0.0,0.0,0.0)
       end if
    end do pebble_loop
    !Note loop inversion.
    index_loop2: do i=1,size(delta_indexs)
       alpha = delta_mults(i)
       l = delta_indexs(i)

       !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size) &
       !$OMP PRIVATE(pp_iterator,in_slip,out_slip,delta_slip,k,i,j,m)
       pebble_loop2: do  m=1,selected_count
          j = selected(m)          
          !XXX is this correct? Basically, this keeps any slips if there 
          ! are any slip deltas.

          !Check for slips 
          call get_pp_slip_single_iterator(pp_iterator, j)
          !Note: this zeros any slips where there is no entry in the pp_slip_delta
          do while(.not. pp_slip_single_iterator_done(deltas(l)%pp_slip_delta,&
               pp_iterator))
             call get_next_pp_slip(deltas(l)%pp_slip_delta,pp_iterator,delta_slip,k)
             if(.not. has_slip(out_state%pp_slip_data,j,k)) then
                in_slip = get_old_slip(in_state%pp_slip_data,j,k)
                out_slip = in_slip + alpha * delta_slip
                call add_new_slip(out_state%pp_slip_data,j,k,out_slip)
             else
                in_slip = get_old_slip(out_state%pp_slip_data,j,k)
                out_slip = in_slip + alpha * delta_slip
                call update_slip(out_state%pp_slip_data,j,k,out_slip)
             end if
             !print *,"Out Slip",j,k,in_slip,delta_slip,out_slip,get_old_slip(out_state%pp_slip_data,j,k)
          end do
       end do pebble_loop2
    end do index_loop2
  end subroutine combine_steps_selected


  subroutine combine_step_selected(in_state, out_state, &
       delta, alpha, selected, selected_count)
    type(pebble_state), intent(in) :: in_state
    type(pebble_state), intent(out) :: out_state
    type(pebble_delta), intent(in) :: delta
    integer,dimension(:),intent(in) :: selected
    integer,intent(in) :: selected_count
    real(kind=d_t), intent(in) :: alpha
    integer :: j, k, m
    type(pp_slip_single_iterator) :: pp_iterator
    type(vector) :: out_slip, delta_slip, in_slip

    call clear_pp_slip(out_state%pp_slip_data)

    !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size) &
    !$OMP PRIVATE(pp_iterator,in_slip,out_slip,delta_slip,k,j, m)
    do m=1,selected_count
       j = selected(m)
       out_state%pp(j) = in_state%pp(j) + alpha * delta%pdp(j)
       out_state%pv(j) = in_state%pv(j) + alpha * delta%pdv(j)
       out_state%pw(j) = in_state%pw(j) + alpha * delta%pdw(j)
       if(delta%keep_pss(j)) then
          out_state%pss(j) = in_state%pss(j) + alpha * delta%pdss(j)
       else
          out_state%pss(j) = vector(0.0,0.0,0.0)
       end if
       !Check for slips 
       call get_pp_slip_single_iterator(pp_iterator, j)
       !Note: this zeros any slips where there is no entry in the pp_slip_delta
       do while(.not. pp_slip_single_iterator_done(delta%pp_slip_delta,&
            pp_iterator))
          call get_next_pp_slip(delta%pp_slip_delta,pp_iterator,delta_slip,k)
          in_slip = get_old_slip(in_state%pp_slip_data,j,k)
          out_slip = in_slip + alpha * delta_slip
          call add_new_slip(out_state%pp_slip_data,j,k,out_slip)
          !print *,"Out Slip",j,k,in_slip,delta_slip,out_slip,get_old_slip(out_state%pp_slip_data,j,k)
       end do

    end do

  end subroutine combine_step_selected

  subroutine get_derivs(nearby_pebble_cache, state, delta, &
       number_of_pebbles, nearby_pebbles_array, nearby_pebbles_size, time)
    type(pebble_state), intent(in) :: state
    type(pebble_delta), intent(out) :: delta
    type(nearby_pebble_cache_type),intent(inout) :: nearby_pebble_cache
    INTEGER, intent(in) :: number_of_pebbles  
    INTEGER, intent(in) ::  nearby_pebbles_size  
    INTEGER, intent(inout) :: nearby_pebbles_array(nearby_pebbles_size)  
    type(vector),dimension(:),pointer :: pp, pv,pw
    type(vector),dimension(:),pointer :: opf, opv, opn
    !type(pp_slip_type),pointer :: pp_slip_data
    type(vector),dimension(:),pointer :: pss
    !type(vector),dimension(:),pointer :: psn
    type(vector) :: old_slip, delta_slip
    real(kind=d_t), intent(in) :: time
    LOGICAL :: changed,keep_slip
    !LOGICAL :: in_old_static_friction
    type(vector) :: t_opf_j, t_opn_j,t_opf_k,t_opn_k

    INTEGER :: l,j,k  
    INTEGER :: nearby_pebbles_count  

    pp => state%pp
    pv => state%pv
    pw => state%pw
    !if(use_static_friction) then
    !Just copy the null.  gfortran gets confused if you don't do this everytime
    !ppsd => state%ppsd
    !pps_index => state%pps_index
    pss => state%pss
    !psn => state%psn
    !else

    opf => delta%pdv
    opv => delta%pdp
    opn => delta%pdw

    call clear_pp_slip(delta%pp_slip_delta)
    !call print_time("#filling_zones")
    call update_nearby_pebble_cache(nearby_pebble_cache,pp)
    !call print_time("#calculating_timesteps")
    !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size) 
    DO j=1,number_of_pebbles  

       if(use_static_friction) then
          call base_force(pp(j),pv(j),pw(j),opf(j),opn(j),delta%keep_pss(j),pss(j),delta%pdss(j),time)
       else 
          call base_force(pp(j),pv(j),pw(j),opf(j),opn(j),delta%keep_pss(j),&
               vector(0.0,0.0,0.0), delta%pdss(j),time)
       end if
       if(is_pebble_of_interest(j)) then
          write(extra_info_unit,*) "#PBF",time,j,"f",opf(j),"n",opn(j),"p",pp(j)
       end if
       call pebble_air_resistance(pv(j),opf(j))
       !         PRINT *,"#base",j,px(j),py(j),pz(j),opfx(j),opfy(j),opfz(j)
       opv(j) = pv(j)
    end do
    !call print_time("#calculating pebble pebble forces")
    !$OMP PARALLEL DO SHARED(opn,opf) &
    !$OMP PRIVATE(nearby_pebbles_array,nearby_pebbles_count,t_opf_j,t_opf_k,t_opn_j,t_opn_k,&
    !$OMP changed,l,k,old_slip,delta_slip,keep_slip) &
    !$OMP SCHEDULE(dynamic,chunk_size)
    do j=1,number_of_pebbles  
       !     Now find the pebbles that are nearby (aka also in same zone)
       call nearby_pebbles_all(nearby_pebble_cache,pp(j), &
            &        nearby_pebbles_array, nearby_pebbles_size, &
            &        nearby_pebbles_count)
       DO l=1,nearby_pebbles_count  
          k = nearby_pebbles_array(l)  
          if (k < j) then 
             old_slip = get_old_slip(state%pp_slip_data,j,k)
             call pebble_pebble_velocity_force4(pp(j),pp(k),t_opf_j,pv(j),&
                  pv(k),pw(j),pw(k),t_opn_j,t_opf_k,t_opn_k,old_slip,&
                  delta_slip,changed,keep_slip)
             if(changed) then
                !print *,"#pp",j,k,old_slip,delta_slip
                !If changed, then update forces.
                call atomic_increment(opf(j),t_opf_j)
                call atomic_increment(opf(k),t_opf_k)
                if(calculate_force_tally) then
                   call tally_pp_force(t_opf_j,t_opf_k,pp(j),pp(k))
                end if
                call atomic_increment(opn(j),t_opn_j)
                call atomic_increment(opn(k),t_opn_k)

                if(is_pebble_of_interest(j)) then
                   write(extra_info_unit,*) "#PPF",time,j,k,"f",t_opf_j,&
                        "n",t_opn_j,"pj",pp(j),"pk",pp(k),"vj",pv(j),&
                        "vk",pv(k),"wj",pw(j),"wk",pw(k)
                end if
                if(is_pebble_of_interest(k)) then
                   write(extra_info_unit,*) "#PPF",time,k,j,"f",t_opf_k,&
                        "n",t_opn_k,"pk",pp(k),"pj",pp(j),"vk",pv(k),&
                        "vj",pv(j),"wk",pw(k),"wj",pw(j)
                end if
                !Note: Slip must be added unless it is to be zeroed.
                !combine step zeros any slip that has no entry in pp_slip_delta
                if(keep_slip) then
                   call add_new_slip(delta%pp_slip_delta,k,j,delta_slip)
                   if(is_pebble_of_interest(j) .or. is_pebble_of_interest(k)) &
                        then 
                      write(extra_info_unit,*) "#PPSD",j,k,old_slip,delta_slip
                   end if
                end if
             end if
          end if
       END DO
    END DO

    !call print_time("#converting forces")
    !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size)
    DO j=1,number_of_pebbles
       !Convert force to acceleration
       opf(j) = (1.0/mass)*opf(j)
       !Convert torque to angular acceleration
       opn(j) = (1.0/moment_of_inertia) * opn(j)
       if(is_pebble_of_interest(j)) then
          write(extra_info_unit,*) "#PPA la",opf(j),"aa",opn(j)
       end if
       if(calculate_force_tally) then
          call tally_force(opf(j),pp,j)
       end if
    END DO

  END SUBROUTINE get_derivs

  subroutine get_derivs_selected(nearby_pebble_cache, state, delta, &
       nearby_pebbles_array, nearby_pebbles_size, &
       &     time,selected, selected_count,&
       provided, provided_count)
    type(pebble_state), intent(in) :: state
    type(pebble_delta), intent(out) :: delta
    type(nearby_pebble_cache_type),intent(inout) :: nearby_pebble_cache
    integer, intent(in),dimension(:) :: selected, provided
    integer, intent(in) :: selected_count, provided_count
    !INTEGER, intent(in) :: number_of_pebbles  
    INTEGER, intent(in) ::  nearby_pebbles_size  
    INTEGER, intent(inout) :: nearby_pebbles_array(nearby_pebbles_size)  
    type(vector),dimension(:),pointer :: pp, pv,pw
    type(vector),dimension(:),pointer :: opf, opv, opn
    !type(pp_slip_type),pointer :: pp_slip_data
    type(vector),dimension(:),pointer :: pss
    !type(vector),dimension(:),pointer :: psn
    type(vector) :: old_slip, delta_slip
    real(kind=d_t), intent(in) :: time
    LOGICAL :: changed,keep_slip
    !LOGICAL :: in_old_static_friction
    type(vector) :: t_opf_j, t_opn_j,t_opf_k,t_opn_k
    logical,dimension(:),allocatable :: is_selected

    INTEGER :: l,j,k, m
    INTEGER :: nearby_pebbles_count  

    pp => state%pp
    pv => state%pv
    pw => state%pw
    !if(use_static_friction) then
    !Just copy the null.  gfortran gets confused if you don't do this everytime
    !ppsd => state%ppsd
    !pps_index => state%pps_index
    pss => state%pss
    !psn => state%psn
    !else

    opf => delta%pdv
    opv => delta%pdp
    opn => delta%pdw

    allocate(is_selected(number_of_pebbles))

    is_selected(:) = .false.

    call write_time(extra_info_unit,"Updating nearby pebble cache")
    call clear_pp_slip(delta%pp_slip_delta)
    !call print_time("#filling_zones")
    call update_nearby_pebble_cache_s(nearby_pebble_cache,pp,&
         provided, provided_count)
   
    do m=1,selected_count
       j = selected(m)
       is_selected(j) = .true.
    end do

    call write_time(extra_info_unit,"Calculating base forces")
    !call print_time("#calculating_timesteps")
    !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size) PRIVATE(m,j)
    DO  m=1,selected_count
       j = selected(m)

       if(use_static_friction) then
          call base_force(pp(j),pv(j),pw(j),opf(j),opn(j),delta%keep_pss(j),pss(j),delta%pdss(j),time)
       else 
          call base_force(pp(j),pv(j),pw(j),opf(j),opn(j),delta%keep_pss(j),&
               vector(0.0,0.0,0.0), delta%pdss(j),time)
       end if
       if(is_pebble_of_interest(j)) then
          write(extra_info_unit,*) "#PBF",time,j,"f",opf(j),"n",opn(j),"p",pp(j)
       end if
       call pebble_air_resistance(pv(j),opf(j))
       !         PRINT *,"#base",j,px(j),py(j),pz(j),opfx(j),opfy(j),opfz(j)
       opv(j) = pv(j)
    end do
    call write_time(extra_info_unit,"Calculating pebble to pebble forces")
    !$OMP PARALLEL DO SHARED(opn,opf) &
    !$OMP PRIVATE(nearby_pebbles_array,nearby_pebbles_count,t_opf_j,t_opf_k,t_opn_j,t_opn_k,&
    !$OMP changed,l,k,j,m,old_slip,delta_slip,keep_slip) &
    !$OMP SCHEDULE(dynamic,chunk_size)
    do m=1,selected_count
       j = selected(m)
       !     Now find the pebbles that are nearby (aka also in same zone)
       call nearby_pebbles_all(nearby_pebble_cache,pp(j), &
            &        nearby_pebbles_array, nearby_pebbles_size, &
            &        nearby_pebbles_count)
       DO l=1,nearby_pebbles_count  
          k = nearby_pebbles_array(l)  
          if (k < j  .or. .not. is_selected(k)) then  
             !TODOXXX is the above if statement correct?
             old_slip = get_old_slip(state%pp_slip_data,j,k)
             call pebble_pebble_velocity_force4(pp(j),pp(k),t_opf_j,pv(j),&
                  pv(k),pw(j),pw(k),t_opn_j,t_opf_k,t_opn_k,old_slip,&
                  delta_slip,changed,keep_slip)
             if(changed) then
                !If changed, then update forces.
                call atomic_increment(opf(j),t_opf_j)
                call atomic_increment(opf(k),t_opf_k)
                if(calculate_force_tally) then
                   call tally_pp_force(t_opf_j,t_opf_k,pp(j),pp(k))
                end if
                call atomic_increment(opn(j),t_opn_j)
                call atomic_increment(opn(k),t_opn_k)

                if(is_pebble_of_interest(j)) then
                   write(extra_info_unit,*) "#PPF",time,j,k,"f",t_opf_j,&
                        "n",t_opn_j,"pj",pp(j),"pk",pp(k),"vj",pv(j),&
                        "vk",pv(k),"wj",pw(j),"wk",pw(k)
                end if
                if(is_pebble_of_interest(k)) then
                   write(extra_info_unit,*) "#PPF",time,k,j,"f",t_opf_k,&
                        "n",t_opn_k,"pk",pp(k),"pj",pp(j),"vk",pv(k),&
                        "vj",pv(j),"wk",pw(k),"wj",pw(j)
                end if
                !Note: Slip must be added unless it is to be zeroed.
                !combine step zeros any slip that has no entry in pp_slip_delta
                if(keep_slip) then
                   call add_new_slip(delta%pp_slip_delta,k,j,delta_slip)
                   if(is_pebble_of_interest(j) .or. is_pebble_of_interest(k)) &
                        then 
                      write(extra_info_unit,*) "#PPSD",j,k,old_slip,delta_slip
                   end if
                end if
             end if
          end if
       END DO
    END DO

    call write_time(extra_info_unit,"Converting force to acceleration")
    !$OMP PARALLEL DO SCHEDULE(dynamic,chunk_size) PRIVATE(m,j)
    DO m=1,selected_count
       j=selected(m)
       !Convert force to acceleration
       opf(j) = (1.0/mass)*opf(j)
       !Convert torque to angular acceleration
       opn(j) = (1.0/moment_of_inertia) * opn(j)
       if(calculate_force_tally) then
          call tally_force(opf(j),pp,j)
       end if
    END DO

  end subroutine get_derivs_selected

  INTEGER FUNCTION find_base_zone(low, increment, size, value)  
    REAL(KIND=D_T), intent(in) :: low,increment,value  
    INTEGER, intent(in) :: size
    INTEGER :: zone  
    zone = find_zone(low, increment, size, value)  
    find_base_zone = iclamp(2,zone,size-1)  
  END FUNCTION find_base_zone

  REAL(kind=d_t) function min_zone(low, increment, bin)
    real(kind=d_t), intent(in) :: low, increment
    integer, intent(in) :: bin
    min_zone = low + increment * bin
  end function min_zone

  real(kind=d_t) function max_zone(low, increment, bin)
    real(kind=d_t), intent(in) :: low, increment
    integer, intent(in) :: bin
    max_zone = low + increment * (bin + 1)
  end function max_zone

  INTEGER FUNCTION find_zone(low, increment, size, value)  
    REAL(KIND=D_T), intent(in) :: low,increment,value  
    INTEGER, intent(in) :: size 
    INTEGER :: bin  
    bin = FLOOR((value - low) / increment)  
    find_zone = iclamp(1,bin,size)  
  END FUNCTION find_zone

  real(kind=d_t) function zone_center(low, increment, bin)
    REAL(KIND=D_T), intent(in) :: low,increment
    integer, intent(in) :: bin
    zone_center = low+increment * (bin + 0.5)
  end function zone_center

  subroutine find_radial_zone(x,y,z,s_z,r_z,z_z)
    REAL(kind=d_t), intent(in) :: x,y,z
    integer, intent(out) :: s_z,r_z,z_z
    REAL(kind=d_t) :: r, theta
    z_z = find_zone(z_low,tally_z_inc, zt_size, z)
    r = sqrt(x**2+y**2)
    theta = atan2(y,x)
    !print *,theta, tally_sections
    s_z = iclamp(1,FLOOR(((theta + pi) / (2* pi)) * tally_sections) + 1,&
         tally_sections)
    r_z = find_zone(0.0_d_t,tally_radial_inc,rt_size, r)
  end subroutine find_radial_zone

  subroutine radial_zone_center(s_z,r_z,z_z,x,y,z) 
    INTEGER, intent(in) :: s_z, r_z, z_z
    REAL(kind=d_t), intent(out) :: x, y, z
    REAL(kind=d_t) :: theta, r
    theta = (real(s_z-1) / real(tally_sections) + 0.5) * 2 * pi
    r = (real(r_z - 1) + 0.5) * tally_radial_inc
    z = z_low+ tally_z_inc * (z_z + 0.5)
    x = cos(theta) * r
    y = sin(theta) * r
  end subroutine radial_zone_center

  subroutine regularize_radial_velocity(pos,velocity,radial_velocity,&
       s_z,r_z,z_z) 
    integer, intent(in) :: s_z, r_z, z_z
    type(vector), intent(in) :: pos, velocity
    type(vector), intent(out) :: radial_velocity
    real(kind=d_t) :: theta, dr,  dtheta, ctheta
    ! x plus velocity, y plus velocity
    type(vector) :: r_hat, theta_hat, c_r_hat, c_theta_hat, z_hat
    real(kind=d_t) :: cx,cy,cz
    theta = atan2(pos%y,pos%x)
    z_hat = vector(0.0,0.0,1.0)
    r_hat = vector(cos(theta),sin(theta),0.0)
    theta_hat = vector(-sin(theta),cos(theta),0.0)
    dr = velocity .dot. r_hat
    dtheta = velocity .dot. theta_hat
    !call find_radial_zone(pos%x,pos%y,pos%z,s_z,r_z,z_z)
    call radial_zone_center(s_z,r_z,z_z,cx,cy,cz)
    ctheta = atan2(cy, cx)
    c_r_hat = vector(cos(ctheta),sin(ctheta),0.0)
    c_theta_hat = vector(-sin(ctheta),cos(ctheta),0.0)
    radial_velocity = dr *  c_r_hat + dtheta * c_theta_hat + velocity%z * z_hat
  end subroutine regularize_radial_velocity

  subroutine random_point_in_sphere(x_cent,y_cent,z_cent,radius,x,y,z,seed)
    !Returns a random point somewhere in the sphere.
    integer(kind=li), intent(inout) :: seed
    real(kind=d_t), intent(in) :: x_cent,y_cent,z_cent,radius
    real(kind=d_t), intent(out) :: x,y,z
    do 
       x = (random(seed)-0.5)*2.0*radius
       y = (random(seed)-0.5)*2.0*radius
       z = (random(seed)-0.5)*2.0*radius
       if (x**2 + y**2 + z**2 <= radius**2) then
          ! found something inside sphere, so now
          ! put in terms of real coordinates and return.
          x = x + x_cent
          y = y + y_cent
          z = z + z_cent
          return
       end if
    end do
  end subroutine random_point_in_sphere

  REAL(KIND=D_T) FUNCTION RANDOM(L)  
    INTEGER(KIND=LI), intent(inout) ::  L      
    L=MOD((7**5)*L,2147483647_LI)  
    RANDOM=REAL(L)*(1.0/REAL(2147483647))  
    RETURN  
  END FUNCTION RANDOM

  function any_pebbles_of_interest()
    logical :: any_pebbles_of_interest
    any_pebbles_of_interest = pebble_of_interest > 0
  end function any_pebbles_of_interest

  function is_pebble_of_interest(pebble_number) 
    integer, intent(in) :: pebble_number
    logical :: is_pebble_of_interest
    if(pebble_number == pebble_of_interest) then
       is_pebble_of_interest = .true.
    else 
       is_pebble_of_interest = .false.
    end if
  end function is_pebble_of_interest

  pure FUNCTION pebble_pebble_overlap_dist(p,op,pebble_radius)
    REAL(KIND=D_T) :: pebble_pebble_overlap_dist
    type(vector), intent(in) :: p,op
    REAL(KIND=D_T), intent(in) :: pebble_radius
    REAL(KIND=D_T) :: dist_sqr, boundry_sqr  
    boundry_sqr = (2.0*pebble_radius)**2  
    dist_sqr = vmag_sqr(p - op)  
    pebble_pebble_overlap_dist = SQRT(boundry_sqr) - SQRT(dist_sqr)  
  END FUNCTION pebble_pebble_overlap_dist

  pure LOGICAL FUNCTION pebble_pebble_overlap(a,b, pebble_radius)
    REAL(KIND=D_T) :: dist_sqr 
    REAL(KIND=D_T), intent(in) :: pebble_radius  
    type(Vector), intent(in) :: a,b
    dist_sqr = vmag_sqr(a - b)
    if ( dist_sqr .ge. (2.0*pebble_radius)**2) THEN  
       pebble_pebble_overlap = .FALSE.  
    ELSE  
       pebble_pebble_overlap = .TRUE.
    END IF
  END FUNCTION pebble_pebble_overlap

  pure SUBROUTINE pebble_air_resistance(v,f)  
    type(vector), intent(in) :: v
    type(vector), intent(inout) :: f
    REAL(KIND=D_T) :: v_mag_sqr  
    v_mag_sqr = vmag_sqr(v)  
    f = f - air_friction*v_mag_sqr*v
  END SUBROUTINE pebble_air_resistance

  function get_static_potential_energy(current_state,i)
    !Note, ignores half the potential since it is only indexed by one 
    ! direction
    type(pebble_state), intent(in) :: current_state
    type(pp_slip_single_iterator) :: pp_iterator
    type(vector) :: slip
    integer, intent(in) :: i
    real(kind=d_t) :: total
    real(kind=d_t) :: get_static_potential_energy
    integer :: j
    total = 0.0
    if(.not. use_static_friction) then
       get_static_potential_energy = 0.0       
       return
    end if
    call get_pp_slip_single_iterator(pp_iterator,i)
    do while(.not. pp_slip_single_iterator_done(current_state%pp_slip_data,&
         pp_iterator))
       call get_next_pp_slip(current_state%pp_slip_data,pp_iterator,slip,j)
       total = total + vmag_sqr(slip) * &
            static_friction_hooke / 2.0
    end do
    get_static_potential_energy = total
  end function get_static_potential_energy

  pure function make_perpendicular(orig, normal) 
    type(vector), intent(in) :: orig, normal
    type(vector) :: make_perpendicular
    make_perpendicular = abs(orig) * normalize(perp_projection(orig,normal))
  end function make_perpendicular

!!$  SUBROUTINE pebble_pebble_velocity_force(pos, other_pos,f,v,other_v, &
!!$       omega, other_omega, torque, other_f, other_torque, &
!!$       old_slip, delta_slip, changed, keep_slip)
!!$    REAL(KIND=D_T) :: two_pebble_radius_sqr,cent_dist_sqr  
!!$    REAL(KIND=D_T) :: inside_dist, cent_dist, raw_force  
!!$    REAL(KIND=D_T) :: fn_mag,vt_mag,ft_mag, current_wear  
!!$    ! v is velocity, other_v is the other vectors velocity, and 
!!$    ! r_v is the relative velocity of the two vectors
!!$    type(vector), intent(in) :: v, other_v
!!$    type(vector) :: r_v
!!$    type(vector), intent(in) :: pos, other_pos
!!$    type(vector), intent(in) :: omega, other_omega
!!$    type(vector), intent(in) :: old_slip
!!$    type(vector) :: delta_pos
!!$    type(vector) :: v_normal, v_transverse
!!$    type(vector) :: f_normal, f_transverse
!!$    type(vector), intent(out) :: f, other_f  
!!$    type(vector), intent(out) :: torque, other_torque, delta_slip
!!$    logical,intent(out) :: changed, keep_slip
!!$
!!$    real(kind=d_t) :: static_mag, slip_mag
!!$    type(vector) :: adjusted_slip
!!$    type(vector) :: normal
!!$    type(vector) :: static_friction_force
!!$    changed = .false.
!!$
!!$    two_pebble_radius_sqr = 1.5*((2.0*pebble_radius)**2)  
!!$    delta_pos = pos - other_pos
!!$    normal = normalize((-1.0) * delta_pos)
!!$    r_v = v - other_v - &
!!$         ((pebble_radius * (omega + other_omega)) .cross. &
!!$         ((1.0/(2.0*pebble_radius)) * delta_pos))
!!$    cent_dist_sqr = vmag_sqr(delta_pos)
!!$    IF(cent_dist_sqr < two_pebble_radius_sqr) THEN  
!!$       cent_dist = SQRT(cent_dist_sqr)  
!!$       inside_dist = cent_dist - 2.0*pebble_radius  
!!$       IF (inside_dist < 0.0) THEN  
!!$          raw_force = pebble_pebble_hooke * inside_dist  
!!$          IF (ABS(cent_dist) < 1.0e-10) THEN  
!!$             cent_dist = 1.0e-10  
!!$
!!$          END IF
!!$          v_normal = r_v .proj. ((-1.0)*delta_pos)
!!$          v_transverse = r_v - v_normal
!!$          !           First figure out the normal force
!!$          f_normal = (raw_force / cent_dist) * ((-1.0) * delta_pos) - &
!!$               normal_dash_pot_constant * v_normal          
!!$          !           Then figure out the transverse force
!!$
!!$
!!$          fn_mag = abs(f_normal)
!!$          vt_mag = abs(v_transverse)
!!$          ft_mag = min(fn_mag*kinetic_friction,&
!!$               &           vt_mag*transverse_dash_pot_constant)
!!$
!!$          if(vmag_sqr(v_transverse) < static_velocity_sqr_max .and. &
!!$               use_static_friction) then
!!$             !calculate static friction
!!$             keep_slip = .true.
!!$             slip_mag = abs(old_slip)
!!$             if(slip_mag*static_friction_hooke > 1.0e-20_d_t) then
!!$                adjusted_slip = perp_projection(old_slip,normal)
!!$                if(linear_static_velocity_clamp) then
!!$                   static_mag = min(slip_mag * static_friction_hooke * &
!!$                        (1.0 - vt_mag/static_velocity_max),&
!!$                        fn_mag * pp_static_friction)
!!$                else 
!!$                   static_mag = min(slip_mag * static_friction_hooke,&
!!$                        fn_mag * pp_static_friction)
!!$                end if
!!$                if(slip_mag * static_friction_hooke >  fn_mag*pp_static_friction*1.01 &
!!$                     .and. zero_slips_gt_mu_fn) then
!!$                   keep_slip = .false.
!!$                end if
!!$                !print *,"#pp_forces lsm",slip_mag * static_friction_hooke * &
!!$                !        (1.0 - vt_mag/static_velocity_max),"sm",&
!!$                !        slip_mag * static_friction_hooke,"mufn",&
!!$                !        fn_mag * pp_static_friction,"os",old_slip,"as",adjusted_slip
!!$                static_friction_force = (-static_mag)* normalize(adjusted_slip)
!!$             else
!!$                static_mag = 0.0
!!$                static_friction_force = vector(0.0,0.0,0.0)
!!$             end if
!!$             if(zero_long_slips .and. slip_mag * static_friction_hooke > &
!!$                  fn_mag * pp_static_friction) then
!!$                ! Slip direct force greater than normal force * static
!!$                keep_slip = .false.
!!$             end if
!!$             delta_slip = (1.0/vmag_sqr(delta_pos))*((v - other_v) .cross. delta_pos .cross. old_slip)+v_transverse
!!$          else 
!!$             static_mag = 0.0
!!$             static_friction_force = (/ 0.0 , 0.0, 0.0 /)
!!$             delta_slip = vector(0.0,0.0,0.0)
!!$             keep_slip = .false.
!!$          end if
!!$          !print *,"#pp_forces ks",keep_slip
!!$
!!$          if(calculate_pebble_wear_vol) then
!!$             current_wear = abs(v_transverse) * (abs(f_normal)**(normal_force_exponent))
!!$             !$OMP CRITICAL (wear_sum)
!!$             pebble_volume_wear_sum = pebble_volume_wear_sum + current_wear
!!$             pebble_wear_sum_tally(find_zone(z_low, zone_increment, z_size, pos%z)) = &
!!$                  pebble_wear_sum_tally(find_zone(z_low, zone_increment, z_size, pos%z)) + current_wear
!!$             pebble_wear_sum_tally(find_zone(z_low, zone_increment, z_size, other_pos%z)) = &
!!$                  pebble_wear_sum_tally(find_zone(z_low, zone_increment, z_size, other_pos%z)) + current_wear
!!$             !$OMP END CRITICAL (wear_sum)
!!$          end if
!!$
!!$          !print *,"#ppvf dump","fn",f_normal,"sf",static_friction_force,"dp",delta_pos,"vt",v_transverse,"fm",ft_mag,"vm",vt_mag
!!$          IF (ABS(ft_mag) + ABS(static_mag) < 1.0e-10) THEN
!!$             f = f_normal
!!$             other_f = (-1.0)*f_normal
!!$             torque = vector(0.0,0.0,0.0)
!!$             other_torque = vector(0.0,0.0,0.0)
!!$             changed = .true.
!!$
!!$          ELSE
!!$             if(abs(ft_mag) < 1.0e-10) then 
!!$                f_transverse = static_friction_force 
!!$             else
!!$                f_transverse = (-ft_mag/vt_mag) * v_transverse +&
!!$                     static_friction_force
!!$             end if
!!$             f = f_normal + f_transverse
!!$             other_f = (-1.0)*(f_normal + f_transverse)
!!$             !Torque is the same for both, since it rotates both the same way
!!$             !Opposite forces cancel since they are on opposite sides.
!!$             torque = (((-1.0/2.0) * delta_pos) .cross. f_transverse)
!!$             other_torque = (((-1.0/2.0) * delta_pos) .cross. f_transverse)
!!$             changed = .true.
!!$          END IF
!!$       END IF
!!$    END IF
!!$  END SUBROUTINE pebble_pebble_velocity_force

  SUBROUTINE pebble_pebble_velocity_force4(pos, other_pos,f,v,other_v, &
       omega, other_omega, torque, other_f, other_torque, &
       old_slip, delta_slip, changed, keep_slip)
    REAL(KIND=D_T) :: two_pebble_radius_sqr,cent_dist_sqr  
    REAL(KIND=D_T) :: inside_dist, cent_dist, raw_force  
    REAL(KIND=D_T) :: fn_mag,vt_mag,ft_mag, ft_sum_mag2
    ! v is velocity, other_v is the other vectors velocity, and 
    ! r_v is the relative velocity of the two vectors
    type(vector), intent(in) :: v, other_v
    type(vector) :: r_v
    type(vector), intent(in) :: pos, other_pos
    type(vector), intent(in) :: omega, other_omega
    type(vector), intent(in) :: old_slip
    type(vector) :: delta_pos
    type(vector) :: v_normal, v_transverse
    type(vector) :: f_normal, f_transverse
    type(vector), intent(out) :: f, other_f  
    type(vector), intent(out) :: torque, other_torque, delta_slip
    logical,intent(out) :: changed, keep_slip

    real(kind=d_t) :: static_mag, slip_mag, max_transverse_force
    type(vector) :: adjusted_slip
    type(vector) :: normal
    type(vector) :: static_friction_force
    changed = .false.

    two_pebble_radius_sqr = 1.5*((2.0*pebble_radius)**2)  
    delta_pos = pos - other_pos
    normal = normalize((-1.0) * delta_pos)
    r_v = v - other_v - &
         ((pebble_radius * (omega + other_omega)) .cross. &
         ((1.0/(2.0*pebble_radius)) * delta_pos))
    !print *,"#r_v",r_v
    cent_dist_sqr = vmag_sqr(delta_pos)
    IF(cent_dist_sqr < two_pebble_radius_sqr) THEN  
       cent_dist = SQRT(cent_dist_sqr)  
       inside_dist = cent_dist - 2.0*pebble_radius  
       IF (inside_dist < 0.0) THEN  
          raw_force = pebble_pebble_hooke * inside_dist  
          IF (ABS(cent_dist) < 1.0e-10) THEN  
             cent_dist = 1.0e-10  

          END IF
          v_normal = r_v .proj. ((-1.0)*delta_pos)
          v_transverse = r_v - v_normal
          !           First figure out the normal force
          f_normal = (raw_force / cent_dist) * ((-1.0) * delta_pos) - &
               normal_dash_pot_constant * v_normal          
          !           Then figure out the transverse force


          fn_mag = abs(f_normal)
          vt_mag = abs(v_transverse)
          ft_mag = vt_mag*transverse_dash_pot_constant
          max_transverse_force = fn_mag*kinetic_friction

          if(vmag_sqr(v_transverse) < static_velocity_sqr_max .and. &
               use_static_friction) then
             !calculate static friction
             keep_slip = .true.
             slip_mag = abs(old_slip)
             if(slip_mag*static_friction_hooke > 1.0e-20_d_t) then
                adjusted_slip = perp_projection(old_slip,normal)
                !print *,"#cos_as_os",(1.0/(abs(adjusted_slip)*abs(old_slip)))*(adjusted_slip .dot. old_slip)
                static_mag = slip_mag * static_friction_hooke                
                !print *,"#pp_forces lsm",slip_mag * static_friction_hooke * &
                !        (1.0 - vt_mag/static_velocity_max),"sm",&
                !        slip_mag * static_friction_hooke,"mufn",&
                !        fn_mag * static_friction,"os",old_slip,"as",adjusted_slip
                static_friction_force = (-static_mag)* normalize(adjusted_slip)
                max_transverse_force = fn_mag*pp_static_friction
             else
                static_mag = 0.0
                static_friction_force = vector(0.0,0.0,0.0)
             end if
             if(zero_long_slips .and. slip_mag * static_friction_hooke > &
                  fn_mag * pp_static_friction * long_slip_mult) then
                ! Slip direct force greater than normal force * static
                !print *,"#ZLS pp",slip_mag,static_friction_hooke,fn_mag,pp_static_friction,long_slip_mult
                keep_slip = .false.
             end if
             delta_slip = (-1.0/vmag_sqr(delta_pos))*((v - other_v) .cross. delta_pos .cross. old_slip)+v_transverse
             if(decrease_long_slips .and. slip_mag * static_friction_hooke > &
                  fn_mag * pp_static_friction * decrease_slip_mult) then
                !basically this add a delta that tries to decrease the 
                ! slip by the amount that it exceeds the decrease_slip_mult
                delta_slip = delta_slip - decrease_slip_slope*(&
                     slip_mag * static_friction_hooke - &
                     fn_mag * pp_static_friction * decrease_slip_mult)*old_slip
             end if
          else 
             static_mag = 0.0
             static_friction_force = vector(0.0,0.0,0.0)
             delta_slip = vector(0.0,0.0,0.0)
             keep_slip = .false.
          end if
          !print *,"#delta_slip",delta_slip
          !print *,"#pp_forces ks",keep_slip

          if(calculate_pebble_wear_vol) then
             call tally_pebble_pebble_wear(pos,other_pos,v_transverse,f_normal)
          end if

          !print *,"#ppvf dump","fn",f_normal,"sf",static_friction_force,"dp",delta_pos,"vt",v_transverse,"fm",ft_mag,"vm",vt_mag
          IF (ABS(ft_mag) + ABS(static_mag) < 1.0e-10) THEN
             f = f_normal
             other_f = (-1.0)*f_normal
             torque = vector(0.0,0.0,0.0)
             other_torque = vector(0.0,0.0,0.0)
             changed = .true.

          ELSE
             if(abs(ft_mag) < 1.0e-10) then 
                f_transverse = static_friction_force 
             else
                f_transverse = (-ft_mag/vt_mag) * v_transverse +&
                     static_friction_force
             end if
             ft_sum_mag2 =  vmag_sqr(f_transverse)
             if(ft_sum_mag2 > max_transverse_force**2) then
                f_transverse = (max_transverse_force/sqrt(ft_sum_mag2))*f_transverse
             end if

             f = f_normal + f_transverse
             other_f = (-1.0)*(f_normal + f_transverse)
             !Torque is the same for both, since it rotates both the same way
             !Opposite forces cancel since they are on opposite sides.
             torque = (((-1.0/2.0) * delta_pos) .cross. f_transverse)
             other_torque = (((-1.0/2.0) * delta_pos) .cross. f_transverse)
             changed = .true.
          END IF
       END IF
    END IF
  END SUBROUTINE pebble_pebble_velocity_force4

  subroutine print_displacement(time)
    real(kind=d_t), intent(in) :: time
    type(vector) :: displacement, wall_velocity
    logical :: displaced
    call calc_earthquake_displacement(displacement,wall_velocity,time, displaced)
    print *,"#EQ_displacement ",time," disp ",displacement," wall ",wall_velocity
  end subroutine print_displacement

  subroutine base_force(pos,velocity,omega,force,torque,has_slip_delta,slip,slip_delta, time)
    type(vector), intent(in) :: pos, velocity, omega, slip
    type(vector), intent(out) :: force, torque, slip_delta
    real(kind=d_t), intent(in) :: time
    logical,intent(out) :: has_slip_delta    
    type(vector) :: displacement, wall_velocity
    logical :: displaced
    call calc_earthquake_displacement(displacement,wall_velocity,time,displaced)
    if(displaced) then
       call default_base_force_sd(pos+displacement,velocity+wall_velocity, omega,&
            force,torque,has_slip_delta,slip,slip_delta,calculate_force_tally)
    else 
       call default_base_force_sd(pos,velocity, omega, force, torque, &
            has_slip_delta, slip, slip_delta,calculate_force_tally)
    end if
  end subroutine base_force

end module marbles9_common_module
