! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module network_domain_module
  use vector_type
  use types
  use number_utils
  implicit none

  type network_domain_type
     private
     integer :: number_of_processes
     !Contains the location of each boundary
     ! z_boundaries(i)*z_delta+z_low is the lower boundary for process i.
     ! z_boundaries(i+1)*z_delta+z_low is the upper boundary for process i.
     integer, dimension(:),pointer :: z_boundaries
     type(network_domain_location_type), dimension(:),pointer :: location_cache
     ! z_boundaries must be increasing
     real(kind=d_t) :: x_low,x_high,y_low,y_high,z_low,z_high
     real(kind=d_t) :: boundary_distance, z_delta, cache_spacing
  end type network_domain_type

  type network_domain_location_type
     private
     integer :: primary_domain
     integer :: boundary_low, boundary_high
  end type network_domain_location_type

  integer,parameter :: NON_DOMAIN = -1

  private
  public :: initialize_network_domain,get_location,primary_domain
  public :: network_domain_type, network_domain_location_type
  public :: number_of_boundary_domains,boundary_domain,balance_domains
  public :: debug_domain,print_domain_balance, balance_domains_work
  public :: primary_domain_for_position
  public :: get_integer_data_save,load_domains_from_integer_array
 
contains 

  ! Initializes the domain.  Domain should be balanced, otherwise
  !  many pebbles can end up on one process
  subroutine initialize_network_domain(number_of_processes,&
       x_high,x_low,y_high,y_low,z_high,z_low,domain,boundary_distance)
    real(kind=d_t),intent(in) :: x_low,x_high,y_low,y_high,z_low,z_high,&
         boundary_distance
    integer,intent(in) :: number_of_processes
    type(network_domain_type),intent(out) :: domain
    integer :: z_size, i
    real(kind=d_t) :: even_spacing
    z_size = max(ceiling((z_high - z_low)/boundary_distance)+1,3*number_of_processes)

    domain%x_low  = x_low
    domain%x_high = x_high
    domain%y_low  = y_low
    domain%y_high = y_high
    domain%z_low  = z_low
    domain%z_high = z_high
    domain%boundary_distance = boundary_distance
    domain%cache_spacing = boundary_distance/16.0
    domain%number_of_processes = number_of_processes
    allocate(domain%z_boundaries(0:number_of_processes))
    allocate(domain%location_cache(int(ceiling((z_high-z_low)/domain%cache_spacing))))
    !domain%location_cache(:)%primary_domain = NON_DOMAIN
    domain%z_delta = (z_high - z_low)/(huge(domain%z_boundaries(0))-1)
    even_spacing = (z_high - z_low)/number_of_processes
    do i=0,number_of_processes
       domain%z_boundaries(i) =get_closest_n_value(domain,even_spacing*i+z_low)
    end do
    call update_cache(domain)
  end subroutine initialize_network_domain
  
  subroutine balance_domains(number_of_processes,domain,positions)
    integer,intent(in) :: number_of_processes
    type(network_domain_type),intent(inout) :: domain
    type(vector),dimension(:),intent(in) :: positions
    integer :: i,mean_node_size
    real(kind=d_t),dimension(:),allocatable :: pebble_heights, work
    integer,dimension(:),allocatable :: primary_count, boundary_count
    
    
    mean_node_size = size(positions)/number_of_processes
    if(size(domain%z_boundaries) /= number_of_processes+1) then
       deallocate(domain%z_boundaries)
       allocate(domain%z_boundaries(0:number_of_processes))
    end if
    !First, start off by balancing the nodes by number of primary pebbles
    !Sort the pebbles by z value, and then start at the bottom and assign 
    ! z boundaries based on when the number of primary pebbles is passed.
    allocate(pebble_heights(size(positions)))
    forall(i=1:size(positions))
       pebble_heights(i) = positions(i)%z
    end forall
    call sort_real(pebble_heights)
    domain%z_boundaries(0) = get_closest_n_value(domain,domain%z_low)
    domain%z_boundaries(number_of_processes) = get_closest_n_value(domain,domain%z_high)
    do i=1,number_of_processes-1
       domain%z_boundaries(i) = get_closest_n_value(domain,pebble_heights(i*mean_node_size))
    end do

    call update_cache(domain)

    !Then, calculate the number of primary + boundary pebbles for each node,
    ! and use the balance domains function that takes a work parameter 
    ! to refine the balance. 
    allocate(primary_count(0:number_of_processes-1),boundary_count(0:number_of_processes-1))
    call calculate_domain_balance(domain,positions,primary_count,boundary_count)

    allocate(work(0:number_of_processes-1))
    
    work(:) = primary_count(:)+boundary_count(:)

    !call print_domain_balance(domain,positions)

    call balance_domains_work(domain,positions,work)

    !call print_domain_balance(domain,positions)

  end subroutine balance_domains

  subroutine balance_domains_work(domain,positions,in_work)
    integer :: number_of_processes
    type(network_domain_type),intent(inout) :: domain
    type(vector),dimension(:),intent(in) :: positions
    real(kind=d_t),dimension(0:domain%number_of_processes-1),&
         intent(in) :: in_work
    integer :: i, used_pebbles
    real(kind=d_t),dimension(:),allocatable :: pebble_heights, transfer_down,&
         work
    integer,dimension(:),allocatable :: primary_count, boundary_count, &
         boundary_move,desired_primary
    real(kind=d_t) :: orig_work_sum

    !call print_time("#start bdw")
    
    number_of_processes = domain%number_of_processes

    allocate(pebble_heights(size(positions)),work(0:number_of_processes-1))
    forall(i=1:size(positions))
       pebble_heights(i) = positions(i)%z
    end forall
    call sort_real(pebble_heights)

    print *,"#Node in work",in_work

    orig_work_sum = sum(in_work)

    if(size(in_work) /= number_of_processes) then
       print *,"#ERROR work array should match number_of_processes",&
            size(in_work),number_of_processes
    end if

    !Make the average work value = 1.0
    forall(i=0:number_of_processes-1)
       work(i) = (number_of_processes/orig_work_sum)*in_work(i)
    end forall

    print *,"#Node Work",work

    allocate(transfer_down(0:number_of_processes-1))

    !Transfer down is an array that tells how much work to move to lower
    ! indexed nodes. 
    ! For example, if transfer_down(1) = 0.2, then node 1 should move about 
    ! 20% of its work to node 0
    ! Basically, we keep a running total of the work that needs to be transfered
    ! down, and at each node add the extra work that node needs(1.0 - work(i-1))
    transfer_down(0) = 0.0
    do i=1,number_of_processes-1
       transfer_down(i) = (1.0 - work(i-1)) + transfer_down(i-1)
    end do

    print *,"#Node Transfer Down",transfer_down

    !Clamp transfer down to between -0.2 and 0.2  
    !Values beyond that are probably wrong
    do i=0,number_of_processes-1
       transfer_down(i) = clamp(-0.2_d_t,transfer_down(i),0.2_d_t)
    end do

    allocate(primary_count(0:number_of_processes-1),boundary_count(0:number_of_processes-1))
    call calculate_domain_balance(domain,positions,primary_count,boundary_count)

    print *,"#Primary count",primary_count

    !Move the boundaries
    allocate(boundary_move(0:number_of_processes),desired_primary(0:number_of_processes-1))
    boundary_move(:) = 0

    !boundary_move(i) means that node i-1 gets that many pebbles from node i
    forall(i=0:number_of_processes-2)
       !boundary_move(i) must not be less than -primary(i-1) 
       ! or more than primary(i)  because that would zero the primary_count
       ! Limit actually is /3 so that no more than 2/3s are removed at 
       ! any given time. (Since this forces the total removed to be less than 
       ! or equal to 2/3*primary_count)
       boundary_move(i+1) = iclamp(-primary_count(i+1-1)/3,&
            int((primary_count(i)+boundary_count(i))*(transfer_down(i+1)/work(i)))/5,&
            primary_count(i+1)/3)
    end forall

    print *,"#Node boundary move",boundary_move

    ! move the boundary down by the given number of pebbles (- for up)
    forall(i=0:number_of_processes-1)
       desired_primary(i) = primary_count(i) + boundary_move(i+1) - boundary_move(i)
    end forall

    print *,"#Node Desired Primary",desired_primary

    used_pebbles = min(size(pebble_heights),desired_primary(0))
    do i=1,number_of_processes-1
       domain%z_boundaries(i) = get_closest_n_value(domain,pebble_heights(used_pebbles))
       used_pebbles=min(used_pebbles + desired_primary(i),size(pebble_heights))
    end do

    call update_cache(domain)

    !call print_time("#end bdw")

  end subroutine balance_domains_work

  subroutine calculate_domain_balance(domain,positions,primary_count,boundary_count)
    type(network_domain_type),intent(in) :: domain
    type(vector),dimension(:),intent(in) :: positions
    integer,dimension(0:domain%number_of_processes-1),intent(inout) :: &
         primary_count,boundary_count
    type(network_domain_location_type) :: location
    integer :: i,j,k

    primary_count(:) = 0
    boundary_count(:) = 0

    do i=1,size(positions)
       call store_location(domain,positions(i),location)
       k = primary_domain(location)
       primary_count(k) = primary_count(k) + 1
       do j=1,number_of_boundary_domains(location)
          k = boundary_domain(location,j)
          boundary_count(k) = boundary_count(k) + 1
       end do
    end do

  end subroutine calculate_domain_balance

  ! Debuging routine to print domain balance
  subroutine print_domain_balance(domain,positions)
    type(network_domain_type),intent(in) :: domain
    type(vector),dimension(:),intent(in) :: positions
    integer :: i
    integer,dimension(0:domain%number_of_processes-1) :: primary_count, boundary_count
    
    call calculate_domain_balance(domain,positions,primary_count,boundary_count)

    do i = lbound(primary_count,1),ubound(primary_count,1)
       print *,"#domain_process_count",i,primary_count(i),boundary_count(i)
    end do
    
    
  end subroutine print_domain_balance

  ! Debuging routine to show state of domain.
  subroutine debug_domain(domain)
    type(network_domain_type),intent(in) :: domain
    integer :: i
    
    print *,"#domain_number_of_processes",domain%number_of_processes
    do i=0,domain%number_of_processes
       print *,"#domain_z_boundary",i,domain%z_boundaries(i),&
            domain%z_boundaries(i)*domain%z_delta + domain%z_low
    end do
  end subroutine debug_domain

  pure function primary_domain_for_position(domain,position)
    type(vector),intent(in) :: position
    type(network_domain_type),intent(in) :: domain
    type(network_domain_location_type) :: location

    integer :: primary_domain_for_position
    call store_location(domain,position,location)
    primary_domain_for_position = primary_domain(location)
  end function primary_domain_for_position

  pure function get_z_value(domain, index)
    real(kind=d_t) :: get_z_value
    type(network_domain_type),intent(in) :: domain
    integer,intent(in) :: index
    get_z_value = domain%z_low + domain%z_delta * index
  end function get_z_value

  pure function get_closest_n_value(domain, z)
    real(kind=d_t),intent(in) :: z
    type(network_domain_type),intent(in) :: domain
    integer :: get_closest_n_value
    get_closest_n_value = nint((z-domain%z_low)/domain%z_delta)
  end function get_closest_n_value
    

  pure function get_index_for_z(domain,z)
    type(network_domain_type),intent(in) :: domain
    real(kind=d_t),intent(in) :: z
    integer :: get_index_for_z
    integer :: low, high, mid, value
    integer :: z_n
    !real(kind=d_t) :: value
    z_n = get_closest_n_value(domain,z)
    low = 0
    high = domain%number_of_processes
    bin_search: do 
       if(high - 1 <= low) then
          exit bin_search
       end if
       mid = (high + low) / 2
       value = domain%z_boundaries(mid)
       if(z_n < value) then
          high = mid
       else if(value <= z_n) then
          low = mid 
       end if       
    end do bin_search
    
    get_index_for_z = low
  end function get_index_for_z

  subroutine update_cache(domain)
    type(network_domain_type),intent(inout) :: domain
    type(network_domain_location_type) :: location
    integer :: cache_index 
    real(kind=d_t) :: normalized_z
  
    do cache_index = 1,size(domain%location_cache)
       normalized_z = (cache_index+0.5)*domain%cache_spacing+domain%z_low

       location%primary_domain = get_index_for_z(domain,normalized_z)
       !primary_domain_for_position(domain,position)

       !Calculate boundary locations
       location%boundary_low = get_index_for_z(domain,normalized_z-domain%boundary_distance)
       location%boundary_high = get_index_for_z(domain,normalized_z+domain%boundary_distance)
       domain%location_cache(cache_index) = location
       
    end do
  end subroutine update_cache

  pure subroutine store_location(domain,position,location)
    type(vector),intent(in) :: position
    type(network_domain_type),intent(in) :: domain
    type(network_domain_location_type),intent(out) :: location
    integer :: cache_index

    cache_index = nint((position%z-domain%z_low)/domain%cache_spacing)
    
    location = domain%location_cache(cache_index)

  end subroutine store_location

  ! Returns a location type for the position
  pure function get_location(domain,position)
    type(vector),intent(in) :: position
    type(network_domain_type),intent(in) :: domain
    type(network_domain_location_type) :: get_location

    call store_location(domain,position,get_location)


  end function get_location

  ! Returns the primary domain for the location
  pure function primary_domain(location)
    type(network_domain_location_type),intent(in) :: location
    integer :: primary_domain
    primary_domain = location%primary_domain
  end function primary_domain

  ! Returns the number of boundary that this location is part of 
  pure function number_of_boundary_domains(location)
    type(network_domain_location_type),intent(in) :: location
    integer :: number_of_boundary_domains
    number_of_boundary_domains = location%boundary_high - location%boundary_low
  end function number_of_boundary_domains

  ! Returns one domain this location is a part of
  pure function boundary_domain(location,n)
    type(network_domain_location_type),intent(in) :: location
    integer,intent(in) :: n
    integer :: boundary_domain
    
    !The boundary domain is between boundary_low and boundary_high
    ! and does not equal primary_domain.  
    boundary_domain = location%boundary_low + n - 1
    if(boundary_domain >= location%primary_domain) then       
       boundary_domain = boundary_domain + 1
    end if
        
  end function boundary_domain
  
  function get_integer_data_save(domain)
    type(network_domain_type),intent(in) :: domain
    integer, dimension(:),pointer :: get_integer_data_save
    
    allocate(get_integer_data_save(size(domain%z_boundaries)))

    get_integer_data_save(:) = domain%z_boundaries
  end function get_integer_data_save

  subroutine load_domains_from_integer_array(domain,array)
    type(network_domain_type),intent(inout) :: domain
    integer, dimension(:),pointer :: array
    domain%z_boundaries(:) = array

    call update_cache(domain)
  end subroutine load_domains_from_integer_array

end module network_domain_module
