! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module network_utils
  use mpi
  use types
  implicit none

  integer, parameter :: DEFAULT_TAG = 2001

  integer, parameter :: NETWORK_ANY_SOURCE = MPI_ANY_SOURCE
  integer :: my_network_rank
  integer, allocatable, dimension(:),private :: buffer
contains

  subroutine init_network() 
    integer :: ierr
    call MPI_INIT( ierr )
  end subroutine init_network

  subroutine close_network()
    integer rc
    call MPI_FINALIZE(rc)
  end subroutine close_network

  subroutine create_buffer(bytes)
    integer,intent(in) :: bytes
    integer :: ierror
    allocate(buffer(bytes*8/bit_size(bytes)+1))
    call MPI_BUFFER_ATTACH(buffer,bytes,ierror)
  end subroutine create_buffer

  subroutine network_buffer_flush()
    integer,allocatable,dimension(:) ::fake_buffer
    integer :: bytes, ierror
    call MPI_BUFFER_DETACH(fake_buffer,bytes,ierror)
    call MPI_BUFFER_ATTACH(buffer,bytes,ierror)
  end subroutine network_buffer_flush

  integer function get_network_size() 
    integer :: numprocs, ierr
    call MPI_COMM_SIZE( MPI_COMM_WORLD, numprocs, ierr )
    get_network_size = numprocs
  end function get_network_size

  integer function get_network_rank()
    integer :: rank, ierr
    call MPI_COMM_RANK( MPI_COMM_WORLD, rank, ierr )
    my_network_rank = rank
    get_network_rank = rank
  end function get_network_rank

  subroutine probe_network_recv(source_request, source)
    integer, dimension(MPI_STATUS_SIZE) :: status    
    integer,intent(in) :: source_request
    integer,intent(out) :: source
    integer :: ierr
    call MPI_PROBE(source_request,DEFAULT_TAG,MPI_COMM_WORLD,status,ierr)
    source = status(MPI_SOURCE)    
  end subroutine probe_network_recv

  subroutine iprobe_network_recv(source_request, source, available)
    integer, dimension(MPI_STATUS_SIZE) :: status    
    integer,intent(in) :: source_request
    integer,intent(out) :: source
    logical,intent(out) :: available
    integer :: ierr
    call MPI_IPROBE(source_request,DEFAULT_TAG,MPI_COMM_WORLD,available,&
         status,ierror=ierr)
    if(available) then 
       source = status(MPI_SOURCE)    
    end if
  end subroutine iprobe_network_recv
    

  subroutine probe_integer_recv(source_request, source, length)
    integer, dimension(MPI_STATUS_SIZE) :: status    
    integer,intent(in) :: source_request
    integer,intent(out) :: source, length
    integer :: ierr
    call MPI_PROBE(source_request,DEFAULT_TAG,MPI_COMM_WORLD,status,ierr)
    source = status(MPI_SOURCE)    
    call MPI_GET_COUNT(status, MPI_INTEGER, length, ierr)
  end subroutine probe_integer_recv

  subroutine probe_double_recv(source_request, source, length)
    integer, dimension(MPI_STATUS_SIZE) :: status    
    integer,intent(in) :: source_request
    integer,intent(out) :: source, length
    integer :: ierr
    call MPI_PROBE(source_request,DEFAULT_TAG,MPI_COMM_WORLD,status,ierr)
    source = status(MPI_SOURCE)    
    call MPI_GET_COUNT(status, MPI_DOUBLE_PRECISION, length, ierr)
  end subroutine probe_double_recv

  subroutine send_int_array(destination,array,length) 
    integer, intent(in) :: length, destination
    integer, dimension(:),intent(in) :: array
    integer :: ierr
    call mpi_send(array, length, MPI_INTEGER, destination, DEFAULT_TAG, &
         MPI_COMM_WORLD, ierr)
    
  end subroutine send_int_array

  subroutine send_int_array_buffered(destination,array,length) 
    integer, intent(in) :: length, destination
    integer, dimension(:),intent(in) :: array
    integer :: ierr
    call mpi_bsend(array, length, MPI_INTEGER, destination, DEFAULT_TAG, &
         MPI_COMM_WORLD, ierr)
    
  end subroutine send_int_array_buffered

  subroutine recv_network_int_array(source, array,length) 
    integer, intent(in) :: length, source
    integer, dimension(:),intent(out) :: array
    integer :: tag, isource, ierr, count
    integer, dimension(MPI_STATUS_SIZE) :: status
    tag = MPI_ANY_TAG
    isource = source

    call mpi_recv(array, length, MPI_INTEGER, isource, tag, MPI_COMM_WORLD, &
         status, ierr)
    call mpi_get_count(status, MPI_INTEGER, count, ierr)
    if(length /= count) then
       print *,"#ERROR Int array mismatch expect ",length," got ",count
       call flush_to_disk_marker_a(my_network_rank,"int array mismatch",length)
       call flush_to_disk_marker_a(my_network_rank,"int array mismatch",count)

    else 
       !print *,"#SUCCESS Int array recieved ",length,count
       !call flush_to_disk_marker_a(my_network_rank,"int array recv",length)
    end if
    
  end subroutine recv_network_int_array

  subroutine send_double_array(destination,array,length) 
    integer, intent(in) :: length, destination
    real(kind=d_t), dimension(:),intent(in) :: array
    integer :: ierr
    call mpi_send(array, length, MPI_DOUBLE_PRECISION, destination, DEFAULT_TAG, &
         MPI_COMM_WORLD, ierr)
    
  end subroutine send_double_array

  subroutine send_double_array_buffered(destination,array,length) 
    integer, intent(in) :: length, destination
    real(kind=d_t), dimension(:),intent(in) :: array
    integer :: ierr
    call mpi_bsend(array, length, MPI_DOUBLE_PRECISION, destination, DEFAULT_TAG, &
         MPI_COMM_WORLD, ierr)
    
  end subroutine send_double_array_buffered

  subroutine recv_network_double_array(source,array,length) 
    integer, intent(in) :: length, source
    real(kind=d_t), dimension(:),intent(out) :: array
    integer :: tag, isource, ierr, count
    integer, dimension(MPI_STATUS_SIZE) :: status
    tag = MPI_ANY_TAG
    isource = source

    call mpi_recv(array, length, MPI_DOUBLE_PRECISION, isource, tag, &
         MPI_COMM_WORLD, status, ierr)

    call mpi_get_count(status, MPI_DOUBLE_PRECISION, count, ierr)
    if(length /= count) then
       print *,"#ERROR Double array mismatch expect ",length," got ",count
       call flush_to_disk_marker_a(my_network_rank,"double array mismatch",length)
       call flush_to_disk_marker_a(my_network_rank,"double array mismatch",count)
       
    else 
       !     print *,"#SUCCESS Double array recieved ",length,count
       !call flush_to_disk_marker_a(my_network_rank,"double array recv",length)

    end if

    
  end subroutine recv_network_double_array

  subroutine send_int(destination, value) 
    integer, intent(in) :: destination, value
    integer :: ierr
    call mpi_send(value, 1, MPI_INTEGER, destination, DEFAULT_TAG, &
         MPI_COMM_WORLD, ierr)

  end subroutine send_int

  subroutine send_int_buffered(destination, value) 
    integer, intent(in) :: destination, value
    integer :: ierr
    call mpi_bsend(value, 1, MPI_INTEGER, destination, DEFAULT_TAG, &
         MPI_COMM_WORLD, ierr)

  end subroutine send_int_buffered

  integer function recv_network_int(source) 
    integer :: received , tag, isource, ierr
    integer, intent(in) :: source
    integer, dimension(MPI_STATUS_SIZE) :: status
    tag = MPI_ANY_TAG
    isource = source
    call mpi_recv(received, 1, MPI_INTEGER, isource, tag, MPI_COMM_WORLD, &
         status, ierr)
    recv_network_int = received
  end function recv_network_int

  subroutine send_double(destination, value) 
    integer, intent(in) :: destination
    real(kind=d_t), intent(in) :: value
    integer :: ierr
    call mpi_send(value, 1, MPI_DOUBLE_PRECISION, destination, DEFAULT_TAG, &
         MPI_COMM_WORLD, ierr)

  end subroutine send_double

  subroutine send_double_buffered(destination, value) 
    integer, intent(in) :: destination
    real(kind=d_t), intent(in) :: value
    integer :: ierr
    call mpi_bsend(value, 1, MPI_DOUBLE_PRECISION, destination, DEFAULT_TAG, &
         MPI_COMM_WORLD, ierr)

  end subroutine send_double_buffered

  real(kind=d_t) function recv_network_double(source) 
    real(kind=d_t) :: received 
    integer :: tag, isource, ierr
    integer, intent(in) :: source
    integer, dimension(MPI_STATUS_SIZE) :: status
    tag = MPI_ANY_TAG
    isource = source
    call mpi_recv(received, 1, MPI_DOUBLE_PRECISION, isource, tag,&
         MPI_COMM_WORLD, status, ierr)
    recv_network_double = received
  end function recv_network_double

  subroutine digit_sub_a(i,c) 
    integer,intent(in) :: i
    character(1),intent(out) :: c
    if(i == 0) then 
       c = "0"
    else if(i == 1) then
       c = "1"       
    else if(i == 2) then
       c = "2"       
    else if(i == 3) then
       c = "3"       
    else if(i == 4) then
       c = "4"       
    else if(i == 5) then
       c = "5"       
    else if(i == 6) then
       c = "6"
    else if(i == 7) then
       c = "7"       
    else if(i == 8) then
       c = "8"       
    else if(i == 9) then
       c = "9"       
    else 
       c = "#"
    end if

  end subroutine digit_sub_a
    

  subroutine flush_to_disk_marker_a(marker,string,int)
    character(*),intent(in) :: string
    integer,intent(in) :: marker
    integer,intent(in),optional :: int
    character(10) :: filename
    filename = "flushed000"
    call digit_sub_a(mod(marker,10),filename(10:10))
    call digit_sub_a(mod(marker/10,10),filename(9:9))
    open(unit=911,file=filename,position="APPEND")
    if(present(int)) then        
       write(911,'(A70,I9)') string,int
    else 
       write(911,'(A79)') string
    end if
    close(911)

  end subroutine flush_to_disk_marker_a


end module network_utils
