! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module pp_slip_module
  use vector_type
  use number_utils
  use types
  use add_and_fetch_module
  !$ use omp_lib

  implicit none

  type pp_slip_type
     private 
     !The number of pebble pebble slips
     integer,pointer,dimension(:) :: num_pps
     !ppsd = pebble pebble slip
     ! ppsd(slip_zone_size,number_of_pebbles)
     type(vector),pointer, dimension(:,:) :: ppsd
     !pps_index = pebble pebble slip index
     ! which slip vector is the pebble pointing to
     ! pps_index(slip_zone_size,number_of_pebbles)
     integer,pointer,dimension(:,:) :: pps_index
  end type pp_slip_type

  ! How many adjecent pebbles can be sliding along this pebble.
  ! For perfectly hard spheres this is 12.  16 is used for paranoia reasons
  INTEGER, PARAMETER :: slip_zone_size = 16

  type pp_slip_single_iterator
     private 
     integer :: i,j
  end type pp_slip_single_iterator

  private
  public :: pp_slip_type,pp_slip_single_iterator
  public :: allocate_pp_slip, deallocate_pp_slip, clear_pp_slip
  public :: reorder_pp_slip, load_pp_slip, display_pp_slips
  public :: get_old_slip, add_new_slip, has_slip, update_slip
  public :: get_pp_slip_single_iterator,pp_slip_single_iterator_done
  public :: get_next_pp_slip, pp_slip_shallow_copy
  public :: get_partial_pp_slip_dump, load_from_partial_pp_slip_dump

contains

  subroutine pp_slip_shallow_copy(out_pp_slip,in_pp_slip)
    type(pp_slip_type),intent(in) :: in_pp_slip
    type(pp_slip_type),intent(out) :: out_pp_slip
    out_pp_slip%num_pps => in_pp_slip%num_pps
    out_pp_slip%ppsd => in_pp_slip%ppsd
    out_pp_slip%pps_index => in_pp_slip%pps_index
  end subroutine pp_slip_shallow_copy

  subroutine allocate_pp_slip(number_of_pebbles,pp_slip)
    type(pp_slip_type),intent(out) :: pp_slip
    integer, intent(in) :: number_of_pebbles
    allocate(pp_slip%num_pps(number_of_pebbles))
    allocate(pp_slip%ppsd(slip_zone_size,number_of_pebbles))
    allocate(pp_slip%pps_index(slip_zone_size,number_of_pebbles))
    call clear_pp_slip(pp_slip)
  end subroutine allocate_pp_slip

  subroutine deallocate_pp_slip(pp_slip)
    type(pp_slip_type),intent(inout) :: pp_slip
    deallocate(pp_slip%ppsd,pp_slip%pps_index,pp_slip%num_pps)
  end subroutine deallocate_pp_slip

  subroutine clear_pp_slip(pp_slip)
    type(pp_slip_type),intent(inout) :: pp_slip
    ! This is faster than only clearing the changed ones
    pp_slip%num_pps(:) = 0
  end subroutine clear_pp_slip

  subroutine reorder_pp_slip(new_ordering,pp_slip)
    type(pp_slip_type),intent(inout) :: pp_slip
    integer,dimension(:),intent(in) :: new_ordering
    type(pp_slip_type) :: temp_pp_slip
    integer :: i,j,old_from,old_to,new_from,new_to,number_of_pebbles,empty
    type(vector) :: old_slip
    number_of_pebbles  = size(pp_slip%pps_index,2)
    call allocate_pp_slip(number_of_pebbles,temp_pp_slip)
    do i=1,number_of_pebbles
       do j=1,pp_slip%num_pps(i)
          if(pp_slip%pps_index(j,i) /= 0) then
             old_from =  i
             old_to = pp_slip%pps_index(j,i)
             new_from = new_ordering(old_from)
             new_to = new_ordering(old_to)
             old_slip = pp_slip%ppsd(j,i)
             if(new_to < new_from) then
                call swap_int(new_to,new_from)
                old_slip = (-1.0)*old_slip
             end if
             empty = fetch_slip_index(temp_pp_slip,new_from)
             temp_pp_slip%pps_index(empty,new_from) = new_to
             temp_pp_slip%ppsd(empty,new_from) = old_slip
          end if
       end do

    end do
    !Note: copying entire arrays, not pointers
    pp_slip%pps_index = temp_pp_slip%pps_index
    pp_slip%ppsd = temp_pp_slip%ppsd
    pp_slip%num_pps = temp_pp_slip%num_pps
    call deallocate_pp_slip(temp_pp_slip)
  end subroutine reorder_pp_slip

  subroutine load_pp_slip(unit_number,pp_slip)
    type(pp_slip_type),intent(out) :: pp_slip
    integer,intent(in) :: unit_number
    type(vector) :: slip
    integer :: i,j,k,length
    character(15) :: junk_slip_str
    character(9) :: junk_pp_slip_str


    call clear_pp_slip(pp_slip)
    slip = vector(0.0, 0.0, 0.0)
    do 
       read(unit_number,*) junk_slip_str,j,length
       if(j == 0) then
          exit
       end if
       pp_slip%num_pps(j) = length
       do i=1,length
          read(unit_number,*) junk_pp_slip_str,k,slip%x,slip%y,slip%z
          pp_slip%pps_index(i,j) = k
          pp_slip%ppsd(i,j) = slip
       end do
    end do    
  end subroutine load_pp_slip
  
  subroutine display_pp_slips(pp_slip)
    type(pp_slip_type),intent(in) :: pp_slip
    type(real_int_pair),dimension(1:slip_zone_size) :: pp_slip_sort
    integer :: number_of_pebbles
    integer :: i, j, k, length
    number_of_pebbles = size(pp_slip%pps_index,2)
    do i=1,number_of_pebbles
       length = pp_slip%num_pps(i)
       if ( length > 0) then
          print *," #P_slip_start",i,length
          do j=1,length
             pp_slip_sort(j)%r = pp_slip%pps_index(j,i)
             pp_slip_sort(j)%i = j
          end do
          !Sort the pp slip so that the order is consistent between 
          ! runs.
          call sort_ri_pair(pp_slip_sort(1:length))          
          do j=1,length
             k = pp_slip_sort(j)%i
             print '(A15,I7,3E25.16E3)',"#PP_slip       ",&
                  pp_slip%pps_index(k,i),pp_slip%ppsd(k,i)%x,&
                  pp_slip%ppsd(k,i)%y,pp_slip%ppsd(k,i)%z
          end do
       end if
    end do
    ! Termination data.
    print *," #P_slip_start",0,0
  end subroutine display_pp_slips

  function get_slip_size(pp_slip, from)
    !integer,dimension(:,:),intent(in) :: pps_index
    type(pp_slip_type),intent(in) :: pp_slip
    integer,intent(in) :: from
    integer :: get_slip_size
    get_slip_size = pp_slip%num_pps(from)
  end function get_slip_size

  !Allocates the next slip index and return it.
  function fetch_slip_index(pp_slip, from)
    type(pp_slip_type),intent(inout) :: pp_slip
    integer,intent(in) :: from
    integer :: fetch_slip_index
    call atomic_add_and_fetch(pp_slip%num_pps(from),1,fetch_slip_index)
    if(fetch_slip_index > slip_zone_size) then       
       print *,"ERROR Overfull slip index",from
       fetch_slip_index = 1
    end if    
  end function fetch_slip_index

  pure function get_slip_index(pp_slip,from_in,to_in) 
    !integer,dimension(:,:),intent(in) :: pps_index
    type(pp_slip_type),intent(in) :: pp_slip
    integer,intent(in) :: from_in,to_in
    integer :: get_slip_index
    integer :: i, from, to
    if(to_in < from_in) then
       ! Swap so that from is always lower than to.
       from = to_in
       to = from_in
    else
       from = from_in
       to = to_in
    end if
    do i=1,pp_slip%num_pps(from)
       if(pp_slip%pps_index(i,from) == to) then
          get_slip_index = i
          return
       end if
    end do

    get_slip_index = 0
  end function get_slip_index

  pure function has_slip(pp_slip,from,to)
    logical :: has_slip
    type(pp_slip_type),intent(in) :: pp_slip
    integer,intent(in) :: from,to 
    integer :: i
    i = get_slip_index(pp_slip,from,to)
    has_slip = i /= 0    
  end function has_slip

  pure function get_old_slip(pp_slip,from_in,to_in)
    type(pp_slip_type),intent(in) :: pp_slip
    !type(vector),dimension(:,:),intent(in) :: ppsd
    !integer,dimension(:,:),intent(in) :: pps_index
    integer,intent(in) :: from_in,to_in    
    type(vector) :: get_old_slip
    integer :: i, from, to
    real(kind=d_t) :: mult
    if(to_in < from_in) then
       ! Swap so that from is always lower than to.
       from = to_in
       to = from_in
       !The vector sould be the opposite direction if these are flipped
       mult = 1.0_d_t
    else
       from = from_in
       to = to_in
       mult = -1.0_d_t
    end if

    i = get_slip_index(pp_slip,from,to)
    if(i /= 0) then
       get_old_slip = mult*pp_slip%ppsd(i,from)
       return
    end if

    get_old_slip = vector(0.0,0.0,0.0)
    

  end function get_old_slip

  subroutine add_new_slip(pp_slip,from_in,to_in,new_slip)
    type(pp_slip_type),intent(inout) :: pp_slip
    integer, intent(in) :: from_in,to_in
    type(vector),intent(in) :: new_slip
    integer :: slip_index
    integer :: from, to
    real(kind=d_t) :: mult
    if(to_in < from_in) then
       ! Swap so that from is always lower than to.
       from = to_in
       to = from_in
       !The vector should be the opposite direction if these are flipped.
       mult = 1.0_d_t
    else
       from = from_in
       to = to_in
       mult = -1.0_d_t
    end if
    slip_index = fetch_slip_index(pp_slip, from)
    pp_slip%pps_index(slip_index,from) = to
    pp_slip%ppsd(slip_index,from) = mult*new_slip
  end subroutine add_new_slip

  subroutine update_slip(pp_slip,from_in,to_in,new_slip)
    type(pp_slip_type),intent(inout) :: pp_slip
    !type(vector),dimension(:,:),intent(in) :: ppsd
    !integer,dimension(:,:),intent(in) :: pps_index
    integer,intent(in) :: from_in,to_in        
    type(vector),intent(in) :: new_slip
    integer :: i, from, to
    real(kind=d_t) :: mult
    if(to_in < from_in) then
       ! Swap so that from is always lower than to.
       from = to_in
       to = from_in
       !The vector sould be the opposite direction if these are flipped
       mult = 1.0_d_t
    else
       from = from_in
       to = to_in
       mult = -1.0_d_t
    end if

    i = get_slip_index(pp_slip,from,to)
    if(i /= 0) then
       pp_slip%ppsd(i,from) = mult*new_slip
       return
    end if

    print *,"#ERROR update_slip called when no slip exists",from_in,to_in,new_slip
  end subroutine update_slip


  subroutine get_pp_slip_single_iterator(pp_slip_iterator,pebble)
    integer,intent(in) :: pebble
    type(pp_slip_single_iterator),intent(out) :: pp_slip_iterator
    pp_slip_iterator%i = pebble
    pp_slip_iterator%j = 0    
  end subroutine get_pp_slip_single_iterator

  logical function pp_slip_single_iterator_done(pp_slip,pp_slip_iterator)
    type(pp_slip_single_iterator),intent(in) :: pp_slip_iterator
    type(pp_slip_type),intent(in) :: pp_slip
    ! Done if j > then the last index
    if(pp_slip_iterator%j + 1 > pp_slip%num_pps(pp_slip_iterator%i)) then
       pp_slip_single_iterator_done = .true.
       return
    end if
    pp_slip_single_iterator_done = .false.
  end function pp_slip_single_iterator_done

  subroutine get_next_pp_slip(pp_slip,pp_slip_iterator,slip,pebble)
    type(pp_slip_single_iterator),intent(inout) :: pp_slip_iterator
    type(pp_slip_type),intent(in) :: pp_slip
    type(vector),intent(out) :: slip
    integer,intent(out) :: pebble
    pp_slip_iterator%j = pp_slip_iterator%j + 1
    slip = pp_slip%ppsd(pp_slip_iterator%j,pp_slip_iterator%i)
    pebble = pp_slip%pps_index(pp_slip_iterator%j,pp_slip_iterator%i)
  end subroutine get_next_pp_slip

  subroutine get_partial_pp_slip_dump(pebble_list, pebble_list_len, pp_slip, &
       int_array, double_array)
    integer, intent(in) :: pebble_list_len
    integer, intent(in), dimension(:) :: pebble_list
    integer, dimension(:), pointer :: int_array
    real(kind=d_t),dimension(:),pointer :: double_array
    type(pp_slip_type),intent(in) :: pp_slip
    integer :: int_array_index, double_array_index, i, j, pebble, len, to
    integer :: num_pebbles_with_info
    !Note, these should not be depended on outside of this module. 
    !Order of the int array
    ! Number of pebbles with information, followed by
    ! For each pebble in the pebble list, 
    ! put Pebble number, number of slips, pebble of the slip
    !Order in the double array
    ! x y z of the slip, in the same order that pebble_number, pebble_of_slip 
    ! appear in the integer array
    int_array_index = 1 !Initialize to 1 for room for num_pebbles_with_info
    double_array_index = 0
    num_pebbles_with_info = 0
    do i = 1,pebble_list_len
       pebble = pebble_list(i)
       len = get_slip_size(pp_slip, pebble)
       if(len > 0) then 
          !If there are slips, then add 2 integers to put in the 
          ! pebble_number and the number of slips, and then the len for the indexs
          int_array_index = int_array_index + 2 + len 
          double_array_index = double_array_index + 3 * len
          num_pebbles_with_info = num_pebbles_with_info + 1
       end if
    end do

    allocate(int_array(int_array_index),double_array(double_array_index))

    int_array_index = 2
    double_array_index = 1
    int_array(1) = num_pebbles_with_info
    do i = 1,pebble_list_len
       pebble = pebble_list(i)
       len = get_slip_size(pp_slip, pebble) 
       if(len > 0) then
          int_array(int_array_index) = pebble
          int_array(int_array_index+1) = len
          int_array_index  = int_array_index + 2
          do j = 1,len
             to = pp_slip%pps_index(j,pebble)
             int_array(int_array_index) = to
             int_array_index = int_array_index + 1
             double_array(double_array_index) = pp_slip%ppsd(j,pebble)%x
             double_array(double_array_index+1) = pp_slip%ppsd(j,pebble)%y
             double_array(double_array_index+2) = pp_slip%ppsd(j,pebble)%z
             double_array_index = double_array_index + 3
          end do
       end if
    end do

  end subroutine get_partial_pp_slip_dump

  subroutine load_from_partial_pp_slip_dump(pp_slip,int_array,double_array)
    integer, dimension(:), pointer :: int_array
    real(kind=d_t),dimension(:),pointer :: double_array
    type(pp_slip_type),intent(inout) :: pp_slip
    integer :: i,j, len, pebble, int_array_index, double_array_index
    integer :: num_pebbles_with_info
    int_array_index = 2
    double_array_index = 1
    num_pebbles_with_info = int_array(1)
    do i=1,num_pebbles_with_info
       pebble = int_array(int_array_index)
       len = int_array(int_array_index+1)
       int_array_index = int_array_index + 2
       pp_slip%num_pps(pebble) = len
       pp_slip%pps_index(:,pebble) = 0
       do j=1,len
          pp_slip%pps_index(j,pebble) = int_array(int_array_index)
          int_array_index = int_array_index  + 1
          pp_slip%ppsd(j,pebble) = vector(double_array(double_array_index),&
               double_array(double_array_index+1),double_array(double_array_index+2))
          double_array_index = double_array_index + 3
       end do
    end do
    
    if(int_array_index - 1 /= size(int_array)) then
       print *,"#ERROR mismatch in pp int array index loading",int_array_index - 1,size(int_array)
    end if

    if(double_array_index - 1 /= size(double_array)) then
       print *,"#ERROR mismatch in pp double array index loading",double_array_index - 1, size(double_array)
    end if
    
  end subroutine load_from_partial_pp_slip_dump
  
end module pp_slip_module
