! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module prob_table 
  use types
  use vector_type
  use number_utils
  implicit none

  type location_index
     ! The index to a location.
     integer :: x,y,z     
  end type location_index

  type count_data_type
     private 
     ! Store all the data
     ! Points go from a x,y,z location to another one
     ! The count stores number of counts that each x,y,z location has
     ! to another x,y,z location.  The other x,y,z location is 
     ! stored in the to_index array.
     integer(kind=li),pointer,dimension(:,:,:,:) :: count
     type(location_index),pointer,dimension(:,:,:,:) :: to_index
     integer,pointer,dimension(:,:,:) :: size
     real(kind=d_t) :: delta
     integer :: x_min,x_max,y_min,y_max,z_min,z_max
  end type count_data_type

  integer,parameter :: TO_ARRAY_SIZE = 100

  private
  public :: initialize_count_data,add_count,print_data,deallocate_count_data
  public :: count_data_type
  
contains
  
  subroutine initialize_count_data(count_data,delta,x_low,x_high,&
       y_low,y_high,z_low,z_high) 
    type(count_data_type),intent(out) :: count_data
    real(kind=d_t),intent(in) :: delta,x_low,x_high,y_low,y_high,z_low,z_high
    integer :: x_min,x_max,y_min,y_max,z_min,z_max
    count_data%delta = delta    
    x_min = floor(x_low/delta)
    x_max = ceiling(x_high/delta)
    y_min = floor(y_low/delta)
    y_max = ceiling(y_high/delta)
    z_min = floor(z_low/delta)
    z_max = ceiling(z_high/delta)

    count_data%x_min = x_min
    count_data%x_max = x_max
    count_data%y_min = y_min
    count_data%y_max = y_max
    count_data%z_min = z_min
    count_data%z_max = z_max
    allocate(count_data%count(x_min:x_max,y_min:y_max,z_min:z_max,TO_ARRAY_SIZE))
    allocate(count_data%to_index(x_min:x_max,y_min:y_max,z_min:z_max,TO_ARRAY_SIZE))
    allocate(count_data%size(x_min:x_max,y_min:y_max,z_min:z_max))
    count_data%count = 0
    count_data%size = 0
  end subroutine initialize_count_data

  integer function compare_location(location1,location2) 
    type(location_index),intent(in) :: location1,location2
    if(location1%x < location2%x) then
       compare_location = -1
    else if(location1%x > location2%x) then
       compare_location = 1
    else if(location1%y < location2%y) then
       compare_location = -1
    else if(location1%y > location2%y) then
       compare_location = 1
    else if(location1%z < location2%z) then 
       compare_location = -1
    else if(location1%z > location2%z) then
       compare_location = 1
    else
       compare_location = 0
    end if
  end function compare_location

  type(location_index) function get_location(count_data,vec) 
    type(count_data_type),intent(in) :: count_data
    type(vector),intent(in) :: vec
    get_location%x = iclamp(count_data%x_min,floor(vec%x/count_data%delta),&
         count_data%x_max)
    get_location%y = iclamp(count_data%y_min,floor(vec%y/count_data%delta),&
         count_data%y_max)
    get_location%z = iclamp(count_data%z_min,floor(vec%z/count_data%delta),&
         count_data%z_max)    
  end function get_location

  subroutine add_count(count_data,old_pos,new_pos)
    type(count_data_type),intent(inout) :: count_data
    type(vector),intent(in) :: old_pos
    type(vector),intent(in) :: new_pos
    !type(location_index),dimension(:) :: loc_array
    type(location_index) :: old_loc,new_loc
    integer :: insertion_point, size, i
    old_loc = get_location(count_data,old_pos)
    new_loc = get_location(count_data,new_pos)
    !loc_array = count_data%to_index(old_loc%x,old_loc%y,old_loc%z,:)
    size = count_data%size(old_loc%x,old_loc%y,old_loc%z)
    insertion_point = get_count_insertion_point(&
         count_data%to_index(old_loc%x,old_loc%y,old_loc%z,:),new_loc,size)
    if(insertion_point < 0) then 
       insertion_point = -insertion_point
       !print *,"#before array",count_data%to_index(old_loc%x,old_loc%y,old_loc%z,1:size)

       if(insertion_point <= size) then 
          !print *,"#array shift old_loc ",old_loc," new_loc ",new_loc,&
          !     "from",insertion_point,size,&
          !     "to",(insertion_point+1),(size+1)
          ! Work around buggy gfortran compiler.
          do i = size,insertion_point,-1
             count_data%to_index(old_loc%x,old_loc%y,old_loc%z,i+1) = &
                  count_data%to_index(old_loc%x,old_loc%y,old_loc%z,i)
             count_data%count(old_loc%x,old_loc%y,old_loc%z,i+1) = &
                  count_data%count(old_loc%x,old_loc%y,old_loc%z,i)
          end do
          
          !count_data%to_index(old_loc%x,old_loc%y,old_loc%z,(insertion_point+1):(size+1)) = &
          !     count_data%to_index(old_loc%x,old_loc%y,old_loc%z,insertion_point:size)
          !count_data%count(old_loc%x,old_loc%y,old_loc%z,(insertion_point+1):(size+1)) = &
          !     count_data%count(old_loc%x,old_loc%y,old_loc%z,insertion_point:size)
       else 
          !print *,"#array extend old_loc ",old_loc," new_loc ",new_loc,&
          !     " size ",size
       end if
       count_data%to_index(old_loc%x,old_loc%y,old_loc%z,insertion_point) = &
            new_loc
       count_data%count(old_loc%x,old_loc%y,old_loc%z,insertion_point) = 1
       count_data%size(old_loc%x,old_loc%y,old_loc%z) = size + 1
       !print *,"#after array",count_data%to_index(old_loc%x,old_loc%y,old_loc%z,1:size+1)

    else        
       count_data%count(old_loc%x,old_loc%y,old_loc%z,insertion_point) = &
            count_data%count(old_loc%x,old_loc%y,old_loc%z,insertion_point) + 1
    end if
    
  end subroutine add_count

  !Returns a positive number if loc is found.
  ! Returns the insertion point as a negative if 
  ! needed.
  integer function get_count_insertion_point(loc_array,to_loc, size)
    ! type(count_data_type),intent(in) :: count_data    
    type(location_index), intent(in),dimension(:) :: loc_array
    type(location_index),intent(in) :: to_loc
    integer,intent(in) :: size
    type(location_index) :: mid_loc, low_loc, high_loc
    integer :: high_index, low_index, mid_index, comp
    integer :: low_comp, mid_comp, high_comp, ip
    logical :: print_data
    print_data = .false.
    high_index = size
    low_index = 1
    if(high_index == 0) then
       ! No data yet, so insertion point first.
       get_count_insertion_point = -1
       return
    end if
    ! comp = -1 if <
    ! comp =  0 if ==
    ! comp =  1 if >
    if(high_index == 1) then
       high_loc = loc_array(high_index)
       high_comp = compare_location(to_loc,high_loc)
       if(high_comp < 0) then ! to_loc < high_loc
          ! insert at front
          get_count_insertion_point = -1          
       else if(high_comp > 0) then ! to_loc > high_loc
          ! insert at end
          get_count_insertion_point = -2
       else 
          ! found it.
          get_count_insertion_point = 1
       end if
       return 
    end if
    if(compare_location(to_loc,loc_array(low_index)) < 0) then
       ! less than everything else, so insert below
       get_count_insertion_point = -1
       return 
    end if
    if(compare_location(to_loc,loc_array(high_index)) > 0) then
       ! greater than everything else, so insert at end
       get_count_insertion_point = -(high_index+1)
       return
    end if
    do       
       mid_index = floor((low_index + high_index) / 2.0)
       mid_loc = loc_array(mid_index)
       comp = compare_location(to_loc,mid_loc)
       if(comp < 0) then ! to_loc < mid_loc
          high_index = mid_index
       else if(comp > 0) then ! to_loc > mid_loc
          low_index = mid_index
       else 
          ! found it.
          get_count_insertion_point = mid_index
          return
       end if
       if(low_index + 1 >= high_index) then
          exit
       end if
    end do
    low_loc = loc_array(low_index)
    low_comp = compare_location(to_loc,low_loc)
    mid_loc = loc_array(mid_index)
    mid_comp = compare_location(to_loc,mid_loc)
    high_loc = loc_array(high_index)
    high_comp = compare_location(to_loc,high_loc)
    
    ! to_loc < low_loc
    if(low_comp < 0) then 
       print *,"#ERROR bad low_comp",low_comp
       print_data = .true.
    end if
    ! to_loc > high_loc
    if(high_comp > 0) then
       print *,"#ERROR bad high_comp",high_comp
       print_data = .true.
    end if
    if(low_comp == 0) then
       get_count_insertion_point = low_index
    else if(mid_comp == 0) then 
       get_count_insertion_point = mid_index
    else if(high_comp == 0) then 
       get_count_insertion_point = high_index
    else if(mid_comp < 0) then
       get_count_insertion_point = -mid_index
    else if(mid_comp > 0) then
       get_count_insertion_point = -(mid_index+1)
    else
       get_count_insertion_point = mid_index
    end if
    if(get_count_insertion_point < 0) then
       ip = -get_count_insertion_point
       ! ip should be greater than the location, and less than 
       ! the next location, and there will be a next location
       if(compare_location(to_loc,loc_array(ip)) /= -1) then
          print *,"#ERROR bad insertion point not greater"
          print_data = .true.
       end if
       if(compare_location(to_loc,loc_array(ip-1)) /= 1) then
          print *,"#ERROR bad insertion point - 1 not less"
          print_data = .true.
       end if
    end if
    if(print_data) then
       print *,"#a",loc_array(1:size)
       print *,"#k",to_loc
       print *,"#Bin_search size ",&
            size," l ",low_index," m ",&
            mid_index," h ",high_index, " l_comp ",&
            low_comp,&
            " mid_comp ",mid_comp,&
            " high_comp ",high_comp,&
            " result ",get_count_insertion_point
    end if
    
  end function get_count_insertion_point

  subroutine print_data(count_data)
    type(count_data_type),intent(in) :: count_data
    type(location_index) :: to
    integer(kind=li) :: count, total
    integer :: x,y,z,i,size
    do x=count_data%x_min,count_data%x_max
       do y=count_data%y_min,count_data%y_max
          do z=count_data%z_min,count_data%z_max
             size = count_data%size(x,y,z)
             total = sum(count_data%count(x,y,z,1:size))
             do i=1,size
                to = count_data%to_index(x,y,z,i)
                count = count_data%count(x,y,z,i)
                print *,"#pt from",x,y,z," to ",to%x,to%y,to%z,count,&
                     real(count)/real(total)," i ",i
             end do
          end do
       end do
    end do
  end subroutine print_data

  subroutine deallocate_count_data(count_data)
    type(count_data_type),intent(inout) :: count_data
    deallocate(count_data%count)
    deallocate(count_data%to_index)
    deallocate(count_data%size)
  end subroutine deallocate_count_data

  

end module prob_table
