! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module rect_grid
  use types
  use vector_type
  implicit none
  type rect_grid_type
     private 
     real(kind=d_t) :: x_delta,y_delta,z_delta
     real(kind=d_t) :: x_min,y_min,z_min
     integer :: x_size, y_size, z_size
  end type rect_grid_type

  interface create_grid
     module procedure create_grid_r
  end interface

  interface zone_in
     module procedure zone_in_r
  end interface

  interface inside_grid 
     module procedure inside_grid_r
  end interface

  interface first_zone
     module procedure first_zone_r
  end interface

  interface number_of_zones
     module procedure number_of_zones_r
  end interface

  interface number_of_surfaces
     module procedure number_of_surfaces_r
  end interface

  interface get_adjacent_zone
     module procedure get_adjacent_zone_r
  end interface

  interface area_of_surface
     module procedure area_of_surface_r
  end interface

  interface grid_zone_center
     module procedure grid_zone_center_r
  end interface

  interface next_zone
     module procedure next_zone_r
  end interface

  interface get_point_normal
     module procedure get_point_normal_r
  end interface

  interface get_point
     module procedure get_point_r
  end interface


  private
  public :: create_grid,number_of_zones,zone_in,maximum_surrounding_zones
  public :: surrounding_zones,number_of_surfaces,area_of_surface
  public :: get_point,get_point_normal,rect_grid_type,first_zone
  public :: get_adjacent_zone,next_zone,grid_zone_center,inside_grid
contains
  ! Create the grid 
  subroutine create_grid_r(grid,x_delta,y_delta,z_delta,x_min,y_min,z_min,&
       x_size,y_size,z_size)
    type(rect_grid_type),intent(out) :: grid
    real(kind=d_t),intent(in) :: x_delta,y_delta,z_delta
    real(kind=d_t),intent(in) :: x_min,y_min,z_min
    integer,intent(in) :: x_size, y_size, z_size
    grid%x_delta = x_delta
    grid%y_delta = y_delta
    grid%z_delta = z_delta
    grid%x_min = x_min
    grid%y_min = y_min
    grid%z_min = z_min
    grid%x_size = x_size
    grid%y_size = y_size
    grid%z_size = z_size
  end subroutine create_grid_r
    

  ! Returns the total number of zones for the grid.
  integer function number_of_zones_r(grid)
    type(rect_grid_type),intent(in) :: grid
    number_of_zones_r = grid%x_size * grid%y_size * grid%z_size
  end function number_of_zones_r

  ! Converts the integer x,y,z location to an interger zone_id
  integer pure function integer_location_to_zone_id(grid,x,y,z)
    type(rect_grid_type),intent(in) :: grid
    ! x ranges from 0 to x_size - 1
    integer,intent(in) :: x,y,z
    integer_location_to_zone_id = x*grid%y_size*grid%z_size + y*grid%z_size + z
  end function integer_location_to_zone_id

  ! Figures out the x,y,z integers for a give zone_id
  subroutine decompose_zone_id(grid,zone_id,x,y,z)
    type(rect_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id
    integer,intent(out) :: x,y,z
    integer :: left
    z = mod(zone_id,grid%z_size)
    left = zone_id/grid%z_size
    y = mod(left,grid%y_size)
    x = left/grid%y_size
  end subroutine decompose_zone_id

  ! Figures out the integer location numbers for the given real valued 
  !  location. 
  pure subroutine get_integer_location(grid,x,y,z,x_i,y_i,z_i)
    type(rect_grid_type),intent(in) :: grid
    real(kind=d_t),intent(in) :: x,y,z
    integer,intent(out) :: x_i,y_i,z_i
    x_i = floor((x - grid%x_min)/grid%x_delta)
    y_i = floor((y - grid%y_min)/grid%y_delta)
    z_i = floor((z - grid%z_min)/grid%z_delta)
  end subroutine get_integer_location

  ! Returns the zone that the location is in.
  integer pure function zone_in_r(grid,x,y,z)
    type(rect_grid_type),intent(in) :: grid
    real(kind=d_t),intent(in) :: x,y,z
    integer :: x_i,y_i,z_i
    call get_integer_location(grid,x,y,z,x_i,y_i,z_i)
    zone_in_r = integer_location_to_zone_id(grid,x_i,y_i,z_i)
  end function zone_in_r

  logical pure function inside_grid_r(grid,x,y,z)
    type(rect_grid_type),intent(in) :: grid
    real(kind=d_t),intent(in) :: x,y,z
    integer :: x_i,y_i,z_i
    call get_integer_location(grid,x,y,z,x_i,y_i,z_i)
    inside_grid_r = (0 <= x_i .and. x_i < grid%x_size) .and. &
         (0 <= y_i .and. y_i < grid%y_size) .and. &
         (0 <= z_i .and. z_i < grid%z_size)
  end function inside_grid_r

  ! Gets the maximum possible number of surrounding zones for the 
  !  given grid
  integer function maximum_surrounding_zones(grid) 
    type(rect_grid_type),intent(in) :: grid
    if(grid%x_size > 0) then 
       !Get rid of unused var warning.
    end if
    maximum_surrounding_zones = 6
  end function maximum_surrounding_zones

  ! Returns a list of all the zones that surround the given zone_id in 
  !  the grid.
  subroutine surrounding_zones(grid,zone_id,zones,number_of_zones)
    type(rect_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id
    integer,intent(out) :: number_of_zones
    integer,dimension(:),intent(out) :: zones
    integer :: x_i,y_i,z_i
    integer :: cx,cy,cz,dx,dy,dz
    number_of_zones = 0
    call decompose_zone_id(grid,zone_id,x_i,y_i,z_i)
    do dx = -1,1
       cx = x_i + dx
       if(0 <= cx .and. cx < grid%x_size) then
          do dy = -1,1
             cy = y_i + dy
             if(0 <= cy .and. cy < grid%y_size) then
                do dz = -1,1
                   cz = z_i + dz
                   if(0 < cz .and. cz < grid%z_size) then
                      if(abs(dx) + abs(dy) + abs(dz) == 1) then
                         number_of_zones = number_of_zones + 1
                         zones(number_of_zones) = &
                              integer_location_to_zone_id(grid,cx,cy,cz)
                      end if
                   end if
                end do
             end if
          end do
       end if
    end do
  end subroutine surrounding_zones

  ! Returns the number of surfaces that given zone_id is surrounded by.
  integer function number_of_surfaces_r(grid,zone_id) 
    integer,dimension(6) :: junk
    integer,intent(in) :: zone_id
    type(rect_grid_type),intent(in) :: grid
    integer :: number_of_zones
    call surrounding_zones(grid,zone_id,junk,number_of_zones)
    number_of_surfaces_r = number_of_zones
  end function number_of_surfaces_r

  ! Utility function to get some values for the given zone_id & surface_number
  !  The values are the dx,y,z which are the deltas to the other zone location,
  !  the cx,y,z that are the other zone location integers
  !  the x_i,y_i,z_i are zone location integers.
  subroutine get_surface_values(grid,zone_id,surface_number,&
       dx,dy,dz,cx,cy,cz,x_i,y_i,z_i)
    type(rect_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id,surface_number
    integer,intent(out) :: x_i,y_i,z_i
    integer,intent(out) :: dx,dy,dz
    integer,intent(out) :: cx,cy,cz
    integer :: number_of_zones
    number_of_zones = 0
    call decompose_zone_id(grid,zone_id,x_i,y_i,z_i)
    do dx = -1,1
       cx = x_i + dx
       if(0 <= cx .and. cx < grid%x_size) then
          do dy = -1,1
             cy = y_i + dy
             if(0 <= cy .and. cy < grid%y_size) then
                do dz = -1,1
                   cz = z_i + dz
                   if(0 < cz .and. cz < grid%z_size) then
                      if(abs(dx) + abs(dy) + abs(dz) == 1) then
                         number_of_zones = number_of_zones + 1
                         if(number_of_zones == surface_number) then
                            return
                         end if
                      end if
                   end if
                end do
             end if
          end do
       end if
    end do
    
  end subroutine get_surface_values

  ! Returns the area of the surface.
  real(kind=d_t) function area_of_surface_r(grid,zone_id,surface_number)
    type(rect_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id,surface_number
    integer :: x_i,y_i,z_i
    integer :: cx,cy,cz,dx,dy,dz
    call get_surface_values(grid,zone_id,surface_number,&
         dx,dy,dz,cx,cy,cz,x_i,y_i,z_i)
    if(abs(dx) == 1) then
       area_of_surface_r = grid%y_delta * grid%z_delta
    else if(abs(dy) == 1) then
       area_of_surface_r = grid%x_delta * grid%z_delta
    else 
       if(abs(dz) /= 1) then 
          print *,"#ERROR invalid area of surface"
       end if
       area_of_surface_r = grid%x_delta * grid%y_delta
    end if
    return
    
  end function area_of_surface_r

  ! Returns a point on the parameterized surface.  The u and v are used
  !  to choose the point on the surface.  u ranges from 0 to max_u and v ranges
  !  ranges from 0 to max_v.  x,y,z are the point on the surface.
  !  area is the area that the point has.  
  subroutine get_point_r(grid,u,v,max_u,max_v,zone_id,surface_number,x,y,z,area)
    type(rect_grid_type),intent(in) :: grid
    integer,intent(in) :: u,v,max_u,max_v,zone_id,surface_number
    real(kind=d_t),intent(out) :: x,y,z,area
    integer,dimension(3) :: in_surf,out_surf, dvect
    real(kind=d_t),dimension(3) :: deltas, new_values, mins
    real(kind=d_t) :: mult
    integer :: x_i,y_i,z_i
    integer :: cx,cy,cz,dx,dy,dz
    integer :: i
    logical :: first
    call get_surface_values(grid,zone_id,surface_number,&
         dx,dy,dz,cx,cy,cz,x_i,y_i,z_i)
    area = area_of_surface(grid,zone_id,surface_number)/(real(max_u*max_v))
    if(u == 0 .or. u == max_u) then
       area = 0.5*area
    end if
    if(v == 0 .or. v == max_v) then
       area = 0.5*area
    end if
    in_surf = (/ x_i,y_i,z_i /)
    out_surf = (/ cx,cy,cz /)
    dvect = (/ dx,dy,dz /)
    deltas = (/ grid%x_delta,grid%y_delta,grid%z_delta /)
    mins = (/ grid%x_min,grid%y_min,grid%z_min /)
    first = .true.
    do i=1,3
       if(dvect(i) == -1) then 
          new_values(i) = mins(i)+in_surf(i)*deltas(i)
       else if(dvect(i) == 1) then
          new_values(i) = mins(i)+(in_surf(i)+1)*deltas(i)
       else
          if(first) then
             first = .false.
             mult = real(u)/real(max_u)
          else 
             mult = real(v)/real(max_v)
          end if
          new_values(i) = mins(i)+(in_surf(i)+mult)*deltas(i)
       end if
    end do
    x = new_values(1)
    y = new_values(2)
    z = new_values(3)
  end subroutine get_point_r

  ! Returns a point on the parameterized surface.  The u and v are used
  !  to choose the point on the surface.  u ranges from 0 to max_u and v ranges
  !  ranges from 0 to max_v.  x,y,z are the point on the surface.
  !  area is the area that the point has.  
  subroutine get_point_normal_r(grid,u,v,max_u,max_v,zone_id,surface_number,&
       x,y,z,area,normal)
    type(rect_grid_type),intent(in) :: grid
    integer,intent(in) :: u,v,max_u,max_v,zone_id,surface_number
    real(kind=d_t),intent(out) :: x,y,z,area
    integer,dimension(3) :: in_surf,out_surf, dvect
    real(kind=d_t),dimension(3) :: deltas, new_values, mins, normal_array
    real(kind=d_t) :: mult
    type(vector),intent(out) :: normal
    integer :: x_i,y_i,z_i
    integer :: cx,cy,cz,dx,dy,dz
    integer :: i
    logical :: first
    call get_surface_values(grid,zone_id,surface_number,&
         dx,dy,dz,cx,cy,cz,x_i,y_i,z_i)
    area = area_of_surface_r(grid,zone_id,surface_number)/(real(max_u*max_v))
    if(u == 0 .or. u == max_u) then
       area = 0.5*area
    end if
    if(v == 0 .or. v == max_v) then
       area = 0.5*area
    end if
    in_surf = (/ x_i,y_i,z_i /)
    out_surf = (/ cx,cy,cz /)
    dvect = (/ dx,dy,dz /)
    deltas = (/ grid%x_delta,grid%y_delta,grid%z_delta /)
    mins = (/ grid%x_min,grid%y_min,grid%z_min /)
    first = .true.
    do i=1,3
       if(dvect(i) == -1) then 
          new_values(i) = mins(i)+in_surf(i)*deltas(i)
          normal_array(i) = -1
       else if(dvect(i) == 1) then
          new_values(i) = mins(i)+(in_surf(i)+1)*deltas(i)
          normal_array(i) = 1
       else
          if(first) then
             first = .false.
             mult = real(u)/real(max_u)
          else 
             mult = real(v)/real(max_v)
          end if
          new_values(i) = mins(i)+(in_surf(i)+mult)*deltas(i)
          normal_array(i) = 0.0
       end if
    end do
    x = new_values(1)
    y = new_values(2)
    z = new_values(3)
    normal = normal_array
  end subroutine get_point_normal_r

  ! Returns the zone id of the adjectent zone that has the given surface number
  integer function get_adjacent_zone_r(grid,zone_id,surface_number) 
    type(rect_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id,surface_number
    integer,dimension(8) :: junk
    integer :: num_of_zones
    call surrounding_zones(grid,zone_id,junk,num_of_zones)
    get_adjacent_zone_r = junk(surface_number)    
  end function get_adjacent_zone_r
  
  ! Returns the first zone_id that can be used for iterating through
  !  all the zones.
  integer function first_zone_r(grid)
    type(rect_grid_type),intent(in) :: grid
    first_zone_r = integer_location_to_zone_id(grid,0,0,0)
  end function first_zone_r

  ! Returns the next zone in the iteration
  integer function next_zone_r(grid,zone_id)
    type(rect_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id
    integer :: x_i,y_i,z_i
    call decompose_zone_id(grid,zone_id,x_i,y_i,z_i)
    x_i = x_i + 1
    if(x_i >= grid%x_size) then
       x_i = 0
       y_i = y_i + 1
       if(y_i >= grid%y_size) then
          y_i = 0
          z_i = z_i + 1
       end if
    end if
    next_zone_r = integer_location_to_zone_id(grid,x_i,y_i,z_i)
  end function next_zone_r

  ! Return the center of the zone with the given id.
  subroutine grid_zone_center_r(grid,zone_id,x,y,z)
    type(rect_grid_type),intent(in) :: grid
    integer,intent(in) :: zone_id
    real(kind=d_t),intent(out) :: x,y,z
    integer :: x_i,y_i,z_i
    call decompose_zone_id(grid,zone_id,x_i,y_i,z_i)
    x = grid%x_min + grid%x_delta * (x_i + 0.5)
    y = grid%y_min + grid%y_delta * (y_i + 0.5)
    z = grid%z_min + grid%z_delta * (z_i + 0.5)
  end subroutine grid_zone_center_r

  
end module rect_grid
