! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module vector_type
  use types
  implicit none 

  type vector
     real(kind=d_t) :: x,y,z
  end type vector

  interface operator(+)
     module procedure plus
  end interface

  interface operator(-)
     module procedure minus
  end interface

  interface operator(*) 
     module procedure mult
  end interface

  interface operator(*) 
     module procedure mult_r
  end interface

  interface assignment(=)
     module procedure assign_r
  end interface

  interface assignment(=)
     module procedure assign_d
  end interface

  interface operator(.dot.)
     module procedure vdot
  end interface

  interface operator(.cross.)
     module procedure vcross
  end interface

  interface abs
     module procedure vmag
  end interface

  interface operator(.proj.)
     module procedure vprojection
  end interface

contains
  
  pure function plus(v1,v2) 
    type(vector) :: plus
    type(vector), intent(in) :: v1,v2
    
    plus%x = v1%x + v2%x
    plus%y = v1%y + v2%y
    plus%z = v1%z + v2%z

  end function plus

  subroutine atomic_increment(l,adder)
    type(vector),intent(inout) :: l
    type(vector),intent(in) :: adder
    !$OMP ATOMIC
    l%x = l%x + adder%x
    !$OMP ATOMIC
    l%y = l%y + adder%y
    !$OMP ATOMIC
    l%z = l%z + adder%z
  end subroutine atomic_increment

  pure function minus(v1, v2)
    type(vector) :: minus
    type(vector), intent(in) :: v1,v2
    
    minus%x = v1%x - v2%x
    minus%y = v1%y - v2%y
    minus%z = v1%z - v2%z

  end function minus

  pure function mult(r, vec)
    type(vector) :: mult
    type(vector), intent(in) :: vec
    real(kind=d_t),intent(in) :: r
    
    mult%x = r * vec%x
    mult%y = r * vec%y
    mult%z = r * vec%z
    
  end function mult

  pure function mult_r(r, vec)
    type(vector) :: mult_r
    type(vector), intent(in) :: vec
    real,intent(in) :: r
    
    mult_r%x = r * vec%x
    mult_r%y = r * vec%y
    mult_r%z = r * vec%z
    
  end function mult_r

  pure function vdot(v1, v2)
    real(kind=d_t) :: vdot
    type(vector), intent(in) :: v1,v2
    
    vdot = v1%x * v2%x + v1%y * v2%y + v1%z * v2%z

  end function vdot

  pure function vcross(v1, v2)
    type(vector) :: vcross
    type(vector), intent(in) :: v1,v2

    vcross%x = v1%y * v2%z - v1%z * v2%y
    vcross%y = v1%z * v2%x - v1%x * v2%z
    vcross%z = v1%x * v2%y - v1%y * v2%x
    
  end function vcross

  pure function vmag_sqr(vec) 
    type(vector), intent(in) :: vec
    real(kind=d_t) :: vmag_sqr
    
    vmag_sqr = vdot(vec,vec)

  end function vmag_sqr
    
  pure subroutine assign_d(V, Ain)
    type(vector), intent(inout) :: v
    real(kind=d_t), dimension(3), intent(in) :: Ain
    
    V%x = Ain(1)
    V%y = Ain(2)
    V%z = Ain(3)

  end subroutine assign_d

  pure subroutine assign_r(V, Ain)
    type(vector), intent(inout) :: v
    real, dimension(3), intent(in) :: Ain
    
    V%x = Ain(1)
    V%y = Ain(2)
    V%z = Ain(3)

  end subroutine assign_r

  pure function vmag(vec)
    type(vector), intent(in) :: vec
    real(kind=d_t) :: vmag
    
    vmag =  sqrt(vmag_sqr(vec))
  end function vmag

  pure function vprojection(from_v, onto_v)
    type(vector), intent(in) :: from_v, onto_v
    type(vector) :: vprojection
    
    vprojection = (vdot(from_v, onto_v)/vmag_sqr(onto_v)) * onto_v
  end function vprojection

  ! Project from_v onto the plane normal to normal_v
  ! length(normal_v) must be 1 for this to work properly
  pure function perp_projection(from_v, normal_v)
    type(vector), intent(in) :: from_v, normal_v
    type(vector) :: perp_projection
    
    perp_projection = from_v - vdot(from_v,normal_v)*normal_v
  end function perp_projection

  pure function normalize(v)
    type(vector),intent(in) :: v
    type(vector) :: normalize
    
    normalize = (1.0_d_t/vmag(v)) * v

  end function normalize

end module vector_type
    
