! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module velocity_func
  use types
  use vector_type
  use number_utils
  implicit none

  integer :: tally_sections
  INTEGER :: rt_size
  INTEGER :: zt_size
  real(kind=d_t) :: tally_radial_inc
  real(kind=d_t) :: tally_z_inc

  INTEGER,allocatable, dimension(:,:,:) :: zone_n
  type(vector),allocatable, dimension(:,:,:) :: zone_vel_sum
  REAL(kind=d_t),allocatable, dimension(:,:,:) :: zone_dot_sum

  integer :: x_size
  integer :: y_size
  integer :: z_size
  real(kind=d_t) :: zone_increment
  real(kind=d_t) :: x_low, y_low, z_low

  !These store the radialized data required to calculate the average velocity
  ! and the variance.  Highly likely to be correct, but harder to interpret.
  INTEGER,allocatable, dimension(:,:,:) :: zone_rect_n
  type(vector),allocatable, dimension(:,:,:) :: zone_rect_vel_sum
  REAL(kind=d_t),allocatable, dimension(:,:,:) :: zone_rect_dot_sum

  private 
  public :: load_data,closest_rect_data,rect_linear_interpolation
  public :: closest_rect_avg,rect_interpolation_avg,rect_linear_density
  !Remove the following from public once testing is done.
  public :: find_zone, zone_center
contains

  subroutine load_data(filename)
    character(*),intent(in) :: filename
    integer :: status,allocate_status
    character(5) :: line_header,junk_chars
    logical :: have_rect_size,have_rect_lows,have_rad_size,have_rad_incs
    real(kind=d_t) :: j1,j2,j3
    real(kind=d_t) :: cx,cy,cz,sx,sy,sz,dot_sum
    integer :: n
    integer :: s_z,r_z,z_z
    integer :: x_z,y_z
    integer :: line
    have_rect_size = .false.
    have_rect_lows = .false.
    have_rad_size = .false.
    have_rad_incs = .false.
    line = 0
    
    open(unit=15,file=filename,status="OLD",iostat=status)

    do
       read(unit=15,advance="NO",FMT='(A5)',end=20) line_header
       line = line + 1
       !print *,"line",line,line_header
       if(line_header == ' #rz ') then
          !read rectangular zone data
          if(.not. (have_rect_size .and. have_rect_lows)) &
               stop "rectangular tally data too early"
          read(15,*) cx,cy,cz,n,sx,sy,sz,j1,j2,j3,dot_sum
          x_z = find_base_zone(x_low, zone_increment, x_size, cx)
          y_z = find_base_zone(y_low, zone_increment, y_size, cy)
          z_z = find_base_zone(z_low, zone_increment, z_size, cz)
          zone_rect_n(x_z,y_z,z_z) = n
          zone_rect_vel_sum(x_z,y_z,z_z) = vector(sx,sy,sz)
          zone_rect_dot_sum(x_z,y_z,z_z) = dot_sum
       else if(line_header == ' #rzs') then
          !read rectangular size data
          read(15,*) x_size,y_size,z_size
          have_rect_size = .true.
          !Allocate the rectangular tally zones
          allocate(zone_rect_n(x_size,y_size,z_size),stat=allocate_status)
          zone_rect_n = 0
          if (allocate_status /= 0) stop "ERROR unable to allocate ram"
          allocate(zone_rect_vel_sum(x_size,y_size,z_size),stat=allocate_status)
          if (allocate_status /= 0) stop "ERROR unable to allocate ram"
          allocate(zone_rect_dot_sum(x_size,y_size,z_size),stat=allocate_status)
          if (allocate_status /= 0) stop "ERROR unable to allocate ram"
          
       else if(line_header == ' #rzl') then
          !read rectangular minimum data
          read(15,*) x_low,y_low,z_low,zone_increment
          have_rect_lows = .true.
       else if(line_header == ' #z  ') then
          !read zone radial data
          if(.not. (have_rad_size .and. have_rad_incs)) &
               stop "radial tally data too early"
          read(15,*) cx,cy,cz,n,sx,sy,sz,j1,j2,j3,dot_sum
          call find_radial_zone(cx,cy,cz,s_z,r_z,z_z)
          if(z_z < 1 .or. z_z > zt_size) then 
             print *,"ERROR bad z_z found ",cx,cy,cz,s_z,r_z,z_z,&
                  tally_sections,rt_size,zt_size
          else if(r_z < 1 .or. r_z > rt_size) then
             print *,"ERROR bad r_z found ",cx,cy,cz,s_z,r_z,z_z,&
                  tally_sections,rt_size,zt_size
          else if(s_z < 1 .or. s_z > tally_sections) then
             print *,"ERROR bad s_z found ",cx,cy,cz,s_z,r_z,z_z,&
                  tally_sections,rt_size,zt_size
          else 
             zone_n(s_z,r_z,z_z) = n
             zone_vel_sum(s_z,r_z,z_z) = vector(sx,sy,sz)
             zone_dot_sum(s_z,r_z,z_z) = dot_sum
          end if
       else if(line_header == ' #zin') then
          !read zone increment data
          read(15,*) tally_radial_inc,tally_z_inc,z_low
          have_rad_incs = .true.
       else if(line_header == ' #zsi') then 
          !read zone size data
          read(15,*) tally_sections,rt_size,zt_size
          !Allocate the radial tally zones
          allocate(zone_n(tally_sections,rt_size,zt_size))
          zone_n = 0
          allocate(zone_vel_sum(tally_sections,rt_size,zt_size))
          allocate(zone_dot_sum(tally_sections,rt_size,zt_size))
          have_rad_size = .true.
       else 
          !print *,"Unhandled line",line_header
          read(15,*) junk_chars
       end if
    end do

    go to 40
20  continue
    !end of file found
40  continue

    close(15)
  end subroutine load_data

  real(kind=d_t) pure function interpolate(bI,aI,bV,aV,s)
    real(kind=d_t),intent(in) :: bI, aI, bV, aV, s
    if(aI == bI) then
       interpolate = bV
    else 
       interpolate = ((aV-bV)/(aI-bI)) * (s-bI) + bV
    end if
  end function interpolate

  type(vector) pure function interpolate_v(bI,aI,bV,aV,s)
    real(kind=d_t),intent(in) :: bI, aI, s
    type(vector),intent(in) :: bV,aV
    !print *,"i_v",bI,aI,"bV",bV,"aV",aV,s
    if(aI == bI) then
       interpolate_v = bV
    else 
       interpolate_v = ((s-bI)/(aI-bI))* (aV-bV) + bV
    end if
  end function interpolate_v

  type(vector) function closest_rect_avg(x,y,z)
    real(kind=d_t),intent(in) :: x,y,z
    integer :: n
    type(vector) :: sum
    real(kind=d_t) :: dot_sum

    call closest_rect_data(x,y,z,n,sum,dot_sum)
    closest_rect_avg = (1.0/n) * sum
  end function closest_rect_avg

  subroutine closest_rect_data(x,y,z,n,sum,dot_sum)
    real(kind=d_t),intent(in) :: x,y,z
    integer, intent(out) :: n
    type(vector), intent(out) :: sum
    real(kind=d_t), intent(out) :: dot_sum
    integer :: x_z,y_z,z_z
    x_z = find_base_zone(x_low, zone_increment, x_size, x)
    y_z = find_base_zone(y_low, zone_increment, y_size, y)
    z_z = find_base_zone(z_low, zone_increment, z_size, z)
    
    n = zone_rect_n(x_z,y_z,z_z)
    sum = zone_rect_vel_sum(x_z,y_z,z_z)
    dot_sum = zone_rect_dot_sum(x_z,y_z,z_z)
  end subroutine closest_rect_data

  type(vector) pure function rect_interpolation_avg(x,y,z)
    real(kind=d_t),intent(in) :: x,y,z
    call rect_linear_interpolation(x,y,z,rect_interpolation_avg)
  end function rect_interpolation_avg

  pure subroutine rect_linear_interpolation(x,y,z,avg)
    real(kind=d_t),intent(in) :: x,y,z
    type(vector), intent(out) :: avg
    integer :: x_l,y_l,z_l,x_h,y_h,z_h
    real(kind=d_t) :: x_lr,y_lr,z_lr,x_hr,y_hr,z_hr
    type(vector) :: &
         avg_xl_yl_zl,&
         avg_xl_yl_zh,&
         avg_xl_yh_zl,&
         avg_xl_yh_zh,&
         avg_xh_yl_zl,&
         avg_xh_yl_zh,&
         avg_xh_yh_zl,&
         avg_xh_yh_zh,&
         avg_xl_yl,&
         avg_xl_yh,&
         avg_xh_yl,&
         avg_xh_yh,&
         avg_xl,&
         avg_xh
    
    x_l = find_low_zone(x_low, zone_increment, x_size, x)    
    y_l = find_low_zone(y_low, zone_increment, y_size, y)
    z_l = find_low_zone(z_low, zone_increment, z_size, z)
    x_h = find_high_zone(x_low, zone_increment, x_size, x)
    y_h = find_high_zone(y_low, zone_increment, y_size, y)
    z_h = find_high_zone(z_low, zone_increment, z_size, z)
    
    x_lr = zone_center(x_low, zone_increment, x_l)
    x_hr = zone_center(x_low, zone_increment, x_h)
    y_lr = zone_center(y_low, zone_increment, y_l)
    y_hr = zone_center(y_low, zone_increment, y_h)
    z_lr = zone_center(z_low, zone_increment, z_l)
    z_hr = zone_center(z_low, zone_increment, z_h)

    avg_xl_yl_zl = get_average(x_l,y_l,z_l)
    avg_xl_yl_zh = get_average(x_l,y_l,z_h)
    avg_xl_yh_zl = get_average(x_l,y_h,z_l)
    avg_xl_yh_zh = get_average(x_l,y_h,z_h)
    avg_xh_yl_zl = get_average(x_h,y_l,z_l)
    avg_xh_yl_zh = get_average(x_h,y_l,z_h)
    avg_xh_yh_zl = get_average(x_h,y_h,z_l)
    avg_xh_yh_zh = get_average(x_h,y_h,z_h)

    avg_xl_yl = interpolate_v(z_lr,z_hr,avg_xl_yl_zl,avg_xl_yl_zh,z)
    avg_xl_yh = interpolate_v(z_lr,z_hr,avg_xl_yh_zl,avg_xl_yh_zh,z)
    avg_xh_yl = interpolate_v(z_lr,z_hr,avg_xh_yl_zl,avg_xh_yl_zh,z)
    avg_xh_yh = interpolate_v(z_lr,z_hr,avg_xh_yh_zl,avg_xh_yh_zh,z)

    avg_xl = interpolate_v(y_lr,y_hr,avg_xl_yl,avg_xl_yh,y)
    avg_xh = interpolate_v(y_lr,y_hr,avg_xh_yl,avg_xh_yh,y)

    avg = interpolate_v(x_lr,x_hr,avg_xl,avg_xh,x)
    
  end subroutine rect_linear_interpolation

  subroutine rect_linear_density(x,y,z,avg)
    real(kind=d_t),intent(in) :: x,y,z
    real(kind=d_t), intent(out) :: avg
    integer :: x_l,y_l,z_l,x_h,y_h,z_h
    real(kind=d_t) :: x_lr,y_lr,z_lr,x_hr,y_hr,z_hr
    real(kind=d_t) :: &
         avg_xl_yl_zl,&
         avg_xl_yl_zh,&
         avg_xl_yh_zl,&
         avg_xl_yh_zh,&
         avg_xh_yl_zl,&
         avg_xh_yl_zh,&
         avg_xh_yh_zl,&
         avg_xh_yh_zh,&
         avg_xl_yl,&
         avg_xl_yh,&
         avg_xh_yl,&
         avg_xh_yh,&
         avg_xl,&
         avg_xh
    
    x_l = find_low_zone(x_low, zone_increment, x_size, x)    
    y_l = find_low_zone(y_low, zone_increment, y_size, y)
    z_l = find_low_zone(z_low, zone_increment, z_size, z)
    x_h = find_high_zone(x_low, zone_increment, x_size, x)
    y_h = find_high_zone(y_low, zone_increment, y_size, y)
    z_h = find_high_zone(z_low, zone_increment, z_size, z)
    
    x_lr = zone_center(x_low, zone_increment, x_l)
    x_hr = zone_center(x_low, zone_increment, x_h)
    y_lr = zone_center(y_low, zone_increment, y_l)
    y_hr = zone_center(y_low, zone_increment, y_h)
    z_lr = zone_center(z_low, zone_increment, z_l)
    z_hr = zone_center(z_low, zone_increment, z_h)

    avg_xl_yl_zl = get_density(x_l,y_l,z_l)
    avg_xl_yl_zh = get_density(x_l,y_l,z_h)
    avg_xl_yh_zl = get_density(x_l,y_h,z_l)
    avg_xl_yh_zh = get_density(x_l,y_h,z_h)
    avg_xh_yl_zl = get_density(x_h,y_l,z_l)
    avg_xh_yl_zh = get_density(x_h,y_l,z_h)
    avg_xh_yh_zl = get_density(x_h,y_h,z_l)
    avg_xh_yh_zh = get_density(x_h,y_h,z_h)

    avg_xl_yl = interpolate(z_lr,z_hr,avg_xl_yl_zl,avg_xl_yl_zh,z)
    avg_xl_yh = interpolate(z_lr,z_hr,avg_xl_yh_zl,avg_xl_yh_zh,z)
    avg_xh_yl = interpolate(z_lr,z_hr,avg_xh_yl_zl,avg_xh_yl_zh,z)
    avg_xh_yh = interpolate(z_lr,z_hr,avg_xh_yh_zl,avg_xh_yh_zh,z)

    avg_xl = interpolate(y_lr,y_hr,avg_xl_yl,avg_xl_yh,y)
    avg_xh = interpolate(y_lr,y_hr,avg_xh_yl,avg_xh_yh,y)

    avg = interpolate(x_lr,x_hr,avg_xl,avg_xh,x)
    
  end subroutine rect_linear_density


  type(vector) pure function get_average(x,y,z)
    integer, intent(in) :: x,y,z
    integer :: n
    type(vector) :: sum
    if(lbound(zone_rect_n,1) <= x .and. x <= ubound(zone_rect_n,1) .and. &
         lbound(zone_rect_n,2) <= y .and. y <= ubound(zone_rect_n,2) .and. &
         lbound(zone_rect_n,3) <= z .and. z <= ubound(zone_rect_n,3)) then
       n =  zone_rect_n(x,y,z)
       sum = zone_rect_vel_sum(x,y,z)
    else
       n = 0
       sum = vector(0.0,0.0,0.0)
    end if
       
    if(n > 0) then
       get_average = (1.0_d_t/n)*sum
    else 
       ! XXX should we be zeroing this?
       get_average = sum
    end if
  end function get_average

  real(kind=d_t) pure function get_density(x,y,z)
    integer, intent(in) :: x,y,z
    integer :: n
    if(lbound(zone_rect_n,1) <= x .and. x <= ubound(zone_rect_n,1) .and. &
         lbound(zone_rect_n,2) <= y .and. y <= ubound(zone_rect_n,2) .and. &
         lbound(zone_rect_n,3) <= z .and. z <= ubound(zone_rect_n,3)) then
       n = zone_rect_n(x,y,z)
    else 
       n = 0
    end if
       

    ! XXX
    !The density is proportional to the number of pebbles 
    ! that were found in the zone
    ! The real density (in mass/volume) depends on zone size, pebble size, and 
    ! number of samples
    get_density = real(n)
    
  end function get_density

  integer pure function find_low_zone(low,increment, size, value)
    REAL(KIND=D_T), intent(in) :: low,increment,value  
    INTEGER, intent(in) :: size
    real(kind=d_t) :: bin
    bin = ((value - low) / increment ) - 0.5
    find_low_zone = iclamp(1,floor(bin),size)
  end function find_low_zone

  integer pure  function find_high_zone(low,increment, size, value)
    REAL(KIND=D_T), intent(in) :: low,increment,value  
    INTEGER, intent(in) :: size
    real(kind=d_t) :: bin
    bin = ((value - low) / increment) - 0.5
    find_high_zone = iclamp(1,ceiling(bin),size)
  end function find_high_zone

  INTEGER pure FUNCTION find_base_zone(low, increment, size, value)  
    REAL(KIND=D_T), intent(in) :: low,increment,value  
    INTEGER, intent(in) :: size
    INTEGER :: zone  
    zone = find_zone(low, increment, size, value)  
    find_base_zone = iclamp(2,zone,size-1)  
  END FUNCTION find_base_zone

 subroutine find_radial_zone(x,y,z,s_z,r_z,z_z)
    REAL(kind=d_t), intent(in) :: x,y,z
    integer, intent(out) :: s_z,r_z,z_z
    REAL(kind=d_t) :: r, theta
    z_z = find_zone(z_low,tally_z_inc, zt_size, z)
    r = sqrt(x**2+y**2)
    theta = atan2(y,x)
    s_z = iclamp(1,FLOOR(((theta + pi) / (2* pi)) * tally_sections) + 1,&
         tally_sections)
    r_z = find_zone(0.0_d_t,tally_radial_inc,rt_size, r)
  end subroutine find_radial_zone

  INTEGER pure FUNCTION find_zone(low, increment, size, value)  
    REAL(KIND=D_T), intent(in) :: low,increment,value  
    INTEGER, intent(in) :: size 
    INTEGER :: bin  
    bin = FLOOR((value - low) / increment)  
    find_zone = iclamp(1,bin,size)  
  END FUNCTION find_zone

  real(kind=d_t) pure function zone_center(low, increment, bin)
    REAL(KIND=D_T), intent(in) :: low,increment
    integer, intent(in) :: bin
    zone_center = low+increment * (bin + 0.5)
  end function zone_center


end module velocity_func
