! Marbles, a distinct element method code for simulating spheres
! Copyright (C) 2013 Joshua J. Cogliati

! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.

! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.

! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

! NOTICE: Portions of this computer software were prepared by Battelle
! Energy Alliance, LLC, hereinafter BEA, under Contract
! No. DE-AC07-05ID14517 with the United States (U.S.) Department of
! Energy (DOE).  NEITHER THE UNITED STATES GOVERNMENT, NOR DOE, NOR
! BEA MAKE ANY WARRANTY, EXPRESSED OR IMPLIED, OR ASSUMES ANY
! LIABILITY OR RESPONSIBILITY FOR THE USE, ACCURACY, COMPLETENESS, OR
! USEFULNESS OR ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
! DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
! OWNED RIGHTS.

! EXPORT RESTRICTIONS. The provider of this computer software and its
! employees and its agents are subject to U.S. export control laws that
! prohibit or restrict (i) transactions with certain parties, and (ii)
! the type and level of technologies and services that may be exported.
! You agree to comply fully with all laws and regulations of the United
! States and other countries (Export Laws) to assure that neither this
! computer software, nor any direct products thereof are (1) exported,
! directly or indirectly, in violation of Export Laws, or (2) are used
! for any purpose prohibited by Export Laws, including, without
! limitation, nuclear, chemical, or biological weapons proliferation.

module wear_tally
  use types
  use vector_type
  use number_utils
  implicit none

  real(kind=d_t) :: pebble_volume_wear_sum = 0.0_d_t
  real(kind=d_t),allocatable,dimension(:) :: pebble_wear_sum_tally
  real(kind=d_t),allocatable,dimension(:) :: surface_wear_sum_tally
  real(kind=d_t) :: pebble_volume_wear_length_sum = 0.0_d_t
  real(kind=d_t),allocatable,dimension(:) :: pebble_wear_length_sum_tally
  real(kind=d_t),allocatable,dimension(:) :: surface_wear_length_sum_tally
  logical :: has_velocity_bins = .false.
  real(kind=d_t),allocatable,dimension(:) :: velocity_bins
  real(kind=d_t),allocatable,dimension(:,:) :: peb_wear_vel_tally,peb_length_vel_tally

  real(kind=d_t) :: normal_force_exponent = 2.0_d_t/3.0_d_t
  real(kind=d_t) :: z_low,z_delta !z_delta is zone_increment
  integer :: z_size

  private
  public :: print_pebble_wear_data,tally_surface_wear,get_wear_force_exponent
  public :: set_wear_force_exponent, initialize_wear_tallies
  public :: tally_pebble_pebble_wear,tally_pebble_surface_wear
  public :: set_wear_velocity_bins
  
contains

  subroutine initialize_wear_tallies(z_low_in,z_delta_in,z_size_in)
    real(kind=d_t),intent(in) :: z_low_in, z_delta_in
    integer, intent(in) :: z_size_in
    z_low = z_low_in
    z_delta = z_delta_in
    z_size = z_size_in
    allocate(pebble_wear_sum_tally(z_size),surface_wear_sum_tally(z_size))  
    allocate(pebble_wear_length_sum_tally(z_size),surface_wear_length_sum_tally(z_size))
    pebble_wear_sum_tally(:) = 0.0
    surface_wear_sum_tally(:) = 0.0
    pebble_wear_length_sum_tally(:) = 0.0
    surface_wear_length_sum_tally(:) = 0.0
    if(has_velocity_bins) then
       allocate(peb_wear_vel_tally(z_size,lbound(velocity_bins,1):&
            ubound(velocity_bins,1)+1))
       peb_wear_vel_tally(:,:) = 0.0
       allocate(peb_length_vel_tally(z_size,lbound(velocity_bins,1):&
            ubound(velocity_bins,1)+1))
       peb_length_vel_tally(:,:) = 0.0
    end if

  end subroutine initialize_wear_tallies

  subroutine set_wear_velocity_bins(velocity_bins_in)
    real(kind=d_t),intent(in),dimension(:) :: velocity_bins_in
    allocate(velocity_bins(size(velocity_bins_in)))
    velocity_bins(:) = velocity_bins_in(:)
    has_velocity_bins = .true.
    call sort_real(velocity_bins)
  end subroutine set_wear_velocity_bins

  subroutine print_pebble_wear_data(time,alpha) 
    real(kind=d_t),intent(in) :: time, alpha
    integer :: z_z,i
    print *,"#pebble_wear_sum",time,pebble_volume_wear_sum*alpha
    do z_z=1,size(pebble_wear_sum_tally)
       if(pebble_wear_sum_tally(z_z) > 0.0) then
          print *,"#pebble_wear_tally",zone_center(z_low,z_delta,z_z),&
               pebble_wear_sum_tally(z_z)*alpha,surface_wear_sum_tally(z_z)*alpha,&
               pebble_wear_length_sum_tally(z_z)*alpha,surface_wear_length_sum_tally(z_z)*alpha
       end if
    end do
    if(has_velocity_bins) then
       print *,"#pebble_wear_vel_tally_head","  z_value             ",(velocity_bins(i),&
            velocity_bins(i),i=lbound(velocity_bins,1),&
            ubound(velocity_bins,1)),"above_length"," above_wear"
       do z_z=1,size(pebble_wear_sum_tally)
          if(pebble_wear_sum_tally(z_z) > 0.0) then
             print *,"#pebble_wear_vel_tally",zone_center(z_low,z_delta,z_z),&
                  (peb_length_vel_tally(z_z,i)*alpha,&
                  peb_wear_vel_tally(z_z,i)*alpha,&
                  i=lbound(peb_length_vel_tally,2),&
                  ubound(peb_length_vel_tally,2))
          end if
       end do
    end if
  end subroutine print_pebble_wear_data

  subroutine tally_pebble_surface_wear(position,v_transverse,f_normal)
    type(vector),intent(in) :: position,v_transverse,f_normal
    real(kind=d_t) :: current_wear, current_length
    !integer :: wearzone
    current_length = abs(v_transverse)
    current_wear = abs(v_transverse) * (abs(f_normal)**(get_wear_force_exponent()))
    call tally_surface_wear(position,current_wear,current_length)
  end subroutine tally_pebble_surface_wear

  subroutine tally_surface_wear(position,current_wear,current_length)
    type(vector),intent(in) :: position
    real(kind=d_t),intent(in) :: current_wear, current_length
    integer :: wearzone
    wearzone = find_zone(z_low,z_delta, z_size, position%z)
    !$OMP ATOMIC
    surface_wear_sum_tally(wearzone) = &
         surface_wear_sum_tally(wearzone) + current_wear
    !$OMP ATOMIC
    surface_wear_length_sum_tally(wearzone) = &
         surface_wear_length_sum_tally(wearzone) + current_length
  end subroutine tally_surface_wear

  subroutine tally_pebble_pebble_wear(pos,other_pos,v_transverse,f_normal)
    type(vector),intent(in) :: pos,other_pos,v_transverse,f_normal
    integer :: wearzone_p,wearzone_o,velocity_bin
    real(kind=d_t) :: current_length,current_wear
    logical :: found
    current_length = abs(v_transverse)
    current_wear = current_length * (abs(f_normal)**(normal_force_exponent))
    wearzone_p = find_zone(z_low, z_delta, z_size, pos%z)
    wearzone_o = find_zone(z_low, z_delta, z_size, other_pos%z)

    !$OMP ATOMIC
    pebble_volume_wear_sum = pebble_volume_wear_sum + current_wear
    !$OMP ATOMIC
    pebble_volume_wear_length_sum = pebble_volume_wear_length_sum + current_length
    !$OMP ATOMIC
    pebble_wear_sum_tally(wearzone_p) = &
         pebble_wear_sum_tally(wearzone_p) + current_wear
    !$OMP ATOMIC
    pebble_wear_length_sum_tally(wearzone_p) = &
         pebble_wear_length_sum_tally(wearzone_p) + current_length
    !$OMP ATOMIC
    pebble_wear_sum_tally(wearzone_o) = &
         pebble_wear_sum_tally(wearzone_o) + current_wear
    !$OMP ATOMIC
    pebble_wear_length_sum_tally(wearzone_o) = &
         pebble_wear_length_sum_tally(wearzone_o) + current_length
    
    if(has_velocity_bins) then
       call search_sorted_real_array(velocity_bins,current_length,velocity_bin,&
            found)
       !$OMP ATOMIC
       peb_length_vel_tally(wearzone_o,velocity_bin) = &
            peb_length_vel_tally(wearzone_o,velocity_bin) + current_length
       !$OMP ATOMIC
       peb_wear_vel_tally(wearzone_o,velocity_bin) = &
            peb_wear_vel_tally(wearzone_o,velocity_bin) + current_wear
       !$OMP ATOMIC
       peb_length_vel_tally(wearzone_p,velocity_bin) = &
            peb_length_vel_tally(wearzone_p,velocity_bin) + current_length
       !$OMP ATOMIC
       peb_wear_vel_tally(wearzone_p,velocity_bin) = &
            peb_wear_vel_tally(wearzone_p,velocity_bin) + current_wear
       
    end if
  end subroutine tally_pebble_pebble_wear

  !XXX also defined in marbles9_common_module
  real(kind=d_t) function zone_center(low, increment, bin)
    REAL(KIND=D_T), intent(in) :: low,increment
    integer, intent(in) :: bin
    zone_center = low+increment * (bin + 0.5)
  end function zone_center

  !XXX also defined in marbles9_common_module
  INTEGER FUNCTION find_zone(low, increment, size, value)  
    REAL(KIND=D_T), intent(in) :: low,increment,value  
    INTEGER, intent(in) :: size 
    INTEGER :: bin  
    bin = FLOOR((value - low) / increment)  
    find_zone = iclamp(1,bin,size)  
  END FUNCTION find_zone

  pure real(kind=d_t) function get_wear_force_exponent()
    get_wear_force_exponent = normal_force_exponent
  end function get_wear_force_exponent

  subroutine set_wear_force_exponent(force_exponent)
    real(kind=d_t),intent(in) :: force_exponent
    normal_force_exponent = force_exponent
  end subroutine set_wear_force_exponent

end module wear_tally
